#!/bin/bash
SOURCE="$0"
source /etc/default/locale 1>/dev/null 2>&1
#set -x

# TMP TODO : we need this in order to have this sourceable and usable everywhere
# elive tools
source /usr/bin/gettext.sh
source /usr/lib/elive-tools/functions
source /usr/lib/elive-tools/functions_hdtools
el_make_environment
TEXTDOMAIN="eliveinstaller"
export TEXTDOMAIN
unset is_console

#===  FUNCTION  ================================================================
#          NAME:  guitool
#   DESCRIPTION:  small tool for interact with the user graphically
#    PARAMETERS:  $1 = mode, $2 = message, $3 = "%s" equivalent vars for $2
#       RETURNS:  value, if needed
#===============================================================================
guitool(){
    functions_pre
    # pre {{{
    local mode message
    #el_debug "\targs '$*'"

    mode="$1"
    message="$2"

    el_check_variables "mode|guitool"

    #el_debug "\targs '$*'"

    # }}}
    # something makes the screen blanking, so force a screen on everytime
    screen_dpms_on

    # This is not going to work here because gettext is unable to get the messages from a dynamic variable
    #    local _translated_message
    #    _translated_message="$( printf "$( eval_gettext "$message" )" "$@" )"

    case "$mode" in
        info)
            el_check_variables "message"
            $guitool --title='Elive Installer' --info --text="$message" 2>/dev/null || true

            $FUNCNAME "working_start"
            ;;
        error)
            el_check_variables "message"
            $guitool --title='Elive Installer' --error --text="$message" 2>/dev/null || true

            $FUNCNAME "working_start"
            ;;
        warning)
            el_check_variables "message"
            $guitool --title='Elive Installer' --warning --text="$message" 2>/dev/null || true

            $FUNCNAME "working_start"
            ;;
        working_start)
            # already running? keep it running instead of creating a new one
            if [[ -e "$TMP_PROGRESS_WORKING_WAITING_f" ]] || [[ -e "$TMP_PROGRESS_WORKING_COPYING_f" ]] || [[ -e "$TMP_PROGRESS_WORKING_WAITING_f" ]] || [[ -e "$TMP_PROGRESS_CONFIGURING_f" ]] || [[ -e "$TMP_PROGRESS_USERS_f" ]] || [[ -e "$TMP_PROGRESS_f" ]] ; then
                return 0
            fi

            # non-invasive status writting to try to not make the bar jumpy
            if ! grep -qs "10" "$TMP_PROGRESS_WORKING_WAITING_f" ; then
                echo "10" > "$TMP_PROGRESS_WORKING_WAITING_f"
            fi


            if [[ -n "$message" ]] ; then
                message_doing_magic="$message"
            else
                # random message generator
                message_doing_magic="$( eval_gettext "Elive is doing some magic..." )"
                # random message generator
                case "$( shuf -i 1-6 -n 1 )" in
                    1)
                        message_doing_magic="$( eval_gettext "Elive is doing some magic..." )"
                        ;;
                    2)
                        message_doing_magic="$( eval_gettext "Transferring bytes at the speed of light..." )"
                        ;;
                    3)
                        message_doing_magic="$( eval_gettext "Arranging some atoms in the universe..." )"
                        ;;
                    4)
                        message_doing_magic="$( eval_gettext "Mutating frogs with unicorns..." )"
                        ;;
                    5)
                        message_doing_magic="Dancing 'Staying Elive' by the Bee Gees..."
                        ;;
                    6)
                        message_doing_magic="$( eval_gettext "Brewing Grog..." )"
                        ;;
                esac
            fi

            # this works correctly this way
            if [[ "$DISPLAY" != "localhost:10"* ]] ; then
                { ( while test -e "$TMP_PROGRESS_WORKING_WAITING_f" ; do cat "$TMP_PROGRESS_WORKING_WAITING_f" || true ; sleep 2 ; done | $guitool --title='Elive Installer' --progress --text="$message_doing_magic" --no-cancel --pulsate --auto-close 2>/dev/null || true ) & disown ; } 2>/dev/null
            fi

            ;;
        working_stop)
            rm -f "$TMP_PROGRESS_WORKING_WAITING_f"
            sync ; LC_ALL=C sleep 0.5

            ;;
        question)
            el_check_variables "message"

            if $guitool --title='Elive Installer' --question --text="$message" 2>/dev/null ; then
                $FUNCNAME "working_start"
                return 0
            else
                $FUNCNAME "working_start"
                return 1
            fi

            ;;
        rebootwanted)
            el_check_variables "message"
            local message_no
            message_no="$( printf "$( eval_gettext "Just continue" )" "" )"
            local message_yes
            message_yes="$( printf "$( eval_gettext "Reboot" )" "" )"

            if $guitool --title='Elive Installer' --question --cancel-label="$message_no" --ok-label="$message_yes" --text="$message" 2>/dev/null ; then
                $FUNCNAME "working_stop"
                return 0
            else
                $FUNCNAME "working_start"
                return 1
            fi

            ;;
        yesnoquestion)
            el_check_variables "message"
            local message_no
            message_no="$( printf "$( eval_gettext "No" )" "" )"
            local message_yes
            message_yes="$( printf "$( eval_gettext "Yes" )" "" )"

            if $guitool --title='Elive Installer' --question --cancel-label="$message_no" --ok-label="$message_yes" --text="$message" 2>/dev/null ; then
                $FUNCNAME "working_start"
                return 0
            else
                $FUNCNAME "working_start"
                return 1
            fi

            ;;
        yesnoquestioninv)
            el_check_variables "message"
            local message_no
            message_no="$( printf "$( eval_gettext "No" )" "" )"
            local message_yes
            message_yes="$( printf "$( eval_gettext "Yes, do that" )" "" )"

            if ! $guitool --title='Elive Installer' --question --cancel-label="$message_yes" --ok-label="$message_no" --text="$message" 2>/dev/null ; then
                $FUNCNAME "working_start"
                return 0
            else
                $FUNCNAME "working_start"
                return 1
            fi

            ;;
        yesnoquestioninvbutcontinue)
            el_check_variables "message"
            local message_no
            message_no="$( printf "$( eval_gettext "No, but continue" )" "" )"
            local message_yes
            message_yes="$( printf "$( eval_gettext "Yes, do that" )" "" )"

            if ! $guitool --title='Elive Installer' --question --cancel-label="$message_yes" --ok-label="$message_no" --text="$message" 2>/dev/null ; then
                $FUNCNAME "working_start"
                return 0
            else
                $FUNCNAME "working_start"
                return 1
            fi

            ;;
    esac
}

screen_dpms_on(){
    # only in E16, seems like it makes E26 crash:
    if [[ -n "$EROOT" ]] && ! [[ -n "$E_HOME_DIR" ]] ; then
        xset dpms force on 2>/dev/null
    fi
}

#===  FUNCTION  ================================================================
#          NAME:  select_filesystem
#   DESCRIPTION:  ask for a filesystem to use, with the optimal selection suggested, returns the FS value
#    PARAMETERS:  -
#       RETURNS:  FS
#===============================================================================
select_filesystem(){
    # pre {{{
    local mode dev answer
    dev="$1"
    mode="$2"

    # }}}

    if [[ "$mode" != "recursive" ]] ; then
        filesystems_help_show
    fi

    # note: these messages are set in the main function start_variables in order to be used in other places

    answer="$(
        {
            if ((is_reiser4_available)) ; then
                echo -e "reiser4\n${msg_fs_reiser4}"
            fi
            if ((is_reiser5_available)) ; then
                echo -e "reiser5\n${msg_fs_reiser5}"
            fi
            if ((is_btrfs_available)) ; then
                echo -e "btrfs\n${msg_fs_btrfs}"
            fi

            echo -e "ext4\n${msg_fs_ext4}"

            if ((is_reiserfs_available)) ; then
                echo -e "reiserfs\n${msg_fs_reiserfs}"
            fi

            echo -e "xfs\n${msg_fs_xfs}"
            echo -e "jfs\n${msg_fs_jfs}"
            echo -e "ext3\n${msg_fs_ext3}"
            echo -e "ext2\n${msg_fs_ext2}"

        } | $guitool --title='Elive Installer' --width=630 --height=320 --list \
            --text="$message_select_fs" \
            --column="$( eval_gettext "Type" )" --column="$( eval_gettext "Description" )" 2>/dev/null || echo cancel )"


    if [[ "$answer" = "btrfs" ]] || [[ "$answer" = "reiser4" ]] || [[ "$answer" = "reiser5" ]] ; then
        el_array_member_add "$answer" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")

        # always
        # is_fs_compression_wanted=1
        is_fs_compression_used=1

        if [[ -z "$is_fs_compression_wanted" ]] ; then
            if guitool yesnoquestion "$msg_fs_supports_compression" ; then
                is_fs_compression_wanted=1
            else
                is_fs_compression_wanted=0
            fi
        fi

        warn_experimental_fs "$answer"
    fi


    filesystems_help_unshow

    # ask again
    if [[ "$answer" = "cancel" ]] || [[ -z "$answer" ]] ; then
        if [[ "$mode" != "recursive" ]] ; then
            guitool warning "$( eval_gettext "You need to select an option from the list." )"
            $FUNCNAME "$dev" recursive
            return 0
        fi
    fi

    # update: this call should be made in the parent function! otherwise we will be unable to exit when cancel
    #verify_result "$answer"

    # reiserfs can be bugged in recent kernels, so warn and confirm:
    if ! ((is_reiserfs_warn)) && [[ "$answer" = "reiserfs" ]] ; then
        is_reiserfs_warn=1

        if ! guitool yesnoquestion "$( eval_gettext "Warning: The ReiserFS filesystem may cause issues in your system and is slated for deprecation. The kernel developers have decided to remove it soon. Are you sure you still want to select it?" )" ; then
            $FUNCNAME "$dev" recursive
            return 0
        fi
        is_reiserfs_used=1
    fi


    if [[ -n "$answer" ]] ; then
        echo "$answer"
    fi
}
#===  FUNCTION  ================================================================
#          NAME:  low_memory_improve
#   DESCRIPTION:  check if we ran out of memory, try to solve it with swap partitions or activate them
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
low_memory_improve(){
    # pre {{{
    el_debug "\targs '$*'"

    # }}}
    sync

    # close some apps to free up memory
    if [[ "$RAM_TOTAL_SIZE_mb" -lt "900" ]]; then
        if ! ((is_message_freeing_up_apps_shown)) ; then
            #guitool info "$( eval_gettext "We are going to close a few applications in order to free up some RAM." )"
            is_message_freeing_up_apps_shown=1
        fi

        # close apps
        for i in conky cairo-dock thunar chromium google-chrome firefox firefox-esr opera opera-stable
        do
            killall "$i" 2>/dev/null 1>&2 || killall -9 "$i" 2>/dev/null 1>&2 ||  true
        done
        # close services
        for i in ssh samba smbd fail2ban cups saned gpm
        do
            service "$i" stop 2>/dev/null 1>&2 || true
        done

        # free some pages
        echo 3 > /proc/sys/vm/drop_caches
    fi

    if dmesg | grep -i "out of memory" ; then
        # try to enable swap partitions
        swap_partitions_all_do "activate"

        if ! swap_partitions_all_do "check" ; then
            guitool warning "$( eval_gettext "It seems an operation ran out of memory. We suggest adding and activating a SWAP partition to proceed avoiding this issue." )"
        fi
    fi
}
#===  FUNCTION  ================================================================
#          NAME:  swap_partitions_all_do
#   DESCRIPTION:  search for existing swap partitions and use the first one found:
#   PARAMETERS:   activate|deactivate
#===============================================================================
swap_partitions_all_do(){
    local mode part
    mode="$1"

    case "$mode" in
        activate)
            while read -ru 3 part
            do
                [[ -z "$part" ]] && continue
                [[ ! -b "$part" ]] && continue

                swapon "$part" 2>/dev/null || true
            done 3<<< "$( EL_DEBUG=0 partitions-list --show-all | awk -v FS="::" '{ if ($3 == "swap") print $1}' || true )"
            ;;
        deactivate)
            # note: we need this entry in low resources computers where was created a temporal swapfile, otherwise the partitions-list will not disable it (because is not a partition) and so the partinioning later will fail
            swapoff -a || true

            sync
            while read -ru 3 part
            do
                [[ -z "$part" ]] && continue
                ! [[ -b "$part" ]] && continue

                swapoff "$part" 1>/dev/null 2>&1 || true
            done 3<<< "$( EL_DEBUG=0 partitions-list --show-all | awk -v FS="::" '{ if ($3 == "swap") print $1}' || true )"
            ;;

        check)
            if LC_ALL=C swapon --noheadings --raw | grep -qs "^/dev" ; then
                return 0
            else
                return 1
            fi
            ;;
    esac
}
#===  FUNCTION  ================================================================
#          NAME:  start_checkers
#   DESCRIPTION:  initialize checkers
#===============================================================================
start_checkers(){
    # pre {{{
    local _seconds_total _seconds_end _seconds_start part p_dev p_filesystem
    functions_pre
    el_debug "\targs '$*'"

    #arg="$1"
    #el_check_variables "arg"

    guitool working_start

    # }}}
    # checks {{{
    el_check_variables "TMP_MAIN_d"
    el_dependencies_check "rsync"

    # root please
    if ! [[ "$UID" = "0" ]] ; then
        guitool error "You need to be root to launch the Elive installer."
        exit 69
    fi


    # - checks }}}
    # cleanups {{{
    #rm -rf "${TMP_MAIN_d}" # do not remove the dir entirely, it will remove things like flags
    #rm -rf /tmp/.eliveinstaller* 2>/dev/null

    mkdir -p "${TMP_MAIN_d}"

    rm -f "$TMP_FSTAB_EXTRA_f" "$DISK_SECTORS" 2>/dev/null || true

    rm -rf /tmp/debug-eliveinstaller 2>/dev/null

    mkdir -p /tmp/debug-eliveinstaller/new /tmp/debug-eliveinstaller/old

    # remove unneeded flags
    #el_flag remove lock_async
    el_flag remove exit_error_reported

    # - cleanups }}}
    # preconfigurations {{{

    # make sure bluetooth works in order to detect it for package selection
    rfkill unblock bluetooth 1>/dev/null 2>&1 || true

    if [[ "$RAM_TOTAL_SIZE_mb" -lt "1200" ]]; then
        # cleanup temporal unneeded files
        service samba stop 1>/dev/null 2>&1 &
        find "${LIVE_MOUNT_OVERLAY_d}/var/lib/samba" -type f -delete 1>/dev/null 2>&1 || true
        find "${LIVE_MOUNT_OVERLAY_d}/var/lib/apt/lists" -type f -delete 1>/dev/null 2>&1 || true
        find "${LIVE_MOUNT_OVERLAY_d}/var/cache/apt/" -type f -delete 1>/dev/null 2>&1 || true

        # browsers cleanup
        killall google-chrome 1>/dev/null 2>&1 || true
        killall chromium 1>/dev/null 2>&1 || true
        killall firefox-esr 1>/dev/null 2>&1 || true
        sync
        find "${LIVE_MOUNT_OVERLAY_d}/home/${eliveuser}/.cache/mozilla/" -type f -delete 1>/dev/null 2>&1 || true
        find "${LIVE_MOUNT_OVERLAY_d}/home/${eliveuser}/.cache/chromium/" -type f -delete 1>/dev/null 2>&1 || true
        find "${LIVE_MOUNT_OVERLAY_d}/home/${eliveuser}/.cache/google-chrome/" -type f -delete 1>/dev/null 2>&1 || true

        # trash
        find "${LIVE_MOUNT_OVERLAY_d}/home/${eliveuser}/.local/share/Trash/" -type f -delete 1>/dev/null 2>&1 || true

    fi

    # FIXME: way to boot parameters the system:  boot=live config autoupgrade image_dir=upgrade/live components findiso=/home/asdf/elive_2.2.6_beta_hybrid.iso
    if grep -hqs "autoupgrade" /proc/cmdline && grep -qs "findiso" /proc/cmdline ; then
        #image_dir="$(grep -h "image_dir" /proc/cmdline | tr ' ' '\n' | grep "image_dir=" | head -1 )"
        #image_dir="/${image_dir#image_dir=}"
        #image_dir="${image_dir%/*}"
        # TODO: what was this for ? not needed aparently, or maybe was the automated install mode?
        #LIVE_MOUNT_MEDIA_d="$LIVE_MOUNT_MAIN_d/findiso/${image_dir}"
        TARGET="$LIVE_MOUNT_MAIN_d/findiso"
        is_autoupgrade=1
        el_explain 0 "Mode __auto-upgrade__ detected to install in __${TARGET}__"

        EL_DEBUG=0 mounts-manager --chroot-layers-umount "$TARGET" "eliveinstaller" 1>/dev/null
        mount -o remount -o rw -o dev -o async "$TARGET"
        # TODO if Autoupgrade we should check if we have enough free space and ask if continue
        # TODO: delete iso in postinstall-elive
    else
        # umount target just in case {{{
        EL_DEBUG=0 mounts-manager --chroot-layers-umount "$TARGET" "eliveinstaller" 1>/dev/null
        target_umount "$TARGET"

        # - umount target just in case }}}
    fi

    # make a temporal copy to the packages to deal in the install with
    # to-remove to-install to-hold to-reconfigure
    for i in remove install reconfigure hold
    do
        rm -f "/tmp/.packages-to-${i}.tmp"

        if [[ -s "/tmp/.packages-to-${i}" ]] ; then
            cp -f "/tmp/.packages-to-$i" "/tmp/.packages-to-${i}.tmp"
        else
            touch "/tmp/.packages-to-${i}.tmp"
        fi
    done


    if [[ -e "/tmp/.eliveinstaller-failed" ]] ; then
        is_previous_installation_failed=1
        rm -f "/tmp/.eliveinstaller-failed"
    fi
    # - preconfigurations }}}

    # check udisks & polkit daemons  {{{

    # check if good reiser4 version
    if installer_requires_newer_kernel_version_than "4.18" silent ; then
        modprobe -q reiser4 2>/dev/null || true

        if grep -qs reiser4 /proc/filesystems ; then
            is_reiser4_available=1
        fi
    fi

    if installer_requires_newer_kernel_version_than "5.5" silent ; then
        modprobe -q reiser5 2>/dev/null || true

        if grep -qs reiser5 /proc/filesystems ; then
            is_reiser5_available=1
        fi
    fi

    # check if we have btrfs
    # update: disable entirely btrfs, it is not safe at all
    if installer_requires_newer_kernel_version_than "6.4" silent ; then
        modprobe -q btrfs 2>/dev/null || true

        if grep -qs btrfs /proc/filesystems ; then
            is_btrfs_available=1
            is_fs_btrfs_prefered=1
        fi
    fi

    # reiserfs is deprecated for newer kernels
    if ((is_wheezy)) ; then
        is_reiserfs_available=1
    fi


    is_ext4_available=1

    # disable all dmssetup mapped devices before to start
    dmsetup remove_all || true

    #if LC_ALL="$EL_LC_EN" udisks --dump 2>&1 | grep -qs "Couldn't enumerate devices" ; then
        #killall udisks-daemon 2>/dev/null || killall -9 udisks-daemon 2>/dev/null || true
        #killall polkitd 2>/dev/null || killall -9 polkitd 2>/dev/null || true

        #{ /usr/lib/udisks/udisks-daemon 2>/dev/null 1>&2 & disown ; } 2>/dev/null
        #{ /usr/lib/policykit-1/polkitd 2>/dev/null 1>&2 & disown ; } 2>/dev/null
        #sync ; sleep 3
    #fi
    if [[ "$( LC_ALL="$EL_LC_EN" udisksctl dump 2>&1 | wc -l )" -lt 10 ]] ; then
        el_warning "restarting udisks2 and polkit services"
        systemctl restart udisks2.service polkit.service  1>/dev/null 2>&1 || true
    fi

    # run if is not running
    if [[ -s "/etc/elive/machine-profile" ]] ; then
        source /etc/elive/machine-profile
    fi

    if [[ "$MACHINE_VIRTUAL" = "yes" ]] ; then
        is_virtualized=1

        case "$MACHINE_VIRTUAL_TYPE" in
            virtualbox)
                is_virtualbox=1
                ;;
            vmware)
                is_vmware=1
                ;;
            qemu)
                is_qemu=1
                ;;
        esac
    fi

    if ! ((is_virtualbox)) ; then
        echo -e "virtualbox-guest-dkms" >> "/tmp/.packages-to-remove.tmp"
        echo -e "virtualbox-guest-x11" >> "/tmp/.packages-to-remove.tmp"
        echo -e "virtualbox-guest-utils" >> "/tmp/.packages-to-remove.tmp"
    fi

    # remove the package virtualbox itself on virtual machine, which also gives problems shutting down the computer
    if ((is_virtualized)) ; then
        echo -e "virtualbox\nvirtualbox-dkms" >> "/tmp/.packages-to-remove.tmp"
    fi

    if ((is_vmware)) ; then
        el_debug "VIRTUAL machine installation: vmware"
        #guitool warning "$( eval_gettext "Installing Elive in VMWARE has been reported to have problems identifying the hard disk. We suggest to install Elive on a real machine instead." )"
    else
        echo -e "open-vm-dkms" >> "/tmp/.packages-to-remove.tmp"
        echo -e "open-vm-tools" >> "/tmp/.packages-to-remove.tmp"
        echo -e "open-vm-tools-desktop" >> "/tmp/.packages-to-remove.tmp"
        echo -e "open-vm-tools-dev" >> "/tmp/.packages-to-remove.tmp"
        echo -e "open-vm-tools-dkms" >> "/tmp/.packages-to-remove.tmp"
        #echo -e "xserver-xorg-input-vmmouse" >> "/tmp/.packages-to-remove.tmp"
        #echo -e "xserver-xorg-video-vmware" >> "/tmp/.packages-to-remove.tmp"
    fi



    if [[ -n "$E_START" ]] ; then
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" "gt" "2.9.10" ; then

            #dbus_get_new_address
            #pong="$( su -c "enlightenment_remote -ping" "$eliveuser" 2>/dev/null )"
            pong="$( su -c "enlightenment_remote -ping" "$eliveuser" 2>/dev/null )"

            # get DBUS address access
            if [[ "$pong" != *pong ]] ; then
                echo "ping: $pong | count: $count | dbus: $DBUS_SESSION_BUS_ADDRESS | DISPLAY: $DISPLAY | XAUTHORITY: $XAUTHORITY | E_IPC_SOCKET: $E_IPC_SOCKET" 1>&2

                if [[ -s "/home/$eliveuser/.cache/dbus-session-address-last.txt" ]] ; then
                    address="$( cat "/home/$eliveuser/.cache/dbus-session-address-last.txt" | tail -1 )"
                    # export result
                    if [[ -n "$address" ]] ; then
                        export DBUS_SESSION_BUS_ADDRESS="$address"
                    fi
                fi
            fi

            pong="$( su -c "enlightenment_remote -ping" "$eliveuser" 2>/dev/null )"
            if [[ "$pong" != *pong ]] ; then
                sleep 1
                pong="$( su -c "enlightenment_remote -ping" "$eliveuser" 2>/dev/null )"
                if [[ "$pong" != *pong ]] ; then
                    el_error "Unable to connect to DBUS via su -c 'enlightenment_remote -ping' $eliveuser"
                fi
            fi
        fi
    fi

    # UEFI mode booted?
    # check efi compatibility:
    if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" "ge" "3.7.2" ; then
        is_efi_compatible=1
    fi
    if mokutil --sb-state 2>&1 | grep -Fqs "SecureBoot enabled" ; then
        is_secureboot_enabled=1
    fi
    if el_check_dir_has_files "/sys/firmware/efi/" 1>/dev/null 2>&1 ; then
        is_efi_booted=1
        modprobe efivarfs 1>/dev/null 2>&1 || true
        sync ; sleep 2
        if ! el_check_dir_hAs_files "/sys/firmware/efi/efivars" 1>/dev/null 2>&1 ; then
            sync ; sleep 3
        fi

        if efibootmgr -v 2>/dev/null | grep -qsiE "(bootia32|grub*32).efi" ; then
            is_efi_32=1
            unset is_efi_64
        fi
        if efibootmgr -v 2>/dev/null | grep -qsiE "(boot.*64|shim*64|grub*64).efi" ; then
            is_efi_64=1
            unset is_efi_32
        fi

        # extra reading from the nvram of efi
        if ! ((is_efi_32)) && ! ((is_efi_64)) ; then
            # extra check
            while read -ru 3 line
            do
                [[ -z "$line" ]] && continue
                if efivar -p -n "$line" 2>/dev/null | grep -qsi "i\.a\.3\.2" ; then
                    is_efi_32=1
                    unset is_efi_64
                fi
            done 3<<<"$( efivar -l 2>/dev/null || true )"
        fi

        # secureboot doesn't works with efi 32bit:
        if ((is_efi_32)) && ((is_secureboot_enabled)) ; then
            el_warning "EFI was detected in 32bit but secureboot is enabled, this is not supported, switching to 64"
            is_efi_64=1
            unset is_efi_32
        fi

        # defaults
        # no way to use 64 on a 32bit system, force it
        # if ((is_i386)) ; then
        #     is_efi_32=1
        #     unset is_efi_64
        # fi
        if ! ((is_efi_32)) && ! ((is_efi_64)) ; then
            if ((is_amd64)) ; then
                # if mismatch we are in the wrong kernel, so use the other one:
                if dmesg | grep -qsiE "efi:.*(mismatch with kernel|No EFI runtime due to)" ; then
                    is_efi_32=1
                    unset is_efi_64
                else
                    is_efi_64=1
                    unset is_efi_32
                fi
            fi
            if ((is_i386)) ; then
                # if mismatch we are in the wrong kernel, so use the other one:
                if dmesg | grep -qsiE "efi:.*(mismatch with kernel|No EFI runtime due to)" ; then
                    is_efi_64=1
                    unset is_efi_32
                else
                    is_efi_32=1
                    unset is_efi_64
                fi
            fi
        fi

        # results
        if ! ((is_efi_32)) && ! ((is_efi_64)) ; then
            el_error "unable to know EFI architecture"
        fi

        if ((is_efi_32)) ; then
            el_debug "EFI architecture: 32 bit, OS arch $arch"
        fi
        if ((is_efi_64)) ; then
            el_debug "EFI architecture: 64 bit, OS arch $arch"
        fi

        # reports to user
        if ((is_efi_32)) && ((is_amd64)) ; then
            guitool warning "$( eval_gettext "You are using the 64-bit version of Elive, but this hardware performs better on a 32-bit system. We recommend downloading the 32-bit version of Elive for a lighter and faster installation." )"
        else
            if [[ "$RAM_TOTAL_SIZE_mb" -lt "4300" ]] && ((is_amd64)) ; then
                guitool warning "$( eval_gettext "Your computer doesn't have much RAM memory. This version of Elive in 64-bit uses more resources than the 32-bit version. You may want to try the 32-bit version of Elive instead to make your computer lighter and faster, consuming less resources so you can work better on this computer. We will continue the installation anyways." )"
            fi
        fi
        if dmesg | grep -qsiE "efi:.*(mismatch with kernel|No EFI runtime due to)" ; then
            if ((is_i386)) ; then
                guitool warning "$( eval_gettext "This version of Elive is for a different architecture. Download the 64-BIT version for this computer." )"
                ((is_efi_64)) && el_warning "Kernel and EFI mismatch, efi is detected for 64 but OS is on 32"
            fi
            if ((is_amd64)) ; then
                guitool warning "$( eval_gettext "This version of Elive is for a different architecture. Download the 32-BIT version for this computer." )"
                ((is_efi_64)) && el_warning "Kernel and EFI mismatch, efi is detected for 32 but OS is on 64"
            fi
        else
            ((is_efi_32)) && ((is_amd64)) && el_warning "EFI is in 32 with OS arch $arch. Boot didn't detected a mismatch in architectures"
            ((is_efi_64)) && ((is_i386)) && el_warning "EFI is in 32 with OS arch $arch. Boot didn't detected a mismatch in architectures"
        fi

        el_debug "is_efi_32 = '$is_efi_32', is_efi_64 = '$is_efi_64', is_i386 = '$is_i386', is_amd64 = '$is_amd64'"

        if el_check_dir_has_files "/sys/firmware/efi/efivars" 1>/dev/null 2>&1 ; then
            el_debug "EFI booted and efivars available (correct efi arch)"
            is_efibootmgr_available=1
        else
            if efibootmgr -v 2>&1 | grep -qsi "EFI variables are not supported on this system" ; then
                el_debug "EFI booted but NOT EFIVARS available"
                local message_uefi_different_arch
                message_uefi_different_arch="$( printf "$( eval_gettext "You need to use a different architecture of Elive for this machine. The system booted in UEFI, but EFIVARS indicates it is not supported for this hardware." )" "" )"
                local message_uefi_different_arch_32
                message_uefi_different_arch_32="$( printf "$( eval_gettext "Your hardware uses a 64-bit UEFI, so you need to use the 64-bit version of Elive to install it on your computer." )" "" )"
                local message_uefi_different_arch_64
                message_uefi_different_arch_64="$( printf "$( eval_gettext "Your hardware uses a 32-bit UEFI, so you need to use the 32-bit version of Elive to install it on your computer." )" "" )"
                local message_uefi_different_arch_end
                message_uefi_different_arch_end="$( printf "$( eval_gettext "Another option is to set your BIOS to boot your computer in %s mode, but this may prevent your other operating systems from booting. Aborting installation..." )" "Legacy" )"
                local message_uefi_different_arch_question
                message_uefi_different_arch_question="$( printf "$( eval_gettext "Would you like to proceed with the installation and attempt to install the correct EFI for this machine?" )" "" )"

                if ((is_i386)) ; then
                    if ! guitool question "${message_uefi_different_arch}\n\n${message_uefi_different_arch_32}\n\n${message_uefi_different_arch_end}\n\n${message_uefi_different_arch_question}" ; then
                        el_error "system needs EFI but we don't have efivars."
                        exit_error
                    fi
                else
                    if ! guitool question "${message_uefi_different_arch}\n\n${message_uefi_different_arch_64}\n\n${message_uefi_different_arch_end}\n\n${message_uefi_different_arch_question}" ; then
                        el_error "system needs EFI but we don't have efivars."
                        exit_error
                    fi
                fi
            else
                el_warning "unexpected state: /sys/firmware/efi/efivars is not available but we have efivars?: $( efibootmgr -v 2>&1 )"
            fi
        fi

    #else
        ## inform about better to switch to uefi
        #if ! ((is_i386)) && ! ((is_notified_efi_boot_suggested)) ; then
            #if ((is_efi_compatible)) ; then
                #guitool info "$( eval_gettext "These new versions of Elive support UEFI, but you booted the computer in LEGACY mode. If you have this option available in your BIOS we recommend switching to it before installing Elive, in order to have an improved disk structure. SecureBoot must be disabled." )"
                    ##guitool info "$( eval_gettext "Installation stopped. Reboot the computer and enter in your BIOS settings to switch to the UEFI mode. You can find howto's on the internet that explains how to enter in the BIOS of your machine." )"
                    ##exit
                ##fi
                #is_notified_efi_boot_suggested=1
            #fi
        #fi
    fi

    # Rare cases where computer boots in Legacy mode but has an EFI partition from which the boot is managed (windows tricky crap), making our install to be installed in BIOS partitions when we need to use the EFI partition instead
    if ! el_flag check warning_efi_notbooted && ! ((is_thanatests)) ; then
        if ! ((is_efi_booted)) ; then
            if [[ -z "$LIVE_MOUNT_MEDIA_DEV_v" ]] ; then
                el_warning "LIVE_MOUNT_MEDIA_DEV_v variable empty"
            fi
            if [[ "$( EL_DEBUG=0 partitions-list --show-all | grep -v "${LIVE_MOUNT_MEDIA_DEV_v}" | awk -v FS="::" '{ if ($5 == "EFI") print $1}' || true )" = "/dev/"* ]] ; then

                if ((is_ventoy)) ; then
                    local message_ventoy_reboot_in_grub2
                    message_ventoy_reboot_in_grub2="$( printf "$( eval_gettext "Your computer has an EFI partition available but it booted in 'Legacy' mode. You need to reboot and make sure to boot your computer in UEFI, and then from Ventoy you should select the boot option named:" )"  ) Grub2 Mode"

                    guitool error "$message_ventoy_reboot_in_grub2"
                else
                    guitool error "$( eval_gettext "Your computer has an EFI partition available but it booted in 'Legacy' mode, probably by accident if you have both options enabled in your BIOS. You must reboot and boot your Elive system in UEFI mode, or change your BIOS settings to boot only in UEFI mode, then you can install Elive correctly. Another solution is to use the entire disk to install Elive using automatic partitioning, thus forcing it to use a Legacy-type boot." )"
                fi

                if guitool rebootwanted "$( eval_gettext "Would you like to reboot now to correctly boot into UEFI mode? If you proceed, you can only install Elive by completely deleting your hard disk and forcing your computer to boot in Legacy mode, which has slight limitations." )" ; then

                    reboot
                fi


                el_flag add warning_efi_notbooted
            fi
        fi
    fi



    # - check udisks & polkit daemons  }}}
    # disable listing of usb device itself for select partitions {{{
    # ignore usb device when showing partitions

    # ignore "ventoy" (usb-creator) devices:
    if ((is_ventoy)) ; then
        buf="$( get_disk_from_partition_complex "/dev/mapper/ventoy" )"

        el_array_member_add "/dev/mapper/ventoy" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")

        if [[ -b "$buf" ]] ; then
            is_live_from_usb=1
            #el_array_member_add "$buf" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; HD_PARTITIONS_IGNORE_UMOUNT_array=("${_out[@]}")
            #el_explain 0 "ventoy: added '$buf' to the list of partitions to ignore umounting"

            #for part in "$buf"*
            for part in "${buf}1" "${buf}2"
            do
                if [[ ! -b "$part" ]] || [[ "$LIVE_MOUNT_MEDIA_DEV_v" = "$part" ]] ; then
                    continue
                fi
                el_array_member_add "$part" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")
                el_explain 0 "ventoy: added '$part' to the list of partitions already used"
            done
        fi
    fi

    # unlist all dvd* devices from listing
    for i in /dev/sr* /dev/dvd* /dev/cdrom* /dev/cdrw* /dev/dvdrw* /dev/dvd* /dev/dvdr* /dev/dvdrw* /dev/dvdram
    do
        if [[ -b "$i" ]] ; then
            el_array_member_add "$i" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")
            el_explain 0 "added '$i' to the list of partitions 'used' in order to ignore it from listing"
        fi
    done

    #if echo "$LIVE_MOUNT_MEDIA_PART_v" | grep -qsE "/dev/[sh]d[a-z][[:digit:]]*" ; then
    # FIXME: move this to a checker function, in places like selecting root partition call it, in case that is a usb suggest the user to not install elive here
    if { LC_ALL="$EL_LC_EN" udevadm info --query=all --name="$LIVE_MOUNT_MEDIA_DEV_v" 2>/dev/null | grep -qs "E: ID_USB_DRIVER=usb-storage" && ! ((is_autoupgrade)) ; } || ((is_ventoy)) ; then

        is_live_from_usb=1

        count=0
        for part in "$LIVE_MOUNT_MEDIA_DEV_v"*
        do
            if [[ ! -b "$part" ]] || [[ "$LIVE_MOUNT_MEDIA_DEV_v" = "$part" ]] ; then
                continue
            fi

            # finish loop after 2 partitions of Elive has been added (we can allow persistence, or other systems like multisystems usb's
            if [[ "$count" -gt 2 ]] ; then
                break
            fi

            if [[ "$( EL_DEBUG=0 partitions-list --show-disks --show-raw --show-only="$part" | grep -iE "(iso9660|:EFI:|:Elive_)" | wc -l )" -gt 0 ]] ; then
                el_array_member_add "$part" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")
                el_explain 0 "usb live: added '$part' to the list of partitions already used"
                count="$(( $count + 1 ))"
            fi
        done

        el_array_member_add "$LIVE_MOUNT_MEDIA_DEV_v" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")
        el_explain 0 "usb live: added disk '$LIVE_MOUNT_MEDIA_DEV_v' to the list of partitions already used"
    fi

    # ignore our own temporal swap files, important
    if LC_ALL=C swapon --noheadings --raw | grep -qs "^/m.*swapfile-tmp" ; then
        while read -ru 3 dir
        do
            dir="${dir%/*}"
            if EL_DEBUG=0 mounts-manager --check-is-mounted "$dir" 1>/dev/null 2>&1 ; then
                part="$( cat /proc/mounts | awk -v dir="$dir" '{if ($2 == dir) print $1}' )"
                if [[ -b "$part" ]] ; then
                    el_array_member_add "$part" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; HD_PARTITIONS_IGNORE_UMOUNT_array=("${_out[@]}")
                    el_explain 0 "added '$part' to the list of partitions used because includes a swap mountpoint"
                    sed -i -e "s|^${dir}.*$|&  # from partition: $part  ## do not delete|" /etc/fstab
                fi
            fi
        done 3<<< "$( swapon --noheadings --raw | grep "^/m.*swapfile-tmp" )"
        unset dir dev
    fi
    #if [[ -n "${HD_PARTITIONS_IGNORE_UMOUNT_array[*]}" ]] ; then
        #guitool warning "$( eval_gettext "You are using a temporary swapfile in a mounted partition, you will not be able to use this partition to install Elive, we recommend that you create a swap partition in your hard disk and reboot in order to free this temporary mount." )"
    #fi


    if ((is_live_from_usb)) ; then
        el_debug "System is from USB"
        if [[ "${LIVE_MOUNT_MEDIA_DEV_v}" = "/dev/sda" ]] && ! [[ "$( ls -1 /dev/nvme* /dev/mmcblk* 2>/dev/null | wc -l )" -gt 0 ]] ; then
            # mark this issue so that the installer knows about it
            is_live_from_usb_detected_as_sda=1

            local message_usb_is_sda
            message_usb_is_sda="$( eval_gettext "Elive booted from a USB thumb drive, but your BIOS detected it as the first hard disk. This shouldn't be a problem for installation, but if the installed system doesn't boot, reboot and change the order in your BIOS." )"

            guitool warning "${message_usb_is_sda}"

            if ! guitool yesnoquestion "$( eval_gettext "Would you like to proceed with the installation?" )" ; then
                guitool info "$( eval_gettext "Hint: Try booting your computer and inserting the USB drive a couple of seconds later. This allows your BIOS to index the hard disk first, then the USB as the secondary drive. If this fails, reset your BIOS to default settings." )"
                reboot
            fi
        fi
    else
        el_debug "System is from DVD"
    fi

    # - disable listing of usb device itself for select partitions }}}
    # check internet connection {{{
    el_debug "verifying internet"

    # wifi drivers ok ? {{{
    el_dependencies_check "laptop-detect|iwlist|iwconfig"
    local wifidev

    # do we use wifi?
    if LC_ALL="$EL_LC_EN" iwconfig 2>/dev/null | grep -qs "IEEE.*ESSID" ; then
        is_wifi_existing=1
    fi

    if ! el_verify_internet fast 2>/dev/null ; then
        if laptop-detect && ! ((is_virtualized)) ; then
            if LC_ALL="$EL_LC_EN" iwconfig 2>/dev/null | grep -qs "IEEE.*ESSID" ; then
                # laptop with wifi (wrongly working)
                wifidev="$( LC_ALL="$EL_LC_EN" iwconfig 2>/dev/null | grep "IEEE.*ESSID" | awk '{print $1}' | psort -- -p "^wlan" | head -1 )"

                if LC_ALL="$EL_LC_EN" iwlist "$wifidev" scan 2>&1 | grep -qs "Network is down" ; then
                    ifconfig "$wifidev" up
                    sleep 3
                fi

                if LC_ALL="$EL_LC_EN" iwlist "$wifidev" scan 2>&1 | grep -qs "Interface doesn't support scanning" ; then
                    ifconfig "$wifidev" up
                    sleep 4

                    if LC_ALL="$EL_LC_EN" iwlist "$wifidev" scan 2>&1 | grep -qs "Interface doesn't support scanning" ; then
                        guitool warning "$( eval_gettext "Wi-Fi not detected. If your laptop has a Wi-Fi switch, ensure it's turned on. There may be no driver for your wireless card. Reboot Elive and select a different kernel version from the boot menu, which might include the necessary drivers." )"
                    fi
                else
                    is_wifi_working=1
                fi
            else
                # laptop without wifi
                if ! guitool rebootwanted "$( eval_gettext "Your computer is a laptop, but no WiFi device could be detected. You can try rebooting and selecting a different kernel version from the first boot menu to see if that works, and/or comment on your issue in our forums. Do you want to continue with the installation?" )" ; then
                    exit
                fi
            fi
        fi
    fi

    # }}}

    if ! el_verify_internet fast 2>/dev/null ; then
        is_forced_installermodule_betas_web=0
    fi

    # know the connections speed
    connection_average_speed_get || true &


    # - check internet connection }}}
    # check APT repositories availability {{{
    # FIX accessibility to debian repos if we are unable to ping them
    repo_url="$( grep "^deb http.*debian.org" /etc/apt/sources.list.d/debian.list | awk '{print $2}' | sed -e 's|^.*://||g' -e 's|/.*$||g' | tail -1 )"
    if [[ -n "$repo_url" ]] ; then
        buf="$( LC_ALL=C ping -W 1 -c 1 "$repo_url" 2>&1 || true )"

        if echo "$buf" | grep -qs "Name or service not known" ; then
            for repo_url in  ftp.$( echo $LANG | sed -e 's|\..*$||g' -e 's|^.*_||g' | tr '[:upper:]' '[:lower:]' ).debian.org ftp.us.debian.org ftp.fr.debian.org ftp.de.debian.org ftp.debian.org
            do
                buf="$( LC_ALL=C ping -W 1 -c 1 "$repo_url" 2>&1 || true )"
                if ! echo "$buf" | grep -qs "Name or service not known" ; then
                    sed -i -e "s|/deb.debian.org/|/$repo_url/|g" /etc/apt/sources.list.d/debian.list
                    echo -e "## Default deb.debian.org repos was moved to $repo_url because we cannot ping the first one" >> /etc/apt/sources.list.d/debian.list
                    el_warning "GOOD TEST (temporal message): moved deb.debian.org repos to $repo_url ; also CHECK if we copy this resulting file to the new /etc installed"
                    break
                fi
            done
        fi
    else
        el_warning "unable to get repo_url variable from debian repos:\n$(grep "^deb http.*debian.org" /etc/apt/sources.list.d/debian.list )"
    fi
    # - check APT repositories availability }}}
    # check for hardware specifics {{{
    if lsmod | grep -qs "^nvidia" && [[ "$( lsmod | grep "^nvidia" | awk '{print $3}' )" != 0 ]]  ; then
        is_nvidia=1
    fi
    if lsmod | grep -qs "^fglrx" && [[ "$( lsmod | grep "^fglrx" | awk '{print $3}' )" != 0 ]]  ; then
        is_fglrx=1
    fi

    # - check for hardware specifics }}}
    # refind needs to be reinstalled? {{{
    if ((is_apple)) ; then
        # empty boot
        if efibootmgr -v | grep -qs "^No BootOrder is set;" ; then
            efi_partition="$( EL_DEBUG=0 partitions-list --show-all | awk -v FS="::" '{ if ($5 == "EFI") print $1}' | head -1 || true )"
            if [[ -b "$efi_partition" ]] ; then
                mkdir /boot/efi
                mount -o dev -o rw "$efi_partition" /boot/efi
                if [[ -d "/boot/efi/EFI/refind" ]] && [[ -d "/boot/efi/EFI/${efi_name}" ]] ; then
                    el_explain 0 "Unbootable system and refind inside, trying to restore it:"
                    cd "/boot/efi/EFI/refind"
                    bkp save  1>/dev/null 2>&1
                    echo -e "refind\trefind/install_to_esp\tboolean\ttrue" | debconf-set-selections
                    TERM=linux DEBIAN_FRONTEND=noninteractive DEBIAN_PRIORITY=critical DEBCONF_NONINTERACTIVE_SEEN=true DEBCONF_NOWARNINGS=true  apt-get install -y refind
                    bkp get  1>/dev/null 2>&1
                    bkp remove  1>/dev/null 2>&1
                    cd /tmp
                    is_refind_restored=1
                fi
                umount /boot/efi
            fi
        fi

        # re-set refind to be the first one to boot
        if efibootmgr | grep -qs "^Boot..... rEFInd Boot Manager" ; then
            local boot_id boot_order
            boot_id="$( efibootmgr | grep "^Boot..... rEFInd Boot Manager" | sed -e 's| rEFInd.*$||g' -e 's|^Boot||g' -e 's|\*$||g' )"
            boot_order="$( efibootmgr | grep "^BootOrder: " | sed -e 's|^BootOrder: ||g' -e "s|${boot_id}||g" -e 's|^,||g' -e 's|,$||g' )"
            if [[ -n "${boot_id}" ]] ; then
                # only suggest to reboot if refind was not the first bootloader
                if ! efibootmgr | grep -qs "^Boot0000\* rEFInd Boot Manager" ; then
                    is_refind_restored=1
                fi
                if [[ -n "${boot_order}" ]] ; then
                    efibootmgr -o "${boot_id},${boot_order}"
                else
                    # XXX not sure yet what makes my macbook boot as "from Preboot" (only, the other option removes refind entirely), but im seeing it working now this way and the only difference I can see is that there's only refind as a boot option (yes, shows macosx from Preboot to pick)
                    efibootmgr -o "${boot_id}"
                fi
            fi
        fi
    else
        echo -e "macfanctld\nfancontrol" >> "/tmp/.packages-to-remove.tmp"
    fi

    if ((is_refind_restored)) ; then
        if guitool rebootwanted "$( eval_gettext "The rEFInd bootloader has been restored. Do you want to reboot now?" )" ; then
            reboot
        fi
    fi


    # -  }}}

    get_disks_number
    encrypted_volumes_open
    list_others_systems
    #warn_windows

    check_datetime
}

#===  FUNCTION  ================================================================
#          NAME:  start_welcome
#   DESCRIPTION:  gives the welcome to the Elive installer
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
start_welcome(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    local answer

    # }}}

    # messages {{{
    echo -e "\tStart installer "

    if [[ "$RAM_TOTAL_SIZE_mb" -lt "900" ]]; then
        guitool info "$( eval_gettext "Your computer does not have much RAM. We recommend removing any unnecessary features for your installed system, such as Bluetooth support, printers, scanners, network services, additional fonts, and languages. This will allow your system to use fewer resources and be faster." )" || true
    fi

    # }}}
    # speak futuristic welcome {{{
    # here because on that way we don't need to wait so long for the update
    if ! ((is_thanatests)) ; then

        #el_speak_text "welcome to the elive installer, please sit in your seat and feel comfortable, we are going to have a magnificent experience together, please enjoy of this futuristic operating system"

        local message_welcome_auto_part
        message_welcome_auto_part="$( printf "$( eval_gettext "Includes correct alignment of partitions to their physical blocks. Special SSD features. Selection of the best combination for your file system to achieve optimal performance. Optional encryption. Delayed data flushes to extend the lifetime of your disk. We strongly recommend this mode, manual partitioning is still possible." )" )"
        local message_welcome_more_than_one_system
        message_welcome_more_than_one_system="$( printf "$( eval_gettext "Your computer may have more than one operating system installed, which you can select at each startup. To achieve this, you need to manually resize your existing partitions to create new ones needed for Elive. Alternatively, you can also run other operating systems like Windows with VirtualBox on Elive, if your CPU includes virtualization features, it will run very fast. Another option is to try installing Windows applications from your Elive, most of which will work without any issues." )" )"
        local message_welcome_migration_upgrade
        message_welcome_migration_upgrade="$( printf "$( eval_gettext "Elive offers unique features: it literally converts another Linux operating system into an Elive one, preserving all your user data and even specific settings like web server configurations. Alternatively, the upgrade mode performs a smart, automated reinstall of your existing Elive system, keeping users and everything necessary intact in a very fast manner. To utilize this feature, you do not need to do anything special. Simply select the existing partition where you wish to install Elive." )" )"
        #local message_welcome_payment
        #message_welcome_payment="$( eval_gettext "Elive can only survive and make progress thanks to donations. During the installation process you will be asked to make a donation, which is a way to motivate users to keep this project alive and evolving. You are not obliged to do this, because we do not want to limit or restrict anybody to use Elive. It is more important for us to give a good system to the world than making a profit from it." )"


        el_dependencies_check "yad"
        guitool "working_stop"

        yad --height=480 --width=450 --form --center --title="Elive Installer" \
            --image-on-top \
            --text="$( eval_gettext "Click to reveal information about the features" ):" \
            --field="$( eval_gettext "Optimizations for automated disk partitioning" )!gtk-dialog-info!Best Option:btn" \
            "@echo \"4:$message_welcome_auto_part\"" \
            --field="$( eval_gettext "More than one operating system found" )!gtk-dialog-info!Use Elive with other systems:btn" \
            "@echo \"4:$message_welcome_more_than_one_system\"" \
            --field="$( eval_gettext "Migration and upgrade modes" )!gtk-dialog-info!Most automated upgrades:btn" \
            "@echo \"4:$message_welcome_migration_upgrade\"" \
            --field="$( eval_gettext "Details" ):txt" "" \
            --field="$( eval_gettext "Guided help during the installation" ):chk" FALSE \
            --field="$( eval_gettext "Configure options (recommended)" ):chk" TRUE \
            --button="gtk-ok" > "$TMP_FILE_1_f" || true

            #--field="$( eval_gettext "Show details of installation progress" ):chk" TRUE \
            #--field="$( eval_gettext "Do I need to pay to install Elive?" )!gtk-dialog-question:btn" \
            #--field="$( eval_gettext "Verify system integrity" ):chk" TRUE \
            #"@echo \"5:$message_welcome_payment\"" \

        if [[ -s "$TMP_FILE_1_f" ]] && grep -qs "|" "$TMP_FILE_1_f" ; then
            if [[ "$( cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' )" = "TRUE" ]] ; then
                el_flag add conf-extra-help
                el_flag add show-suggestions
            else
                el_flag remove conf-extra-help
                el_flag remove show-suggestions
            fi

            if [[ "$( cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $6}' )" = "TRUE" ]] ; then
                el_flag add conf-configure-options
            else
                el_flag remove conf-configure-options
            fi
            #if [[ "$( cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $7}' )" = "FALSE" ]] ; then
                #el_flag add show-progress-information
            #fi
            #if [[ "$( cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $8}' )" = "FALSE" ]] ; then
                #EL_DEBUG=0 el_flag add ignore-checksum-verification
            #fi
        else
            el_flag remove conf-extra-help
            el_flag remove show-suggestions

            el_flag add conf-configure-options
        fi


        rm -f "$TMP_FILE_1_f"


        # there's a small pause before to show the next dialog
        guitool "working_start"
    fi

    # get confs even if we restart the installer
    # extra help
    if el_flag check "conf-extra-help" ; then
        conf_is_extra_help_wanted=1
        conf_is_novice_user=1
        el_config_save "conf_is_novice_user"
        el_config_save "conf_is_extra_help_wanted"

        help_extra_show &
        sync ; sleep 1
    fi

    # customized / advanced mode
    if el_flag check "conf-configure-options" ; then
        conf_configure_options=1
        el_config_save "conf_configure_options"
    fi


    # we want to see debug info (tmux terminals)
    #if el_flag check "show-progress-information" && [[ -x "$( which tmux || true )" ]] ; then
    if [[ -x "$( which tmux || true )" ]] ; then
        unset is_tmux_unwanted
        is_tmux_wanted=1
    else
        unset is_tmux_wanted
        is_tmux_unwanted=1
    fi
    if el_flag check "show-suggestions" ; then
        ignore_show_suggestions=0
    else
        ignore_show_suggestions=1
    fi

    if EL_DEBUG=0 el_flag check "ignore-checksum-verification" ; then
        is_checksum_verification_needed=0
    fi

    # }}}
    # autoupgrade {{{
    if ((is_autoupgrade)) ; then
        message="$( eval_gettext "About to upgrade your Elive system. This feature will preserve your users and files. Do you wish to proceed?" )"

        guitool info "$( eval_gettext "Elive installer started" )"
        if ! guitool yesnoquestion "${message}" ; then
            exit
        fi

        HD_PART_ROOT_dev="$( LC_ALL="$EL_LC_EN" mount | grep -- " on $TARGET" | awk '{print $1}' | head -1 )"

        el_array_member_add "$HD_PART_ROOT_dev" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")

        old_hostname="$(cat "$TARGET/etc/hostname" 2>/dev/null | tail -1 )"
        old_eliveversion="$(cat "$TARGET/etc/elive-version" | grep elive-version | sed -e 's|^.*version: ||g' -e 's|+$||g' )"
        old_elivecodename="$(cat "$TARGET/etc/elive-version" | grep elive-codename | sed 's|^.*codename: ||g')"
        #old_autologin_user="$(cat "$TARGET/etc/elive-version" | grep "^autologin-user:" | sed 's|^.*autologin-user: ||g' | grep "[[:alpha:]]*" | tail -1 )"
        old_autologin_user="$( grep "^autologin-user=" "$TARGET/etc/lightdm/lightdm.conf" 2>/dev/null | awk -v FS="=" '{print $2}' | tail -1 )"

        if [[ -n "$old_autologin_user" ]] ; then
            is_autologin_wanted=1
            autologin_username="$old_autologin_user"
        fi

        # set real home directory
        if [[ -s "$point/etc/adduser.conf" ]] ; then
            dhome="$( grep "^DHOME" "$point/etc/adduser.conf" | sed -e 's|^.*=||g' )"
        fi
        if [[ -z "$dhome" ]] ; then
            dhome="/home"
        fi

        is_old_elive_upgrade_request_accepted=1
        is_migration=1

        # we dont come from usb so just remove this possibly enabled flag
        unset is_live_from_usb
    fi

    # - autoupgrade }}}
    # tell the user that this is a beta version
    #state_development
}

#===  FUNCTION  ================================================================
#          NAME:  encrypted_volumes_open
#   DESCRIPTION:  function to ask for unlock encrypted volumes
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
encrypted_volumes_open(){
    # pre {{{
    el_debug "\targs '$*'"
    el_security_function_loop || return 0
    local is_part_unencrypted

    # }}}
    # open encrypted / lvm volumes {{{
    # remove all encrypted devices first
    dmsetup remove_all || true
    # rescan lvm2's, just in case
    lvscan 1>/dev/null 2>&1 || true
    vgchange --refresh 1>/dev/null 2>&1 || true
    vgchange -ay 1>/dev/null 2>&1 || true

    # warn the user to not unlock multiple disks containing Elive, so the /dev/mapper/elive_vg-root_lv names will be the same and only the first unlocked one will be shown to list, and not all of them
    if [[ "$( EL_DEBUG=0 partitions-list --show-all-raw 2>/dev/null | grep -v "::persistence::" | grep "::crypto_LUKS::" | wc -l || true )" -gt 1 ]] ; then
        guitool warning "$( eval_gettext "More than one encrypted disk available. If you have multiple Elive systems installed, ensure you unlock only the one you want to upgrade." )"
    fi

    # unlock encrypted devices to be able to show them to the user
    while read -ru 3 line
    do
        unset p_dev p_filesystem p_enc_uuid p_enc_name

        p_filesystem="$( echo "$line" | awk -v FS="::" '{print $3}' )"
        p_dev="$( echo "$line" | awk -v FS="::" '{print $1}' )"
        p_size_h="$( echo "$line" | awk -v FS="::" '{print $7}' )"


        case $p_filesystem in
            "crypto_LUKS")
                local message_encrypted_part
                message_encrypted_part="$( printf "$( eval_gettext "Your device '%s' is encrypted.\nPlease enter your password to unlock it." )" "$p_dev (${p_size_h})" )"

                p_enc_uuid="$( LC_ALL="$EL_LC_EN" cryptsetup luksDump "$p_dev" 2>/dev/null | grep "^UUID:" | head -1 | awk '{print $2}' )"
                p_enc_name="${p_dev#/dev/}_crypt"

                # skip if already mounted
                if [[ -b "/dev/mapper/$p_enc_name" ]] ; then
                    continue
                fi

                # luksOpen requires a minimum of ram in order to continue
                if [[ "$RAM_TOTAL_SIZE_mb" -lt 1500 ]] ; then
                    swap_partitions_all_do "activate"
                fi

                for i in 1 2 3
                do
                    # ask again for password if needed
                    if [[ -z "$encrypted_disk_password" ]] || [[ "$encrypted_disk_password" = "cancel" ]] ; then
                        encrypted_disk_password="$( $guitool --title='Elive Installer' --entry --hide-text --text="$message_encrypted_part ($i / 3)" 2>/dev/null || echo cancel )"
                    fi

                    if [[ "$encrypted_disk_password" = "cancel" ]] ; then
                        unset encrypted_disk_password
                        # skip to next partition
                        continue 2
                    fi

                    # try to unlock
                    if [[ -n "$encrypted_disk_password" ]] ; then
                        if printf "%s" "$encrypted_disk_password" | cryptsetup luksOpen "$p_dev" "$p_enc_name" 1>/dev/null ; then
                            is_part_unencrypted=1
                            continue 2
                        else
                            # note: sometimes this process is killed because of running out of memory
                            low_memory_improve

                            printf "%s" "$encrypted_disk_password" | cryptsetup luksOpen "$p_dev" "$p_enc_name" 2>"$TMP_LOGS_ERR_f"  || true
                            if grep -qsi "Not enough available memory to open a keyslot" "$TMP_LOGS_ERR_f" ; then
                                guitool warning "$( eval_gettext "Your computer does not have enough memory to manage this encryption layer. Skipping it..." )"
                            fi

                            if printf "%s" "$encrypted_disk_password" | cryptsetup luksOpen "$p_dev" "$p_enc_name" ; then
                                is_part_unencrypted=1
                                continue 2
                            else
                                unset encrypted_disk_password
                            fi
                        fi
                    fi
                done

                if ! ((is_part_unencrypted)) ; then
                    guitool warning "$( eval_gettext "Encrypted partitions have not been unlocked. The installation will proceed, but you will not be able to select the partitions within the encrypted layer. You also need to unlock them if you wish to have other operating systems listed on your boot menu." )"
                fi

                ;;
        esac
    done 3<<< "$( EL_DEBUG=0 partitions-list --show-all-raw 2>/dev/null | grep -v "::persistence::" || true )"


    # rescan lvm2's, just in case
    lvscan 1>/dev/null 2>&1 || true
    vgchange --refresh 1>/dev/null 2>&1 || true
    vgchange -ay 1>/dev/null 2>&1 || true
    # - open encrypted / lvm volumes }}}
}

#===  FUNCTION  ================================================================
#          NAME:  efi_detect_partition
#   DESCRIPTION:  scans partitions for our uefi bootloader wanted to use
#    PARAMETERS:  $1 = HD_DEV (opt)
#       RETURNS:  -
#===============================================================================
efi_detect_partition(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    local match line disk

    disk="$1"

    # }}}
    # detect if our machines boots in uefi mode
    if ! ((is_efi_booted)) ; then
        return 0
    fi

    # TODO: check "gdisk -l /dev/sda | grep "GPT: present", if not, this disk is not compatible with uefi? (should switch to mbr), only if we install grub on this disk
    # TODO: if our disk is GPT and we end the install without an EFI partition, show an error to the user telling that he must need one
        # TODO: check that the system includes a bios fallback partition if has not, and not uefi too
    # TODO: support for gpt + bios partition fallback too
    # TODO: autopart with GPT mode
    # TODO: add the requried efi* package dependencies to the installer too
    # TODO: detect if bios-partition exist when GPT + non EFI
    # TODO: if amd64 & legacy: recommend to switch to UEFI

    # scan efi partitions from our os-probed results, example:  /dev/sda2@/efi/Microsoft/Boot/bootmgfw.efi:Windows Boot Manager:Windows:efi
    while read -ru 3 line
    do
        unset match
        [[ -z "$line" ]] && continue
        # ignore our USB results (we have an EFI partition but we don't it to be considered!)
        if [[ -n "$LIVE_MOUNT_MEDIA_DEV_v" ]] && [[ "$line" = "${LIVE_MOUNT_MEDIA_DEV_v}"* ]] ; then
            continue
        fi

        if [[ "$line" = *":efi" ]] || [[ "$line" = *"@/efi/"* ]] ; then
            match="$line"
            #match="${match%%:*}"
            match="${match%%@/efi/*}"
            match="${match%%@*}"
            if [[ -b "$match" ]] ; then
                el_array_member_add "$match" "${efi_partitions_maybe[@]}" ; efi_partitions_maybe=("${_out[@]}")
                el_explain 0 "found a possible EFI partition from os-prober (${match}):  $line"
            fi
        fi
    done 3<<< "$( timeout 90 os-prober 2>/dev/null || true )"


    # scan selected disk for efi partition
    if [[ -b "$disk" ]] ; then
        # fdisk:
        while read -ru 3 line
        do
            unset match
            [[ -z "$line" ]] && continue

            # search for efi partitions reading in msdos mode OR gpt mode:
            if echo "$line" | grep -qs "/dev/.*EFI System$" || echo "$line" | grep -qs "/dev/.*ef EFI" ; then
                match="$( echo "$line" | awk '{print $1}' )"
                if [[ -b "$match" ]] \
                    && [[ "$( EL_DEBUG=0 partitions-list --show-only="$match" | grep -viE -- "::(extended|crypto_LUKS|persistence|udf|iso9660|squashfs|Elive_)::" | wc -l )" -ge 1 ]] \
                    && ! el_array_member_check "$match" "${efi_partitions[@]}" \
                    && ! el_array_member_check "$match" "${HD_PARTITIONS_USED_array[@]}" ; then

                        el_array_member_add "$match" "${efi_partitions[@]}" ; efi_partitions=("${_out[@]}")
                        el_explain 0 "found EFI partition from fdisk (${match}):  $line"
                fi
            fi
        done 3<<< "$( LC_ALL=C fdisk -l "$disk" || true )"

        # gdisk:
        # second option, less reliable disk listing:
        if ((is_disk_table_gpt)) ; then
            while read -ru 3 line
            do
                unset match
                [[ -z "$line" ]] && continue

                if echo "$line" | grep -qsi "EFI System" ; then
                    match="$( echo "$line" | awk '{print $1}' )"
                    if [[ -b "${disk}$match" ]] \
                        && ! el_array_member_check "${disk}$match" "${efi_partitions[@]}" \
                        && ! el_array_member_check "${disk}$match" "${HD_PARTITIONS_USED_array[@]}" ; then

                            el_array_member_add "${disk}$match" "${efi_partitions[@]}" ; efi_partitions=("${_out[@]}")
                            el_explain 0 "found EFI partition from gdisk (${disk}${match}):  $line"
                    fi
                fi
            done 3<<< "$( LC_ALL=C gdisk -l "$disk" || true )"
        fi
    fi

    # parted results:
    # note: we wont implement it for now since its not realiable (not showing device, just partition numbers), bu there's demo data:
    # msdos type disk:
    #5:128GB:129GB:262MB:fat32::esp;

    # fdisk (all):
    while read -ru 3 line
    do
        unset match
        [[ -z "$line" ]] && continue
        # ignore our USB results (we have an EFI partition but we don't it to be considered!)
        if [[ -n "$LIVE_MOUNT_MEDIA_DEV_v" ]] && [[ "$line" = "${LIVE_MOUNT_MEDIA_DEV_v}"* ]] ; then
            continue
        fi

        # search for efi partitions reading in msdos mode OR gpt mode:
        if echo "$line" | grep -qs "/dev/.*EFI System$" || echo "$line" | grep -qs "/dev/.*ef EFI" ; then
            match="$( echo "$line" | awk '{print $1}' )"
            if [[ -b "$match" ]] \
                && [[ "$( EL_DEBUG=0 partitions-list --show-only="$match" | grep -viE -- "::(extended|crypto_LUKS|persistence|udf|iso9660|squashfs|Elive_)::" | wc -l )" -ge 1 ]] \
                && ! el_array_member_check "$match" "${efi_partitions[@]}" \
                && ! el_array_member_check "$match" "${HD_PARTITIONS_USED_array[@]}" ; then

                    el_array_member_add "$match" "${efi_partitions[@]}" ; efi_partitions=("${_out[@]}")
                    el_explain 0 "found EFI partition from fdisk (${match}):  $line"
            fi
        fi
    done 3<<< "$( LC_ALL=C fdisk -l | grep "^/dev" || true )"

    if ! [[ -n "${efi_partitions[*]}" ]] ; then
        #/dev/sda5::filesystem::vfat::::EFI::262144000::244 MB::691A-F63C
        while read -ru 3 line
        do
            # ignore our USB results (we have an EFI partition but we don't it to be considered!)
            if [[ -n "$LIVE_MOUNT_MEDIA_DEV_v" ]] && [[ "$line" = "${LIVE_MOUNT_MEDIA_DEV_v}"* ]] ; then
                continue
            fi

            match="$( echo "$line" | awk -v FS="::" '{if ($5 == "EFI" ) print $1}' )"

            if [[ -n "$match" ]] \
                && ! el_array_member_check "$match" "${efi_partitions[@]}" \
                && ! el_array_member_check "$match" "${HD_PARTITIONS_USED_array[@]}" ; then

                    el_array_member_add "$match" "${efi_partitions[@]}" ; efi_partitions=("${_out[@]}")
                    el_explain 0 "found EFI partition from partitions-list (${match}):  $line"
            fi
        done 3<<< "$( EL_DEBUG=0 partitions-list --show-all || true )"
    fi
}

efi_enforce_boot_elive_entry(){
    # NOTE: do not run this function, by default is correctly set, but this function can take "any" elive (or debian) entry to put it as the default, and also depends to be "elive" and not debian
    if efibootmgr | grep -qs "^Boot..... ${efi_name}$" ; then
        local boot_id boot_order
        boot_id="$( efibootmgr | grep "^Boot..... ${efi_name}$" | sed -e 's| elive.*$||g' -e 's| debian.*$||g' -e 's|^Boot||g' -e 's|\*$||g' )"
        boot_order="$( efibootmgr | grep "^BootOrder: " | sed -e 's|^BootOrder: ||g' -e "s|${boot_id}||g" -e 's|^,||g' -e 's|,$||g' )"
        if [[ -n "${boot_id}" ]] ; then
            # only suggest to reboot if refind was not the first bootloader
            # if ! efibootmgr | grep -qs "^Boot0000\* ${efi_name}$" ; then
            #     is_refind_restored=1
            # fi
            if [[ -n "${boot_order}" ]] ; then
                efibootmgr -o "${boot_id},${boot_order}"
            fi
        fi
    fi
}

efi_fix_sparsefiles(){
    # only if a previous installation failed to boot
    if [[ "$( grep "Status: " "$point/etc/elive/installation-status" 2>/dev/null | sed -e 's|^.*Status: ||g' | tail -1 )" != "installed" ]] ; then
        return 0
    fi
    local message_sparse
    message_sparse="$( printf "$( eval_gettext "Did your previous installation fail to boot on any of your systems due to the boot message '%s'? Choose 'Yes' to fix this particular error." )" "Sparse Files not allowed" )"

    if guitool question "$message_sparse" ; then
        is_fix_grub_sparsefiles_wanted=1
    fi

}
efi_fix_debian_dir(){
    list_others_systems
    if ! grep -qsi "debian" /tmp/os-probed ; then
        if [[ -d "$TARGET/boot/efi/EFI" ]] ; then
            if [[ -d "$TARGET/boot/efi/EFI/debian" ]] ; then
                if guitool question "$( eval_gettext "You have a 'debian' entry in your EFI boot, but no 'debian' system is installed. We recommend removing this directory to avoid conflicts with Elive. Remove it?" )" ; then
                    is_fix_grub_remove_efi_debian_dir_wanted=1
                fi
            fi
        fi
    fi

    if ((is_secureboot_enabled)) ; then
        efi_boot_file_elive="\\\\EFI\\\\${efi_name}\\\\shimx64.efi"
    else
        if ((is_efi_32)) ; then
            efi_boot_file_elive="\\\\EFI\\\\${efi_name}\\\\grubia32.efi"
        else
            efi_boot_file_elive="\\\\EFI\\\\${efi_name}\\\\grubx64.efi"
        fi
    fi
}

efi_fix_windows_monopol(){
    # tell the user how to fix his boot menu if has windows
    efi_boot_file_elive="$( efibootmgr -v | grep "\\\\EFI\\\\${efi_name}\\\\" | sed -e 's|^.*File(||g' -e 's|\.efi).*$|.efi|g' -e 's|\\|\\\\|g' | sort -u | tail -1 )"
    if [[ ! -n "$efi_boot_file_elive" ]] ; then
        if ((is_secureboot_enabled)) ; then
            efi_boot_file_elive="\\\\EFI\\\\${efi_name}\\\\shimx64.efi"
        else
            if ((is_efi_32)) ; then
                efi_boot_file_elive="\\\\EFI\\\\${efi_name}\\\\grubx64.efi"
            else
                efi_boot_file_elive="\\\\EFI\\\\${efi_name}\\\\grubia32.efi"
            fi
        fi
    fi
    # temporal debug to know how the windows boot manager is set, some computers like bobby's one doesn't shows the grub menu at all after install, we need to make sure that there's not the default boot entry showing "Windows Boot Manager" incorrectly, or see if elive is new default one, or what... this message is needed to know what is the default values from new installed elives where we have not YET fixed the bcdedit thingy
    # or, we should simply include the elive in the boot order (-o) ? and set it the first default one probably
    # update: seems like changing the boot order (-o) doesn't fix anything, we really need to use bcdedit from windows to fix it when it fails
    el_info "efibootmgr result:"
    efibootmgr -v 1>&2

    local message_efi_win_1
    message_efi_win_1="$( printf "$( eval_gettext "Important: If you can't access Elive after rebooting and the computer only boots into Windows, it's due to Microsoft's monopoly. The solution is to access Windows and follow these steps: Click the start icon, type CMD (Command Prompt), right-click the option 'Run as administrator', then enter the following command in the terminal:" )" "" )"
    local message_efi_win_2
    message_efi_win_2="$( printf "$( eval_gettext "Make sure to take a photo of this command with your phone before rebooting in order to having saved these instructions." )" "" )"
    message_checkbox="$( printf "$( eval_gettext "I confirm I took the photo." )" "" )"
    local message_efi_win_button_ok
    # message_efi_win_button_ok="$( printf "$( eval_gettext "Important - Read First" )" "" )"
    message_efi_win_button_ok="$( printf "$( eval_gettext "Important - Read First" )" "" )"
    local message_must_select
    message_must_select="$( printf "$( eval_gettext "The previous message contains critical information. You must read it fully and take a photo with your phone to continue." )" "" )"
    local message_command
    message_command="\nbcdedit  /set  {bootmgr}  path  ${efi_boot_file_elive}\n"


    # $guitool --title='Elive Installer' --warning --ok-label="${message_efi_win_button_ok}" --text="${message_efi_win_1}\n\nbcdedit /set {bootmgr} path ${efi_boot_file_elive}\n\n${message_efi_win_2}" 2>/dev/null

    while true ; do
        result="$( yad --on-top --center \
            --form \
            --image=gtk-dialog-warning \
            --button=gtk-ok:0 \
            --field="${message_efi_win_1}":LBL \
            --field="${message_command}":LBL \
            --field="${message_efi_win_2}":LBL \
            --field="${message_checkbox}":CHK "TRUE" \
            $NULL || true )"

        result="$( echo "$result" | awk -v FS="|" '{print $4}' )"

        # user don't agree
        # if [[ -z "$result" ]] ; then
        #     $guitool --title='Elive Installer' --image=gtk-dialog-error --button=gtk-ok:0 --text="${message_declined}" 2>/dev/null
        #     exit 1
        # fi
        if [[ "$result" = "FALSE" ]] ; then
            # $guitool --title='Elive Installer' --error --text="${message_must_select}" 2>/dev/null
            guitool error "${message_must_select}"
            continue
        fi

        break
    done

}

#########################################################################
parter_select(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"

    local answer i

    if ((is_old_elive_upgrade_request_accepted)) || ((ignore_parter_select)) || ((is_migration)) ; then
        return 0
    fi

    # - pre }}}
    # prompt user {{{
    #guitool "working_stop"

    answer="$( $guitool --title='Elive Installer' --height=260 --width=700 --list --text="$( eval_gettext "Select a hard disk partitioning option" )" --column="$( eval_gettext "Program" )" --column="$( eval_gettext "Description" )" "DONE" "$( eval_gettext "My disks are already partitioned." )" "Automatic" "$( eval_gettext "Use the entire disk, smart partitioning with optimal options, can be encrypted" )" "Windows" "$( eval_gettext "Resize your Windows partition to have both systems (tutorial)" )" "gparted" "$( eval_gettext "Graphical and user-friendly partitioner" )" "gnome-disks" "$( eval_gettext "Advanced partitioner, allowing encryption and LVM layers" )" "Help" "$( eval_gettext "Tell me how I should partition my hard disk manually" )" 2>/dev/null || echo cancel )"

    if [[ "$1" != "again" ]] ; then
        verify_result "$answer"
    fi

    el_explain 2 "Partitioner selected is $answer"

    guitool "working_start"

    # TODO: if another OS (linux) is detected in os-prober, suggest the user to migrate it instead of repartition?


    # - prompt user }}}
    # umount partitions for being able to repartition {{{
    # disable swap ones
    el_debug "Umounting Swap partitions"

    if swap_partitions_all_do "check" ; then
        # always umount is the desired behaviour
        if [[ "$RAM_TOTAL_SIZE_mb" -gt 1400 ]] ; then
            is_swap_umount_wanted=1
        else
            if guitool question "$( eval_gettext "The swap partition will be unmounted. Close all applications to minimize memory usage. If the installation fails, choose NO next time you reach this step." )" ; then
                is_swap_umount_wanted=1
            fi
        fi

        if ((is_swap_umount_wanted)) ; then
            killall google-chrome chromium firefox cairo-dock opera skype skypeforlinux zoom libreoffice gimp inkscape 1>/dev/null 2>&1 || true
            swap_partitions_all_do "deactivate"
        fi
    fi

    # umount disks before to try to repartition them
    el_explain 0 "Umounting first all partitions, needed for correctly use the parter tools"

    guitool "working_start"

    #if [[ "$EL_DEBUG" -gt 2 ]] ; then
        #el_debug "$FUNCNAME going to processing partitions to umount:"
        #EL_DEBUG=0 partitions-list --show-all-mountable | grep -vE -- "^${LIVE_MOUNT_MEDIA_PART_v}$" || true
    #fi

    #swap_partitions_all_do "deactivate"

    while read -ru 3 line
    do
        # doesn't exist? continue
        if [[ -z "$line" ]] ; then
            continue
        fi
        if ! [[ -b "$line" ]] ; then
            el_debug "skipping $line for umounting because it's not a block device"
            continue
        fi

        # persistence partition should be not umounted (not even the buggy one)
        # update: we don't support installing with persistence, so ignore this:
        #if EL_DEBUG=0 mounts-manager --get-mountpoint "$line" 2>/dev/null | grep -qs "mount/persistence" ; then
            #el_debug "skipping $line for umounting (persistence partition)"
            #continue
        #fi

        #el_debug "at $line"
        if ! el_array_member_check "$line" "${HD_PARTITIONS_USED_array[@]}" && ! el_array_member_check "$line" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; then
            # re-read partition table
            device_check_is_busy "$line"

            if el_umount "$line" 1>/dev/null ; then
                el_debug "$line is not in our array of partitions to use, umounted"
            else
                el_warning "Unable to umount $line"
            fi
        fi
    done 3<<< "$( EL_DEBUG=0 partitions-list --show-all-mountable | grep -viE -- "::(persistence|udf|iso9660|squashfs|Elive_)::" | awk -v FS="::" '{print $1}' | grep -vE -- "^${LIVE_MOUNT_MEDIA_PART_v}$" || true )"

    # TODO: show a message before to ask to partitioner explaining all the options and specially which is the partitionment for Elive

    # - umount partitions for being able to repartition }}}
    # run parters {{{
    case "$answer" in
        "gparted")
            # needed for not hangs on some computers
            rmmod floppy 2>/dev/null || true
            if ! el_dependencies_check "gparted" ; then
                el_dependencies_install gparted
            fi
            # TODO: show an "info" message to the user explaining how to optimally partition for elive (crypted partition etc.. too)
            guitool info "$( eval_gettext "Repartition your hard disk according to your needs. The installer will continue after you close the partitioning application. If you have important data on this hard disk, do a backup BEFORE proceeding with any partitioning procedure." )"

            dmsetup remove_all || true
            swap_partitions_all_do "deactivate"

            # run parter
            gparted

            encrypted_volumes_open

            $FUNCNAME again
            ;;
        "palimpsest"|"gnome-disks")
            if ! el_dependencies_check "gnome-disks" ; then
                el_dependencies_install "gnome-disk-utility"
            fi
            guitool info "$( eval_gettext "Repartition your hard disk according to your needs. The installer will continue after you close the partitioning application. If you have important data on this hard disk, do a backup BEFORE proceeding with any partitioning procedure." )"

            dmsetup remove_all || true
            swap_partitions_all_do "deactivate"

            # run parter
            gnome-disks 2>/dev/null || palimpsest 2>/dev/null

            encrypted_volumes_open

            # include back possible daemons {{{
            if LC_ALL="$EL_LC_EN" vgchange -ay 2>/dev/null | grep -q "logical volume.*active" ; then
                sed -i 's|lvm2||g' "/tmp/.packages-to-remove.tmp"
                # sed -i 's|mdadm||g' "/tmp/.packages-to-remove.tmp"
            fi
            # if cat /proc/mdstat | grep -qs " : active.*raid" \
            #     || cat /proc/mdstat | grep Personalities | sed -e 's|Personalities :||g' | grep -qs "[[:alpha:]]" \
            #     || ls -1 /dev/md* /dev/md/* 1>/dev/null 2>&1 ; then
            if [[ -n "$( awk '/^md/ {printf "%s: ", $1}; /blocks/ {print $NF}' /proc/mdstat )" ]] ; then
                sed -i 's|mdadm||g' "/tmp/.packages-to-remove.tmp"
            fi
            if dmraid -r 1>/dev/null 2>/dev/null ; then
                sed -i 's|dmraid||g' "/tmp/.packages-to-remove.tmp"
            fi

            # - include back possible daemons }}}
            #if guitool yesnoquestion "$( eval_gettext "Do you want to run the Gparted partitioner now?" )" ; then
                #dmsetup remove_all || true
                #gparted
            #fi

            $FUNCNAME again
            ;;
        "none"|"done"|"DONE"|"cancel")
            true

            ;;
        "Automatic")
            dmsetup remove_all || true

            swap_partitions_all_do "deactivate"

            swapfile_temporal_live "deactivate"
            swapfile_temporal_live "remove"

            do_automated_partitioning
            is_automated_partitioned=1

            ;;
        "Windows")
            if ! el_verify_internet ; then
                guitool info "$( eval_gettext "This option requires a connection to the internet. Close this window after the connection is established." )"
            fi
            su -c "web-launcher --incognito --temp-profile \"https://forum.elivelinux.org/t/how-to-resize-ms-windows-to-install-linux-and-have-both-systems/2455\"" "$eliveuser" 1>/dev/null 2>&1  &

            sleep 20

            $FUNCNAME again
            ;;

        "Help")
            if ! el_verify_internet ; then
                guitool info "$( eval_gettext "This option requires a connection to the internet. Close this window after the connection is established." )"
            fi
            su -c "web-launcher --incognito --temp-profile \"https://forum.elivelinux.org/t/how-to-manually-partition-your-hard-disk-to-install-linux/2456\"" "$eliveuser" 1>/dev/null 2>&1  &
            sleep 20

            $FUNCNAME again
            ;;
        *)
            true
            ;;
    esac

    # show new partitions structure
    if [[ "$EL_DEBUG" -gt 2 ]] && [[ "$1" != "again" ]] ; then
        el_debug "Partitions new structure: \n$( EL_DEBUG=0 partitions-list --show-all --show-raw --show-disks || true )"
    fi

    # do not remove: example code for run a "progress" window for misc things
    #{ sleep 1 ; echo 10 ; /etc/rc2.d/S*disks_fstab ; echo 100 ; } | $guitool --title='Elive Installer' --progress --pulsate --auto-close 2>/dev/null
    #service udev restart

    # run it again, if they changed partition numbers
    # update: not needed, its already run
    #list_others_systems

    # - run parters }}}
}
#########################################################################
do_automated_partitioning(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    local answer is_autopart_hd_small autopart_hd_type_size_small p_num p_last_size p_size_mb answer disk size luksformat_options

    el_explain 2 "__Automated Partitioning__ selected"

    is_partitioning_auto=1

    # - pre }}}
    # select disk {{{
    unset menu
    while read -ru 3 entry
    do
        unset disk size name
        [[ "$entry" = "/dev/loop"* ]] && continue
        [[ "$entry" = "/dev/sr"* ]] && continue
        if ! [[ -b "$entry" ]] ; then
            el_debug "skipping $entry because is not a block device"
            continue
        fi

        # skip usb device
        if [[ -n "$LIVE_MOUNT_MEDIA_DEV_v" ]] && [[ "$entry" = "${LIVE_MOUNT_MEDIA_DEV_v}" ]] ; then
            el_explain 0 "$entry is our USB bootable stick ($LIVE_MOUNT_MEDIA_DEV_v), we dont want to install elive here"
            continue
        fi

        size="$( LC_ALL="$EL_LC_EN" lsblk -rino PATH,TYPE,SIZE | awk -v item="${entry}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' )"
        [[ -z "$size" ]] && size="(unknown)"

        name="$( udevadm info --query=property --name="$entry" | grep "ID_VENDOR_ID=" | sed -e 's|^.*VENDOR_ID=||g' | tail -1 )"
        [[ -z "$name" ]] && name="$( udevadm info --query=property --name="$entry" | grep "ID_MODEL=" | sed -e 's|^.*ID_MODEL=||g' | tail -1 )"
        [[ -z "$name" ]] && name="$( LC_ALL="$EL_LC_EN" lsblk -rino PATH,TYPE,LABEL | awk -v item="${entry}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' )"
        [[ -z "$name" ]] && name="disk"

        menu+=("$( echo "$entry" )")
        menu+=("$( echo "$size" )")
        menu+=("$( echo "$name" )")

    done 3<<< "$( { LC_ALL="$EL_LC_EN" lsblk -d -rino PATH ; ls -1 /dev/[shv]d[a-z] /dev/xvd[a-z] /dev/nvme[0-9]*n[0-9] /dev/md[0-9][0-9][0-9] /dev/mmcblk[0-9] /dev/mmcblk[0-9][0-9]   2>/dev/null || true ; }  | grep -vE "^/dev/(fd|sr)" | sort -u )"
    # note: nvme0p1 is not a partition but a disk, so is correct to use with *n[0-9] then


    # if we got only one result for the menu is because we have only one device found, don't ask and just use it
    if [[ "${#menu[@]}" -eq 3 ]] ; then
        answer="$( echo "${menu[0]}" | grep "/dev/" | head -1 | awk '{print $1}' )"
        el_explain 0 "Selected '$answer' as disk because is the only one"
    fi


    # if there was more than one disk found, ask
    if ! [[ -b "$answer" ]] ; then
        answer="$( $guitool --title='Elive Installer' --list --height=300 --width=500 --text="$( eval_gettext "Choose the hard disk where to install Elive.\n\nSelect the first one if you're unsure." )" --column="$( eval_gettext "Device" )" --column="$( eval_gettext "Size" )" --column="$( eval_gettext "Name" )" "${menu[@]}" 2>/dev/null || echo cancel )"
        if [[ "$answer" = "cancel" ]] || [[ -z "$answer" ]] ; then
            guitool error "$( eval_gettext "Select a disk for the installation of Elive" )"

            answer="$( $guitool --title='Elive Installer' --list --height=300 --width=500 --text="$( eval_gettext "Choose the hard disk where to install Elive.\n\nSelect the first one if you're unsure." )" --column="$( eval_gettext "Device" )" --column="$( eval_gettext "Size" )" --column="$( eval_gettext "Name" )" "${menu[@]}" 2>/dev/null || echo cancel )"
        fi
    fi

    if [[ "$answer" = "cancel" ]] || [[ -z "$answer" ]] ; then
        guitool error "$( eval_gettext "No disk selected. Aborting installation..." )"
        el_error "no HD selected, aborting install: ${menu[*]}"
        exit
    fi

    if [[ -b "$answer" ]] ; then
        HD_DEV="$( get_disk_from_partition_complex "$answer" )"
        if [[ "$answer" != "$HD_DEV" ]] ; then
            el_debug "disk selected $answer converted to $HD_DEV"
        fi
    fi

    # ask to erase
    if ! ((is_thanatests)) ; then
        if ! guitool question "$( eval_gettext "You are about to ERASE your entire hard disk to install Elive. Continue?" )" ; then
            el_error "user don't want to erase entirely the disk to continue, aborting"
            guitool error "$( eval_gettext "You chose not to automatically partition your hard disk for Elive installation. Stopping the installation..." )"
            exit
        fi
    fi


    if [[ -b "$HD_DEV" ]] ; then
        # unblock possible things on disk
        # note: ventoy doesn't shows up from here
        if [[ "$( ls -1 /dev/mapper/ 2>/dev/null | grep -vE "(control|ventoy)" | wc -l )" -ge 1 ]] ; then
            el_debug "dmsetup remove '$HD_DEV'"
            dmsetup remove "$HD_DEV" 2>/dev/null || true

            el_debug "dmsetup remove_all"
            dmsetup remove_all || true
        fi

        # just a check to get the size of the disk
        parted -ms -a optimal "$HD_DEV" mklabel msdos 2>/dev/null || true

        device_check_is_busy "$HD_DEV"

        HD_SIZE_mb="$( hd_device_get_size_mb "$HD_DEV" )"
        # note: if we are unable to determine the hard disk size, it may be due that it is broken (badblocks, etc)

        # enable smartmon to scan for possible bad blocks on the disk
        if ! smartctl -s on "$HD_DEV" 1>/dev/null 2>&1 ; then
            echo -e "smartmontools" >> "/tmp/.packages-to-remove.tmp"
        fi

        # fail if no disk size known
        if ! echo "$HD_SIZE_mb" | grep -qs "[[:digit:]]" ; then
            guitool error "$( eval_gettext "The selected partition is not optimally aligned, which may result in poor performance.\n\nYou can proceed with the installation, and you likely won't notice any issues. However, it's advised to repartition your hard disk using the %s application and recreate your partitions with the alignment option set to '%s'.\n\nAlternatively, use Elive's automated partitioner, which picks the best options for you." )"
            el_error "Unable to know the size of your hard disk to partition: $HD_DEV - $HD_SIZE_mb"
            exit_error
        fi
    fi

    if [[ -b "$HD_DEV" ]] ; then
        el_explain 2 "Selected Disk is __${HD_DEV}__ with size __${HD_SIZE_mb}__"
    else
        guitool error "$( eval_gettext "No disk selected. Aborting installation..." )"
        el_error "HD_DEV is empty, user didn't selected any disk for automated partitioning"
        exit
    fi

    device_check_is_busy "$HD_DEV"

    # check and repair errors in HD (fast check)
    hd_errors_check

    # improve ram if needed
    low_memory_improve

    # installer module ?
    #instcod_request


    # finished to select disk
    unset answer



    # - select disk }}}
    # Request a security code and erase the entire disk {{{
    #if [[ "$( cat /tmp/os-probed | wc -l )" -gt 0 ]] ; then
        #if ! guitool question "$( eval_gettext "You are about to ERASE your entire hard disk to install Elive. Continue?" )" ; then
    # Erase the disk
    dd if=/dev/zero of=${HD_DEV} bs=4M count=100 1>/dev/null 2>&1
    # update values (remove partitions known)
    hd_refresh "$HD_DEV"

        # full erasing for disk boosting
        # update: im not sure that this improves in any sense, and it's quite annoying, also bad wording it used (should be "filling with zeroes)
        #if device_is_ssd "$HD_DEV" ; then
            #if guitool question "$( eval_gettext "Do you want to erase your entire hard disk? This will take a quite a while, but can make your SSD run four times faster." )" ; then
                #guitool info "$( eval_gettext "This process can take several hours. You can abort it at any time, but it is recommended to wait until it is finished." )"
                #{ echo 10 ; dd if=/dev/zero of=${HD_DEV} bs=1M 2>/dev/null || true ; sync ; echo 100 ; } | $guitool --title='Elive Installer' --progress --text="$( eval_gettext "Overwriting hard disk with zeroes..." )" --pulsate --auto-close --auto-kill 2>/dev/null
            #fi
        #fi

    device_check_is_busy "$HD_DEV"

    echo 20 > "$TMP_PROGRESS_CONFIGURING_f"


    # }}}
    # calculate automated sizes and configurations {{{
    # Calculate sizes for: swap partition
    # TODO: swap is needed for not break desktop for example if ram is full (and things dying), but we need a detector of that we are using SWAP and warn the user
    # TODO: swap should be not used or never writed when ssd
    # TODO: see if the user really needs swap depending of the RAM amount
    # TODO: improve overall partitioning, for example:
    #  - root partition small with reiserfs (small = faster boot), home partition separed with ext4 for faster boot?
    #  - align HD (ssd?) cilinders and stuff, see other HD configurations
    #    - https://wiki.debian.org/SSDOptimization TODO


    # maintain only for tests:
    #if ((is_virtualized)) ; then
        #if ((is_thanatests)) ; then
            #autopart_hd_type_size_small=9000
        #else
            #autopart_hd_type_size_small=18000
        #fi
    #else
        ## we dont need /home for disks less than in net-tops (160 GB), we can deal with full partition, and would be a waste of space
        #autopart_hd_type_size_small=75000
    #fi

    # we dont need a /home for disks less than in net-tops (160 GB), we can deal with full partition, and would be a waste of space - UPDATE: not less than 500 gb, we don't need it at all in fact! - or at least 250gb:
    #autopart_hd_type_size_small=140000
    if ((is_thanatests)) ; then
        # force the disk to be a "big disk" when doing tests
        autopart_hd_type_size_small=2000
    else
        # IMPORTANT: the min size should always have a big value, otherwise the sizes selector will make a too big value and the installer will fail
        # so it needs to be min the same sizes used for the "big options" configuration (root + extra home + boot.. + efi )
        autopart_hd_type_size_small=280000
    fi

    # notify user about small HD
    if [[ "$HD_SIZE_mb" -lt "15000" ]] ; then
        guitool warning "$( eval_gettext "Your hard disk doesn't have much space. We will try to use the optimal sizes for disk partitioning in order to correctly use its space. You can also try to use the 32-bit version of Elive, which is lighter in resources." )"
    fi

    # compress by default if small
    if [[ "$HD_SIZE_mb" -lt "200000" ]] ; then
        if ((is_btrfs_available)) || ((is_reiser4_available)) || ((is_reiser5_available)) ; then
            #answer="btrfs"
            needs_boot_partition=1
            # update: always enabled from start_checkers
            is_fs_compression_wanted=1
            is_fs_btrfs_prefered=1
            el_explain 0 "pre-selecting btrfs and compression for the FS by default because the disk has less than 200 GB"
        fi
    fi


    # if we only have ext4 featured for SSD's, use it if our disk is SSD
    if ! ((is_reiser4_available)) && ! ((is_btrfs_available)) && ! ((is_reiser5_available)) ; then
        if device_is_ssd "$HD_DEV" && ((is_ext4_available)) ; then
            answer="ext4"
        fi
    fi

    if ((is_btrfs_available)) && ((is_fs_btrfs_prefered)) ; then
        needs_boot_partition=1
        if ((is_thanatests)) ; then
            unset is_fs_compression_wanted
        else
            is_fs_compression_wanted=1
        fi
        answer="btrfs"
    fi

    # ask for FS {{{
    if [[ -z "$answer" ]] ; then

        if ((is_reiser4_available)) || ((is_reiser5_available)) || ((is_btrfs_available)) || ((is_reiserfs_available)) ; then
            answer="$(
            {
                echo -e "ext4\n${msg_fs_ext4}"
                #echo -e "ext2\next2"

                if ((is_reiser4_available)) ; then
                    echo -e "reiser4\n${msg_fs_reiser4}"
                fi
                if ((is_reiser5_available)) ; then
                    echo -e "reiser5\n${msg_fs_reiser5}"
                fi
                if ((is_btrfs_available)) ; then
                    echo -e "btrfs\n${msg_fs_btrfs}"
                fi
                if ((is_reiserfs_available)) ; then
                    echo -e "reiserfs\n${msg_fs_reiserfs}"
                fi

            } | $guitool --title='Elive Installer' --width=720 --height=360 --list \
                --text="$message_select_fs" \
                --column="$( eval_gettext "File system" )" --column="$( eval_gettext "Description" )" 2>/dev/null || echo cancel )"

            if [[ "$answer" = "cancel" ]] || [[ -z "$answer" ]] ; then
                answer="ext4"
            fi

            if [[ "$answer" = "btrfs" ]] || [[ "$answer" = "reiser4" ]] || [[ "$answer" = "reiser5" ]] ; then
                warn_experimental_fs "$answer"
                needs_boot_partition=1

                # always
                # is_fs_compression_wanted=1
                is_fs_compression_used=1
                el_array_member_add "$answer" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")

                # compression?
                if [[ -z "$is_fs_compression_wanted" ]] ; then
                    if guitool yesnoquestion "$msg_fs_supports_compression" ; then
                        is_fs_compression_wanted=1
                    else
                        is_fs_compression_wanted=0
                    fi
                fi
            fi

        else
            # we only have ext4 available
            answer="ext4"
        fi
    fi

    # set it
    HD_PART_ROOT_FS_v="$answer"

    # select default mount options
    case "$HD_PART_ROOT_FS_v" in
        btrfs)
            if ((is_fs_compression_wanted)) ; then
                # Note: to use max compression use instead:  compress-force=zstd:15 , default is 3, it is recommended to not use the -force so it will compress anything (even uncompressable files)
                HD_PART_ROOT_FS_OPTS_v="commit=300,compress=${compressor_algorithm},lazytime"
            else
                HD_PART_ROOT_FS_OPTS_v="commit=300,lazytime"
            fi
            el_array_member_add "$HD_PART_ROOT_FS_v" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")
            ;;
        reiser4)
            # compression is set in format time
            if device_is_ssd "$HD_DEV" ; then
                HD_PART_ROOT_FS_OPTS_v="defaults,relatime,txmod=wa"
            else
                HD_PART_ROOT_FS_OPTS_v="defaults,relatime"
            fi
            el_array_member_add "$HD_PART_ROOT_FS_v" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")
            ;;
        reiser5)
            el_error "reiser5 needs to implemented"
            exit_error
            ;;
        reiserfs)
            HD_PART_ROOT_FS_OPTS_v="defaults,commit=300,relatime,notail"
            is_reiserfs_used=1
            ;;
        ext2)
            HD_PART_ROOT_FS_OPTS_v="defaults,commit=60,relatime"
            ;;
        ext3)
            HD_PART_ROOT_FS_OPTS_v="defaults,commit=300,relatime"
            ;;
        ext4)
            HD_PART_ROOT_FS_OPTS_v="defaults,commit=300,lazytime"
            ;;
        *)
            el_error "unknown FS selected in ${FUNCNAME}: $HD_PART_ROOT_FS_OPTS_v"
            exit_error
            ;;
    esac


    # }}}

    guitool "working_start"

    # small disk setup:
    if [[ "$HD_SIZE_mb" -lt "$autopart_hd_type_size_small" ]] ; then
        is_autopart_hd_small=1
    fi

    # set default sizes
    if ((is_autopart_hd_small)) ; then
        el_explain 0 "HD SMALL: is not very big ($HD_SIZE_mb), so we will use a traditional and simple partitioning instead, RAM is ${RAM_TOTAL_SIZE_mb}"

        if [ "$HD_SIZE_mb" -ge "60000" ]; then
            part_size_swap="$(( ${RAM_TOTAL_SIZE_mb} + 500 ))"
        elif [ "$HD_SIZE_mb" -ge "38000" ]; then
            part_size_swap="$(( ${RAM_TOTAL_SIZE_mb} + 300 ))"
        elif [ "$HD_SIZE_mb" -ge "21500" ]; then
            part_size_swap=1600
        elif [ "$HD_SIZE_mb" -ge "17000" ]; then
            part_size_swap=1024
        elif [ "$HD_SIZE_mb" -ge "13500" ]; then
            part_size_swap=800
        elif [ "$HD_SIZE_mb" -ge "9500" ]; then
            part_size_swap=512
        elif [ "$HD_SIZE_mb" -ge "7500" ]; then
            part_size_swap=450
        elif [ "$HD_SIZE_mb" -ge "5500" ]; then
            part_size_swap=320
        elif [ "$HD_SIZE_mb" -ge "4500" ]; then
            part_size_swap=180
        else
            part_size_swap=no_swap
        fi

        if echo "$part_size_swap" | grep -qs "[[:digit:]]" && [[ "$RAM_TOTAL_SIZE_mb" -gt "$part_size_swap" ]] ; then
            if ! ((is_virtualized)) ; then
                el_debug "not enough swap to hibernate: swap ${part_size_swap}, ram ${RAM_TOTAL_SIZE_mb}, disk ${HD_SIZE_mb}"
            fi
        else
            el_explain 2 "Swap partition estimated size is $part_size_swap MB"
        fi

    else
        # set a swap size
        if [ "$HD_SIZE_mb" -lt "230000" ]; then
            part_size_swap="$(( ${RAM_TOTAL_SIZE_mb} + 300 ))"
        else
            part_size_swap="$(( ${RAM_TOTAL_SIZE_mb} + 800 ))"
        fi
        # GPT needed because disk is big {{{
        if [[ "$HD_SIZE_mb" -gt "1400000" ]] ; then
            if ! ((is_efi_booted)) && ((is_efi_compatible)) ; then
                if guitool question "$( eval_gettext "Your disk is almost 2TB. It is recommended to switch your BIOS to UEFI mode for better compatibility. Do you want to stop the installer now and change this setting in your BIOS?" )" ; then
                    guitool info "$( eval_gettext "Installation halted. Reboot your computer and set your BIOS to boot in UEFI mode. Disable SecureBoot. Refer to online guides for accessing BIOS settings." )"
                    exit
                else
                    is_gpt_needed=1
                fi
            fi
        fi

        # - GPT needed (to implement) }}}
    fi

    # XXX force to always use the simple partitioning mode since having a /home is somewhat useless:
    is_autopart_hd_small=1


    # boot partition sizes
    if [ "$HD_SIZE_mb" -ge "620000" ]; then
        # bigger than 250 GB disks
        part_size_boot=450
    elif [ "$HD_SIZE_mb" -ge "70000" ]; then
        part_size_boot=400
    else
        part_size_boot=300
    fi
    # EFI partition sizes
    if [ "$HD_SIZE_mb" -ge "620000" ]; then
        # bigger than 250 GB disks
        part_size_efi=80
    elif [ "$HD_SIZE_mb" -ge "70000" ]; then
        part_size_efi=60
    else
        part_size_efi=40
    fi
    # root partition sizes
    # real size used for Elive is actually 13 GB, and 4.5 GB in reiser4+lzo (2.6.3)
    if ((is_fs_compression_used)) ; then
        if [ "$HD_SIZE_mb" -ge "280000" ]; then
            # bigger than 250 GB disks
            part_size_root=28000
        elif [ "$HD_SIZE_mb" -ge "220000" ]; then
            # like 250 GB ssd's
            part_size_root=19000
        elif [ "$HD_SIZE_mb" -ge "140000" ]; then
            # like 160 GB nettops
            part_size_root=14000
        elif [ "$HD_SIZE_mb" -ge "75000" ]; then
            # very rare cases, 80 GB
            part_size_root=12000
        else
            part_size_root=8000
        fi
    else
        if [ "$HD_SIZE_mb" -ge "280000" ]; then
            # bigger than 250 GB disks
            part_size_root=50000
        elif [ "$HD_SIZE_mb" -ge "220000" ]; then
            # like 250 GB ssd's
            part_size_root=35000
        elif [ "$HD_SIZE_mb" -ge "140000" ]; then
            # like 160 GB nettops
            part_size_root=28000
        elif [ "$HD_SIZE_mb" -ge "75000" ]; then
            # very rare cases, 80 GB
            part_size_root=24000
        else
            part_size_root=16000
        fi
    fi


    # ask for encryption
    # TODO: add this feature as a checkbox (must the entire dialog be rewrited):
    # https://forum.elivelinux.org/t/elive-3-0-0-installation-feedback/551 -  ^ Here, the No/Yes buttons can easily do something that the user doesn’t want, if they click in a hurry. Considering the encryption is something the user should really take time to read about, it would be better if you presented a window with a checkbox and a “Continue” button that works even without checking the box, but only enables encryption if you check the box.
    # in my vbox tests, 64bit build requires ram on 700 in order to make it working, 600 is not enough
    if [[ "${RAM_TOTAL_SIZE_mb}" -lt "800" ]] && ! swap_partitions_all_do "check" ; then
        unset is_encryption_wanted
    else
        if guitool yesnoquestion "$( eval_gettext "Would you like to enable encryption? Your hard disk will be secured with a password, ensuring that no one can access your data without it. You will need to enter this password every time the system boots. IF YOU FORGET YOUR PASSWORD, YOUR DATA WILL BE INACCESSIBLE." )" ; then
            is_encryption_wanted=1
            needs_boot_partition=1
            is_fs_mapper=1
            # warn the user that extra ram is needed
            #if ! LC_ALL=C swapon --noheadings --raw | grep -qs "^/" ; then
                #if [[ "$RAM_TOTAL_SIZE_mb" -lt "950" ]] ; then
                    #guitool warning "$( eval_gettext "Sometimes adding encryption can fail in computers with low amount of RAM." )"
                #fi
                #if [[ "$RAM_TOTAL_SIZE_mb" -lt "1400" ]] ; then
                    #guitool warning "$( eval_gettext "Your computer doesn't has much RAM memory. Enabling encryption is not recommended with these low resources. But if you want to use it we recommend to add and enable a SWAP partition before to continue in order to avoid issues." )"
                #fi
            #fi
            el_explain 2 "Encryption wanted: __YES__"
        else
            el_explain 2 "Encryption wanted: __NO__"
        fi
    fi


    # ask if wants swap
    if [[ "$HD_SIZE_mb" -lt "40000" ]] || [[ "${RAM_TOTAL_SIZE_mb}" -lt "5700" ]] || ! device_is_ssd "$HD_DEV" ; then
        local message_ask_for_swap
        message_ask_for_swap="$( printf "$( eval_gettext "Would you like to create a SWAP partition? While not strictly necessary, it is particularly useful if you have limited RAM, especially when it becomes full. SWAP is also required for hibernation. The optimal size for your case is %s MB." )" "$(( $RAM_TOTAL_SIZE_mb + 800 ))" )"

        if ! guitool yesnoquestion "$message_ask_for_swap" ; then
            part_size_swap="no_swap"
        fi
    fi


    # - calculate automated sizes and configurations }}}

    # betatest in vbox: add swap in low-sized disks:
    #_ask_for_swap=1
    #part_size_swap=700

    # create structure of disk {{{

    # show progress (after to request instmod)
    { ( while test -f "$TMP_PROGRESS_CONFIGURING_f" ; do cat "$TMP_PROGRESS_CONFIGURING_f" || true ; LC_ALL="$EL_LC_EN" sleep 0.6 ; done | $guitool --title='Elive Installer' --progress --text="$( eval_gettext "Partitioning your hard disk..." )" --pulsate --auto-close ) & disown ; } 2>/dev/null

    # Create DOS table
    # We will not use gpt yet, will implement after wheezy (and bigger compatibility, integration, bugs fixed, etc too)
    #parted -ms -a optimal "$HD_DEV" mklabel gpt
    device_check_is_busy "$HD_DEV"

    if ((is_efi_booted)) || ((is_amd64)) || ((is_gpt_needed)) ; then
        parted -ms -a optimal "$HD_DEV" mklabel gpt  2>"$TMP_LOGS_ERR_f" || partitioner_failed
        el_debug "parted -ms -a optimal "$HD_DEV" mklabel gpt"
        el_debug "Created partition table in GPT with HD $HD_DEV aligned correctly"

        is_disk_table_gpt=1
        unset is_disk_table_msdos
    else
        parted -ms -a optimal "$HD_DEV" mklabel msdos  2>"$TMP_LOGS_ERR_f" || partitioner_failed
        el_debug "parted -ms -a optimal "$HD_DEV" mklabel msdos"
        el_debug "Created partition table in MSDOS (legacy boot --> MBR) with HD $HD_DEV aligned correctly"
        is_disk_table_msdos=1
        unset is_disk_table_gpt
    fi
    sync ; sleep 2

    device_check_is_busy "$HD_DEV"

    echo 40 > "$TMP_PROGRESS_CONFIGURING_f"

    # decide and set a partitioning structure
    p_num=1
    if [[ "$HD_DEV" = "/dev/nvme"* ]] || [[ "$HD_DEV" = "/dev/mmcblk"* ]] || [[ "$HD_DEV" = "/dev/md"*p* ]] ; then
        el_debug "Device $HD_DEV is SSD over PCIe (nvme) or eMMC, or md"
        partition_number_prefix="p"
    else
        partition_number_prefix=""
    fi

    # 8MiB is the start needed for alignment
    p_last_size=32
    p_size_mb=0  # set a default value just to avoid bugs

    # create an EFI / BIOS partition first if needed {{{
    # create an EFI partition when we boot on UEFI, always
    if ((is_efi_booted)) ; then
        efi_detect_partition
        # do we have an efi partition?
        if [[ -n "${efi_partitions_maybe[*]}" ]] ; then
            if [[ "$HD_DISKS_NUMBER_v" -eq 1 ]] ; then
                # select the already-existing EFI partition for our bootloader
                efi_select
            else
                if guitool yesnoquestion "$( eval_gettext "Would you like to use the existing EFI partition and add our boot entry? You should do this if you currently boot your computer from that disk. Select NO if you wish to boot from this disk independently or if you do not want to modify the existing boot loader." )" ; then
                    # use the actual EFI partition for our bootloader
                    efi_select
                fi
            fi
        fi
        # create an efi partition if we don't have one
        if [[ ! -b "$HD_PART_EFI_dev" ]] ; then
            el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary fat32 ${p_last_size}MiB $(( ${part_size_efi} + ${p_last_size} ))MiB"
            parted -ms -a optimal "$HD_DEV" mkpart primary fat32 ${p_last_size}MiB $(( ${part_size_efi} + ${p_last_size} ))MiB  2>"$TMP_LOGS_ERR_f" || partitioner_failed
            partition_format "${HD_DEV}${partition_number_prefix}${p_num}" EFI "EFI"
            # set variables
            sync ; sleep 2
            hd_refresh "$HD_DEV"
            if hd_device_is_blockdevice "${HD_DEV}${partition_number_prefix}${p_num}" ; then
                if fdisk -l "${HD_DEV}" | grep -qs "${HD_DEV}${partition_number_prefix}${p_num}.*EFI " || EL_DEBUG=0 partitions-list --show-only="${HD_DEV}${partition_number_prefix}${p_num}" | grep -qs "EFI" ; then
                    HD_PART_EFI_dev="${HD_DEV}${partition_number_prefix}${p_num}"
                    HD_PART_EFI_FS_v="vfat"
                fi
            fi

            # update values
            device_check_is_busy "$HD_DEV"
            el_explain 2 "Created partition __${p_num}__ for __EFI__ with size __${part_size_efi} MB__ in fat32"

            p_num="$(( $p_num + 1 ))"
            p_last_size="$(( $p_last_size + $part_size_efi ))"
        fi

        is_disk_table_gpt=1
        unset is_disk_table_msdos

    else
        # we need a bios partition (no-efi + big disk or similar)
        if ((is_gpt_needed)) || ((is_disk_table_gpt)) ; then
            # note: p_last_size is set to 8, so it will format "from 8mb to (8+8)mb", and its correct
            el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary fat32 ${p_last_size}MiB $(( ${p_last_size} + ${p_last_size}))MiB"
            parted -ms -a optimal "$HD_DEV" mkpart primary fat32 ${p_last_size}MiB $(( ${p_last_size} + ${p_last_size}))MiB  2>"$TMP_LOGS_ERR_f" || partitioner_failed
            partition_format "${HD_DEV}${partition_number_prefix}${p_num}" BIOS "BIOS"
            # update values
            device_check_is_busy "$HD_DEV"
            el_explain 2 "Created partition __${p_num}__ for __BIOS__ with size __8 MB__"

            p_num="$(( $p_num + 1 ))"
            p_last_size="$(( $p_last_size + ${p_last_size} ))"
        fi
    fi
    sync ; sleep 2

    # efi partitions can allow to have encryptied root, but not in non-efi mode, so require it in these cases:
    if ((is_encryption_wanted)) && ! ((is_efi_booted)) ; then
        needs_boot_partition=1
    fi

    # }}}
    if ((is_autopart_hd_small)) ; then
        # SMALL partitioning:
        if ((is_encryption_wanted)) ; then
            el_debug "PARTITIONING STRUCTURE: 1 [/boot?] + 5 extended ([LUKS + LVM]  6 [swap?], 7 root ))"
            # boot {{{
            # we wanted to use reiser4 for all the system so we need a /boot
            # Update: do not use JFS since sometimes fails to mount at boot (like unclean FS), this is a major bug
            # Update: do not use reiserfs too, grub installing fails for not be a stable option for /boot, use ext3/4 which is also smaller
            if ((needs_boot_partition)) ; then
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_boot} + ${p_last_size} ))MiB"
                parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_boot} + ${p_last_size} ))MiB  2>"$TMP_LOGS_ERR_f" || partitioner_failed
                if ((is_disk_table_msdos)) ; then
                    el_debug "parted -ms -a optimal "$HD_DEV" set "${p_num}" boot on"
                    parted -ms -a optimal "$HD_DEV" set "${p_num}" boot on || partitioner_failed
                fi

                # format and configure
                partition_format "${HD_DEV}${partition_number_prefix}${p_num}" ext4 "Elive_boot"
                HD_PART_BOOT_dev="${HD_DEV}${partition_number_prefix}${p_num}"
                HD_PART_BOOT_fs="ext4"
                # fstab will be added in pre_copying
                needs_boot_partition=1

                # update values
                device_check_is_busy "$HD_DEV"
                el_explain 2 "Created partition __${p_num}__ for __/boot__ with size __${part_size_boot} MB__ in ext4"

                p_num="$(( $p_num + 1 ))"
                p_last_size="$(( $p_last_size + $part_size_boot ))"
            fi

            # }}}
            # extended partition layer {{{
            if ((is_disk_table_msdos)) ; then
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart extended ${p_last_size}MiB -> 100%"
                parted -ms -a optimal "$HD_DEV" mkpart extended ${p_last_size}MiB 100% || partitioner_failed

                # remove lba to change ID from "f" to "5", just because this is the type of extended id that default's debian uses in their setup, and also looks like to show sda2 as a "wrong disk" if is set
                el_debug "parted -ms -a optimal "$HD_DEV" set "${p_num}" lba off"
                parted -ms -a optimal "$HD_DEV" set "${p_num}" lba off || partitioner_failed

                # another shift of 8 MiB is needed for correct alignment and no error showing up
                p_last_size="$(( $p_last_size + 8 ))"

                # create the logical layer
                p_num="5" # now the number is 5 for the logical partition and so luks one, always

                el_debug "parted -ms -a optimal "$HD_DEV" mkpart logical ${p_last_size}MiB -> 100%"
                parted -ms -a optimal "$HD_DEV" mkpart logical ${p_last_size}MiB 100% || partitioner_failed
            fi

            if ((is_disk_table_gpt)) ; then
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary ${p_last_size}MiB -> 100%"
                parted -ms -a optimal "$HD_DEV" mkpart primary ${p_last_size}MiB 100% || partitioner_failed
            fi

            # - extended partition layer }}}
            # LUKS layer {{{
            # make sure that we have the computer's (generic) password
            password_root_ask

            # default to all formats to 256 which is faster than the default of 512
            luksformat_options="-s 256"

            if [[ "$( el_ram_usage_show "total" )" -lt 2400 ]] ; then
                luksformat_options="--pbkdf pbkdf2 $luksformat_options"
            fi

            # this defaults to aes-xts-plain(64?) with sha 256, which is the best option and for performance
            if ! printf "%s" "$rootpass" | cryptsetup $luksformat_options --use-urandom --batch-mode luksFormat "${HD_DEV}${partition_number_prefix}${p_num}" ; then
                el_error "error creating luks layer with opts '${luksformat_options}':\n$( printf "%s" "$rootpass" | cryptsetup $luksformat_options --use-urandom --batch-mode luksFormat "${HD_DEV}${partition_number_prefix}${p_num}" 2>&1 )"

                # format again with fallback options
                printf "%s" "$rootpass" | cryptsetup -c aes-xts-plain -s 256 --use-urandom --type luks1 --batch-mode luksFormat "${HD_DEV}${partition_number_prefix}${p_num}"

            fi

            sync

            # the name structure should match everywhere in the installer, using like debian does: sda5_crypt
            hd_part_luks_name="${HD_DEV##*/}${partition_number_prefix}${p_num}_crypt"

            if ! printf "%s" "$rootpass" | cryptsetup luksOpen "${HD_DEV}${partition_number_prefix}${p_num}" "$hd_part_luks_name" ; then
                swap_partitions_all_do "activate"
                sync ; sleep 2

                if ! printf "%s" "$rootpass" | cryptsetup luksOpen "${HD_DEV}${partition_number_prefix}${p_num}" "$hd_part_luks_name" ; then
                    low_memory_improve

                    if ! printf "%s" "$rootpass" | cryptsetup luksOpen "${HD_DEV}${partition_number_prefix}${p_num}" "$hd_part_luks_name" ; then
                        if dmesg | grep -aiqs "out of memory.*cryptsetup" ; then
                            guitool error "$( eval_gettext "Unable to encrypt your hard disk due to a lack of RAM. To solve this, first create a swap partition and activate it, then manually partition the disk to add an encrypted partition. Alternatively, do not use encryption in the automatic partitioner. Aborting installation..." )"
                            el_error "Unable to encrypt partition (out of memory)"
                            exit_error
                        else
                            guitool error "$( eval_gettext "Unable to encrypt your hard disk using this password. Exiting..." )"
                            el_error "Unable to encrypt partition with the provided password '$rootpass'"
                            exit_error
                        fi
                    fi
                fi

                swap_partitions_all_do "deactivate"
            fi

            # note: discard is not needed if we use fstrim
            if device_is_ssd "$HD_DEV" ; then
                el_array_member_add "${hd_part_luks_name}    UUID=$( LC_ALL="$EL_LC_EN" cryptsetup luksDump "${HD_DEV}${partition_number_prefix}${p_num}" | grep "^UUID:" | head -1 | awk '{print $2}' )    none    luks,discard" "${hd_dev_to_add_in_crypttab[@]}" ; hd_dev_to_add_in_crypttab=("${_out[@]}")
            else
                el_array_member_add "${hd_part_luks_name}    UUID=$( LC_ALL="$EL_LC_EN" cryptsetup luksDump "${HD_DEV}${partition_number_prefix}${p_num}" | grep "^UUID:" | head -1 | awk '{print $2}' )    none    luks" "${hd_dev_to_add_in_crypttab[@]}" ; hd_dev_to_add_in_crypttab=("${_out[@]}")
            fi

            el_debug "Encrypted partition added to crypttab list: '${HD_DEV}${partition_number_prefix}${p_num}' "


            sync

            # - LUKS layer }}}
            # LVM2 layer {{{
            # create the LVM physical volume from the luks layer
            pvcreate "/dev/mapper/$hd_part_luks_name"
            # create the virtual group from the phisical device
            vgcreate elive_vg "/dev/mapper/$hd_part_luks_name"

            # create logical volumes now

            # swap volume
            el_check_variables "part_size_swap"

            if [[ "$part_size_swap" != "no_swap" ]] && echo "$part_size_swap" | grep -qs "[[:digit:]]" ; then
                lvcreate -L ${part_size_swap}M -n swap_lv elive_vg
                el_debug "LV Partition Swap: ${part_size_swap} MB"
            fi

            # root volume
            el_check_variables "HD_SIZE_mb"

            #lvcreate -L ${p_size_mb}M -n root_lv elive_vg
            lvcreate -l 100%FREE -n root_lv elive_vg
            el_debug "LV Partition Root: 100%FREE"

            sync
            if ! [[ -b "/dev/mapper/elive_vg-root_lv" ]] ; then
                guitool error "$( eval_gettext "Failed to create an LVM2 partition. This could be due to insufficient free space or partitions already in use. We recommend manually removing them or selecting the automated partitioning option." )"
                el_error "Unable to create LVM2 partition, not enough space?"
                exit_error
            fi

            # home volume
            #lvcreate -l 100%FREE -n home_lv elive_vg
            #el_debug "LV Partition Swap: ${part_size_swap} MB"

            is_lvm2_configured=1

            # - LVM2 layer }}}

            # lvm Swap {{{
            # format and configure
            sync
            if [[ -b "/dev/mapper/elive_vg-swap_lv" ]] ; then
                swapoff "/dev/mapper/elive_vg-swap_lv" 2>/dev/null || true

                el_debug "Formating swap partition"
                partition_format "/dev/mapper/elive_vg-swap_lv" swap swap

                if [[ "$RAM_TOTAL_SIZE_mb" -lt 1400 ]] ; then
                    swapon "/dev/mapper/elive_vg-swap_lv"  2>/dev/null 1>&2 || true
                fi

                # update values
                el_explain 2 "Created LVM2 partition for __swap__ with size __${part_size_swap} MB__"
            fi


            # }}}
            # lvm root {{{
            # format and configure

            partition_format "/dev/mapper/elive_vg-root_lv" "$HD_PART_ROOT_FS_v" "Elive_OS"

            # fstab entries is not needed because it is added in post-install (and only the root one)
            HD_PART_ROOT_dev="/dev/mapper/elive_vg-root_lv"

            # update values
            el_explain 2 "Created LVM2 partition for __/__ with the rest of free space"
            partition_alignment_check "/dev/mapper/elive_vg-root_lv" report

            # }}}

        else
            el_debug "PARTITIONING STRUCTURE: [1 swap?] + [2 boot?] + 1-2-3 root"
            # Swap {{{
            if [[ "$part_size_swap" != "no_swap" ]] && echo "$part_size_swap" | grep -qs "[[:digit:]]" ; then
                # create swap partition from 8mb to disk-position (it doesn't append the size)
                p_size_mb="$part_size_swap"
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary linux-swap ${p_last_size}MiB $(( ${p_size_mb} + ${p_last_size} ))MiB"
                parted -ms -a optimal "$HD_DEV" mkpart primary linux-swap ${p_last_size}MiB $(( ${p_size_mb} + ${p_last_size} ))MiB || partitioner_failed

                # format and configure
                swapoff "${HD_DEV}${partition_number_prefix}${p_num}" 2>/dev/null || true

                partition_format "${HD_DEV}${partition_number_prefix}${p_num}" swap swap
                el_debug "Formated swap partition"

                if [[ "$RAM_TOTAL_SIZE_mb" -lt 1400 ]] ; then
                    swapon "${HD_DEV}${partition_number_prefix}${p_num}"  2>/dev/null 1>&2 || true
                fi


                # update values
                device_check_is_busy "$HD_DEV"
                el_explain 2 "Created partition __${p_num}__ for __swap__ with size __${p_size_mb} MB__"
                p_num="$(( $p_num + 1 ))"
                p_last_size="$(( $p_last_size + $p_size_mb ))"
            fi

            # }}}
            # boot {{{
            if ((needs_boot_partition)) ; then
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_boot} + ${p_last_size} ))MiB"
                parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_boot} + ${p_last_size} ))MiB || partitioner_failed
                if ((is_disk_table_msdos)) ; then
                    el_debug "parted -ms -a optimal "$HD_DEV" set "${p_num}" boot on"
                    parted -ms -a optimal "$HD_DEV" set "${p_num}" boot on || partitioner_failed
                fi

                # format and configure
                partition_format "${HD_DEV}${partition_number_prefix}${p_num}" ext4 "Elive_boot"
                HD_PART_BOOT_dev="${HD_DEV}${partition_number_prefix}${p_num}"
                HD_PART_BOOT_fs="ext4"
                # fstab will be added in pre_copying

                # update values
                device_check_is_busy "$HD_DEV"
                el_explain 2 "Created partition __${p_num}__ for __/boot__ with size __${part_size_boot} MB__ in ext4"

                p_num="$(( $p_num + 1 ))"
                p_last_size="$(( $p_last_size + $part_size_boot ))"
            fi

            # }}}
            # root {{{
            #p_size_mb=140
            el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB -> 100%"
            parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB 100% || partitioner_failed

            # format and configure
            partition_format "${HD_DEV}${partition_number_prefix}${p_num}" "$HD_PART_ROOT_FS_v" "Elive_OS"

            # fstab entries is not needed because it is added in post-install (and only the root one)
            HD_PART_ROOT_dev="${HD_DEV}${partition_number_prefix}${p_num}"

            # update values
            #device_check_is_busy "$HD_DEV"
            el_explain 2 "Created partition __${p_num}__ for __/__ with the rest of the disk in __${HD_PART_ROOT_FS_v}__ ($HD_PART_ROOT_FS_OPTS_v)"
            partition_alignment_check "$HD_PART_ROOT_dev" report
            p_num="$(( $p_num + 1 ))"
            p_last_size="$(( $p_last_size + $p_size_mb ))"

            # }}}
        fi
    else
        if ((is_encryption_wanted)) ; then
            el_debug "PARTITIONING STRUCTURE: 1 [/boot?] + 5 extended ([LUKS + LVM]  6 [swap?], 7 root, 8 home))"
            # boot {{{
            # we wanted to use reiser4 for all the system so we need a /boot
            if ((needs_boot_partition)) ; then
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_boot} + ${p_last_size} ))MiB"
                parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_boot} + ${p_last_size} ))MiB || partitioner_failed
                if ((is_disk_table_msdos)) ; then
                    el_debug "parted -ms -a optimal "$HD_DEV" set "${p_num}" boot on"
                    parted -ms -a optimal "$HD_DEV" set "${p_num}" boot on || partitioner_failed
                fi

                # format and configure
                partition_format "${HD_DEV}${partition_number_prefix}${p_num}" ext4 "Elive_boot"
                HD_PART_BOOT_dev="${HD_DEV}${partition_number_prefix}${p_num}"
                HD_PART_BOOT_fs="ext4"
                # fstab will be added in pre_copying
                needs_boot_partition=1

                # update values
                #device_check_is_busy "$HD_DEV"
                el_explain 2 "Created partition __${p_num}__ for __/boot__ with size __${part_size_boot} MB__ in ext4"

                p_num="$(( $p_num + 1 ))"
                p_last_size="$(( $p_last_size + $part_size_boot ))"
            fi

            # }}}
            # extended partition layer {{{
            if ((is_disk_table_msdos)) ; then
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart extended ${p_last_size}MiB -> 100%"
                parted -ms -a optimal "$HD_DEV" mkpart extended ${p_last_size}MiB 100% || partitioner_failed

                # remove lba to change ID from "f" to "5", just because this is the type of extended id that default's debian uses in their setup
                if ((is_disk_table_msdos)) ; then
                    el_debug "parted -ms -a optimal "$HD_DEV" set "${p_num}" lba off"
                    parted -ms -a optimal "$HD_DEV" set "${p_num}" lba off || partitioner_failed
                fi

                # another shift of 8 MiB is needed for correct alignment and no error showing up
                p_last_size="$(( $p_last_size + 8 ))"

                # create the logical layer
                p_num="5" # now the number is 5 for the logical partition and so luks one, always
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart logical ${p_last_size}MiB -> 100%"
                parted -ms -a optimal "$HD_DEV" mkpart logical ${p_last_size}MiB 100% || partitioner_failed
            fi

            if ((is_disk_table_gpt)) ; then
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary ${p_last_size}MiB -> 100%"
                parted -ms -a optimal "$HD_DEV" mkpart primary ${p_last_size}MiB 100% || partitioner_failed
            fi


            # - extended partition layer }}}
            # LUKS layer {{{
            # make sure that we have the computer's (generic) password
            password_root_ask

            # update: not needed if we already use the -s 256 option
            #if [[ "${CPU_MHZ_CORE_v}" -lt "1300" ]] && [[ -n "$CPU_MHZ_CORE_v" ]] && [[ "$( el_ram_usage_show "total" )" -lt 2200 ]] ; then
                #luksformat_options="-c aes-xts-plain -s 256 $luksformat_options"
            #else
                ##uksformat_options="-s 512 $luksformat_options"
                ## always use 256 for better performance over 512
                ## FIXME: debug if format failed with el_error and show resulting example, then fallback to a predefined format to always give it correctly
                luksformat_options="-s 256"
            #fi

            # this option is needed to fix the bug about "Killed out of memory" problem (machines with less than 1GB ram)
            if [[ "$( el_ram_usage_show "total" )" -lt 2400 ]] ; then
                luksformat_options="--pbkdf pbkdf2 $luksformat_options"
            fi

            if ! printf "%s" "$rootpass" | cryptsetup $luksformat_options --use-urandom --batch-mode luksFormat "${HD_DEV}${partition_number_prefix}${p_num}" ; then
                el_error "error creating luks layer with opts '${luksformat_options}':\n$( printf "%s" "$rootpass" | cryptsetup $luksformat_options --use-urandom --batch-mode luksFormat "${HD_DEV}${partition_number_prefix}${p_num}" 2>&1 )"

                # format again with fallback options
                printf "%s" "$rootpass" | cryptsetup -c aes-xts-plain -s 256 --use-urandom --type luks1 --batch-mode luksFormat "${HD_DEV}${partition_number_prefix}${p_num}"

            fi


            sync

            # the name structure should match everywhere in the installer, using like debian does: sda5_crypt
            hd_part_luks_name="${HD_DEV##*/}${partition_number_prefix}${p_num}_crypt"

            if ! printf "%s" "$rootpass" | cryptsetup luksOpen "${HD_DEV}${partition_number_prefix}${p_num}" "$hd_part_luks_name" ; then
                swap_partitions_all_do "activate"
                sync ; sleep 2

                if ! printf "%s" "$rootpass" | cryptsetup luksOpen "${HD_DEV}${partition_number_prefix}${p_num}" "$hd_part_luks_name" ; then
                    sync ; sleep 2
                    if ! printf "%s" "$rootpass" | cryptsetup luksOpen "${HD_DEV}${partition_number_prefix}${p_num}" "$hd_part_luks_name" ; then
                        guitool error "$( eval_gettext "Unable to encrypt your hard disk using this password. Exiting..." )"
                        el_error "Unable to encrypt partition with the provided password '$rootpass'"
                        exit_error
                    fi
                fi

                swap_partitions_all_do "deactivate"
            fi

            if device_is_ssd "$HD_DEV" ; then
                el_array_member_add "${hd_part_luks_name}    UUID=$( LC_ALL="$EL_LC_EN" cryptsetup luksDump "${HD_DEV}${partition_number_prefix}${p_num}" | grep "^UUID:" | head -1 | awk '{print $2}' )    none    luks,discard" "${hd_dev_to_add_in_crypttab[@]}" ; hd_dev_to_add_in_crypttab=("${_out[@]}")
            else
                el_array_member_add "${hd_part_luks_name}    UUID=$( LC_ALL="$EL_LC_EN" cryptsetup luksDump "${HD_DEV}${partition_number_prefix}${p_num}" | grep "^UUID:" | head -1 | awk '{print $2}' )    none    luks" "${hd_dev_to_add_in_crypttab[@]}" ; hd_dev_to_add_in_crypttab=("${_out[@]}")
            fi

            el_debug "Encrypted partition added to crypttab list: '${HD_DEV}${partition_number_prefix}${p_num}' "

            sync

            # - LUKS layer }}}
            # LVM2 layer {{{
            # create the LVM physical volume from the luks layer
            pvcreate "/dev/mapper/$hd_part_luks_name"
            # create the virtual group from the phisical device
            vgcreate elive_vg "/dev/mapper/$hd_part_luks_name"

            # create logical volumes now

            # swap volume
            el_check_variables "part_size_swap"

            if [[ "$part_size_swap" != "no_swap" ]] && echo "$part_size_swap" | grep -qs "[[:digit:]]" ; then
                lvcreate -L ${part_size_swap}M -n swap_lv elive_vg
                el_debug "LV Partition Swap: ${part_size_swap} MB"
            fi

            el_check_variables "HD_SIZE_mb"

            # root volume
            lvcreate -L ${part_size_root}M -n root_lv elive_vg
            el_debug "LV Partition Root: ${part_size_root} MB"

            sync
            if ! [[ -b "/dev/mapper/elive_vg-root_lv" ]] ; then
                el_error "Unable to create LVM2 partition, not enough space?"
                exit_error
            fi

            # home volume
            lvcreate -l 100%FREE -n home_lv elive_vg
            el_debug "LV Partition Home: 100%FREE"

            is_lvm2_configured=1

            # - LVM2 layer }}}

            # lvm Swap {{{
            # format and configure
            sync
            if [[ -b "/dev/mapper/elive_vg-swap_lv" ]] ; then
                swapoff "/dev/mapper/elive_vg-swap_lv" 2>/dev/null || true

                partition_format "/dev/mapper/elive_vg-swap_lv" swap swap
                el_debug "Formated swap partition"

                if [[ "$RAM_TOTAL_SIZE_mb" -lt 1400 ]] ; then
                    swapon "/dev/mapper/elive_vg-swap_lv"  2>/dev/null 1>&2 || true
                fi


                # update values
                el_explain 2 "Created LVM2 partition for __swap__ with size __${part_size_swap} MB__"
            fi


            # }}}
            # lvm root {{{
            # format and configure
            partition_format "/dev/mapper/elive_vg-root_lv" "$HD_PART_ROOT_FS_v" "Elive_OS"

            # fstab entries is not needed because it is added in post-install (and only the root one)
            HD_PART_ROOT_dev="/dev/mapper/elive_vg-root_lv"

            # update values
            el_explain 2 "Created LVM2 partition for __/__ with the rest of free space"
            partition_alignment_check "/dev/mapper/elive_vg-root_lv" report

            # }}}
            # lvm home {{{
            # format and configure
            partition_format "/dev/mapper/elive_vg-home_lv" "$HD_PART_ROOT_FS_v" "Elive_home"
            HD_PART_HOME_FS_v="$HD_PART_ROOT_FS_v"
            HD_PART_HOME_FS_OPTS_v="$HD_PART_ROOT_FS_OPTS_v"

            # fstab entries
            HD_PART_HOME_dev="/dev/mapper/elive_vg-home_lv"
            echo -e "$( hd_uuid_encode_fstab "${HD_PART_HOME_dev}" )  /home    ${HD_PART_HOME_FS_v}    ${HD_PART_HOME_FS_OPTS_v}    0  2" >> "$TMP_FSTAB_EXTRA_f"
            el_debug "$( hd_uuid_encode_fstab "${HD_PART_HOME_dev}" )  /home    ${HD_PART_HOME_FS_v}    ${HD_PART_HOME_FS_OPTS_v}    0  2"

            # update values
            el_explain 2 "Created LVM2 partition for __/home__ with the rest of the disk"
            partition_alignment_check "$HD_PART_HOME_dev" report

            # }}}

        else
            el_debug "PARTITIONING STRUCTURE: 1 [swap?] + [2 boot?]  2-3 root + 3-4 home]"
            # Swap {{{
            if [[ "$part_size_swap" != "no_swap" ]] && echo "$part_size_swap" | grep -qs "[[:digit:]]" ; then
                # create swap partition from 8mb to disk-position (it doesn't append the size)
                p_size_mb="$part_size_swap"
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary linux-swap ${p_last_size}MiB $(( ${p_size_mb} + ${p_last_size} ))MiB"
                parted -ms -a optimal "$HD_DEV" mkpart primary linux-swap ${p_last_size}MiB $(( ${p_size_mb} + ${p_last_size} ))MiB || partitioner_failed

                # format and configure
                swapoff "${HD_DEV}${partition_number_prefix}${p_num}" 2>/dev/null || true

                partition_format "${HD_DEV}${partition_number_prefix}${p_num}" swap swap
                el_debug "Formated swap partition"

                if [[ "$RAM_TOTAL_SIZE_mb" -lt 1400 ]] ; then
                    swapon "${HD_DEV}${partition_number_prefix}${p_num}"  2>/dev/null 1>&2 || true
                fi


                # update values
                device_check_is_busy "$HD_DEV"
                el_explain 2 "Created partition __${p_num}__ for __swap__ with size __${p_size_mb} MB__"
                p_num="$(( $p_num + 1 ))"
                p_last_size="$(( $p_last_size + $p_size_mb ))"
            fi

            # }}}
            # boot {{{
            if ((needs_boot_partition)) ; then
                # we wanted to use reiser4 for all the system so we need a /boot
                el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_boot} + ${p_last_size} ))MiB"
                parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_boot} + ${p_last_size} ))MiB || partitioner_failed
                if ((is_disk_table_msdos)) ; then
                    el_debug "parted -ms -a optimal "$HD_DEV" set "${p_num}" boot on"
                    parted -ms -a optimal "$HD_DEV" set "${p_num}" boot on || partitioner_failed
                fi

                # format and configure
                partition_format "${HD_DEV}${partition_number_prefix}${p_num}" ext4 "Elive_boot"
                HD_PART_BOOT_dev="${HD_DEV}${partition_number_prefix}${p_num}"
                HD_PART_BOOT_fs="ext4"
                # fstab will be added in pre_copying
                needs_boot_partition=1

                # update values
                device_check_is_busy "$HD_DEV"
                el_explain 2 "Created partition __${p_num}__ for __/boot__ with size __${part_size_boot} MB__ in ext4"

                p_num="$(( $p_num + 1 ))"
                p_last_size="$(( $p_last_size + $part_size_boot ))"
            fi

            # }}}
            # root {{{
            el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_root} + ${p_last_size} ))MiB"
            parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB $(( ${part_size_root} + ${p_last_size} ))MiB || partitioner_failed

            # format and configure
            partition_format "${HD_DEV}${partition_number_prefix}${p_num}" "${HD_PART_ROOT_FS_v}" "Elive_OS"

            # fstab entries is not needed because it is added in post-install (and only the root one)
            HD_PART_ROOT_dev="${HD_DEV}${partition_number_prefix}${p_num}"

            # update values
            device_check_is_busy "$HD_DEV"
            el_explain 2 "Created partition __${p_num}__ for __/__ with size __${part_size_root} MB__"
            partition_alignment_check "$HD_PART_ROOT_dev" report
            p_num="$(( $p_num + 1 ))"
            p_last_size="$(( $p_last_size + $part_size_root ))"

            # }}}
            # home {{{
            #p_size_mb=40000
            el_debug "parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB -> 100%"
            parted -ms -a optimal "$HD_DEV" mkpart primary reiserfs ${p_last_size}MiB 100% || partitioner_failed

            # format and configure
            partition_format "${HD_DEV}${partition_number_prefix}${p_num}" ${HD_PART_ROOT_FS_v} "Elive_home"
            HD_PART_HOME_FS_v="$HD_PART_ROOT_FS_v"
            HD_PART_HOME_FS_OPTS_v="$HD_PART_ROOT_FS_OPTS_v"

            # fstab entries
            HD_PART_HOME_dev="${HD_DEV}${partition_number_prefix}${p_num}"
            echo -e "$( hd_uuid_encode_fstab "${HD_PART_HOME_dev}" )  /home    ${HD_PART_HOME_FS_v}    ${HD_PART_HOME_FS_OPTS_v}   0  2" >> "$TMP_FSTAB_EXTRA_f"
            el_debug "$( hd_uuid_encode_fstab "${HD_PART_HOME_dev}" )  /home    ${HD_PART_HOME_FS_v}    ${HD_PART_HOME_FS_OPTS_v}   0  2"

            # update values
            device_check_is_busy "$HD_DEV"
            el_explain 2 "Created partition __${p_num}__ for __/home__ with the rest of the disk"

            partition_alignment_check "$HD_PART_HOME_dev" report
            p_num="$(( $p_num + 1 ))"
            #p_last_size="$(( $p_last_size + $p_size_mb ))" # we don't need this variable anymore and since we don't have p_size_mb on this part (no size, we used 'rest of the disk'), it will fail

            # }}}
        fi
    fi


    echo 70 > "$TMP_PROGRESS_CONFIGURING_f"

    # re-read partition table, remove ghost partitions
    #device_check_is_busy "$HD_DEV"

    # - create structure of disk }}}
    echo 90 > "$TMP_PROGRESS_CONFIGURING_f"

    # delete gui and parent process
    rm -f "$TMP_PROGRESS_CONFIGURING_f"
    guitool "working_start"
}
#########################################################################
check_datetime(){
    functions_pre
    el_debug "\targs '$*'"
    # check bios clock {{{
    # check if clock is working correctly
    actual_date="$(date +%F -d '+1 days' )"

    el_debug "Date builded: '$datebuilded', Date actual: '${actual_date}'"

    if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$datebuilded" "gt" "$actual_date" ; then
        # try to fix
        if el_verify_internet fast 2>/dev/null ; then
            /etc/NetworkManager/dispatcher.d/timezone-update -f
        fi
        actual_date="$(date +%F -d '+1 days' )"

        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$datebuilded" "gt" "$actual_date" ; then

            local message_clock
            message_clock="$( printf "$( eval_gettext "There is a problem with your internal clock; it shows the time as '%s'. Correct clock settings are crucial; incorrect settings can cause issues with your Elive system, especially affecting web browsers. If your computer forgets the date at every shutdown, replace the BIOS battery. Take your computer to a shop for replacement, then do a new installation of Elive." )" "$(date)" )"

            guitool error "$message_clock"

            # try to fix
            [[ -x /usr/bin/datetimeconf ]] && datetimeconf || true

            local message_clock
            message_clock="$( printf "$( eval_gettext "The new date has been set. To ensure your BIOS saves it correctly, turn off your computer for a few minutes and try reinstalling Elive. If the issue persists, it is strongly recommended to replace your BIOS battery." )" "" )"

            guitool info "$message_clock"
        fi
    fi
    # - clock bios check }}}
}
#########################################################################
verify_speeds_timing(){
    functions_pre
    el_debug "\targs '$*'"
    unset speed_disk_num speed_disk speed_disk_slow
    local disk
    disk="$1"

    #speed_disk="$( LC_ALL="$EL_LC_EN" hdparm -t "$disk" 2>/dev/null | grep "=.*/sec" | sed 's/^.*seconds = //')"
    speed_disk="$( LC_ALL="$EL_LC_EN" hdparm -t "$disk" 2>/dev/null | grep "buffered disk read" | sed -e 's|^.* = ||g' )"
    speed_disk_num="$speed_disk"
    speed_disk="$( echo "$speed_disk" | awk '{print $1}')"

    case "$speed_disk_num" in
        *MB/sec|*mb/sec)
            LC_ALL="$EL_LC_EN" dpkg --compare-versions "$speed_disk" "lt" "12" && speed_disk_slow=2
            LC_ALL="$EL_LC_EN" dpkg --compare-versions "$speed_disk" "lt" "4" && speed_disk_slow=1
            ;;
        *KB/sec|*kb/sec)
            speed_disk_slow="kb"
            ;;
        *GB/sec|*gb/sec|*tb/sec|*TB/sec)
            true
            ;;
        *)
            el_error "Unknown speed results: $speed_disk, continuing...\ndemo data for disk '$disk':\n$( hdparm -t "$disk" 2>&1 )"
            ;;
    esac

    el_explain 2 "Speed of your partition $disk is $speed_disk_num"
}
#########################################################################
verify_speeds(){
    functions_pre
    el_debug "\targs '$*'"
    local disk
    disk="$1"

    if ((ignore_hd_speeds_verify)) ; then
        return 0
    fi

    guitool "working_start"
    verify_speeds_timing "$disk"

    for i in 1 2
    do
        if echo "$speed_disk_slow" | grep -qsE "^(kb|1|2)$" ; then
            # Slow, repeat test to be sure
            sync
            verify_speeds_timing "$disk"
        fi
    done

    # create a message with results to use later
    local message_slow_hd
    message_slow_hd="$( printf "$( eval_gettext "WARNING: The disk you wish to install Elive on is extremely slow (%s). This might be due to a bug. Consider updating your BIOS, using another disk, or updating the kernel. Review this before proceeding. Continuing will result in Elive being installed on a VERY SLOW disk!" )" "$speed_disk_num" )"

    case "$speed_disk_slow" in
        kb|1)
            if test -f /tmp/.eeepc-detected
            then
                guitool warning "$( eval_gettext "WARNING: The disk for installing Elive is extremely slow. This appears to be a bug. If your computer is an EeePC, there's a known issue with EeePCs and USB connectors. Update your BIOS to resolve this. Alternatively, disable SD Media Card devices in your BIOS settings (you won't be able to read memory cards unless you re-enable this setting)." )"
                guitool info "$( eval_gettext "To update the BIOS on your EeePC, download the BIOS from asus.com. On the website, go to 'Global' and click 'Support'. Follow the instructions to download a zip file containing the ROM. After unpacking, rename it to 701.ROM (for 4 GB models, 701 series) and copy it to a FAT16-formatted USB thumb drive. FAT16 is mandatory. Reboot your EeePC with the USB drive inserted (and no other USB devices connected), then press Alt-F2 during the BIOS splash screen. This will cause the BIOS to search for a file named 701.ROM on your USB thumb drive and upgrade the BIOS using it. (Note: 2 GB models look for the file 700.ROM). Proceed with caution and patience. Do not interrupt the process or shut down your computer before the upgrade is complete." )"
            else
                guitool warning "$message_slow_hd"

            fi
            ;;
        2)
            guitool warning "$message_slow_hd"
            ;;
    esac
}
#########################################################################
verify_result(){
    functions_pre
    #el_debug "\targs '$*'"

    if [[ "$1" = "cancel" ]] ; then
        guitool warning "$( eval_gettext "You exited the installer." )"
        el_error "user cancel a dialog to exit the installer"
        exit
    fi

    # FIXME: in things like the parter selector, the user don't know how to pick a choice, instead of exit the installer just run the options again (everytime for every function? mmmh)
    if [[ -z "$1" ]] || [[ "$1" = "(null)" ]] ; then
        guitool info "$( eval_gettext "No option selected. Launch the installer again and select an option." )" || true
        el_error "user didnt choose option so we exited <- ${FUNCNAME[1]}"
        exit
    fi

    if ! [[ "${FUNCNAME[1]}" = *"pass"* ]] ; then
        el_debug "verify_result with value: $*"
    fi
}
#########################################################################
error_found(){
    functions_pre
    el_debug "\targs '$*'"
    # indent
    local message_error
    message_error="$( printf "$( eval_gettext "An error occurred during the installation process. Please reboot and try again. If the problem persists, report it to Elive. Your reports are important to us.\n\nError message: %s" )" "$( cat "$TMP_LOGS_ERR_f" )" )"

    el_error "REPORTED to the user:  $( cat "$TMP_LOGS_ERR_f" )" # report logs to elive
    guitool error "$message_error \n$( cat "$TMP_LOGS_ERR_f" )"

    rm -f "$TMP_LOGS_ERR_f"
    exit_error
}
#########################################################################
list_others_systems(){
    [[ -s "/tmp/os-probed" ]] && return
    functions_pre
    el_debug "\targs '$*'"
    guitool working_start

    # required to list correctly the names of the OS's when selecting partitions
    timeout 90 os-prober 1>/tmp/os-probed 2>/dev/null || true

    if [[ -n "$(cat /tmp/os-probed)" ]] ; then
        el_explain 0 "os-prober result:"
        cat /tmp/os-probed 2>&1
    fi
    if grep -qs "Windows" /tmp/os-probed ; then
        is_windows_installed=1
    fi
}
#########################################################################
hooks_post_install(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    local answer i file entry

    # }}}
    # variables {{{
    # set correct home directory
    if [[ -s "$TARGET/etc/adduser.conf" ]] ; then
        dhome="$( grep "^DHOME" "$TARGET/etc/adduser.conf" | sed -e 's|^.*=||g' )"
    fi
    if [[ -z "$dhome" ]] ; then
        dhome="/home"
    fi

    # }}}
    # cleanups {{{
    # remove useless/broken files
    rm -f "$TARGET"/vmlinuz*
    rm -f "$TARGET"/initrd*
    # rm -f "$TARGET/etc/default/grub" 2>/dev/null 1>&2 || true # just in case there's an old one, do not prompt debconf
    rm -f "$TARGET/etc/apt/apt.conf.d/00_disable-cache-files.conf" # make the pkgcache.bin and srcpkgcache.bin files working by default (not in live mode)


    # migration mdoe:
    if ((is_migration)) || ((is_old_elive_upgrade_request_accepted)) ; then

        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "3.8.8" ; then
            rm -rf "$TARGET/etc/elive-tools/geolocation/timezones.conf"
        fi
    fi
    # }}}

    # Configurations
    # fixes {{{
    if ! [[ -s "$TARGET/etc/apt/preferences.d/elive_priority.pref" ]] ; then
        el_debug "elive_priority.pref file not found, creating it..."
        cat >> "$TARGET/etc/apt/preferences.d/elive_priority.pref" << EOF
Package:             *
Pin:                 release o=Elive
Pin-Priority:        1500
EOF
    fi
    # - fixes }}}
    # PRESEEDS {{{

    echo 10 > "$TMP_PROGRESS_f"
    # copy all possible needed preseeds
    if [[ -s "/tmp/preseeds.cfg" ]] ; then
        cat /tmp/preseeds.cfg | sort -u > "$TARGET/root/preseeds.cfg"
    fi

    if [[ -d "/tmp/preseeds.d" ]] ; then
        mkdir -p "$TARGET/root/preseeds.d"
        cp -a /tmp/preseeds.d/* "$TARGET/root/preseeds.d/" 2>/dev/null || true
    fi

    # MBR ask and install (includes a preseed dumped to TARGET, so REQUIRES to be after the preseed copy)
    efi_select
    # efi selection has preference (called before) mbr:
    mbr_select


    # }}}
    # reconfigures {{{

    # append some default needed ones too:
    echo -e "x11-common" >> /tmp/.packages-to-reconfigure.tmp
    echo -e "dash" >> /tmp/.packages-to-reconfigure.tmp
    if [[ -x "/usr/bin/gksu" ]] ; then
        echo -e "gksu" >> /tmp/.packages-to-reconfigure.tmp
    fi

    # - reconfigures }}}
    # Fstab entries add {{{
    rm -f "$TARGET/etc/fstab" 2>/dev/null
    el_debug "Creating FSTAB"

    echo -e "# Generated by the Elive installer\n" >> "$TARGET/etc/fstab"
    # not needed anymore since wheezy:
    #echo -e "proc        /proc    proc        defaults    0    0" >> "$TARGET/etc/fstab"
    #echo -e "sysfs        /sys    sysfs        defaults    0    0" >> "$TARGET/etc/fstab"

    # }}}
    # swap entries for hebernate {{{
    # TODO: verify if this is still needed / working
    if [[ -n "$SWAPDEVICES" ]] ; then
        el_explain 2 "Swap devices list: $SWAPDEVICES"
        for i in $SWAPDEVICES
        do
            if [[ -b "$i" ]] ; then
                swaptab_add "$i"
            else
                el_error "$i is not a block device and we wanted to add it as a swap device"
            fi
        done
    fi

    # - swap entries for fstab }}}
    # root partition entry {{{
    if [[ -z "$HD_PART_ROOT_FS_OPTS_v" ]] ; then
        HD_PART_ROOT_FS_OPTS_v="defaults,commit=300,relatime"
    fi

    if [[ -n "$HD_PART_HOME_dev" ]] ; then
        el_explain 0 "Changing __relatime__ to __lazytime__ in your root partition of your __fstab__ because you have either a separed /home or /var partition"
        HD_PART_ROOT_FS_OPTS_v="${HD_PART_ROOT_FS_OPTS_v/relatime/lazytime}"
    fi

    echo -e "$( hd_uuid_encode_fstab "${HD_PART_ROOT_dev}" )    /    ${HD_PART_ROOT_FS_v}    $HD_PART_ROOT_FS_OPTS_v    0    1" >> "$TARGET/etc/fstab"
    el_debug "$( hd_uuid_encode_fstab "${HD_PART_ROOT_dev}" )    /    ${HD_PART_ROOT_FS_v}    $HD_PART_ROOT_FS_OPTS_v    0    1"


    # - root partition entry }}}
    # extra mountpoints for fstab {{{
    if [[ -s "$TMP_FSTAB_EXTRA_f" ]] ; then

        el_debug "Adding the extra mountpoints to fstab:"

        while read -ru 3 line
        do
            #if grep -qsE "^$( echo "${line}" | awk '{print $1}' )\s+" "$TARGET/etc/fstab" ; then
            # skip if ID or mountpoint already exists in fstab
            #if cat "$TARGET/etc/fstab" | grep -v "^#" | awk '{print $1}' | grep -qs "^$( echo "${line}" | awk '{print $1}' )$" ; then
                #el_warning "SKIPP: F1 in '$( echo "$line" | awk '{print $1}')' matches in:\n$(cat $TARGET/etc/fstab | awk '{print $1}' )"
            #fi
            #if cat "$TARGET/etc/fstab" | grep -v "^#" | awk '{print $2}' | grep -qs "^$( echo "${line}" | awk '{print $2}' )$" ; then
                #el_warning "SKIPP: F2 in '$( echo "$line" | awk '{print $2}')' matches in:\n$(cat $TARGET/etc/fstab | awk '{print $2}' )"
            #fi

            if cat "$TARGET/etc/fstab" | grep -v "^#" | awk '{print $1}' | grep -qs "^$( echo "${line}" | awk '{print $1}' )$" || cat "$TARGET/etc/fstab" | grep -v "^#" | awk '{print $2}' | grep -qs "^$( echo "${line}" | awk '{print $2}' )$" ; then
                el_debug "Skipped fstab entry because is already included: $line"
            else
                echo "$line" >> "$TARGET/etc/fstab"
                el_debug "Extra fstab entry added: $line"
            fi
        done 3<<< "$( cat "$TMP_FSTAB_EXTRA_f" | grep -viE "(^#|/boot/efi)" )"
    fi

    cp "$TARGET/etc/fstab" "${TMP_MAIN_d}/installed-fstab"

    # checks
    if [[ "$( cat "$TARGET/etc/fstab" | grep -v "^#" | awk '{print $2}' | sort | uniq -d | wc -l )" -gt 0 ]]  ; then
        el_warning "duplicated mountpoints in fstab:\n$( cat "$TARGET/etc/fstab" | grep -v "^#" | awk '{print $2}' | sort | uniq -d )"
    fi
    if [[ "$( cat "$TARGET/etc/fstab" | grep -v "^#" | grep -E "^(/dev|UUID|LABEL)" | awk '{print $1}' | sort | uniq -d | wc -l )" -gt 0 ]] ; then
        el_warning "duplicated partitions in fstab:\n$( cat "$TARGET/etc/fstab" | grep -v "^#" | grep -E "^(/dev|UUID|LABEL)" | awk '{print $1}' | sort | uniq -d )"
    fi

    # clean some files
    rm -f "$TARGET/etc/udev/rules.d/92-hide-partitions-eliveinstaller.rules"

    # }}}
    # crypttab entries {{{

    # not needed
    #_head="$( grep "^#" "$TARGET/etc/crypttab" 1>/dev/null 2>&1 | head -1 )"
    #echo "$_head" > "$TARGET/etc/crypttab"

    # add the Elive needed crypttab file
    counter=0
    if [[ -n "${hd_dev_to_add_in_crypttab[*]}" ]] ; then
        echo -e "##  This list includes all the encrypted disks that needs to be decrypted at boot time, like if your Elive system requires the root / partition that is encrypted in order to boot correctly.\n##  Remove any entry that is not Elive depending\n\n" > "${TARGET}/etc/crypttab"

        while read -ru 3 line
        do
            # add crypttab entry
            echo "${hd_dev_to_add_in_crypttab[$counter]}" >> "${TARGET}/etc/crypttab"
            el_debug "added crypttab entry: ${hd_dev_to_add_in_crypttab[$counter]}"

            is_cryptsetup_configured=1

            p_enc_uuid="$( echo "${hd_dev_to_add_in_crypttab[$counter]}" | tr ' ' '\n' | grep "UUID=" | sed -e 's|UUID=||g' | tail -1 )"

            if [[ -n "$p_enc_uuid" ]] ; then
                # also blacklist it in thunar, we don't want to see it at all, we want just to -use- it in the installed system
                if ((is_wheezy)) ; then
                    echo -e "ENV{ID_FS_UUID}==\"${p_enc_uuid}\", ENV{UDISKS_PRESENTATION_HIDE}:=\"1\"" >> "$TARGET/etc/udev/rules.d/92-hide-partitions-eliveinstaller.rules"
                else
                    echo -e "ENV{ID_FS_UUID}==\"${p_enc_uuid}\", ENV{UDISKS_IGNORE}:=\"1\"" >> "$TARGET/etc/udev/rules.d/92-hide-partitions-eliveinstaller.rules"
                fi
            else
                el_error "UUID not included in $line"

            fi

            counter="$(( $counter + 1 ))"
        done 3<<< "$( seq "${#hd_dev_to_add_in_crypttab[@]}")"
    fi


    # create crypttab-extra-entries file
    while read -ru 3 line
    do
        unset p_dev p_filesystem p_passwd p_enc_uuid

        [[ -z "$line" ]] && continue

        p_filesystem="$( echo "$line" | awk -v FS="::" '{print $3}' )"
        p_dev="$( echo "$line" | awk -v FS="::" '{print $1}' )"
        p_size_h="$( echo "$line" | awk -v FS="::" '{print $7}' )"


        case "$p_filesystem" in
            crypto_LUKS)
                p_enc_uuid="$( LC_ALL="$EL_LC_EN" cryptsetup luksDump "$p_dev" | grep "^UUID:" | head -1 | awk '{print $2}' )"
                #p_enc_name="${p_dev#/dev/mapper/}"
                #p_enc_name="${p_dev#/dev/md/}"
                #p_enc_name="${p_dev#/dev/}_crypt"
                p_enc_name="${p_dev##*/}"

                # already added in default crypttab so we don't need this entry
                if grep -qs "UUID=$p_enc_uuid" "$TARGET/etc/crypttab" ; then
                    continue
                fi

                # header
                if ! [[ -s "$TARGET/etc/crypttab-extra-entries" ]] ; then
                    echo -e "##  This is an example list of encrypted partitions with their configuration that needs to be decrypted at boot-time in order to boot correctly the entire system, like if your Elive system requires the root / partition that is encrypted in order to boot correctly.\n##  You can use this list to add devices in your real /etc/crypttab file\n\n" > "${TARGET}/etc/crypttab-extra-entries"
                fi

                # add them in /etc/crypttab-extra-entries
                if device_is_ssd "$p_dev" ; then
                    el_explain 0 "crypttab entry added: ${p_enc_name}_crypt    UUID=${p_enc_uuid}    none    luks,discard"
                    echo -e "# Auto entry added by the Elive installer for: ${p_dev}\n${p_enc_name}_crypt    UUID=${p_enc_uuid}    none    luks,discard\n" >> "$TARGET/etc/crypttab-extra-entries"
                else
                    el_explain 0 "crypttab entry added: ${p_enc_name}_crypt    UUID=${p_enc_uuid}    none    luks"
                    echo -e "# Auto entry added by the Elive installer for: ${p_dev}\n${p_enc_name}_crypt    UUID=${p_enc_uuid}    none    luks\n" >> "$TARGET/etc/crypttab-extra-entries"
                fi

                is_cryptsetup_configured=1

                ;;
        esac
    done 3<<< "$( EL_DEBUG=0 partitions-list --show-all || true )"

    # if multiple lines found, ask if manually edit it
    if ! ((conf_is_novice_user)) ; then
        if [[ "$( cat "$TARGET/etc/crypttab" | grep -v "^#" | grep -i luks | sort | wc -l )" -gt "1" ]] ; then
            if guitool yesnoquestioninv "$( eval_gettext "There are several encrypted disks listed in your Crypttab file. It's best to remove entries unrelated to Elive; otherwise, you'll be prompted for their passwords at every boot. Do you want to edit your Crypttab configuration now?" )" ; then
                if ! el_dependencies_check scite ; then
                    el_dependencies_install scite
                fi
                scite "$TARGET/etc/crypttab"
            fi
        fi
    fi

    # FIXME if is_live_from_usb_detected_as_sda we could have problems with this part

    # - crypttab entries }}}
    # lvm2 changes {{{
    # if [[ -b "$HD_DEV" ]] ; then
    #     if device_is_ssd "$HD_DEV" ; then
    #         if ((is_lvm2_configured)) || ((is_cryptsetup_configured)) || [[ -n "$( LC_ALL="$EL_LC_EN" LVM_SUPRESS_FD_WARNINGS=1 lvs --noheadings "$HD_PART_ROOT_dev" )" ]] ; then
    #             sed -i "s|^.*issue_discards = 0.*$|    issue_discards = 1|g" "$TARGET/etc/lvm/lvm.conf"
    #             echo -e "lvm2" >> /tmp/.packages-to-reconfigure.tmp
    #         fi
    #     fi
    # fi

    # - lvm2 changes }}}
    # don't show some partitions to udev {{{
    if [[ -n "$HD_PART_BOOT_dev" ]] ; then
        if hd_uuid_encode "$HD_PART_BOOT_dev" 1>/dev/null ; then
            if ((is_wheezy)) ; then
                echo -e "ENV{ID_FS_UUID}==\"$( hd_uuid_encode "${HD_PART_BOOT_dev}" )\", ENV{UDISKS_PRESENTATION_HIDE}:=\"1\"" >> "$TARGET/etc/udev/rules.d/92-hide-partitions-eliveinstaller.rules"
            else
                echo -e "ENV{ID_FS_UUID}==\"$( hd_uuid_encode "${HD_PART_BOOT_dev}" )\", ENV{UDISKS_IGNORE}:=\"1\"" >> "$TARGET/etc/udev/rules.d/92-hide-partitions-eliveinstaller.rules"
            fi
            el_debug "hidding partition ${HD_PART_BOOT_dev} to udev in our system"
        fi
    fi

    # don't show windows recovery partitions in our thunar
    while read -ru 3 line
    do
        [[ -z "$line" ]] && continue
        if hd_uuid_encode "$line" 1>/dev/null ; then
            if ((is_wheezy)) ; then
                echo -e "ENV{ID_FS_UUID}==\"$( hd_uuid_encode "${line}" )\", ENV{UDISKS_PRESENTATION_HIDE}:=\"1\"" >> "$TARGET/etc/udev/rules.d/92-hide-partitions-eliveinstaller.rules"
            else
                echo -e "ENV{ID_FS_UUID}==\"$( hd_uuid_encode "${line}" )\", ENV{UDISKS_IGNORE}:=\"1\"" >> "$TARGET/etc/udev/rules.d/92-hide-partitions-eliveinstaller.rules"
            fi
            el_debug "hidding partition ${line} to udev in our system"
        fi

        # run os-prober ignoring results like the EFI partition of windows, like:
            #  /dev/sda2@/efi/Microsoft/Boot/bootmgfw.efi:Windows Boot Manager:Windows:efi
    done 3<<< "$( timeout 90 os-prober 2>/dev/null | grep -v "/dev/.*@/" | grep -iE "Windows.*(setup|Recovery|Boot Manager)" | grep "^/dev/" | sed -e 's|:.*$||g' )"

    # - ignore /boo too }}}
    # set if we use ssd {{{
    if [[ -b "$HD_PART_HOME_dev" ]] ; then
        if device_is_ssd "$HD_PART_HOME_dev" ; then
            echo "ssd_on_home: yes" >> "$TARGET/etc/elive-version"
        else
            echo "ssd_on_home: no" >> "$TARGET/etc/elive-version"
        fi
    else
        # if no extra home, we should set it to yes if root is set to yes
        if [[ -b "$HD_PART_ROOT_dev" ]] && device_is_ssd "$HD_PART_ROOT_dev" ; then
            echo "ssd_on_home: yes" >> "$TARGET/etc/elive-version"
        fi
    fi

    if [[ -b "$HD_PART_ROOT_dev" ]] && device_is_ssd "$HD_PART_ROOT_dev" ; then
        echo "ssd_on_root: yes" >> "$TARGET/etc/elive-version"
    else
        echo "ssd_on_root: no" >> "$TARGET/etc/elive-version"
    fi
    # - tell if we use ssd }}}
    # hosts configurationa {{{
    host_ask
    # }}}
    # refind extra boot system {{{
    #if ((is_efi_booted)) && [[ -d "$TARGET/boot/efi/EFI" ]] && [[ "$( find "$TARGET/boot/efi/EFI/" -mindepth 1 -maxdepth 1 -type d | grep -viE "/(OC|elive|BOOT|tools|refind|Windows|Windows_RE|WIN_RE|WinRE|Microsoft|grub)$" | wc -l )" -ge 1 ]] ; then
        #if guitool yesnoquestion "$( eval_gettext "You have other operating systems on this computer, we suggest installing REFIND to have an improved boot manager, on it you can select which operating system to boot in a better way. Do you want to install REFIND?" )" ; then
            #is_refind_wanted=1
        #fi
    #fi
    # - refind extra boot system }}}


    # Remove extra packages / kernels /modules
    # add extra packages to install {{{
    # always include elive-tools to be updated in the system
    echo "elive-tools" >> "/tmp/.packages-to-install.tmp"
    # - extra packages to install }}}
    # packages not needed anymore in the installed system {{{
    # FIXME TODO: move this to the installed system first boot, we still depend of the eliveinstaller (and maybe others?) before to proceed
    # remove extra non-needed kernels
    # include also the modules to remove, they should be removed before the kernels if possible
    for package in deliver live-config live-config-sysvinit live-config-systemd live-tools live-boot live-boot-initramfs-tools elive-welcome
    do
        echo "$package" >> "/tmp/.packages-to-remove.tmp"
    done

    # do not use plymouth when is not needed
    if ((is_i386)) && ! ((is_wheezy)) ; then
        if ! ((is_encryption_wanted)) && [[ "$RAM_TOTAL_SIZE_mb" -lt 1300 ]] ; then
            echo "plymouth" >> "/tmp/.packages-to-remove.tmp"
            echo "plymouth-label" >> "/tmp/.packages-to-remove.tmp"
        fi
    fi

    # buggy slamr...
    if grep -qs sl-modem "/tmp/.packages-to-remove.tmp" 2>/dev/null ; then
        rm -f "$TARGET/"lib/modules/*/misc/slamr.ko 2>/dev/null
    fi
    # remove annoying things, this gives dkms errors when updating kernel (postinst kernel dkms)
    if ! LC_ALL="$EL_LC_EN" lsmod | grep -qs "ndiswrapper" ; then
        echo "ndiswrapper-common" >> "/tmp/.packages-to-remove.tmp"
        echo "ndiswrapper-dkms" >> "/tmp/.packages-to-remove.tmp"
        echo "ndiswrapper-source" >> "/tmp/.packages-to-remove.tmp"
    fi

    # remove virtualbox packages if we dont need them
    if ! ((is_virtualbox)) ; then
        echo -e "virtualbox-guest-dkms" >> "/tmp/.packages-to-remove.tmp"
        echo -e "virtualbox-guest-utils" >> "/tmp/.packages-to-remove.tmp"
        echo -e "virtualbox-guest-x11" >> "/tmp/.packages-to-remove.tmp"
    fi

    # remove unneeded efi things
    if ! ((is_efi_booted)) ; then
        echo -e "efibootmgr" >> "/tmp/.packages-to-remove.tmp"
        echo -e "efivar" >> "/tmp/.packages-to-remove.tmp"
        echo -e "fwupd" >> "/tmp/.packages-to-remove.tmp"
        killall -9 fwupd 1>/dev/null 2>&1 || true
        # before, if we removed this package it lead to the package eliveinstaller-6g being removed too, making the system wrongly installed
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" "ge" "3.8.42" ; then
            echo -e "mokutil" >> "/tmp/.packages-to-remove.tmp" # do not remove this package, so it will remove the installer too
            echo -e "sbsigntool" >> "/tmp/.packages-to-remove.tmp"
        fi
    fi

    if ((is_virtualized)) ; then
        # remove useless things for virtual machines
        for i in laptop-mode-tools pcmciautils powertop vbetool hdapsd tp-smapi-dkms task-laptop rsyslog rfkill urfkill cpufrequtils thermald linux-cpupower hddtemp v86d libfprint0 madfuload madwimax usb-modeswitch alsa-firmware-loaders

        do
            echo "$i" >> "/tmp/.packages-to-remove.tmp"
        done
    fi

    # remove uneeded features
    if grep -qs btrfs "$TARGET/etc/fstab" ; then
        echo -e "btrfsmaintenance" >> "/tmp/.packages-to-install.tmp"
    else
        echo -e "btrfsmaintenance" >> "/tmp/.packages-to-remove.tmp"
    fi

    # encrypted HD requires this pacakge
    if ((is_encryption_wanted)) ; then
        echo -e "cryptsetup-initramfs" >> "/tmp/.packages-to-install.tmp"
    else
        echo -e "cryptsetup-initramfs" >> "/tmp/.packages-to-remove.tmp"
    fi
    if ((is_nvme_available)) ; then
        echo -e "nvme-cli" >> "/tmp/.packages-to-install.tmp"
    else
        echo -e "nvme-cli" >> "/tmp/.packages-to-remove.tmp"
    fi

    if [[ -x "$(which laptop-detect || true )" ]] ; then
        if ! laptop-detect ; then
            # disable laptop features if we don't need them
            echo -e "laptop-mode-tools" >> "/tmp/.packages-to-remove.tmp"

            if ! ((is_wifi_existing)) ; then
                echo "rfkill" >> "/tmp/.packages-to-remove.tmp"
                echo "urfkill" >> "/tmp/.packages-to-remove.tmp"
            fi
        fi
    else
        el_warning "no laptop-detect tool included, we need it"
    fi
    # toshiba keys
    if ! [[ -e /proc/acpi/toshiba/keys ]] ; then
        echo -e "fnfxd" >> "/tmp/.packages-to-remove.tmp"
    fi
    # hdaps (mac / thinkpad hard disk park header when movements)
    if ! LC_ALL="$EL_LC_EN" ps aux | grep -v grep | grep -qs "sbin/hdapsd" ; then
        el_explain 0 "removing package __hdaps__ because this computer don't need it"
        echo -e "hdapsd" >> "/tmp/.packages-to-remove.tmp"
    fi
    if ! LC_ALL="$EL_LC_EN" lsmod | grep -qs "tp_smapi" ; then
        el_explain 0 "removing modules for __tp-smapi__ because this computer don't need it"
        echo -e "tp-smapi-dkms" >> "/tmp/.packages-to-remove.tmp"
        echo -e "tp-smapi-source" >> "/tmp/.packages-to-remove.tmp"
    fi
    # amd firmwares that wastes space
    # update: don't remove, otherwise will remove all the firmwares
    #if [[ -x "$( which fglrx-privative-drivers-install || true )" ]] && ! fglrx-privative-drivers-install -q -d ; then
        #echo "firmware-amd-graphics" >> "/tmp/.packages-to-remove.tmp"
    #fi
    # big size
    _module="nfp"
    if ! LC_ALL="$EL_LC_EN" lsmod | grep -qsE "^${_module}\s+" || LC_ALL="$EL_LC_EN" lsmod | grep -qsE "^${_module}\s+.*\s+0$" ; then
        if [[ -x "$(which hwinfo || true )" ]] && ! hwinfo --network | grep -qsi "netronome" ; then
            echo "firmware-netronome" >> "/tmp/.packages-to-remove.tmp"
        fi
    fi
    # remove unneeded modules that slows updating kernels
    _module="adv17v35x"
    if ! LC_ALL="$EL_LC_EN" lsmod | grep -qsE "^${_module}\s+" || LC_ALL="$EL_LC_EN" lsmod | grep -qsE "^${_module}\s+.*\s+0$" ; then
        echo "adv-17v35x-dkms" >> "/tmp/.packages-to-remove.tmp"
    fi

    # reconfigures
    # reconfigure ping to work from user
    # update: moved to post-install
    #echo "iputils-ping" >> "/tmp/.packages-to-reconfigure.tmp"
    ## reconfigure every package that needs setcap command:
    #for i in $( grep -El "\W+setcap\W+" /var/lib/dpkg/info/*postinst | sed -e 's|\:.*$||g' -e 's|\.postinst||g' -e 's|^.*/||g' )
    #do
        #for j in /var/lib/dpkg/info/${i}.list /var/lib/dpkg/info/${i}:*.list
        #do
            #if [[ -f "$j" ]] ; then
                #j="${j##*/}"
                #j="${j%.list}"
                #echo "$j" >> "/tmp/.packages-to-reconfigure.tmp"
            #fi
        #done
    #done

    # raid / lvm to remove if not needed {{{
    if ! LC_ALL="$EL_LC_EN" vgchange -ay 2>/dev/null | grep -q "logical volume.*active" ; then
        echo 'lvm2' >> "/tmp/.packages-to-remove.tmp"
    fi
    # if ! cat /proc/mdstat | grep -qs " : active.*raid" && ! cat /proc/mdstat | grep Personalities | sed -e 's|Personalities :||g' | grep -qs "[[:alpha:]]" && ! ls -1 /dev/md* /dev/md/* 1>/dev/null 2>&1 ; then
    if [[ -z "$( awk '/^md/ {printf "%s: ", $1}; /blocks/ {print $NF}' /proc/mdstat )" ]] ; then
        echo 'mdadm' >> "/tmp/.packages-to-remove.tmp"
    fi
    # fi
    if ! dmraid -r 1>/dev/null 2>/dev/null ; then
        echo 'dmraid' >> "/tmp/.packages-to-remove.tmp"
    fi

    # - raid / lvm to remove if not needed }}}

    # all the extra kernels and metapackages should be removed:
    #LC_ALL="$EL_LC_EN" COLUMNS=1000 dpkg -l | grep -E "^(hi|ii|rc).*linux-image-" | grep -vE -- "^(hi|ii).*$(uname -r | sed -e 's|+|\\\+|g' )(-unsigned)?\s+" | grep -v "(meta-package)" | awk '{print $2}' >> "/tmp/.packages-to-remove.tmp"
    LC_ALL="$EL_LC_EN" COLUMNS=1000 dpkg -l | grep -E "^(hi|ii|rc).*linux-image-" | grep -vE -- "^(hi|ii).*$(uname -r | sed -e 's|+|\\\+|g' )(-unsigned)?\s+" | awk '{print $2}' >> "/tmp/.packages-to-remove.tmp"
    # security check for possible bugs
    if [[ "$( cat /tmp/.packages-to-remove.tmp | grep "linux-image" | wc -l )" -gt 1 ]] ; then
        el_warning "more than one kernel wanted to be removed, is this correct?\n$( cat /tmp/.packages-to-remove.tmp | grep "linux-image" )"
    fi
    #LC_ALL="$EL_LC_EN" COLUMNS=1000 dpkg -l | grep -E "^(hi|ii|rc).*linux-headers-" | grep -vE -- "^(hi|ii).*$(uname -r | sed -e 's|+|\\\+|g' )\s+" | grep -v "linux-headers-generic" | grep -v "(meta-package)" | grep -v "[[:digit:]]*-common\s*" | awk '{print $2}' >> "/tmp/.packages-to-remove.tmp"
    LC_ALL="$EL_LC_EN" COLUMNS=1000 dpkg -l | grep -E "^(hi|ii|rc).*linux-headers-" | grep -vE -- "^(hi|ii).*$(uname -r | sed -e 's|+|\\\+|g' )\s+" | grep -v "linux-headers-generic" | grep -v "[[:digit:]]*-common\s*" | awk '{print $2}' >> "/tmp/.packages-to-remove.tmp"

    # if we have a /boot partition, make sure that we never update the kernels in order to not run out of space. Otherwise it should not matter updating the kernel of the same version (and so we will have a more stable system)
    if [[ -n "$HD_PART_BOOT_dev" ]] ; then
        LC_ALL="$EL_LC_EN" COLUMNS=1000 dpkg -l | grep -E "^(hi|ii|rc).*linux-(image|headers)-" | grep -E -- "^(hi|ii).*$(uname -r | sed -e 's|+|\\\+|g' | sed -e 's|-[^-]*$||g' ).*(-unsigned)?\s+" | awk '{print $2}' >> "/tmp/.packages-to-hold.tmp"
        LC_ALL="$EL_LC_EN" COLUMNS=1000 dpkg -l | grep -E "^(hi|ii|rc).*linux-kbuild-" | grep -E -- "^(hi|ii).*$(uname -r | sed -e 's|+|\\\+|g' | sed -e 's|-[^-]*$||g' ).*(-unsigned)?\s+" | awk '{print $2}' >> "/tmp/.packages-to-hold.tmp"
    fi


    # }}}

    # CUSTOM options and packages {{{
    if ! ((conf_configure_options)) ; then
        el_explain 0 "AUTOMATED selection of fine-tune options"
    fi
    if ! ((ignore_remove_install_selected_packages)) ; then
        guitool "working_stop"

        # defaults {{{
        _app_bluetooth="TRUE"
        _app_disknotify="TRUE"
        _app_fail2ban="FALSE"
        _app_nssmdns="FALSE"
        _app_printerscan="TRUE"
        _app_samba="TRUE"
        _app_systemd="TRUE"

        # pre-enable ssh for advanced users only:
        if ((conf_is_extra_help_wanted)) || ((conf_is_novice_user)) ; then
            _app_ssh="FALSE"
        else
            _app_ssh="TRUE"
        fi

        if [[ "$( LC_ALL=C hcitool dev 2>/dev/null | grep -v "^Devices:$" | grep ":.*:.*:" | wc -l )" -gt 0 ]] ; then
            _app_bluetooth="TRUE"
        else
            _app_bluetooth="FALSE"
        fi

        if ((is_virtualized)) ; then
            _app_bluetooth="FALSE"
            _app_disknotify="FALSE"
            _app_nssmdns="FALSE"
            _app_printerscan="FALSE"
            _app_samba="FALSE"
            _app_ssh="TRUE"
        fi

        #_app_="TRUE"
        # - defaults }}}

        # select features ( bluetooth / printer / scanner / samba / ssh / systemd ) {{{
        if ! ((conf_configure_options)) ; then
            # XXX non-interactive:
            echo "${_app_bluetooth}|${_app_printerscan}|${_app_printerscan}|${_app_samba}|${_app_ssh}|${_app_fail2ban}|${_app_disknotify}|${_app_systemd}|" > "$TMP_FILE_1_f"
        else
            yad --form --center --width=460 --title="$( eval_gettext "Features of Elive:" )" --image=gnome-settings --image-on-top \
                --text="$( eval_gettext "Choose the features you want to have installed. Deselecting them can lighten your system." )" \
                --field="$( eval_gettext "Bluetooth support" ):chk" $_app_bluetooth \
                --field="$( eval_gettext "Printer support" ):chk" $_app_printerscan \
                --field="$( eval_gettext "Scanner support" ):chk" $_app_printerscan \
                --field="$( eval_gettext "SAMBA shared directories and printers" ):chk" $_app_samba \
                --field="$( eval_gettext "SSH remote access" ):chk" $_app_ssh \
                --field="$( eval_gettext "Block IP's to prevent continuously attempted attacks" ):chk" $_app_fail2ban \
                --field="$( eval_gettext "Disks utility" ):chk" $_app_disknotify \
                --field="$( eval_gettext "Use SYSTEMD (recommended). Uncheck to switch to SYSVINIT." ):chk" $_app_systemd \
                --field="$( printf "$( eval_gettext "Directly access this machine as '%s' on your local network" )" "${hostname}.local" ):chk" $_app_nssmdns \
                --button="gtk-ok" > "$TMP_FILE_1_f" || true

        fi

        #--field="$( eval_gettext "Printer support" ):chk" "$(if LC_ALL="$EL_LC_EN" lpstat -s | grep -v "system default destination: PDF" | grep -qs "system default destination:" ; then echo TRUE ; else echo FALSE ; fi )" \

        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "FALSE" ; then
            echo -e "blueman\ngnome-bluetooth\nbluez-.*\nbluez\nobex-data-server" >> "/tmp/.packages-to-remove.tmp" || true
        else
            is_bluetooth_wanted=1
        fi
        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "FALSE" && echo -e "cups\ncups-.*\nfoomatic-.*\ncups-driver-.*\nprinter-driver-.*\nlibcupsimage2\nhplip\nhp-ppd\nopenprinting-ppds\npython-cups\nghostscript-cups" >> "/tmp/.packages-to-remove.tmp" || true
        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "FALSE" && echo -e "sane-utils\nlibsane\nlibsane-.*\nsimple-scan" >> "/tmp/.packages-to-remove.tmp" || true
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $4}' | grep -qs "FALSE" ; then
            rm -rf "$TARGET/var/spool/samba" 2>/dev/null || true
            echo -e "samba\nsmb-admin\nsmb-browser\nsamba-common" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' | grep -qs "FALSE" ; then
            echo -e "ssh\nopenssh-server" >> "/tmp/.packages-to-remove.tmp" || true
        else
            echo -e "ssh\nopenssh-server" >> "/tmp/.packages-to-install.tmp" || true
            if ((is_virtualized)) ; then
                is_ssh_password_enabled=1
                is_ssh_root_enabled=1
            else
                if guitool yesnoquestioninvbutcontinue "$( eval_gettext "Would you like to enable password-based SSH logins? This is the default on many systems, but it carries a security risk due to brute force attacks. If you disable this option, you will no longer be able to log in using passwords; you will need to use SSH keys instead." )" ; then
                    is_ssh_password_enabled=1
                fi
            fi
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $6}' | grep -qs "TRUE" ; then
            # with systemd?
            if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $8}' | grep -qs "TRUE" ; then
                echo -e "fail2ban\npython3-systemd" >> "/tmp/.packages-to-install.tmp" || true
            else
                echo -e "fail2ban" >> "/tmp/.packages-to-install.tmp" || true
            fi
        else
            echo -e "fail2ban" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $7}' | grep -qs "FALSE" && echo -e "gpm" >> "/tmp/.packages-to-remove.tmp" || true
        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $7}' | grep -qs "FALSE" && echo -e "gnome-disk-utility" >> "/tmp/.packages-to-remove.tmp" || true
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $8}' | grep -qs "FALSE" ; then
            if ((is_apple)) ; then
                guitool warning "$( eval_gettext "It seems removing SystemD is not compatible with Mac machines" )"
            fi
            if guitool yesnoquestioninvbutcontinue "$( eval_gettext "Warning: You have chosen to replace systemd with sysvinit. This should only be done by advanced users who understand the implications and purpose of this action. It is crucial to know how to resolve any system bugs that may arise from this significant change. We encourage you to share your experiences in our forums or provide any information that can help enhance this feature for improved stability. Are you sure you wish to switch to sysvinit?" )" ; then

                is_sysvinit_wanted=1
            fi
        fi
        #if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $9}' | grep -qs "TRUE" ; then
            #echo -e "libnss-mdns" >> "/tmp/.packages-to-install.tmp"
        #else
            #echo -e "libnss-mdns" >> "/tmp/.packages-to-remove.tmp"
        #fi

        rm -f "$TMP_FILE_1_f"


        # }}}
        # auto: remove unneeded firmwares / drivers {{{
        #if ((conf_configure_options)) ; then
            #yad --form --center --width=460 --title="$( eval_gettext "Features of Elive:" )" --image=gtk-harddisk --image-on-top \
                #--text="$( eval_gettext "Select the special hardware features you want to have installed. By deselecting them you can make your system lighter." )" \
                #--field="$( eval_gettext "Firmware for Broadcom NetXtremeII network" ):chk" "$( if lsmod | grep -qs "bnx2" ; then echo TRUE ; else echo FALSE ; fi )" \
                #--field="$( eval_gettext "Firmware for Qlogic QLA2xxx fibre channel" ):chk" "$( if lsmod | grep -qs "qla2" ; then echo TRUE ; else echo FALSE ; fi )" \
                #--field="$( eval_gettext "PCMCIA support" ):chk" FALSE \
                #--button="gtk-ok" > "$TMP_FILE_1_f" || true
        #fi

        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "FALSE" && echo -e "firmware-bnx2" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "FALSE" && echo -e "firmware-qlogic" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "FALSE" && echo -e "pcmciautils" >> "/tmp/.packages-to-remove.tmp" || true
        #rm -f "$TMP_FILE_1_f"

        if ! lsmod | grep -qs "bnx2" ; then
            echo -e "firmware-bnx2" >> "/tmp/.packages-to-remove.tmp"
        fi
        if ! lsmod | grep -qs "qla2" ; then
            echo -e "firmware-qlogic" >> "/tmp/.packages-to-remove.tmp"
        fi
        if [[ "$RAM_TOTAL_SIZE_mb" -gt "1200" ]] || [[ "$CPU_MHZ_CORE_v" -gt 1200 ]]  ; then
            echo -e "pcmciautils" >> "/tmp/.packages-to-remove.tmp"
        fi

        if ((is_i386)) || ((is_virtualized)) ; then
            _app_gnomesoftware="FALSE"
        else
            _app_gnomesoftware="TRUE"
        fi


        # - firmwares and hardware }}}
        # select applications {{{
        # defaults {{{
        _app_blender="FALSE"
        _app_mediacenter="TRUE"
        _app_office="TRUE"
        _app_thunderbird="FALSE"
        _app_calibre="TRUE"
        _app_gamesemulators="TRUE"
        if [[ -e /var/lib/dpkg/info/virtualbox-dkms.list ]] ; then
            _app_virtualbox="TRUE"
        else
            _app_virtualbox="FALSE"
        fi

        if { [[ "$RAM_TOTAL_SIZE_mb" -lt 1350 ]] || [[ "$CPU_MHZ_CORE_v" -lt 800 ]] ; } && ! ((is_virtualized))  ; then
            _app_mediacenter="FALSE"
            _app_thunderbird="TRUE"
        else
            _app_mediacenter="TRUE"
        fi

        if ((is_virtualized)) ; then
            _app_gamesemulators="FALSE"
            _app_blender="FALSE"
            _app_mediacenter="FALSE"
            _app_virtualbox="FALSE"
            _app_calibre="FALSE"
            _app_inkscape="FALSE"
            _app_office="FALSE"
            _app_gimp="FALSE"
        fi

        if ((is_i386)) ; then
            _app_thunderbird="TRUE"
            _app_calibre="FALSE"
        fi

        # - defaults }}}
        if ! ((conf_configure_options)) ; then
            # XXX non-interactive:
            echo "${_app_gnomesoftware}|${_app_blender}|${_app_mediacenter}|${_app_office}|${_app_inkscape}|${_app_gimp}|${_app_thunderbird}|${_app_calibre}|${_app_gamesemulators}|${_app_virtualbox}" > "$TMP_FILE_1_f"
        else
            yad --form --center --width=460 --title="$( eval_gettext "Features of Elive:" )" --image=synaptic --image-on-top \
                --text="$( eval_gettext "Choose the applications you want to have. By deselecting, you can save disk space, but they will not be available." )" \
                --field="$( eval_gettext "Software Center: Switch to a more comprehensive and user-friendly one, support Flatpaks, though it consumes more resources." ):chk" $_app_gnomesoftware \
                --field="Blender $( eval_gettext "3D Modelling" ):chk" $_app_blender \
                --field="XBMC / Kodi $( eval_gettext "Media center" ):chk" $_app_mediacenter \
                --field="$( eval_gettext "Office applications" ) (libreoffice):chk" $_app_office \
                --field="Inkscape $( eval_gettext "Vector drawing" ):chk" $_app_inkscape \
                --field="Gimp $( eval_gettext "Photoshop equivalent" ):chk" $_app_gimp \
                --field="Thunderbird / Sylpheed $( eval_gettext "Desktop email application" ):chk" $_app_thunderbird \
                --field="Calibre: $( eval_gettext "Reader of Ebooks, transfer and manager of collections" ):chk" $_app_calibre \
                --field="$( eval_gettext "Games and emulators" ) (50 MB):chk" $_app_gamesemulators \
                --field="VirtualBox $( eval_gettext "Virtual computer emulator" ) (50 MB):chk" $_app_virtualbox \
                --field="$( eval_gettext "Note: You can manually remove more software later in your installed system." ):lbl" TRUE \
                --button="gtk-ok" > "$TMP_FILE_1_f" || true
        fi

        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" ; then
            is_software_center_wanted=1
        fi

        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "TRUE" ; then
            echo -e "blender" >> "/tmp/.packages-to-install.tmp" || true
        else
            echo -e "blender\nblender-data\nblender-last" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "TRUE" ; then
            echo -e "kodi" >> "/tmp/.packages-to-install.tmp" || true
        else
            echo -e "kodi\nxbmc\nxbmc-.*" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $4}' | grep -qs "TRUE" ; then
            echo -e "libreoffice" >> "/tmp/.packages-to-install.tmp" || true
        else
            echo -e "libreoffice\nlibreoffice-.*\nlibrecad\nscribus\nscribus-data" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' | grep -qs "TRUE" ; then
            echo -e "inkscape" >> "/tmp/.packages-to-install.tmp" || true
        else
            echo -e "inkscape" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $6}' | grep -qs "TRUE" ; then
            echo -e "gimp\ngimp-ufraw\ngimp-plugin-registry\ngimp-lensfun\ngimp-gap\ngimp-texturize\ngtkam-gimp\ngimp-gmic" >> "/tmp/.packages-to-install.tmp" || true
        else
            echo -e "gimp\ngimp-.*\ngtkam-gimp" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        # seems like in buster, icedove (dependency of enigmail) is not available
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $7}' | grep -qs "TRUE" ; then
            if ((is_i386)) ; then
                echo -e "sylpheed\nsylpheed-plugins" >> "/tmp/.packages-to-install.tmp" || true
            else
                if ((is_buster)) ; then
                    echo -e "thunderbird\nenigmail\nxul-ext-gcontactsync" >> "/tmp/.packages-to-install.tmp" || true
                else
                    echo -e "thunderbird\nenigmail" >> "/tmp/.packages-to-install.tmp" || true
                fi
            fi
        else
            echo -e "thunderbird.*\nenigmail\nxul-ext-gcontactsync\nsylpheed\nsylpheed-plugins" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $8}' | grep -qs "TRUE" ; then
            echo -e "calibre" >> "/tmp/.packages-to-install.tmp" || true
        else
            echo -e "calibre" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $9}' | grep -qs "FALSE" ; then
            #echo -e "$( COLUMNS=1000 dpkg -l | grep -E "^(hi|ii)" | awk '{print $2}' | grep "gnome-icon-theme-" | grep -v "^gnome-icon-theme$" )" >> "/tmp/.packages-to-remove.tmp" || true
            for i in $( dpkg -S /usr/games/ | tr ',' '\n' | grep -v "^/usr/games$" | grep -vE "(patchutils|base-files|cowsay|cowthink)" | sed -e 's|:.*$||g' ) ; do
                echo "$i" >> "/tmp/.packages-to-remove.tmp"
            done
            echo "uplinkdemo" >> "/tmp/.packages-to-remove.tmp"

            # just a little fallback, for things like dosbox
            echo -e "scummvm\nscummvm-data\ndosbox\nfury-furries\npacman-arena\npcsxr\nvisualboyadvance\nzsnes\nsteam-debian\ntorus-trooper\nfoobillard\nsupermariowar\nsupermariowar-data\nretroarch\nlutris" >> "/tmp/.packages-to-remove.tmp" || true
            # extra ones needed
            echo -e "steam\nsteam:i386\nsteam-installer\nsyasokoban\nsyasokoban:i386\nhigan\nsupermariowar:i386\nsupermariowar:amd64\nzsnes:i386\nsupermariowar-data:i386\nsupermariowar-data:amd64" >> "/tmp/.packages-to-remove.tmp" || true
            # just a little beautifier bad back:
            rm -f "$TARGET/usr/share/xdgeldsk/applications/game-and-watch.desktop" 2>/dev/null || true
            rm -f "$TARGET/usr/share/xdgeldsk/applications/msdos-games-archiveorg.desktop" 2>/dev/null || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $10}' | grep -qs "TRUE" && ! ((is_virtualbox)) ; then
            echo -e "virtualbox\nvirtualbox-qt" >> "/tmp/.packages-to-install.tmp" || true
            # Fix to remove a useless conf file before to install virtualbox
            rm -f "$TARGET/etc/default/virtualbox"
            # in bookworm seems like it is needed to load the module manually
            echo -e "vboxdrv" >> "$TARGET/etc/modules" || true
        else
            echo -e "virtualbox\nvirtualbox-qt" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        rm -f "$TMP_FILE_1_f"


        # }}}
        # select designs {{{
        # defaults {{{
        _app_themes="TRUE"
        #_app_themes_e16="TRUE"

        if ((is_e16)) ; then
            _app_themes_e16="TRUE"
        else
            _app_themes_e16="FALSE"
        fi

        if ((is_virtualized)) ; then
            _app_themes="FALSE"
            _app_themes_e16="FALSE"
        fi

        # always suggest to reinstall the desktop
        if ((is_thanatests)) ; then
            _app_desktop_reinstall="FALSE"
        else
            _app_desktop_reinstall="TRUE"
        fi

        _app_user_confs_reinstall="FALSE"

        if ((is_retrowave)) ; then
            local message_default_designs
            message_default_designs="$( printf "$( eval_gettext "Install the default Elive designs instead of those in this version." )" "" )"

            _app_designs_defaults="FALSE"
        else
            _app_designs_defaults="TRUE"
        fi


        # - defaults }}}
        if ! ((conf_configure_options)) ; then
            # XXX non-interactive:
            echo "${_app_themes}|${_app_themes_e16}|${_app_desktop_reinstall}|${_app_user_confs_reinstall}|${_app_designs_defaults}" > "$TMP_FILE_1_f"
        else

            if ((is_retrowave)) ; then
                yad --form --center --width=460 --title="$( eval_gettext "Desktop customizations and designs" )" --image=desktop-effects --image-on-top \
                    --text="$( eval_gettext "Include specially selected designs and customizations" )" \
                    --field="$( eval_gettext "Install recommended additional themes and icons" ) (400 MB):chk" $_app_themes \
                    --field="$( eval_gettext "Install additional E16 desktop themes" ) (50 MB):chk" $_app_themes_e16 \
                    --field="$( eval_gettext "Update my desktop configurations with new features" ):chk" $_app_desktop_reinstall \
                    --field="$( eval_gettext "Update my user configurations (recommended)" ):chk" $_app_user_confs_reinstall \
                    --field="${message_default_designs}:chk" $_app_designs_defaults \
                    --button="gtk-ok" > "$TMP_FILE_1_f" || true

            else
                yad --form --center --width=460 --title="$( eval_gettext "Desktop customizations and designs" )" --image=desktop-effects --image-on-top \
                    --text="$( eval_gettext "Include specially selected designs and customizations" )" \
                    --field="$( eval_gettext "Install recommended additional themes and icons" ) (400 MB):chk" $_app_themes \
                    --field="$( eval_gettext "Install additional E16 desktop themes" ) (50 MB):chk" $_app_themes_e16 \
                    --field="$( eval_gettext "Update my desktop configurations with new features" ):chk" $_app_desktop_reinstall \
                    --field="$( eval_gettext "Update my user configurations (recommended)" ):chk" $_app_user_confs_reinstall \
                    --button="gtk-ok" > "$TMP_FILE_1_f" || true

            fi
        fi


        # icons & gtk
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" ; then
            # Icon themes
            echo -e "flatwoken-icon-theme\ngnome-icon-theme-beautyline\nelementary-icon-theme\ngnome-icon-theme-emerald\ngnome-icon-theme-faenza\ngnome-icon-theme-fs-icons\ngnome-icon-theme-meliae\ngnome-icon-theme-sweet-blue\ngnome-icon-theme-sweet-purple\ngnome-icon-theme-sweet-rainbow\nluv-icon-theme\npaper-icon-theme" >> "/tmp/.packages-to-install.tmp"
            # GTK themes:
            echo -e "gnome-themes-retrowave\ngnome-themes-azure\ngnome-themes-qogir\ngnome-themes-sweet-dark\nadapta-gtk-theme" >> "/tmp/.packages-to-install.tmp"
        else
            # Icon themes
            echo -e "$( COLUMNS=1000 dpkg -l | grep -E "^(hi|ii)" | awk '{print $2}' | grep -E "(gnome-icon-theme-|\-icon-theme$)" | grep -vE "^(gnome-icon-theme|adwaita-icon-theme|hicolor-icon-theme|beautyline|sweet)$" )" >> "/tmp/.packages-to-remove.tmp" || true
            # GTK themes:
            echo -e "$( COLUMNS=1000 dpkg -l | grep -E "^(hi|ii)" | awk '{print $2}' | grep -E "(\-themes?-|\-gtk-theme$)" | grep -vE "(e16|elive|gnome-icon-theme|gnome-themes-extra|darkone|retrowave|elivedarK|sound-theme-freedesktop)" )" >> "/tmp/.packages-to-remove.tmp" || true

        fi

        # e16 themes
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "TRUE" ; then
            # E16 themes:
            echo -e "e16-theme-retrowave\ne16-theme-23ozglass\ne16-theme-black-e\ne16-theme-bluesteel\ne16-theme-detroit\ne16-theme-e13\ne16-theme-evolution\ne16-theme-lcars\ne16-theme-orangejuice" >> "/tmp/.packages-to-install.tmp"
        else
            # E16 themes:
            echo -e "$( COLUMNS=1000 dpkg -l | grep -E "^(hi|ii)" | awk '{print $2}' | grep -E "e16-themes?-" | grep -vE "(elive|gnome-icon-theme|gnome-themes-extra|darkone|retrowave|elivedarK|sound-theme-freedesktop)" )" >> "/tmp/.packages-to-remove.tmp" || true
        fi

        # reinstall desktop
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "TRUE" ; then
            is_desktop_upgrade_wanted=1
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $4}' | grep -qs "TRUE" ; then
            is_user_confs_upgrade_wanted=1
        fi

        # force updates for migrations:
        if ((is_migration_otheros)) ; then
            is_desktop_upgrade_wanted=1
            is_user_confs_upgrade_wanted=1
        fi

        # switch to default elive designs
        if ((is_retrowave)) ; then
            if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' | grep -qs "TRUE" ; then
                echo -e "elive-skel-default-all\nelive-skel-default-e16" >> "/tmp/.packages-to-install.tmp"
                is_gtk_theme_default_wanted=1
            fi
        fi

        rm -f "$TMP_FILE_1_f"

        # fixes:
        # do not uninstall special elive version designs:
        if ((is_retrowave)) ; then
            sed -i -e '/retrowave/d' "/tmp/.packages-to-remove.tmp"
            sed -i -e '/sweet/d' "/tmp/.packages-to-remove.tmp"
            sed -i -e '/beautyline/d' "/tmp/.packages-to-remove.tmp"
        fi


        # }}}
        # select browsers {{{
        # defaults {{{
        _app_netsurf="FALSE"
        _app_links2="FALSE"

        if which chromium 1>/dev/null 2>&1 ; then
            _app_chromium="TRUE"
        else
            _app_chromium="FALSE"
        fi
        if which google-chrome 1>/dev/null 2>&1 ; then
            _app_chrome="TRUE"
        else
            _app_chrome="FALSE"
        fi
        if ((is_bookworm)) && [[ "$_app_chromium" = "TRUE" ]] ; then
            _app_chrome="FALSE"
        fi

        _app_opera="FALSE"
        _app_dillo="FALSE"
        _app_palemoon="FALSE"


        if ((is_i386)) || [[ "$RAM_TOTAL_SIZE_mb" -lt "4800" ]] ; then
            _app_opera="FALSE"
            _app_palemoon="TRUE"
            _app_netsurf="TRUE"
        fi

        if ((is_wheezy)) ; then
            _app_opera="TRUE"
            _app_chrome="TRUE"
            _app_palemoon="FALSE"
        else
            _app_opera="FALSE"
        fi

        _app_firefox="FALSE"

        # virtualmachines can make chrom* fail, so let's use firefox instead
        if ((is_virtualized)) ; then
            # _app_chrome="FALSE"
            # _app_chromium="FALSE"
            # _app_firefox="TRUE"

            _app_palemoon="FALSE"
            _app_netsurf="FALSE"
        fi


        # - defaults }}}
        if ! ((conf_configure_options)) ; then
            # XXX non-interactive:
            echo "${_app_chromium}|${_app_chrome}|${_app_firefox}|${_app_opera}|${_app_palemoon}|${_app_netsurf}|" > "$TMP_FILE_1_f"
        else
            yad --form --center --width=460 --title="$( eval_gettext "Features of Elive:" )" --image=web-browser --image-on-top \
                --text="$( eval_gettext "Choose the web browsers you want to use. One or the other works better on different computers, and some alternatives use fewer resources." )" \
                --field="$( eval_gettext "Google Chromium (Open Source version of Chrome)" ):chk" $_app_chromium \
                --field="$( eval_gettext "Google Chrome (Official version)" ) (64bit only, not opensource):chk" $_app_chrome \
                --field="$( eval_gettext "Firefox (Slower, consumes less RAM, newer)" ):chk" $_app_firefox\
                --field="$( eval_gettext "Opera (Alternative browser, fast)" ) (64bit only, not opensource):chk" $_app_opera \
                --field="$( eval_gettext "Palemoon, a truly faster alternative, supports old Firefox extensions" ):chk" $_app_palemoon \
                --field="$( eval_gettext "Netsurf (extremely low resources, with menus, no JavaScript)" ):chk" $_app_netsurf \
                --field="Links2 $( eval_gettext "extremely low resources, CLI featured, no JavaScript" ):chk" $_app_links2 \
                --button="gtk-ok" > "$TMP_FILE_1_f" || true

                # surf takes almost the same as iceweasel, so don't use it by deafult
                #--field="$( eval_gettext "Google Chromium (Open Source version)" ):chk" "$( if lsmod | grep -qs "^nvidia" ; then echo FALSE ; else echo TRUE ; fi )" \
                #--field="$( eval_gettext "Surf (low resources, no menus, some JavaScript features)" ):chk" FALSE \
        fi


        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" ; then
            echo -e "chromium\nchromium-browser\nchromium-l10n" >> "/tmp/.packages-to-install.tmp"
        else
            echo -e "chromium\nchromium-browser\nchromium-l10n" >> "/tmp/.packages-to-remove.tmp"
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "TRUE" ; then
            if ((is_amd64)) ; then
                echo -e "google-chrome-stable" >> "/tmp/.packages-to-install.tmp"
                _app_chrome="TRUE"
            else
                guitool warning "$( eval_gettext "The next package/software you selected is only available for 64-bit, so we cannot include it. If you want this software, you should use the 64-bit version of Elive." )\n\nGoogle Chrome" &
                rm -f /etc/apt/sources.list.d/google-chrome.list 2>/dev/null || true
                # but include chromium as an alternative if was not selected yet
                if ! cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" ; then
                    echo -e "chromium\nchromium-browser\nchromium-l10n" >> "/tmp/.packages-to-install.tmp"
                fi
            fi
        else
            echo -e "google-chrome-stable" >> "/tmp/.packages-to-remove.tmp"
            # remove the repo too from the system
            rm -f /etc/apt/sources.list.d/google-chrome.list 2>/dev/null || true
            _app_chrome="FALSE"
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "TRUE" ; then
            echo -e "firefox-esr" >> "/tmp/.packages-to-install.tmp"
        else
            echo -e "firefox-esr\nfirefox-.*" >> "/tmp/.packages-to-remove.tmp"
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $4}' | grep -qs "TRUE" ; then
            if ((is_amd64)) ; then
                echo -e "opera-stable" >> "/tmp/.packages-to-install.tmp" || true
            else
                guitool warning "$( eval_gettext "The next package/software you selected is only available for 64-bit, so we cannot include it. If you want this software, you should use the 64-bit version of Elive." )\n\nOpera" &
            fi
        else
            echo -e "opera-stable" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' | grep -qs "TRUE" ; then
            echo -e "palemoon" >> "/tmp/.packages-to-install.tmp" || true
        else
            echo -e "palemoon" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $6}' | grep -qs "TRUE" ; then
            echo -e "netsurf-common\nnetsurf-gtk" >> "/tmp/.packages-to-install.tmp" || true
        else
            echo -e "netsurf-common\nnetsurf-gtk" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $7}' | grep -qs "TRUE" ; then
            echo -e "links2" >> "/tmp/.packages-to-install.tmp" || true
        else
            echo -e "links2" >> "/tmp/.packages-to-remove.tmp" || true
        fi
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $6}' | grep -qs "FALSE" && echo -e "" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $7}' | grep -qs "FALSE" && echo -e "" >> "/tmp/.packages-to-remove.tmp" || true
        rm -f "$TMP_FILE_1_f"


        # }}}
        # free up space {{{
        # defaults {{{
        _app_demofiles="FALSE"
        _app_compile="FALSE"
        _app_mandb="FALSE"
        _app_fonts="FALSE"
        # this is a bit huge and not so needed, only for davinci resolve
        _app_opencl="FALSE"
        # uninstall extra fonts if we know we don't need them
        if echo "$LANG" | grep -qsiE "^(ca|de|en|es|fr|fi|gl|it|nl|pt|ro)_" ; then
            # but only if is a low end machine, we don't want broken chars in the OS
            # TODO: we need a checker of hard disk size, this depends more of the size on disk than in resources
            if [[ "$RAM_TOTAL_SIZE_mb" -lt 1700 ]] || [[ "$CPU_MHZ_CORE_v" -lt 1200 ]] ; then
                _app_fonts="TRUE"
            fi
        fi

        if ((is_virtualized)) ; then
            _app_demofiles="TRUE"
            _app_compile="TRUE"
            _app_mandb="TRUE"
        fi
        if ((is_i386)) ; then
            _app_mandb="TRUE"
            _app_languages="TRUE"
            _app_opencl="TRUE"
        else
            _app_languages="FALSE"
        fi
        # - defaults }}}
        if ! ((conf_configure_options)) ; then
            # XXX non-interactive:
            echo "${_app_languages}|${_app_fonts}|${_app_mandb}|FALSE|${_app_demofiles}|${_app_compile}|FALSE|TRUE|${_app_opencl}" > "$TMP_FILE_1_f"
        else
            yad --form --center --width=460 --title="$( eval_gettext "Features of Elive:" )" --image=gnome-devel --image-on-top \
                --text="$( eval_gettext "Choose some special features to make your Elive system lighter." )" \
                --field="$( eval_gettext "Remove translations of unnecessary languages" ) (450 MB):chk" $_app_languages \
                --field="$( eval_gettext "Remove unnecessary international fonts" ):chk" $_app_fonts \
                --field="$( eval_gettext "Remove manpages (makes APT faster)" ):chk" $_app_mandb \
                --field="$( eval_gettext "Remove the directory:" ) /usr/share/doc (210 MB):chk" FALSE \
                --field="$( eval_gettext "Remove Elive demo files" ) (400 MB):chk" $_app_demofiles \
                --field="$( eval_gettext "Remove dependencies used for compilation" ) (240 MB):chk" $_app_compile \
                --field="$( eval_gettext "Remove drivers compilation feature (not recommended)" ) (200 MB):chk" FALSE \
                --field="$( eval_gettext "Remove debugging packages" ) (30 MB):chk" TRUE \
                --field="$( eval_gettext "Remove OpenCL acceleration (for advanced GPU processing)" ) (530 MB):chk" $_app_opencl \
                --button="gtk-ok" > "$TMP_FILE_1_f" || true
        fi

        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" && echo -e "localepurge" >> "/tmp/.packages-to-install.tmp" || true
        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "TRUE" && echo -e "fonts-lklug-sinhala\nfonts-farsiweb\nfonts-khmeros\nfonts-lao\nfonts-lohit-telu\nfonts-samyak\nfonts-sil-abyssinica\nfonts-sil-padauk\nfonts-smc\nfonts-tibetan-machine\nfonts-tlwg-.*\nfonts-ukij-uyghur\nfonts-indic\nfonts-knda\nfonts-telu" >> "/tmp/.packages-to-remove.tmp" || true

        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "TRUE" ; then
            echo -e "man-db-notinstalled" >> "/tmp/.packages-to-install.tmp"
            is_cleanup_wanted_manpages=1
        fi
        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $4}' | grep -qs "TRUE" && is_cleanup_wanted_dir_usrsharedoc=1 || true


        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' | grep -qs "TRUE" && is_cleanup_wanted_dir_elivedemofiles=1 || true
        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $6}' | grep -qs "TRUE" && echo -e "ccache\nautopoint\nautoconf.*\nlibefl-dev\nlibc6-dev\ne17-stable-dev\nenlightenment-dev\nautotools-dev\nxorg-dev\nlibxext-dev\nx11proto-core-dev\nlibxml2-dev\nxorg-dev" >> "/tmp/.packages-to-remove.tmp" || true
        # Warning: these next dependencies are verified that doesn't break on wheezy, but we don't know in the future!
        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $7}' | grep -qs "TRUE" && echo -e "g++\ngcc\ngcc-4.6-base\ngcc-4.7\nlibdbus-1-dev\nlibxml2-dev\nlibc6-dev\nlinux-compiler-.*\nlinux-headers.*\nlinux-kbuild.*linux-libc.*\nkbuild\nlibc-dev-bin" >> "/tmp/.packages-to-remove.tmp" || true
        cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $8}' | grep -qs "TRUE" && echo -e "gdb\nvalgrind\nlibefl1-dbg\ne17-stable-dbg\nenlightenment-dbgsym\nterminology-dbg\n$( COLUMNS=1000 dpkg -l | grep "^ii.*lib.*-dbg" | awk '{print $2}' | grep "dbg$" | sed -e 's|:.*$||g' )" >> "/tmp/.packages-to-remove.tmp" || true
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $9}' | grep -qs "TRUE" ; then
            echo -e "intel-opencl-icd\nbeignet-opencl-icd\nmesa-opencl-icd" >> "/tmp/.packages-to-remove.tmp" || true
        else
            echo -e "mesa-opencl-icd" >> "/tmp/.packages-to-install.tmp" || true
            # extra package for intel cards:
            if inxi -G -c 0 | grep -qsiE "(Intel|i915)" ; then
                echo -e "intel-opencl-icd" >> "/tmp/.packages-to-install.tmp" || true
            fi
        fi
        rm -f "$TMP_FILE_1_f"


        # }}}
        # very low resources {{{
        # defaults {{{
        _app_alsa="FALSE"
        _app_abiword="TRUE"

        if ((is_virtualized)) ; then
            _app_abiword="FALSE"
        fi
        # always set it for lightness if this menu opens
        _app_gnomepackagekit="TRUE"

        # - defaults }}}
        if [[ "$RAM_TOTAL_SIZE_mb" -lt "1850" ]] || [[ "$CPU_MHZ_CORE_v" -lt 1550 ]]  ; then
            yad --form --center --width=460 --title="$( eval_gettext "Computer with limited resources" )" --image=weather-few-clouds --image-on-top \
                --text="$( eval_gettext "Modifications making the system faster will sacrifice some features. Keeping the defaults is suggested." )" \
                --field="$( eval_gettext "Replace the office suite with one that takes fewer resources?" ) (abiword/gnumeric):chk" $_app_abiword \
                --field="$( eval_gettext "Replace the Software-center with a much lighter alternative for your RAM" ):chk" $_app_gnomepackagekit \
                --field="$( eval_gettext "Switch from" ) Zsh to Bash:chk" FALSE \
                --field="$( eval_gettext "Switch Terminal" ) Terminology to Urxvt:chk" FALSE \
                --field="$( eval_gettext "Uninstall" ) Network-Manager $( eval_gettext "(not recommended)" ):chk" FALSE \
                --field="$( eval_gettext "Remove login manager (not recommended)" ):chk" FALSE \
                --field="$( eval_gettext "Remove Pulseaudio (use Alsa)" ), +33 MB RAM:chk" $_app_alsa \
                --field="$( eval_gettext "Uninstall" ) command-not-found:chk" FALSE \
                --button="gtk-ok" > "$TMP_FILE_1_f" || true

            if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" ; then
                echo -e "abiword\ngnumeric\nabiword-plugin-grammar" >> "/tmp/.packages-to-install.tmp"
                echo -e "libreoffice\nlibreoffice-.*" >> "/tmp/.packages-to-remove.tmp"
                echo -e "librecad\nscribus\nscribus-data" >> "/tmp/.packages-to-remove.tmp"

                # other dialogs added it:
                sed -i -e '/libreoffice/d' "/tmp/.packages-to-install.tmp"
            fi
            if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "TRUE" ; then
                unset is_software_center_wanted
            fi
            if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "TRUE" ; then
                sed -i "s|:/bin/zsh|:/bin/bash|g" "$TARGET/etc/passwd"
                sed -i "s|:/bin/zsh|:/bin/bash|g" "$TARGET/etc.old-to-${eliveversion}/passwd" 2>/dev/null || true
                sed -i "/^DSHELL=/s|zsh|bash|" "$TARGET/etc/adduser.conf"
                is_shell_in_bash_wanted=1
            fi
            if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $4}' | grep -qs "TRUE" ; then
                chroot "$TARGET" update-alternatives --set x-terminal-emulator /usr/bin/urxvt 1>/dev/null 2>&1 || true
            else
                chroot "$TARGET" update-alternatives --set x-terminal-emulator /usr/bin/terminology 1>/dev/null 2>&1 || true
            fi
            cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' | grep -qs "TRUE" && echo -e "network-manager\nstrongswan*" >> "/tmp/.packages-to-remove.tmp" || true

            if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $6}' | grep -qs "TRUE" ; then
                echo -e "lightdm\nlightdm-gtk-greeter" >> "/tmp/.packages-to-remove.tmp"
                is_lightdm_disable_wanted=1
                echo -e "echo -e \"\n$( eval_gettext "To log into the graphical system, run:" ) startx\"" >> "$TARGET/etc/profile.d/login_message_howto.sh"
                echo -e "echo -e \"\n$( eval_gettext "To log into the graphical system, run:" ) startx\"" >> "$TARGET/etc/zsh/zprofile.d/login_message_howto.sh"
            fi
            if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $7}' | grep -qs "TRUE" ; then
                if ((is_bluetooth_wanted)) ; then
                    if guitool yesnoquestioninvbutcontinue "$( eval_gettext "Bluetooth support selected, but not PulseAudio. For Bluetooth speaker connection, PulseAudio is needed. Not having it saves RAM. Remove PulseAudio?" )" ; then
                        echo -e "pulseaudio-compat-notinstalled" >> "/tmp/.packages-to-install.tmp" || true
                        echo -e "pavucontrol\npulseaudio\npulsemixer" >> "/tmp/.packages-to-remove.tmp" || true
                    fi
                else
                    echo -e "pulseaudio-compat-notinstalled" >> "/tmp/.packages-to-install.tmp" || true
                    echo -e "pavucontrol\npulseaudio\npulsemixer" >> "/tmp/.packages-to-remove.tmp" || true
                fi
            fi
            if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $8}' | grep -qs "FALSE" ; then
                # make the system featured with command-not-found which also includes apt-file dependency
                echo "command-not-found" >> "/tmp/.packages-to-install.tmp"
            else
                sed -i -e '/command-not-found/d' "/tmp/.packages-to-install.tmp"
                echo -e "command-not-found" >> "/tmp/.packages-to-remove.tmp"
            fi

            rm -f "$TMP_FILE_1_f"
        else
            # make sure that we always have terminology set as default terminal (when this dialog is not run)
            if [[ -x "/usr/bin/terminology" ]] ; then
                chroot "$TARGET" update-alternatives --set x-terminal-emulator /usr/bin/terminology 1>/dev/null 2>&1
            fi
        fi



        # }}}
        # compatibility / deprecated {{{
        # elive features
        # Update: this feature is deprecated, but let's continue use it in older versions of elive:
        if ! grep -qs "sudoers.d/sudo_nopasswd_" "/usr/sbin/user-manager" ; then

            if grep -qs "first-user:" "$TARGET/etc.old-to-${eliveversion}/elive-version" 2>/dev/null ; then
                user_first="$( cat "$TARGET/etc.old-to-${eliveversion}/elive-version" | grep "first-user:" | sed -e 's|^.*first-user: ||g' | tail -1 )"
            fi
            if [[ -z "$user_first" ]] ; then
                username_user_ask
                user_first="$username"
            fi

            if [[ -n "$user_first" ]] ; then
                yad --form --center --width=460 --title="$( eval_gettext "Elive extra features" )" --image=logo-elive --image-on-top \
                    --text="$( eval_gettext "Choose extra custom features by Elive" )" \
                    --field="$( eval_gettext "Never ask password for 'SUDO' (not recommended)" ):chk" FALSE \
                    --field="$( eval_gettext "Do not ask for a password for package management" ):chk" TRUE \
                    --button="gtk-ok" > "$TMP_FILE_1_f" || true

                if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" ; then
                    rm -f "$TARGET/etc/sudoers.d/sudo_nopasswd_all_${user_first}"
                    echo "$user_first ALL=NOPASSWD: ALL" > "$TARGET/etc/sudoers.d/sudo_nopasswd_all_${user_first}"
                    chmod 0440 "$TARGET/etc/sudoers.d/sudo_nopasswd_all_${user_first}"
                fi
                if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "TRUE" ; then
                    rm -f "$TARGET/etc/sudoers.d/sudo_nopasswd_packagers_${user_first}"
                    echo "$user_first ALL=(ALL) NOPASSWD: /usr/bin/apt" >> "$TARGET/etc/sudoers.d/sudo_nopasswd_packagers_${user_first}"
                    echo "$user_first ALL=(ALL) NOPASSWD: /usr/bin/apt-get" >> "$TARGET/etc/sudoers.d/sudo_nopasswd_packagers_${user_first}"
                    echo "$user_first ALL=(ALL) NOPASSWD: /usr/bin/apt-cache" >> "$TARGET/etc/sudoers.d/sudo_nopasswd_packagers_${user_first}"
                    echo "$user_first ALL=(ALL) NOPASSWD: /usr/bin/dpkg" >> "$TARGET/etc/sudoers.d/sudo_nopasswd_packagers_${user_first}"
                    echo "$user_first ALL=(ALL) NOPASSWD: /usr/bin/aptitude" >> "$TARGET/etc/sudoers.d/sudo_nopasswd_packagers_${user_first}"
                    echo "$user_first ALL=(ALL) NOPASSWD: /usr/sbin/synaptic" >> "$TARGET/etc/sudoers.d/sudo_nopasswd_packagers_${user_first}"
                    chmod 0440 "$TARGET/etc/sudoers.d/sudo_nopasswd_packagers_${user_first}"
                fi

                rm -f "$TMP_FILE_1_f"
            fi
        fi


        # - compatibility / deprecated }}}
        # remove closed-source software {{{
        # defaults {{{
        _app_skype="FALSE"
        _app_steam="FALSE"
        _app_dropbox="FALSE"
        if ((is_amd64)) && ! ((is_virtualized)) ; then
            _app_zoom="TRUE"
        else
            _app_zoom="FALSE"
        fi

        if ((is_virtualized)) ; then
            _app_dropbox="FALSE"
        fi

        if ((is_software_center_wanted)) ; then
            _app_flatpaks="TRUE"
        else
            _app_flatpaks="FALSE"
        fi
        if [[ -z "$_app_snaps" ]] ; then
            _app_snaps="FALSE"
        fi

        # - defaults }}}
        if ! ((conf_configure_options)) ; then
            # XXX non-interactive:
            echo "${_app_skype}|${_app_zoom}|${_app_steam}|${_app_dropbox}|${_app_flatpaks}|${_app_snaps}" > "$TMP_FILE_1_f"
        else
            yad --form --center --width=460 --title="$( eval_gettext "Non-open or extra software" )" --image=emblem-package --image-on-top \
                --text="$( eval_gettext "Choose the closed-source software you want to add:" )" \
                --field="Skype, $( eval_gettext "video call application" ) (64bit only):chk" $_app_skype \
                --field="Zoom, $( eval_gettext "video call application" ):chk" $_app_zoom \
                --field="Steam, $( eval_gettext "video game distribution service" ):chk" $_app_steam \
                --field="Dropbox, $( eval_gettext "cloud storage and file synchronization" ):chk" $_app_dropbox \
                --field="Flatpak $( eval_gettext "packages installation support" ):chk" $_app_flatpaks \
                --field="Snap $( eval_gettext "packages support (not recommended, requires extra resources)" ):chk" $_app_snaps \
                --button="gtk-ok" > "$TMP_FILE_1_f" || true

                #--field="Google Chrome:chk" $_app_chrome \
                #--field="Firmwares, $( eval_gettext "some drivers for your computer (not recommended to remove!)" ):chk" TRUE \
                #--field="$( eval_gettext "Note: If you booted selecting the proprietary Nvidia drivers, you must reinstall again Elive by without choosing them first" ):lbl" TRUE \
        fi


        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" ; then
            if ((is_amd64)) ; then
                echo -e "skypeforlinux" >> "/tmp/.packages-to-install.tmp"
            else
                guitool warning "$( eval_gettext "The next package/software you selected is only available for 64-bit, so we cannot include it. If you want this software, you should use the 64-bit version of Elive." )\n\nSkype" &
            fi
        else
            echo -e "skypeforlinux" >> "/tmp/.packages-to-remove.tmp"
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "TRUE" ; then
            echo -e "zoom" >> "/tmp/.packages-to-install.tmp"
        else
            echo -e "zoom" >> "/tmp/.packages-to-remove.tmp"
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "TRUE" ; then
            if ((is_bullseye)) ; then
                echo -e "steam:i386" >> "/tmp/.packages-to-install.tmp"
            else
                echo -e "steam-installer" >> "/tmp/.packages-to-install.tmp"
            fi
        else
            echo -e "steam:i386\nsteam-installer\nsteam-compat-notinstalled" >> "/tmp/.packages-to-remove.tmp"
        fi
        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $4}' | grep -qs "TRUE" ; then
            echo -e "dropbox" >> "/tmp/.packages-to-install.tmp"
        else
            echo -e "dropbox" >> "/tmp/.packages-to-remove.tmp"
        fi

        if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' | grep -qs "TRUE" ; then
            is_flatpaks_wanted=1
        fi
        if cat "$TMP_FILE_2_f" | awk -v FS="|" '{print $6}' | grep -qs "TRUE" ; then
            is_snaps_wanted=1
        fi
        #if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' | grep -qs "TRUE" ; then
            #echo -e "google-chrome-stable" >> "/tmp/.packages-to-install.tmp"
        #else
            #echo -e "google-chrome-stable" >> "/tmp/.packages-to-remove.tmp"
            ## remove the repo too from the system
            #rm -f /etc/apt/sources.list.d/google-chrome.list 2>/dev/null || true
        #fi
        #if cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" ; then
            #echo -e "" >> "/tmp/.packages-to-install.tmp"
        #else
            #echo -e "" >> "/tmp/.packages-to-remove.tmp"
        #fi
        # remove the ms core fonts if user dont want any closed-source software:
        if \
            cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "TRUE" && \
            cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "TRUE" && \
            cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "TRUE" && \
            cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $4}' | grep -qs "TRUE" ; then
                rm -rf "$TARGET/usr/share/fonts/truetype/msttcorefonts"
        fi

        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $1}' | grep -qs "FALSE" && echo -e "skypeforlinux" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $2}' | grep -qs "FALSE" && echo -e "steam\nsteam:i386" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $3}' | grep -qs "FALSE" && echo -e "dropbox" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $4}' | grep -qs "FALSE" && echo -e "" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $5}' | grep -qs "FALSE" && echo -e "" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $6}' | grep -qs "FALSE" && echo -e "" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $7}' | grep -qs "FALSE" && echo -e "" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $8}' | grep -qs "FALSE" && echo -e "" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $9}' | grep -qs "FALSE" && echo -e "" >> "/tmp/.packages-to-remove.tmp" || true
        #cat "$TMP_FILE_1_f" | awk -v FS="|" '{print $10}' | grep -qs "FALSE" && echo -e "" >> "/tmp/.packages-to-remove.tmp" || true



        rm -f "$TMP_FILE_1_f"
        # }}}

        # manage results
        # software center options {{{
        if ((is_software_center_wanted)) || ((is_flatpaks_wanted)) || ((is_snaps_wanted)) ; then
            if ((is_i386)) && ((is_bullseye)) ; then
                if guitool yesnoquestion "$( eval_gettext "Warning: Gnome-Software in a 32-bit OS has a bug causing crashes when selecting the categories menu. Despite being more user-friendly for software installation and supporting flatpaks, it requires more resources. Would you prefer to replace it with gnome-packagekit, a simpler and lighter application?" )" ; then
                    # use gnome-packagekit (light one)

                    for i in gnome-software* apt-config-icons-hidpi packagekit-tools fwupd ; do
                        echo -e "$i" >> "/tmp/.packages-to-remove.tmp"
                        sed -i -e "/${i%\*}/d" "/tmp/.packages-to-install.tmp"
                    done
                    echo -e "gnome-packagekit" >> "/tmp/.packages-to-install.tmp"
                    sed -i -e '/gnome-packagekit/d' "/tmp/.packages-to-remove.tmp"
                else
                    # switch to gnome-software (bloated)
                    for i in gnome-software apt-config-icons-hidpi packagekit-tools fwupd ; do
                        echo -e "${i%\*}" >> "/tmp/.packages-to-install.tmp"
                        sed -i -e "/$i/d" "/tmp/.packages-to-remove.tmp"
                    done
                    #echo -e "gnome-packagekit" >> "/tmp/.packages-to-remove.tmp"
                    #sed -i -e '/gnome-packagekit/d' "/tmp/.packages-to-install.tmp"
                    # update: leave the light one as an option when on 32bit so its lighter and no buggy
                    echo -e "gnome-packagekit" >> "/tmp/.packages-to-install.tmp"
                    sed -i -e '/gnome-packagekit/d' "/tmp/.packages-to-remove.tmp"
                fi
            else
                # switch to gnome-software (bloated)
                for i in gnome-software apt-config-icons-hidpi packagekit-tools fwupd ; do
                    echo -e "${i%\*}" >> "/tmp/.packages-to-install.tmp"
                    sed -i -e "/$i/d" "/tmp/.packages-to-remove.tmp"
                done
                echo -e "gnome-packagekit" >> "/tmp/.packages-to-remove.tmp"
                sed -i -e '/gnome-packagekit/d' "/tmp/.packages-to-install.tmp"
            fi

            if ((is_flatpaks_wanted)) ; then
                if ((is_bullseye)) ; then
                    echo -e "gnome-software-flatpak" >> "/tmp/.packages-to-install.tmp"
                else
                    echo -e "gnome-software-plugin-flatpak" >> "/tmp/.packages-to-install.tmp"
                fi
            fi

            if ((is_snaps_wanted)) ; then
                echo -e "gnome-software-plugin-snap\nsnapd" >> "/tmp/.packages-to-install.tmp"
            fi

        else
            for i in gnome-software* apt-config-icons-hidpi packagekit-tools fwupd ; do
                echo -e "$i" >> "/tmp/.packages-to-remove.tmp"
                sed -i -e "/${i%\*}/d" "/tmp/.packages-to-install.tmp"
            done
            echo -e "gnome-packagekit" >> "/tmp/.packages-to-install.tmp"
            sed -i -e '/gnome-packagekit/d' "/tmp/.packages-to-remove.tmp"
        fi

        # }}}

        # elive demo files removal, before user is created
        if ((is_cleanup_wanted_dir_elivedemofiles)) ; then
            rm -rf "$TARGET/usr/share/elive-demo-files" 2>/dev/null || true
            rm -rf "$TARGET/usr/share/elive-demo-files-skel" 2>/dev/null || true
        fi
    fi


    # - custom options and packages }}}

    # manual selection of packages to maintain installed {{{

    mkdir -p "$TARGET/etc/elive"
    touch "$TARGET/etc/elive/packages_to_maintain_installed"

    # list of packages to maintain installed from old system, any ?
    maintain_installed_packages="$( cat "$TARGET/etc.old-to-${eliveversion}/elive/packages_to_maintain_installed" 2>/dev/null | grep -v "^#" | tr ',' ' ' | tr ' ' '\n' | sort -u | tr '\n' ' ' | sed '/^$/d' )"
    # remove extra leading blank chars
    read -r maintain_installed_packages <<< "$maintain_installed_packages"

    # remove some unwanted packages
    if grep -qs "libreoffice" "/tmp/.packages-to-remove.tmp" ; then
        maintain_installed_packages="$( echo "$maintain_installed_packages" | grep -v "^#" | tr ',' ' ' | tr ' ' '\n' | grep -v "libreoffice" | sort -u | tr '\n' ' ' | sed '/^$/d' )"
        read -r maintain_installed_packages <<< "$maintain_installed_packages"
    fi
    if grep -qs "iceweasel" "/tmp/.packages-to-remove.tmp" ; then
        maintain_installed_packages="$( echo "$maintain_installed_packages" | grep -v "^#" | tr ',' ' ' | tr ' ' '\n' | grep -v "iceweasel" | sort -u | tr '\n' ' ' | sed '/^$/d' )"
        read -r maintain_installed_packages <<< "$maintain_installed_packages"
    fi
    if grep -qs "firefox-esr" "/tmp/.packages-to-remove.tmp" ; then
        maintain_installed_packages="$( echo "$maintain_installed_packages" | grep -v "^#" | tr ',' ' ' | tr ' ' '\n' | grep -v "firefox-esr" | sort -u | tr '\n' ' ' | sed '/^$/d' )"
        read -r maintain_installed_packages <<< "$maintain_installed_packages"
    fi

    # remove conflicting ones
    maintain_installed_packages="$( echo "$maintain_installed_packages" | grep -v "^#" | tr ',' ' ' | tr ' ' '\n' | grep -vE "(myspell)" | sort -u | tr '\n' ' ' | sed '/^$/d' )"
    read -r maintain_installed_packages <<< "$maintain_installed_packages"

    local message_maintain_installed
    message_maintain_installed="$( eval_gettext "This is a list of the packages that you want to keep installed across re-installations (in upgrade mode). You can edit this list in the future at '/etc/elive/packages_to_maintain_installed'." )"

    { echo -e "# ${message_maintain_installed}\n\n# Packages list:" ; echo "$maintain_installed_packages" ; echo "$maintain_installed_packages_extra" ; } > "$TARGET/etc/elive/packages_to_maintain_installed"

    # install libreoffice and other intl packages {{{
    local lang_pkg lang_small
    lang_small="${LANG%%.*}"
    lang_small="${lang_small%%_*}"

    if ! grep -qs "libreoffice" "/tmp/.packages-to-remove.tmp" && grep -qs "libreoffice" "/tmp/.packages-to-install.tmp" ; then
        for lang_pkg in libreoffice-help-$lang_small libreoffice-l10n-$lang_small hyphen-$lang_small hunspell-$lang_small hunspell
        do
            # skip conflicting ones
            if [[ "$lang_pkg" = "hyphen-hr" ]] || [[ "$lang_pkg" = "libreoffice-help-"* ]] \
                || [[ "$lang_pkg" = "libreoffice-l10n-gd" ]] \
                || [[ "$lang_pkg" = "libreoffice-l10n-kk" ]] \
                ; then
                continue
            fi
            if LC_ALL="$EL_LC_EN" apt-cache show "$lang_pkg" 2>/dev/null | grep -qs "^Installed-Size:" ; then
                if ! cat "$TARGET/etc/elive/packages_to_maintain_installed" 2>/dev/null | grep -v "^#" | grep -qs "$lang_pkg" ; then
                    echo "$lang_pkg" >> "$TARGET/etc/elive/packages_to_maintain_installed"
                fi
            fi
        done
    fi

    if LC_ALL="$EL_LC_EN" COLUMNS=1000 dpkg -l | grep -qs "^ii.*iceweasel" && ! grep -qs "iceweasel" "/tmp/.packages-to-remove.tmp" ; then
        for lang_pkg in iceweasel-l10n-$lang_small iceweasel-l10n-${lang_small}-${lang_small} iceweasel-l10n-$( echo "${LANG%%.*}" | sed -e 's|_|-|g' | awk '{print tolower($0)}' )
        do
            if LC_ALL="$EL_LC_EN" apt-cache show "$lang_pkg" 2>/dev/null | grep -qs "^Installed-Size:" ; then
                if ! cat "$TARGET/etc/elive/packages_to_maintain_installed" 2>/dev/null | grep -v "^#" | grep -qs "$lang_pkg" ; then
                    echo "$lang_pkg" >> "$TARGET/etc/elive/packages_to_maintain_installed"
                fi
            fi
        done
    fi

    if LC_ALL="$EL_LC_EN" COLUMNS=1000 dpkg -l | grep -qs "^ii.*firefox-esr" && ! grep -qs "firefox-esr" "/tmp/.packages-to-remove.tmp" ; then
        for lang_pkg in firefox-l10n-$lang_small firefox-l10n-${lang_small}-${lang_small} firefox-l10n-$( echo "${LANG%%.*}" | sed -e 's|_|-|g' | awk '{print tolower($0)}' ) firefox-esr-l10n-$lang_small firefox-esr-l10n-${lang_small}-${lang_small} firefox-esr-l10n-$( echo "${LANG%%.*}" | sed -e 's|_|-|g' | awk '{print tolower($0)}' )

        do
            if LC_ALL="$EL_LC_EN" apt-cache show "$lang_pkg" 2>/dev/null | grep -qs "^Installed-Size:" ; then
                if ! cat "$TARGET/etc/elive/packages_to_maintain_installed" 2>/dev/null | grep -v "^#" | grep -qs "$lang_pkg" ; then
                    echo "$lang_pkg" >> "$TARGET/etc/elive/packages_to_maintain_installed"
                fi
            fi
        done
    fi
    for lang_pkg in  aspell-$lang_small hunspell-$lang_small hunspell
    do
        if LC_ALL="$EL_LC_EN" apt-cache show "$lang_pkg" 2>/dev/null | grep -qs "^Installed-Size:" ; then
            if ! cat "$TARGET/etc/elive/packages_to_maintain_installed" 2>/dev/null | grep -v "^#" | grep -qs "$lang_pkg" ; then
                echo "$lang_pkg" >> "$TARGET/etc/elive/packages_to_maintain_installed"
            fi
        fi
    done


    # }}}
    # now, show to the user the list
    # TODO: show also a diff of packages installed in the previous system and the new one to see if he want's to include those packages too
    if ((conf_configure_options)) && ! ((ignore_ask_list_packages_to_maintain)) && ! ((is_virtualized)) ; then
        maintain_packages_list="$( $guitool --title='Elive Installer' --width="550" --height="360" --text-info --editable --filename "$TARGET/etc/elive/packages_to_maintain_installed" 2>/dev/null || echo cancel )"

        if ! [[ "$maintain_packages_list" = "cancel" ]] ; then
            echo "$maintain_packages_list" > "$TARGET/etc/elive/packages_to_maintain_installed"
        fi
    fi


    # get the list to install now
    unset maintain_packages_list
    maintain_installed_packages="$( cat "$TARGET/etc/elive/packages_to_maintain_installed" 2>/dev/null | grep -v "^#" | tr ',' ' ' | tr ' ' '\n' | sort -u | tr '\n' ' ' | sed '/^$/d' )"
    # remove extra leading blank chars
    read -r maintain_installed_packages <<< "$maintain_installed_packages"



    # - packages to maintain installed }}}
    # add a list of packages to install in chroot {{{
    if [[ -n "$maintain_installed_packages" ]] ; then
        el_debug "Packages list desired to maintain installed: $maintain_installed_packages"

        if ! el_verify_internet fast 2>/dev/null ; then
            guitool info "$( eval_gettext "No internet connection was found. Connect to the internet before closing this message; otherwise, the packages listed in '/etc/elive/packages_to_maintain_installed' will fail to install. If you proceed, you'll need to install them manually afterward (retrieve the list and add it to an 'apt-get install' command in a terminal)." )"

        else

            # internet to slow? ask what to do
            if [[ -s "/tmp/.connection-speed-average.txt" ]] ; then
                if [[ "$( cat /tmp/.connection-speed-average.txt | sed -e 's|\..*$||g' | awk '{print $1}' )" -lt 30 ]] ; then
                    # show to the user first which packages are exactly

                    guitool warning "$( eval_gettext "Your internet connection is very slow, and the installation of the packages you want to keep may take longer than desired. In case it fails, here you can find a list of these packages that you can install manually later:" ) '/etc/elive/packages_to_maintain_installed'\n\nList: ${maintain_installed_packages} " &
                fi
            fi

        fi
    fi

    # - add a list of packages to install in chroot }}}
    # upgrade installed system needed? {{{

    # if enough internet speed:
    if [[ -e "/tmp/.installation-full-upgrade-required" ]] ; then
        if [[ "$( cat /tmp/.connection-speed-average.txt | sed -e 's|\..*$||g' | awk '{print $1}' )" -gt 1300 ]] ; then
            if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" "gt" "3.8.40"  &&  LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" "lt" "3.9.20" ; then
                if el_check_version_development_is_days_recent 120 ; then
                    if el_verify_internet 2>/dev/null ; then
                        is_full_upgrade_wanted=1
                    fi
                fi
            fi
        fi
    fi

    #if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" "le" "2.6.4"  && el_verify_internet 2>/dev/null ; then
        #echo "# shared home fixes" >> "$TARGET/tmp/temp"
        #echo "if el_verify_internet 2>/dev/null ; then apt-get update ; apt-get install -y  os-prober elive-tools e17-stable    ; fi" >> "$TARGET/tmp/temp"
    #fi


    # }}}

    # CHROOT remove & install packages, install user {{{

    # checks {{{
    if ! { LC_ALL=C cat /var/lib/apt/lists/*Packages 2>/dev/null ; LC_ALL=C zcat /var/lib/apt/lists/*Packages.gz 2>/dev/null ; LC_ALL=C lz4cat /var/lib/apt/lists/*Packages.lz4 2>/dev/null ; } | LC_ALL=C grep -aqs "^Package: grub" && ! ((is_thanatests)) ; then
        el_error "Seems like our lists of apt-get are not updated by default, Elive should ship with it already ready to use so please report this problem to Elive, if you want to continue with the installation you need to be connected to internet"
    fi

    # }}}

    cp -f "/tmp/.packages-to-install.tmp"         "$TARGET/tmp/.packages-to-install"
    cp -f "/tmp/.packages-to-remove.tmp"          "$TARGET/tmp/.packages-to-remove"
    cp -f "/tmp/.packages-to-hold.tmp"            "$TARGET/tmp/.packages-to-hold"
    cp -f "/tmp/.packages-to-reconfigure.tmp"     "$TARGET/tmp/.packages-to-reconfigure"
    cp -f "/tmp/os-probed"                        "$TARGET/tmp/" 1>/dev/null 2>&1 || true

    # first of all, remove extra kernels, which saves space and build-time for later, also we need this to be able to mount /boot extra partition without wasting space
    # clean & prepare
if [[ -s "$TARGET/tmp/temp" ]] ; then
    el_warning "file '$TARGET/tmp/temp' was not empty, but we are going to remove it before to run its previous contents:\n$(cat $TARGET/tmp/temp)"
fi
    cat > "$TARGET/tmp/temp" << EOF
#!/bin/bash
export LC_ALL="$EL_LC_EN"
source /usr/lib/elive-tools/functions
export EL_DEBUG="$EL_DEBUG"
set -E
set +e

# ignore accidentals Ctrl + C
trap '' SIGINT

export TERM=xterm-256color

if (($is_previous_installation_failed)) ; then
    export DEBIAN_FRONTEND=gnome
else
    export DEBCONF_NONINTERACTIVE_SEEN=true
    export DEBCONF_NOWARNINGS=true
    export DEBIAN_FRONTEND=noninteractive
    export DEBIAN_PRIORITY=critical
fi

trap "exit_ok" EXIT

exit_ok(){
    touch /tmp/.eliveinstaller-chroot-tasks-done
}


echo -e "\n\n      Working, do not close this terminal...\n\n"
sync

# remove dummy user
userdel -r eliveuser 1>/dev/null 2>&1 || true
rm -rf /home/eliveuser 1>/dev/null 2>&1 || true

# delete annoying files
rm -rf "/etc/skel/.e_thanatests"  1>/dev/null 2>&1 || true


# grub: install or remove special grub for secureboot
if (($is_secureboot_enabled)) ; then
    rm -rf "/boot/efi/EFI/elive/grub"*.efi 1>/dev/null 2>&1 || true
    if (($is_efi_32)) ; then
        package_grub_extra="grub-efi-ia32-signed \$package_grub_extra"
        echo "grub-efi-amd64" >> "/tmp/.packages-to-remove"
        el_warning "installing 32bit signed / secureboot efi, this should be uncompatible and not supposed to work"
        zenity --warning --text="installing grub as 32bit with signed / secureboot efi, this should be uncompatible and not supposed to work"
    fi
    if (($is_efi_64)) ; then
        package_grub_extra="grub-efi-amd64-signed grub-efi-ia32-signed \$package_grub_extra"
        # echo "grub-efi-ia32" >> "/tmp/.packages-to-remove"
    fi

    package_grub_extra="shim-signed \$package_grub_extra"
    echo -e "shim-signed-common\tshim/disable_secureboot\tboolean\tfalse" | debconf-set-selections 2>/dev/null

else
    rm -rf "/boot/efi/EFI/elive/shim"*.efi 1>/dev/null 2>&1 || true
    echo "shim-signed" >> "/tmp/.packages-to-remove"
fi

# pre-delete it, just in case. Update: don't do that, or windows will mark the bootloader to not include the linux one
# rm -rf "/boot/efi/EFI/${efi_name}" 1>/dev/null 2>&1 || true

# delete possibly conflicting ones
if (($is_fix_grub_remove_efi_debian_dir_wanted)) ; then
    el_info "removing old debian directory from EFI partition"
    rm -rf "/boot/efi/EFI/debian"
fi


# temporal moving of update-initramfs:
if ! [[ "\$( readlink -f "/usr/sbin/update-initramfs" )" = "/bin/true" ]] \
    && ! [[ -e "/usr/sbin/update-initramfs.orig" ]] \
    && ! [[ -e "/usr/sbin/update-initramfs.orig.initramfs-tools" ]] ; then

    mv -f "/usr/sbin/update-initramfs" "/usr/sbin/update-initramfs.orig"
    ln -s /bin/true "/usr/sbin/update-initramfs"
fi


el_debug "cleaning apt state"
apt-get -f -y install 1>/dev/null
el_debug "checking packages"

# disable update-initramfs for faster apt:
if grep -qs "^live-tools$" /tmp/.packages-to-remove ; then
    el_debug "optimizing apt"
    apt-get remove --purge -y live-tools deliver  1>/dev/null
fi


# before to remove packages, install alternative ones in order to not break things like pulseaudio:
if grep -qs "compat-notinstalled" "/tmp/.packages-to-install" ; then
    if ! apt-get update 2>/dev/null ; then
        sleep 20
        apt-get update || true
    fi
    list="\$( grep "compat-notinstalled" "/tmp/.packages-to-install" | tr '\n' ' ' )"
    list="\$( echo \$list )"
    if apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$list ; then
        apt-get -f -y install 1>/dev/null
    else
        el_error "failed to install pre-needed packages: \$list"
    fi
fi

# before to remove or install packages, switch to sysvinit if wanted
if (($is_sysvinit_wanted)) ; then
    list="sysvinit-core libpam-elogind orphan-sysvinit-scripts"
    if (($is_amd64)) && dpkg -l | grep -qs "^ii.*:i386" ; then
        list="\$list libpam-elogind:i386"
    fi
    if apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$list ; then
        apt-get -f -y install 1>/dev/null
    else
        el_error "failed to install pre-needed packages: \$list"
    fi

else
    # fix ssh masked mode on the installed system before to proceed with packages
    systemctl unmask ssh.service 1>/dev/null 2>&1 || true
    systemctl enable ssh.service 1>/dev/null 2>&1 || true
fi


# cleanup the package list with only the ones that we have installed in the system
j="\$( COLUMNS=1000 LC_ALL=C dpkg -l | grep -E -- "^(hi|ii)" | awk '{print \$2}' )"

set -o noglob
while read -ru 3 i
do
    [[ -z "\$i" ]] && continue

    # generic matches are allowed in the form like xfonts-thai-.*  (using a dot, not only * )
    # matches of packages that are included in the installed system
    printf "%s\n" "\$j" | LC_ALL=C grep -- "^\${i}$" || true

done 3<<< "\$( cat "/tmp/.packages-to-remove" | sort -u )" > "/tmp/.packages-to-remove"
set +o noglob

# needs to be since here or we will have a flood of the previous command
[[ "$EL_DEBUG" -gt 3 ]] && set -x || true

if [[ "\$EL_DEBUG" -gt 1 ]] ; then
    el_debug "Packages to remove: \n\$( cat "/tmp/.packages-to-remove" | sort -u )\n"
fi


# Remove / uninstall Packages:
if (($is_packages_remove_ignored)) ; then
    el_debug "IGNORING Removal of some packages for Thanatests mode"
    cat "/tmp/.packages-to-remove" | grep -E "(^live|^linux|initramfs|crypto)" | sort -u > "/tmp/.packages-to-remove.tmp"
    mv -f "/tmp/.packages-to-remove.tmp" "/tmp/.packages-to-remove"
fi

if grep -qs "[[:alnum:]]" "/tmp/.packages-to-remove" ; then
    if apt-get remove --purge -y \$( cat "/tmp/.packages-to-remove" | sort -u | psort -- -p "^linux" | tac | tr '\n' ' ' ) ; then
        apt-get -f -y install 1>/dev/null
    else
        echo 'Failed the removing of some not-needed packages, so: Removing Packages independently'
        apt-get -f -y install 1>/dev/null

        while read -ru 3 i
        do
            [[ -z "\$i" ]] && continue
            apt-get remove --purge -y "\$i"
        done 3<<< "\$( cat "/tmp/.packages-to-remove" | sort -u | psort -- -p "^linux" | tac )"
    fi
fi


# regenerate icons cache, after to have maybe-removed some of them (not shipping anymore by default in the isos, too huge)
if [[ "$HD_SIZE_mb" -gt "220000" ]] ; then
    if ! [[ -s "/usr/share/icons/gnome/icon-theme.cache" ]] ; then
        #el_info "Updating caches for icon themes..."

        while read -ru 3 dir
        do
            dir="/usr/share/icons/\$dir"

            if [[ -d "\$dir" ]] && [[ -s "\$dir/index.theme" ]] ; then
                # Paper will always fail in 32 bits
                if [[ "\$dir" = *"Paper" ]] && [[ -d "/usr/lib/i386-linux-gnu/zsh" ]] ; then
                    continue
                fi
                # only generate it for the gnome one, because we don't use the others and it takes disk space and also delayed installation
                if [[ "\$dir" != *"/gnome" ]] ; then
                    continue
                fi

                if [[ "\$dir" = *"/gnome" ]] ; then
                    el_info "Updating icon cache for Elive icon theme"
                else
                    el_info "Updating icon cache for: \$(basename \$dir )"
                fi

                # don't exit if fails, but at least remove the uless cache's
                if ! gtk-update-icon-cache-3.0 -f -t -q "\$dir" 2>/dev/null && ! gtk-update-icon-cache -f -t -q "\$dir" 2>/dev/null ; then
                    rm -rf "\$dir"/.icon-theme.cache || true
                    rm -rf "\$dir"/icon-theme.cache  || true
                fi
            fi
        done 3<<< "\$( ls -1 "/usr/share/icons" )"
    fi
else
    rm -f "/usr/share/icons/gnome/icon-theme.cache" 2>/dev/null || true
fi



# set generic preseeds
if [[ -s "/root/preseeds.cfg" ]] ; then
    debconf-set-selections < "/root/preseeds.cfg" || true
    el_debug "debconf set selections: \n\$( cat /root/preseeds.cfg | sort -u )\n"
fi



#
# install required things
#

export DEBIAN_FRONTEND=noninteractive
if [[ -n "$HD_PART_EFI_dev" ]] ; then
    if (($is_efi_32)) ; then
        package_grub="grub-efi-ia32"
    fi
    if (($is_efi_64)) ; then
        package_grub="grub-efi-amd64"
    fi
    package_grub="grub-common grub2-common \$package_grub"
    # always ask how to install efi, if forced or not
else
    package_grub="grub-pc grub2 grub-common grub2-common"
    if [[ -z "$mbr_device_id_install" ]] ; then
        export DEBIAN_FRONTEND=gnome
    fi
fi

# install refind if needed (more than one OS in EFI mode)
if (($is_apple)) ; then
    package_grub_extra="refind refind-theme-regular-elive \$package_grub_extra"
fi

# reconfigure each needed preseed which is packagename
if [[ -d "/root/preseeds.d" ]] ; then
    for file in /root/preseeds.d/*
    do
        if [[ -s "\$file" ]] ; then
            debconf-set-selections < "\$file"
            if ! echo "\$file" | grep -qsi "noreconfigure" ; then
                el_info "Reconfiguring \$( basename \${file%%_*} )"
                cat "\${file}" 1>&2
                DEBIAN_FRONTEND="" dpkg-reconfigure -fnoninteractive -pcritical --no-reload "\$( basename \${file%%_*} )" 2>/dev/null || true
            fi
        fi
    done
fi

# INSTALL grub & login-manager
# we need to have this list, always
if ! { LC_ALL=C cat /var/lib/apt/lists/*Packages 2>/dev/null ; LC_ALL=C zcat /var/lib/apt/lists/*Packages.gz 2>/dev/null ; LC_ALL=C lz4cat /var/lib/apt/lists/*Packages.lz4 2>/dev/null ; } | LC_ALL=C grep -aqs "^Package: grub" ; then
    # If we don't have a *Packages files, we will be unable to use apt-get to install grub and lightdm (but we can still do it with dpkg)

    el_warning "Lists not found, we should have them available in the isos, trying to obtain them from internet instead..."

    # update packages
    for try in 2 6 12
    do
        if el_verify_internet fast 1>/dev/null 2>&1 ; then
            if ! apt-get update 2>/dev/null ; then
                sleep 20
                apt-get update || true
            fi
            break
        else
            sleep "\$try"
        fi
    done
fi

# Install from apt-get (with or without internet should work if we have them cache'd)
# install grub without asking with debconf
el_info "Installing Grub and Login manager..."
if apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$package_grub \$package_grub_extra lightdm ; then
    apt-get -f -y install 1>/dev/null || true
    el_info "Installed Grub and Login manager via apt"
else
    if el_verify_internet fast 1>/dev/null 2>&1 ; then
        if ! apt-get update 2>/dev/null ; then
            sleep 8
            apt-get update || true
        fi
        apt-get -f -y install 1>/dev/null || true

        if apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$package_grub \$package_grub_extra lightdm ; then
            el_info "Installed Grub and Login manager via apt after to have updated packages fist"
        else
            el_warning "Unable to install grub2 and lightdm from apt-get, trying another option..."
        fi
    fi
fi

# verify and if they are not, install them manually:
if ! COLUMNS=1000 dpkg -l | grep -qsE "^ii\s+lightdm\s+" || ! COLUMNS=1000 dpkg -l | grep -qsE "^ii\s+(grub2|grub-pc|grub-efi)" ; then
    el_info "Packages (grub2|lightdm) has not been installed correctly using apt-get, probably due to a lack of connection to internet, switching to dpkg instead..."

    # No access to /var/lib/apt/lists/*Packages files? install with dpkg -i
    if ls /var/cache/apt/archives/*deb 1>/dev/null ; then

        cd /var/cache/apt/archives

        if dpkg -i \${package_grub}*deb \${package_grub_extra}*deb lightdm*deb liblightdm*deb libxklavier*deb 1>/dev/null ; then
            apt-get -f -y install 1>/dev/null || true
            el_info "Installed using dpkg"
        else
            # there's also libraries to install, but they can change in the future, so let's just install all of them:
            if dpkg -i \${package_grub}*deb \${package_grub_extra}*deb lightdm*deb lib*deb ; then
                apt-get -f -y install 1>/dev/null || true
                el_info "Installed using dpkg"
            else
                el_warning "Unable to install grub2 and lightdm from dpkg"
            fi
        fi

        cd
    fi
fi

# set grub configuration
elive_version_grub="\$( cat /etc/os-release | grep "^PRETTY_NAME=" | sed -e 's|^.*NAME="||' -e 's|"$||g' | tail -1 | grep "[[:alnum:]]" || echo "Elive" )"
sed -i -e "s|^GRUB_DISTRIBUTOR=.*$|GRUB_DISTRIBUTOR=\"\${elive_version_grub}\"|g" /etc/default/grub


# not needed, should be auto
#export DEBIAN_FRONTEND=noninteractive

# checks:
elinst_package_name="\$( COLUMNS=1000 dpkg -l | grep "^ii\s*eliveinstaller-" | awk '{print $2}' | sort -u | sed -e 's|:.*||g' | tail -1 )"
if ! [[ -n "\$elinst_package_name" ]] ; then
    el_warning "Eliveinstaller package was deinstalled? review the installer code to fix this bug, fixing it manually... List of packages wanted to be removed: \$( cat /tmp/.packages-to-remove | sort -u | psort -- -p '^linux' | tac | tr '\n' ' ' )"
    echo -e "$elinst_package_name" >> "/tmp/.packages-to-install"
fi





#
# Install extra packages
#

if [[ -s "/tmp/.packages-to-install" ]] ; then
    apt-get -f -y install 1>/dev/null || true

    # cleanup the package list with only the ones that we can install in the system
    j="\$( { cat /var/lib/apt/lists/*Packages 2>/dev/null ; zcat /var/lib/apt/lists/*Packages.gz 2>/dev/null ; } | grep "^Package:" | sed -e 's|^Package: ||g' | sort -u )"

    set -o noglob
    while read -ru 3 i
    do
        [[ -z "\$i" ]] && continue

        # do not install already-installed packages (faster especially for single-run mode)
        [[ -e "/var/lib/dpkg/info/\${i}.list" ]] && continue

        # generic matches are allowed in the form like xfonts-thai-.*  (using a dot, not only * )
        # matches of packages that are instalable in the installed system
        printf "%s\n" "\$j" | grep -- "^\${i}$" || true

    done 3<<< "\$( cat "/tmp/.packages-to-install" | sort -u )" > "/tmp/.packages-to-install"
    set +o noglob



    # localepurge needs to be preconfigured
    if grep -qs "^localepurge" "/tmp/.packages-to-install" ; then
        if [[ "\$( cat "/etc/default/locale" | sed -e 's|^.*=||g' | grep "^[[:alpha:]]" | tail -1 )" = "en_US.UTF-8" ]] ; then
            echo -e "localepurge\tlocalepurge/nopurge\tmultiselect\ten, en_US, en_US.UTF-8" | debconf-set-selections 2>/dev/null

        else
            echo -e "localepurge\tlocalepurge/nopurge\tmultiselect\ten, en_US, en_US.UTF-8, \$( cat "/etc/default/locale" | sed -e 's|^.*=||g' -e 's|_.*$||g' -e 's|-.*$||g' | grep "^[[:alpha:]]" | tail -1 ), \$( cat "/etc/default/locale" | sed -e 's|^.*=||g' -e 's|\..*$||g' | grep "^[[:alpha:]]" | tail -1 ), \$( cat "/etc/default/locale" | sed -e 's|^.*=||g' | grep "^[[:alpha:]]" | tail -1 )" | debconf-set-selections
        fi
        el_debug "localepurge: \$(debconf-get-selections | grep "localepurge/nopurge" )"
    fi

    if [[ "\$EL_DEBUG" -gt 1 ]] ; then
        el_debug "Packages to install: \n\$( cat "/tmp/.packages-to-install" | sort -u )\n"
    fi

    # try to install them in one shot
    if grep -qs "[[:alnum:]]" "/tmp/.packages-to-install" ; then

        if ! apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$( cat "/tmp/.packages-to-install" | sort -u | psort -- -p "^linux" | tr '\n' ' ' ) 2>/dev/null ; then

            # the main reason will be packages has been changed, so update is needed first:
            if el_verify_internet fast 1>/dev/null 2>&1 ; then
                if ! apt-get update 2>/dev/null ; then
                    sleep 30
                    if ! apt-get update 2>/dev/null ; then
                        sleep 1m
                        if ! apt-get update 2>/dev/null ; then
                            sleep 2m
                            apt-get update
                        fi
                    fi
                fi
            fi

            # do a few attempts first
            if el_verify_internet  1>/dev/null 2>&1 ; then
                for i in \$( seq 4 )
                do
                    echo -e "Trying another (\$i) attempt"
                    if apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$( cat "/tmp/.packages-to-install" | sort -u | psort -- -p "^linux" | tr '\n' ' ' ) 2>/dev/null ; then
                        break
                    else
                        sleep 10
                        apt-get update
                        apt-get clean
                    fi
                done

                # try to clean apt lists
                if ! apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$( cat "/tmp/.packages-to-install" | sort -u | psort -- -p "^linux" | tr '\n' ' ' ) 2>/dev/null ; then
                    # delete lists, sometimes they fail updating
                    find /var/lib/apt/lists/ -type f -delete

                    # the main reason will be packages has been changed, so update needed first:
                    if el_verify_internet fast 1>/dev/null 2>&1 ; then
                        if ! apt-get update 2>/dev/null ; then
                            sleep 20
                            apt-get update
                        fi
                    fi
                fi


                # may be a network problem:
                sleep 30

                if ! apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$( cat "/tmp/.packages-to-install" | sort -u | psort -- -p "^linux" | tr '\n' ' ' )  2>/tmp/failed-logs.txt 1>&2 ; then

                    el_warning "Unable to install all packages in one shot: \$( cat "/tmp/.packages-to-install" | sort -u | psort -- -p "^linux" | tr '\n' ' ' )\n\n\$( cat /tmp/failed-logs.txt )"

                    # install them independently if failed the one shot
                    while read -ru 3 i
                    do
                        if [[ -n "\$i" ]] ; then
                            if ! apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y "\$i" ; then
                                sleep 6

                                if el_verify_internet fast 1>/dev/null 2>&1 ; then
                                    if ! apt-get update 2>/dev/null ; then
                                        sleep 8
                                        apt-get update
                                    fi
                                fi
                                apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y "\$i" || sleep 2
                            fi
                        fi
                    done 3<<< "\$( cat "/tmp/.packages-to-install" | sed -e '/^$/d' | sort -u )"
                fi
            fi
        fi
    fi


    # update entire OS first if wanted
    if (($is_full_upgrade_wanted)) ; then
        el_info "UPGRADING the whole OS now ..."
        sleep 2
        if ! apt-get dist-upgrade -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y   2>/dev/null ; then
            if apt-get update 2>/dev/null ; then
                apt-get dist-upgrade -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y
            fi
        fi
        apt-get -f -y install 1>/dev/null
    fi



    rm -f /tmp/failed-logs.txt

    # run localepurge just in case (processing triggers of apt should already do it)
    if grep -qs "^localepurge" "/tmp/.packages-to-install" ; then
        el_debug "Removing unused locales (like translations):"

        localepurge
    fi
else
    el_debug "No packages extra to install"
fi

# Install packages to maintain installed
if [[ -n "$maintain_installed_packages" ]] ; then
    if el_verify_internet fast 1>/dev/null 2>&1 ; then

        el_debug "Packages to install (user maintain-packages selection): \n\$( echo "$maintain_installed_packages" | sort -u )\n"

        if ! apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y $maintain_installed_packages ; then
            if el_verify_internet fast 1>/dev/null 2>&1 ; then
                if ! apt-get update 2>/dev/null ; then
                    sleep 8
                    apt-get update
                fi
            fi
        fi

        if apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y $maintain_installed_packages ; then
            apt-get -f -y install 1>/dev/null
        else
            for package in $maintain_installed_packages
            do
                if ! apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$package ; then
                    sed -i "s|\$package||g" /etc/elive/packages_to_maintain_installed
                    apt-get -f -y install || apt-get -f -y install || apt-get -f -y install --force-yes
                fi
            done
            apt-get -f -y install 1>/dev/null
        fi
    fi
else
    el_debug "no packages to maintain installed from previous system"
fi

# cleanup left packages
# list the packages that we have autoremoved for report
apt-get -f --purge -s autoremove | grep -E "^(Remv|Purg)\s+" | awk '{print \$2}' | sort -u > "/tmp/.packages-to-remove_autoremoved"
apt-get -f -y --purge autoremove
apt-get -f -y install 1>/dev/null

# purge removed packages left
i="\$( COLUMNS=1000 LC_ALL=C dpkg -l | grep -E -- "^rc" | awk '{print \$2}' )"
apt-get remove --purge -y \$i
apt-get -f -y install 1>/dev/null

# Packages to hold
# they are going to be holded when installing, so not needed here
sed -i "/nvidia/d" "/tmp/.packages-to-hold"
sed -i "/fglrx/d" "/tmp/.packages-to-hold"

if [[ -s "/tmp/.packages-to-hold" ]] ; then
    while read -ru 3 i
    do
        if [[ -n "\$i" ]] ; then
            # note: sometimes it fails, like holding headers when they have been deinstalled from the fine-tune step
            if [[ -e "/var/lib/dpkg/info/\${i}.list" ]] ; then
                el_debug "holding \$i"
                echo "\$i hold" | dpkg --set-selections 2>/dev/null || true
            fi
        fi
    done 3<<< "\$( cat "/tmp/.packages-to-hold" | sort -u )"
else
    el_debug "no packages wanted to hold"
fi


# final verification
if ! COLUMNS=1000 dpkg -l | grep -qsE "^ii\s+lightdm\s+" || ! COLUMNS=1000 dpkg -l | grep -qsE "^ii\s+(grub2|grub|grub-pc|grub-efi)" ; then
    el_error "Packages (grub2|lightdm) entirely unable to install, please report this bug to Elive"
    zenity --error --text="Unable to install grub and/or lightdm required packages, please report the issue to Elive"
    exit
fi





# reconfigure locales, also needed before to run some things to the user like user-manager
if [[ -s /etc/locale.gen.orig ]] ; then
    mv -f /etc/locale.gen.orig /etc/locale.gen
fi
locale-gen --keep-existing 2>/dev/null || true
update-locale 2>/dev/null || true


# reconfigure packages
if [[ -s "/tmp/.packages-to-reconfigure" ]] ; then
    while read -ru 3 package
    do
        if [[ -e "/var/lib/dpkg/info/\${package}.list" ]] ; then
            dpkg-reconfigure -fnoninteractive -pcritical --no-reload "\$package"
        fi
    done 3<<< "\$( cat "/tmp/.packages-to-reconfigure" | sort -u )"
else
    el_debug "no packages wanted to reconfigure"
fi



# reconfigure SSH keys to have unique ones, ssh will not work too otherwise
if [[ -e "/var/lib/dpkg/info/openssh-server.list" ]] && ! (($is_migration)) ; then
    rm -f /etc/ssh/ssh_host_*_key /etc/ssh/ssh_host_*_key.pub 2>/dev/null || true
    dpkg-reconfigure -fnoninteractive -pcritical --no-reload openssh-server
fi

# generate default snakeoil
if [[ -e "/var/lib/dpkg/info/sslcert.list" ]] && ! (($is_migration)) ; then
    make-ssl-cert generate-default-snakeoil --force-overwrite || true
fi

# configure default desktop
if [[ -n "$(pidof e16)" ]] ; then
    update-alternatives --set x-window-manager /usr/bin/starte16
else
    if [[ -n "$(pidof enlightenment_start )" ]] ; then
        update-alternatives --set x-window-manager /usr/bin/enlightenment_start
    fi
fi

# if command-not-found is installed and we don't have the contents generated, generate them
if [[ -x "/usr/bin/apt-file" ]] || [[ -x "/usr/bin/command-not-found" ]] ; then
    if ! find "/var/lib/apt/lists/" -name '*'_Contents-'*' | grep -qs "Contents" ; then
        apt-file update
    fi
fi




touch /tmp/.eliveinstaller-chroot-tasks-done
echo -e "\nDONE"

EOF
chmod +x "$TARGET/tmp/temp"


# run gui
sync
#echo 10 > "$TMP_PROGRESS_CONFIGURING_f"
#{ ( while test -f "$TMP_PROGRESS_CONFIGURING_f" ; do cat "$TMP_PROGRESS_CONFIGURING_f" || true ; sleep 1 ; done | $guitool --title='Elive Installer' --progress --pulsate --text="$( eval_gettext "Installing and removing packages. This process may take some time. Please wait..." )" --auto-close ) & disown ; } 2>/dev/null

echo "30" > "$TMP_PROGRESS_f"
echo "# $( eval_gettext "Installing and removing packages. This process may take some time. Please wait..." )" >> "$TMP_PROGRESS_f"

    # free some memory in order to not bottleneck the HD
    swapfile_installer deactivate
    swapfile_installer activate

    # mount with bind first
    EL_DEBUG=0 mounts-manager --chroot-layers-mount "$TARGET" "eliveinstaller" bind 1>/dev/null

    # run all!
    # update: we always want tmux here, so the installing of packages can be a slow process
    #if { ((is_tmux_unwanted)) || ((conf_is_novice_user)) || ((conf_is_extra_help_wanted)) ; } && ! ((is_tmux_wanted)) ; then
    if ((is_tmux_unwanted)) ; then
        chroot "$TARGET" "/tmp/temp" 2>&1 | tee "${TMP_LOGS_REPORT_f}"
    else
        if [[ -x /usr/bin/tmux ]] && [[ -x /usr/bin/tmux-attach-jobs ]] ; then
            sed -i -e 's|^if-shell.*message_welcome.*$|#&|g' "/root/.tmux.conf"
            sed -i -e 's|^if-shell.*message_welcome.*$|#&|g' "/home/$eliveuser/.tmux.conf"
            timeout 10 tmux-attach-jobs eliveinstaller start "touch '${TMP_MAIN_d}/test-tmux-attach.txt'" 1>/dev/null
            sleep 1 ; sync
        fi
        # run it, undetached or direct if we don't have tmux feature
        if [[ -e "${TMP_MAIN_d}/test-tmux-attach.txt" ]] ; then
            rm -f "${TMP_MAIN_d}/test-tmux-attach.txt"

            sed -i -e 's|^if-shell.*message_welcome.*$|#&|g' "/root/.tmux.conf"
            sed -i -e 's|^if-shell.*message_welcome.*$|#&|g' "/home/$eliveuser/.tmux.conf"
            tmux-attach-jobs eliveinstaller "Managing_Packages" "chroot '$TARGET' bash -c '/tmp/temp' 2>&1 | tee '${TMP_LOGS_REPORT_f}' " 1>/dev/null
            sleep 2 ; sync

            # visualize tasks if we have it already not opened
            if ! LC_ALL="$EL_LC_EN" ps aux | grep -v grep | grep -qs "rxvt.*tmux attach" ; then
                urxvt -e bash -c "unset TMUX ; tmux attach -t eliveinstaller" &
            fi
        else
            urxvt -e bash -c "chroot '$TARGET' bash -c '/tmp/temp' 2>&1 | tee '${TMP_LOGS_REPORT_f}' " &
        fi
    fi

    # continue after to have finished the tasks
    # FIXME: this is crap, we need to put all the processes in a single chroot one
    while ! test -e "$TARGET/tmp/.eliveinstaller-chroot-tasks-done"
    do
        sleep 2
    done

    # inform to the user when a package has failed to install
    if cat "/tmp/eliveinstaller-debug-process.txt" "${TMP_LOGS_REPORT_f}" 2>/dev/null | grep -qs "E: Unable to locate package " ; then
        packages_failed_to_install="$( cat "/tmp/eliveinstaller-debug-process.txt" "${TMP_LOGS_REPORT_f}" | colors-remove | grep "E: Unable to locate package " | sed 's|^.*locate package ||g' | sort -u | tr '\n' ' ' )"
        el_warning "packages failed to install: $packages_failed_to_install"
        guitool warning "$( eval_gettext "Some packages failed to install. You may need to install them manually later:" ) $packages_failed_to_install"
    fi

    # create users and other configurations
    # fixes for older versions:
    if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" "le" "3.0.6" ; then
        chroot "$TARGET" bash -c "source /usr/lib/elive-tools/functions ; el_package_update_last_version elive-tools --force "
    fi

    # create new user:
    if ! ((is_migration)) && ! ((is_shared_home_partition)) ; then
        ( xsettingsd 2>/dev/null & )
        sleep 1

        if ((is_encryption_wanted)) ; then
            user_manager_result="$( chroot "$TARGET" /usr/sbin/user-manager --quiet --user-add --want-sudo --want-sudo-apt --show-output --force-request --password="${rootpass}" --email="${conf_send_debug_reports_email}" )"
        else
            user_manager_result="$( chroot "$TARGET" /usr/sbin/user-manager --quiet --user-add --want-sudo --want-sudo-apt --want-autologin --show-output --force-request --password="${rootpass}" --email="${conf_send_debug_reports_email}" )"
        fi

        if ! eval "$user_manager_result" ; then
            el_error "failed evaluating resuls:\n|$user_manager_result|"
        fi

        if [[ -n "$username" ]] ; then
            user_first="$username"
        else
            # el_error "Unable to get new user created (first attempt), result is:\n$user_manager_result"
            # try again
            sync

            # second attempt:
            if ((is_encryption_wanted)) ; then
                user_manager_result="$( chroot "$TARGET" /usr/sbin/user-manager --quiet --user-add --want-sudo --want-sudo-apt --show-output --password="${rootpass}" --email="${conf_send_debug_reports_email}" )"
            else
                user_manager_result="$( chroot "$TARGET" /usr/sbin/user-manager --quiet --user-add --want-sudo --want-sudo-apt --want-autologin --show-output --password="${rootpass}" --email="${conf_send_debug_reports_email}" )"
            fi

            # get resulting data
            if ! eval "$user_manager_result" ; then
                el_error "Unable to eval results, maybe we need to escape a strange character? data is:\n$user_manager_result"
            fi
            eval "$user_manager_result"

            if [[ -n "$username" ]] ; then
                user_first="$username"
            else
                guitool warning "$( eval_gettext "There was a problem creating your new user. Please try again. If you've enabled sending installer reports, we can receive useful information to help resolve this. You can also use our forums for assistance." )"
                # temporal attempt to show the possible errors:
                el_error "unable to get new user created, previous results:\n${user_manager_result}\nFULL new debug:\n"
                sed -i -e '1aset -x' "$TARGET/usr/sbin/user-manager"
                #user_manager_result="$( chroot "$TARGET" /usr/sbin/user-manager --quiet --user-add --want-sudo --want-sudo-apt --want-autologin --show-output --password="${rootpass}" --email="${conf_send_debug_reports_email}" 2>&1 )"
                chroot "$TARGET" /usr/sbin/user-manager --quiet --user-add --want-sudo --want-sudo-apt --want-autologin --show-output --password="${rootpass}" --email="${conf_send_debug_reports_email}"
                sed -i -e '/^set -x$/d' "$TARGET/usr/sbin/user-manager"

                el_error "unable to get new user created, results:\n${user_manager_result}"

                guitool error "$( eval_gettext "Failed to create your new user. Use standard English characters or report the issue to Elive if you believe there's a problem with the installer." )"
                exit_error
            fi
        fi

        killall xsettingsd 1>/dev/null 2>&1 || true

        # reciclate the same user password for root:
        if [[ -z "$rootpass" ]] && [[ "$user_password" ]] ; then
            rootpass="$user_password"
        fi
        if [[ -z "$rootpass" ]] ; then
            password_root_ask
            #el_warning "user_password $user_password was not possible to reciclate for rootpass"
        fi

        if [[ "$autologin" = "TRUE" ]] ; then
            is_autologin_wanted=1
            autologin_username="$username"
        else
            unset is_autologin_wanted
            unset autologin_username
        fi
    fi

    # when using a shared home (reciclated users), we still need a system's password
    if ((is_shared_home_partition)) ; then
        password_root_ask
    fi

    killall -9 fwupd 1>/dev/null 2>&1 || true


    rm -f "$TARGET/tmp/.eliveinstaller-chroot-tasks-done"
    rm -f "$TARGET/tmp/temp"

    # report and debug
    #if [[ "$( cat "$TARGET/tmp/.packages-to-remove_autoremoved" 2>/dev/null | grep -vE '^(lib|gir1|linux-header|linux-kbuild|syslinux-common|fonts-|kodi-|python-|cryptsetup|chromium|dmevent|openssh|tdb-tools|thin-)' | wc -l )" -ge 2 ]] ; then
        #el_warning "AutoRemoved Packages from installed system: $( cat "$TARGET/tmp/.packages-to-remove_autoremoved" | grep -vE '^(lib|gir1|linux-header|linux-kbuild|syslinux-common|fonts-|kodi-|python-|cryptsetup|chromium|dmevent|openssh|tdb-tools|thin-)' | tr '\n' ' ' )"
    #fi

    if [[ -s "$TMP_LOGS_REPORT_f" ]] ; then
        cat "${TMP_LOGS_REPORT_f}" 1>&2
        rm -f "${TMP_LOGS_REPORT_f}"
    fi


    # close the gui
    #rm -f "$TMP_PROGRESS_CONFIGURING_f"


    # }}}

    #
    # Post Configurations
    #
    # get optimal resolution to use for grub {{{

    while read -ru 3 res_fb
    do
        [[ -z "$res_fb" ]] && continue
        # preference for widescreen resolutions if we have it available:
        for res in 1280x720 1280x800 1360x768 1366x768 1440x900 1536x864 1024x576
        do
            if [[ "$res" = "$res_fb" ]] ; then
                grub_resolutions="${res},${grub_resolutions}"
            fi
        done
    done 3<<< "$( LC_ALL="$EL_LC_EN" hwinfo --framebuffer 2>/dev/null | grep -E "^\s+Mode\s+" | awk '{print $3}' | sort -uV )"

    read -r grub_resolutions <<< "$grub_resolutions"
    grub_resolutions="${grub_resolutions%,}"

    # set fallback defaults
    #grub_resolutions="${grub_resolutions},1024x768,auto"
    # 1024 looks too ugly, dont use it
    grub_resolutions="${grub_resolutions},auto"

    # fix when there's only ",auto" value
    grub_resolutions="${grub_resolutions#,}"

    # force this resolution for virtual machines
    if ((is_virtualized)) ; then
        grub_resolutions="800x600,auto"
    fi

    if LC_ALL="$EL_LC_EN" xrandr -q 2>/dev/null | grep -i connected | grep -qsE "(HDMI|DVI).* disconnected" ; then
        if LC_ALL="$EL_LC_EN" xrandr -q 2>/dev/null | grep -i connected | grep -qs "VGA.* connected" ; then
            guitool warning "$( eval_gettext "Your monitor is connected using a VGA cable. Check the connectors to see if you can use a DVI or HDMI cable instead, as this will improve graphics quality. The operating system will also detect your monitor more reliably, reducing eye strain. Note that this is merely a suggestion and optional, not required." )" &
        fi
    fi

    el_debug "optimal resolutions for grub: $grub_resolutions"

    # reconfigure resolution for grub:
    if [[ -n "$grub_resolutions" ]] ; then
        if ((is_wheezy)) ; then
            sed -i "s|^.*GRUB_GFXMODE=.*$|GRUB_GFXMODE=$grub_resolutions\nGRUB_GFXPAYLOAD_LINUX=keep|g" "$TARGET/etc/default/grub"
        else
            # we should not need this entry anymore, but if we failed our previous install, try again with this option
            if ((is_previous_installation_failed)) ; then
                if ! guitool question "$( eval_gettext "Your previous installation failed. Would you like to disable the default graphical resolution for the boot process? Select this option if your previous installation resulted in a blank boot screen, where you were unable to see anything during the boot." )" ; then
                    sed -i "s|^.*GRUB_GFXMODE=.*$|## Specify your desired resolution for your Grub.\n## Note: if you feel the grub menu slow to load you can use 1024x768 which is most of the times faster, at the cost to be less beautiful specially on widescreens.\nGRUB_GFXMODE=$grub_resolutions\nGRUB_GFXPAYLOAD_LINUX=keep|g" "$TARGET/etc/default/grub"
                else
                    sed -i "s|^.*GRUB_GFXMODE=.*$|#GRUB_GFXMODE=$grub_resolutions\n#GRUB_GFXPAYLOAD_LINUX=keep|g" "$TARGET/etc/default/grub"
                fi
            fi
        fi
    fi


    # }}}
    # hardware grub configurations {{{
    #if lshal | grep -qsiE "samsung.*n150\s+" ; then
    #if [[ "$( dmidecode -s system-product-name )" = *"samsung"*"n150"* ]] && [[ "$( dmidecode -s system-manufacturer )" = "ASUS"* ]] ; then
        #if ((is_wheezy)) ; then
            #el_array_member_add "acpi_backlight=vendor" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
        #else
            #guitool warning "$( eval_gettext "You have a Samsung N150 computer, we fixed a problem with the screen in the past and we are ignoring it for these new versions, if you see any screen problem please ask in our Elive forums how to solve it." )"
        #fi
    #fi

    # include nvidia / ati-fglrx confs
    # update: not needed, installing nvidia adds the blacklist in /etc/modprobe.d/
    #if ((is_nvidia)) ; then
        #el_array_member_add "modprobe.blacklist=nouveau" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
    #fi
    #if ((is_fglrx)) ; then
        #el_array_member_add "modprobe.blacklist=radeon,radeonhd" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
    #fi
    # not tested:
    #if ((is_vmware)) ; then
        #el_array_member_add "modprobe.blacklist=vmwgfx" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
    #fi

    # - hardware grub configurations }}}
    # add swap partition in grub to have hibernation {{{
    size=0
    for i in $SWAPDEVICES
    do
        if [[ -b "$i" ]] ; then
            _size="$( partitions-list --show-only=$i | awk -v FS="::" '{print $6}' )"
            # select the biggest swap partition
            if [[ "$_size" -gt "$size" ]] ; then
                _swap_partition_hibernator="$i"
                size="$_size"
            fi
        fi
    done

    # fallback to any hibernation partition
    if ! [[ -b "$_swap_partition_hibernator" ]] ; then
        _swap_partition_hibernator="$( EL_DEBUG=0 partitions-list --show-all | awk -v FS="::" '{if ( $3 == "swap" ) print $6" ::"$1}' | sort -g | sed -e 's|^.* ::||g' | head -1 || true )"
    fi


    for i in $( grep "GRUB_CMDLINE_LINUX_DEFAULT=" "$TARGET/etc/default/grub" 2>/dev/null | sed -e 's|^GRUB_CMDLINE_LINUX_DEFAULT=||g' -e 's|^"||g' -e 's|"$||g' | tail -1 )
    do
        [[ -z "$i" ]] && continue
        el_array_member_add "$i" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
    done
    # add default grub entries
    el_array_member_add "quiet" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")

    el_array_member_add "splash" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
    # ecore_no_vsync
    if [[ "$ECORE_NO_VSYNC" = "1" ]] || grep -qsE "(failsafe|bootmode=graphicalproblems)" /proc/cmdline ; then
        el_array_member_add "ecorenovsync" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
    fi

    if [[ -b "$_swap_partition_hibernator" ]] ; then
        if ((is_wheezy)) ; then
            el_array_member_add "resume=$_swap_partition_hibernator" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
        else
            el_array_member_add "resume=/dev/disk/by-uuid/$( hd_uuid_encode "$_swap_partition_hibernator" )" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
        fi

    else
        # no swap partition, but we can still add one
        swapfile_installer remove

        freespaceroot="$( LC_ALL="$EL_LC_EN" df -m | grep "^${HD_PART_ROOT_dev}" | awk '{print $4}' | tail -1 )"
        if [[ -n "$freespaceroot" ]] ; then
            # if the free space in the root partition is bigger than the ram size and we have at least X more gb free

            if [[ "$freespaceroot" -ge "$(( $RAM_TOTAL_SIZE_mb + 10000 ))" ]] && ! ((is_thanatests)) && [[ "$HD_PART_ROOT_FS_v" = ext* ]] ; then
                swapfile_add_auto

                # get values for add to grub:
                #fileswap_uuid="$( findmnt / -o UUID -n | tail -1 )"
                if [[ -e "$TARGET/swapfile.swp" ]] ; then
                    swapon "$TARGET/swapfile.swp" 2>/dev/null || true

                    fileswap_uuid="$( hd_uuid_encode "$HD_PART_ROOT_dev" )"
                    fileswap_offset="$( filefrag -v "$TARGET/swapfile.swp" | awk 'NR==4{gsub(/\./,"");print $4;}' )"

                    el_array_member_add "resume=UUID=$fileswap_uuid resume_offset=$fileswap_offset" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")

                    swapoff "$TARGET/swapfile.swp" 2>/dev/null || true
                fi
            fi
        fi
    fi

    # set 20 seconds instead of 5 if multiple boot options, or disable grub entirely if only elive!
    if [[ "$( timeout 90 os-prober 2>/dev/null | wc -l )" -gt 1 ]] || timeout 90 os-prober 2>/dev/null | grep -qsE "(Windows|Microsoft)" ; then
        sed -i "s|^.*GRUB_TIMEOUT.*=.*$|GRUB_TIMEOUT=22|g" "$TARGET/etc/default/grub"
    else
        # always show grub at least 2 seconds, we would want to run memtest (not working in ISO mode, only USB, wheezy), or the recovery mode (nurse) to recover configuration states or similar
        #sed -i "s|^.*GRUB_TIMEOUT.*=.*$|GRUB_TIMEOUT=0\nGRUB_HIDDEN_TIMEOUT=0\nGRUB_HIDDEN_TIMEOUT_QUIET=true\nGRUB_FORCE_HIDDEN_MENU=true|g" "$TARGET/etc/default/grub"
        #sed -i "s|^.*GRUB_TIMEOUT.*=.*$|GRUB_TIMEOUT=2|g" "$TARGET/etc/default/grub"
        # updated due to bilibop / impermutable feature
        sed -i "s|^.*GRUB_TIMEOUT.*=.*$|GRUB_TIMEOUT=3|g" "$TARGET/etc/default/grub"
    fi

    # Elive special confs:
    if ! grep -qs "^GRUB_DISABLE_SUBMENU=.y." "$TARGET/etc/default/grub" || ! grep -qs "^GRUB_SAVEDEFAULT=.true." "$TARGET/etc/default/grub" ; then
        cat >> "$TARGET/etc/default/grub" << EOF

# Elive extra confs:
# Disable recovery modes, we don't use them at all and if we do, we should have our own reparation tool like elive-nurse, by other side, the easiest and fastest way to fix an elive is by just reinstalling it (upgrade-mode integrated feature)
GRUB_DISABLE_RECOVERY="true"
# Disable submenus: they are pretty useless and annoying, we want to have our selected kernels well visually and easy to select since the first moment:
GRUB_DISABLE_SUBMENU="y"
# Always save our last selected boot: in any of the cases, is what we should want
GRUB_DEFAULT="saved"
GRUB_SAVEDEFAULT="true"

EOF
    fi

    # set grub elive themes by default:
    if ls "$TARGET"/usr/share/grub/themes/elive-*/theme.txt 1>/dev/null 2>&1 ; then
        if [[ "$( ls -1 "$TARGET"/usr/share/grub/themes/elive-*/theme.txt | wc -l )" -gt 1 ]] ; then
            el_warning "multiple grub themes included by default, selecting the last one..."
        fi
        grub_theme="$( ls -1 "$TARGET"/usr/share/grub/themes/elive-*/theme.txt | tail -f | sed -e "s|^${TARGET}||g" )"
        if [[ -n "$grub_theme" ]] ; then
            echo "GRUB_THEME=\"${grub_theme}\"" >> "$TARGET/etc/default/grub"
        fi
    fi


    # failsafe needed options?
    if cat /proc/cmdline | grep -qsE "(vga=|b43|noapic|noacpi|nomodeset|modeset=|nosmp|nolapic|nodma|noapm|nomce|failsafe)" ; then
        # add failsafe entries
        while read -ru 3 line
        do
            read -r line <<< "$line"
            [[ -z "$line" ]] && continue

            for i in "vga=" b43 noapic noacpi nomodeset "modeset=" nosmp nolapic nodma noapm nomce failsafe
            do
                if echo "$line" | grep -qs "$i" ; then
                    el_array_member_add "$line" "${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}" ; GRUB_CMDLINE_LINUX_DEFAULT_arr=("${_out[@]}")
                fi
            done
        done 3<<< "$( cat "/proc/cmdline" | tr ' ' '\n' | sort -u )"
        # real example options in version 2.9.14: username=eliveuser boot=live config components union=aufs vga=normal b43.blacklist=yes noapic noacpi nomodeset nosmp nolapic nodma noapm nomce failsafe loglevel=7 initrd=/live/initrd1.img BOOT_IMAGE=/live/vmlinuz1
    fi

    # mark a special boot mode
    if cat /proc/cmdline | grep -qsE "bootmode=" ; then
        if ! grep -qs "^special_boot_mode:" "$TARGET/etc/elive-version" ; then
            echo "special_boot_mode: $( grep "bootmode=" "/proc/cmdline" | tr ' ' '\n' | sed -e 's|^.*=||g' | grep bootmode | tail -1 )" >> "$TARGET/etc/elive-version"
        fi
    fi



    # convert variable to be compatible for sed:
    GRUB_CMDLINE_LINUX_DEFAULT="${GRUB_CMDLINE_LINUX_DEFAULT_arr[@]}"


    # user edits
    if cat /proc/cmdline | grep -qsE "(modeset|vga=|b43|noapic|noacpi|nomodeset|nosmp|nolapic|nodma|noapm|nomce|failsafe)" && ! ((is_thanatests)) ; then
        # ask to edit failsafe entries
        if ! ((conf_is_novice_user)) ; then

            if guitool yesnoquestion "$( eval_gettext "Elive booted with special options. These have been added to your installed system's boot options. If you know which options your computer needs, use only those. Edit the list to remove unnecessary ones?" )" ; then

                i="$( $guitool --title='Elive Installer' --entry --width=400 --text="$( eval_gettext "Change your default boot options? It's best to leave them as they are if you're unsure of what this means." )" --entry-text="${GRUB_CMDLINE_LINUX_DEFAULT}" )"
                if [[ -n "$i" ]] && [[ "$i" != "$GRUB_CMDLINE_LINUX_DEFAULT" ]] ; then
                    el_warning "user changed grub options from-to: ${GRUB_CMDLINE_LINUX_DEFAULT} --> ${i}"
                    GRUB_CMDLINE_LINUX_DEFAULT="$i"
                fi
            fi
        fi
    fi

    # write final GRUB_CMDLINE_LINUX_DEFAULT
    el_explain 0 "Configuring grub default options with: $GRUB_CMDLINE_LINUX_DEFAULT"
    cp -f "$TARGET/etc/default/grub" "$TARGET/etc/default/grub.bak"
    sed -i -e "s|^GRUB_CMDLINE_LINUX_DEFAULT.*$|GRUB_CMDLINE_LINUX_DEFAULT=\"${GRUB_CMDLINE_LINUX_DEFAULT}\"|g" "$TARGET/etc/default/grub"
    # el_warning "Diff for grub conf:\n$( diff -u "$TARGET/etc/default/grub.bak" "$TARGET/etc/default/grub" 2>&1 || true )"
    #el_warning "grub options: $( grep GRUB_CMDLINE_LINUX_DEFAULT "$TARGET/etc/default/grub" )"


    # do a backup of grub to see if something goes wrong:
    rm -f "$TARGET/etc/default/grub.bak"
    cp -fa "$TARGET/etc/default/grub" "$TARGET/etc/default/grub.bak"

    # }}}
    # Microsoft Windows hooks {{{
    # if windows set times to be LocalTime
    # UPDATE: we should have already set this from the geolocation script so.. disable it from here
    #if os-prober 2>/dev/null | grep -qs "Windows" ; then
        ## save the mode to use
        #mkdir -p "/etc/elive-tools/geolocation"
        #sed -e "/^bios mode:::/d" "/etc/elive-tools/geolocation/timezones.conf"

        #echo "# set to LOCAL just because you have a windows system installed" >>  "/etc/elive-tools/geolocation/timezones.conf"
        #echo "bios mode:::LOCAL" >>  "/etc/elive-tools/geolocation/timezones.conf"
        ## indent
        #ed -s /etc/adjtime << EOF
#3d
#a
#LOCAL
#.
#w
#EOF
        ## update the files on the installed system too
        #mkdir -p "$TARGET/etc/elive-tools/geolocation"
        #cp -a /etc/elive-tools/geolocation/timezones.conf "$TARGET/etc/elive-tools/geolocation/"
    #fi

    # - Microsoft Windows hooks }}}
    # sudo configurations {{{
    # Update: this feature is deprecated, but let's continue use it in older versions of elive:
    if ! grep -qs "sudoers.d/sudo_nopasswd_" "/usr/sbin/user-manager" ; then
        rm -f "$TARGET/etc/sudoers.d/timeout_inmediate" # delete old version name

        # set a fast timeout because of security reasons (1 minute)
        echo -e "# Timeout to ask again password for next sudo commands: 1 minute\nDefaults:ALL timestamp_timeout=1" > "$TARGET/etc/sudoers.d/timeout_password"
        # network privileges
        echo -e "# User should always have access to rfkill and ifconfig, elive has an unblock wifi feature in desktop startup that requires it too\n%sudo   ALL=(ALL) NOPASSWD: /sbin/ifconfig\n%sudo   ALL=(ALL) NOPASSWD: /usr/sbin/rfkill" > "$TARGET/etc/sudoers.d/privileges_network"
        # elive upgrader privileges
        echo -e "# We don't want to ask for admin password in every security update needed of the system\n%sudo   ALL=(ALL) NOPASSWD: /usr/sbin/elive-upgrader-root" > "$TARGET/etc/sudoers.d/privileges_upgrader"
        # brightness control
        echo -e "# User should always be able to change the brightness of the screen\n%sudo   ALL=(ALL) NOPASSWD: /usr/bin/brightnessctl" > "$TARGET/etc/sudoers.d/privileges_brightness"
        # profile-sync-daemon
        echo -e "# profile-sync-daemon which is needed for the browsers require privileges\n%sudo   ALL=(ALL) NOPASSWD: /usr/bin/psd-overlay-helper" > "$TARGET/etc/sudoers.d/privileges_psd"

        chmod 0440 "$TARGET/etc/sudoers.d/timeout_password"
        chmod 0440 "$TARGET/etc/sudoers.d/privileges_network"
        chmod 0440 "$TARGET/etc/sudoers.d/privileges_upgrader"
        chmod 0440 "$TARGET/etc/sudoers.d/privileges_psd"
        chmod a+r "$TARGET/etc/sudoers.d/privileges_upgrader" # we need this for the tool itself to know that we can have access
    fi

    # }}}

    # modules {{{
    for module in "${modules_extra_initramfs[@]}"
    do
        echo "$module" >> "$TARGET/etc/initramfs-tools/modules"
    done

    # }}}
    # Copy things from Live configured {{{
    # update our time no-matter-what, because the boot could not boot if thinks that FS was last mounted in the future
    cp -a /etc/adjtime "$TARGET/etc/" 2>/dev/null


    # enable password-based ssh because thats the default one, now that we have own real passwords
    # update: ssh should not allow logins with password by default due to security reasons
    if ((is_ssh_password_enabled)) ; then
        sed -i -E -e "s|#?PasswordAuthentication .*$|PasswordAuthentication yes|g" "$TARGET/etc/ssh/sshd_config" 2>/dev/null || true
    else
        sed -i -E -e "s|#?PasswordAuthentication .*$|PasswordAuthentication no|g" "$TARGET/etc/ssh/sshd_config" 2>/dev/null || true
    fi
    if ((is_ssh_root_enabled)) ; then
        sed -i -E -e "s|#?PermitRootLogin .*$|PermitRootLogin yes|g" "$TARGET/etc/ssh/sshd_config" 2>/dev/null || true
    else
        sed -i -E -e "s|#?PermitRootLogin .*$|#PermitRootLogin yes\nPermitRootLogin prohibit-password|g" "$TARGET/etc/ssh/sshd_config" 2>/dev/null || true
    fi

    # encrypted disks require this option (/boot is not needed anymore)
    if ((is_encryption_wanted)) ; then
        echo -e "# This option is required to make the nice theme appear and also to not raise errors on installation of grub when the root partition is encrypted:\nGRUB_ENABLE_CRYPTODISK=y" >> "$TARGET/etc/default/grub"
    fi

    echo "70" > "$TMP_PROGRESS_f"
    echo -e "# Generating Grub and Initrds" >> "$TMP_PROGRESS_f"

    # copy the lists of packages installed
    mkdir -p "$TARGET/etc/elive/system/packages"
    cp -f "$TARGET/tmp/.packages-to-reconfigure" "$TARGET/etc/elive/system/packages/packages-to-reconfigure" 2>/dev/null || true
    cp -f "$TARGET/tmp/.packages-to-install"     "$TARGET/etc/elive/system/packages/packages-to-install"     2>/dev/null || true
    cp -f "$TARGET/tmp/.packages-to-remove"      "$TARGET/etc/elive/system/packages/packages-to-remove"      2>/dev/null || true
    cp -f "$TARGET/tmp/.packages-to-hold"        "$TARGET/etc/elive/system/packages/packages-to-hold"        2>/dev/null || true

    # save audio configurations:
    cp -f "/usr/share/alsa/alsa.conf" "$TARGET/usr/share/alsa/"


    # enable HD verifications to notify users if errors were found
    if [[ -s "$TARGET/etc/default/smartmontools" ]] && ! ((is_virtualized)) ; then
        sed -i "s|^#start_smartd=yes|start_smartd=yes|g" "$TARGET/etc/default/smartmontools"
    fi

    # checks
    if grep -qEi "(Xinerama|TwinView)" "$TARGET/etc.old-to-${eliveversion}/X11/xorg.conf" 2>/dev/null ; then
        guitool info "$( eval_gettext "An Xorg configuration has been found for multiple screens. We are assuming that you do not need it anymore. Settings of multiple screens will work with the new Elive settings. In case you need the old version, you can find it in a backup copy of your:" ) /etc"
    fi


    # copy back the SSH keys of the host identifiers if any
    cp -fa $TARGET/etc.old-to-${eliveversion}/ssh/ssh_host_*key* "$TARGET/etc/ssh/" 2>/dev/null || true

    # copy back the list of WIFI connections that we remember, but not the entire conf! (only for elive)
    if ((is_old_elive_upgrade_request_accepted)) ; then
        cp -fa $TARGET/etc.old-to-${eliveversion}/NetworkManager/system-connections/* "$TARGET/etc/NetworkManager/system-connections/" 2>/dev/null || true
    fi

    # }}}
    # set machine-id {{{
    # create /etc/machine-id
    if ((is_sysvinit_wanted)) ; then
        # moved to the full chroot command
        true
    else
        #rm -f "$TARGET/var/lib/dbus/machine-id" "$TARGET/etc/machine-id"
        #chroot "$TARGET" systemd-machine-id-setup 1>/dev/null 2>&1 || true
        #ln -fs /etc/machine-id "$TARGET/var/lib/dbus/machine-id"

        # update: blanking out the machine-id will generate a new one
        : > "$TARGET/etc/machine-id"
    fi
    # - set machine-id }}}
    # set /etc confs {{{
    # know if the user wants to send reports
    if [[ -n "$conf_send_debug_reports" ]] ; then
        mkdir -p "$TARGET/etc/elive"
        sed -i "/^conf_send_debug_reports=/d" "$TARGET/etc/elive/settings" 2>/dev/null || true
        echo -e "conf_send_debug_reports=\"$conf_send_debug_reports\"" >> "$TARGET/etc/elive/settings"
    fi

    if [[ -n "$conf_send_debug_reports_email" ]] ; then
        sed -i "/^conf_send_debug_reports_email=/d" "$TARGET/etc/elive/settings" 2>/dev/null || true
        echo -e "conf_send_debug_reports_email=\"$conf_send_debug_reports_email\"" >> "$TARGET/etc/elive/settings"

        # verify if our reporter email is a patreon member
        if [[ -z "$patreon_email" ]] ; then
            patreon_email_checksum="$( echo "$conf_send_debug_reports_email" | sha1sum | awk '{print $1}' )"

            if curl -Ls -m 10 -A "Mozilla/5.0" https://www.elivecd.org/files/patreon_members.txt | grep -qs "^${patreon_email_checksum}$" ; then
                sed -i "/^patreon_email=/d" "$TARGET/etc/elive/settings" 2>/dev/null || true
                echo -e "patreon_email=\"$conf_send_debug_reports_email\"" >> "$TARGET/etc/elive/settings"
            fi
        fi
    fi

    # make a copy to /boot
    mkdir -p "$TARGET/boot/etc/elive"
    cp -f "$TARGET/etc/elive/settings" "$TARGET/boot/etc/elive/" 2>/dev/null || true

    if [[ -s "/tmp/.connection-speed-average.txt" ]] ; then
        echo -e "internet-average-speed: $( cat "/tmp/.connection-speed-average.txt" )" >> "$TARGET/etc/elive-version"
    fi

    sed -i '/^kernel:.*$/d' "$TARGET/etc/elive-version"
    echo -e "kernel: $(uname -r)" >> "$TARGET/etc/elive-version"
    sed -i '/^machine-id:.*$/d' "$TARGET/etc/elive-version"
    echo -e "machine-id: $( el_get_machine_id )" >> "$TARGET/etc/elive-version"

    # make virtualbox be silent:
    if [[ -s "$TARGET/etc/default/virtualbox" ]] && ! grep -qs "^log_progress_msg" "$TARGET/etc/default/virtualbox" ; then
        cat >> "$TARGET/etc/default/virtualbox" << EOF
log_begin_msg(){
    echo "\$@"
}
log_progress_msg(){
    echo "\$@"
}
log_failure_message(){
    echo "\$@" 1>&2
}
log_success_msg(){
    echo "\$@"
}
log_end_msg(){
    echo "\$@"
}
EOF
    fi

    # - set /etc confs }}}
    # fixes on the system {{{
    # sometimes these files has wrong permissions, for example if they come from a migraton-mode:
    chmod 0600 /etc/ssh/ssh_host_*key* 2>/dev/null || true

    # - fixes on the system }}}

    # /boot mount {{{

    if [[ -b "$HD_PART_EFI_dev" ]] ; then
        el_umount "$HD_PART_EFI_dev" 1>/dev/null
    fi
    if [[ -b "$HD_PART_BOOT_dev" ]] ; then
        el_umount "$HD_PART_BOOT_dev" 1>/dev/null
    fi

    # /boot should be mounted now, after to have removed extra kernels and before to install grub2
    if [[ -b "$HD_PART_BOOT_dev" ]] ; then

        # prepare to move data
        #rm -f "${TARGET}/boot/"initrd*old-dkms 2>/dev/null
        rm -f "${TARGET}/boot/"initrd* 2>/dev/null

        mv "$TARGET/boot" "$TARGET/boot_temporal"
        mkdir -p "$TARGET/boot"

        if [[ -n "$HD_PART_BOOT_fs" ]] ; then
            if ! mount -o dev -t "${HD_PART_BOOT_fs}" "${HD_PART_BOOT_dev}" "$TARGET/boot" ; then
                el_warning "Unable to mount /boot partition (wrongly FS detected?):  mount -o dev -t '${HD_PART_BOOT_fs}' '${HD_PART_BOOT_dev}' '$TARGET/boot'"
                HD_PART_BOOT_fs="$( hdtools_get_fs "${HD_PART_BOOT_dev}" )"

                if ! mount -o dev -t "${HD_PART_BOOT_fs}" "${HD_PART_BOOT_dev}" "$TARGET/boot" ; then
                    el_error "Unable to mount /boot partition to your installed system as:  mount -o dev -t '${HD_PART_BOOT_fs}' '${HD_PART_BOOT_dev}' '$TARGET/boot'"
                    guitool error "Unable to mount /boot partition to your installed system"
                    exit_error
                fi
            fi
        else
            el_warning "unknown FS to mount the /boot ${HD_PART_BOOT_dev} partition? ok assume auto..."
            if ! mount -o dev "${HD_PART_BOOT_dev}" "$TARGET/boot" ; then
                el_error "Unable to mount /boot partition to your installed system as:  mount -o dev '${HD_PART_BOOT_dev}' '$TARGET/boot'"
                guitool error "Unable to mount /boot partition to your installed system"
                exit_error
            fi
        fi

        # move files
        if rsync -aq --delete --delete-before "$TARGET/boot_temporal/" "$TARGET/boot/" ; then
            rm -rf "$TARGET/"boot_temporal
        else
            el_error "error in /boot partition?:\n$(dmesg | tail -n 14)"
            guitool error "$( eval_gettext "Your boot partition has run out of free space. The installation will continue, but your system may not be able to boot. Consider increasing the size of your boot partition and installing Elive again. Suggested size:" ) 400 MB"
        fi
    fi


    # include EFI setup
    if [[ -n "$HD_PART_EFI_dev" ]] ; then
        mkdir -p "$TARGET/boot/efi"

        # mount it
        if mount -o dev -o rw -t "${HD_PART_EFI_FS_v}" "${HD_PART_EFI_dev}" "$TARGET/boot/efi" ; then
            # verify / create efi directory
            if ! ls -1 "$TARGET/boot/efi/" | grep -qsi "^EFI$" ; then
                mkdir -p "$TARGET/boot/efi/EFI"
            fi

            # add to fstab
            echo -e "$( hd_uuid_encode_fstab "${HD_PART_EFI_dev}" )    /boot/efi    ${HD_PART_EFI_FS_v}    defaults         0    2" >> "$TARGET/etc/fstab"
            el_debug "$( hd_uuid_encode_fstab "${HD_PART_EFI_dev}" )    /boot/efi    ${HD_PART_EFI_dev}    defaults         0    2"
        else
            el_error "unable to mount EFI partition on $TARGET/boot/efi as:   mount -o dev -o rw -t '${HD_PART_EFI_FS_v}' '${HD_PART_EFI_dev}' '$TARGET/boot/efi'\n$( mount | grep "$TARGET" )"
            guitool warning "Unable to mount EFI partition on the target system, switching to MBR mode. But the installation can fail."
            unset HD_PART_EFI_dev
        fi

    fi



    # - /boot mount }}}
    # CHROOT post: rebuild INITRD's (deinstalled kernels) {{{
local message_initrd_regeneration_failed
message_initrd_regeneration_failed="$( printf "$( eval_gettext "Issue encountered while regenerating bootloaders. Possible causes include a full boot partition or insufficient free space." )" "" )"
local message_efi_disk_full
message_efi_disk_full="$( printf "$( eval_gettext "Your EFI NVRAM is full. You need to manually delete unused entries to free up space. The identifiers are the 4 numbers following 'BOOT'. If unsure, close the terminal. Use a command like: %s" )" "efibootmgr -b IDENTIFIER -B" )"
local message_efi_delete_all
message_efi_delete_all="$( printf "$( eval_gettext "Your previous attempt to install Elive failed. Do you want to remove all boot entries from your UEFI's NvRam? This will prevent other installed operating systems from booting, but Elive will still list them in its boot menu. Select this option especially if you are unable to boot correctly on this machine." )" "efibootmgr -b IDENTIFIER -B" )"

# nvidia confs
if [[ -s "/tmp/.nvidia-version" ]] ; then
    nvidia_version="$( cat /tmp/.nvidia-version )"

    nvidia_installer_extraoptions=""
    # UPDATE: this option is a bit broken: if you install legacy-390xx and your driver is not supported by that cuda, the install fails, but it makes nvidia also fail to install! leaving an entirely broken system, cuda may be an advanced feature for users to install manually
    if guitool yesnoquestioninv "$( eval_gettext "Would you like to install the full suite of packages for CUDA and other components for your Nvidia graphics card? This will add an additional 2GB to your installation. We do not recommend this option as it is highly likely that these components are not necessary." )" ; then
        nvidia_installer_extraoptions="-x"
    fi
fi

    if [[ -s "$TARGET/tmp/temp" ]] ; then
        el_warning "file '$TARGET/tmp/temp' was not empty, but we are going to remove it before to run its previous contents:\n$(cat $TARGET/tmp/temp)"
    fi
    cat > "$TARGET/tmp/temp" << EOF
#!/bin/bash
export LC_ALL="$EL_LC_EN"
export EL_DEBUG="$EL_DEBUG"
[[ "$EL_DEBUG" -gt 3 ]] && set -x || true
source /usr/lib/elive-tools/functions
set -E

# ignore accidentals Ctrl + C
trap '' SIGINT


export TERM=xterm-256color

if (($is_previous_installation_failed)) ; then
    export DEBIAN_FRONTEND=gnome
else
    export DEBCONF_NONINTERACTIVE_SEEN=true
    export DEBCONF_NOWARNINGS=true
    export DEBIAN_FRONTEND=noninteractive
    export DEBIAN_PRIORITY=critical
fi

trap "exit_ok" EXIT

exit_ok(){
    touch /tmp/.eliveinstaller-chroot-tasks-done
}


echo -e "\n\n      Working, do not close this terminal...\n\n"
sync

# always update the correct time clock, and show info to user
if [[ -x "/etc/NetworkManager/dispatcher.d/timezone-update" ]] ; then
    el_info "Setting your timezone"
    EL_DEBUG=2 /etc/NetworkManager/dispatcher.d/timezone-update -f
fi

if [[ -s "/root/preseeds.cfg" ]] ; then
    debconf-set-selections < "/root/preseeds.cfg"
    el_debug "debconf set selections: \$( cat /root/preseeds.cfg | sort -u ) "
fi

# install required things?
#apt-get -f -y install 1>/dev/null

if (($is_nvidia)) ; then
    el_info "Updating Nvidia installer..."

    # update elive-tools
    if NOREPORTS=1 el_package_upgrade_fast 220 force "elive-tools" ; then
        el_info "Updated Nvidia installer"
        # update functions always
        source /usr/lib/elive-tools/functions
    fi

    # avoid fails (but in such case, it is preselected debconf to "yes continue install" ?)
    # update: we want in fact to fail, so that the installation fail and we re-install again without it
    #if [[ "$nvidia_installer_extraoptions" = *"-x"* ]] ; then
        #export OVERRIDE_NVIDIA_LEGACY_CHECK_UNSUPPORTED_DEVICES=1
    #fi

    if [[ -n "$nvidia_version" ]] ; then
        el_info "Installing Nvidia using version '${nvidia_version}' ..."
        if ! nvidia-privative-drivers-install -s -f -a -q -m "$nvidia_version" $nvidia_installer_extraoptions ; then
            NOREPORTS=1 el_warning "failed installing nvidia, trying again without extra options '$nvidia_installer_extraoptions' ..."
            nvidia-privative-drivers-install -s -f -a -q -u || true
            nvidia-privative-drivers-install -s -f -a -q -m "$nvidia_version" || true
        fi

    else
        el_info "Installing Nvidia..."
        if ! nvidia-privative-drivers-install -s -f -a -q $nvidia_installer_extraoptions ; then
            nvidia-privative-drivers-install -s -f -a -q -u || true
            nvidia-privative-drivers-install -s -f -a -q || true
        fi
    fi
fi
if (($is_fglrx)) ; then
    fglrx-privative-drivers-install -f -a -q || true
fi


# reconfigure each needed preseed which is packagename
if [[ -d "/root/preseeds.d" ]] ; then
    for file in /root/preseeds.d/*
    do
        if [[ -s "\$file" ]] ; then
            debconf-set-selections < "\$file"
            if ! echo "\$file" | grep -qsi "noreconfigure" ; then
                el_info "Reconfiguring \$( basename \${file%%_*} ):"
                cat "\${file}" 1>&2
                DEBIAN_FRONTEND="" dpkg-reconfigure -fnoninteractive -pcritical --no-reload "\$( basename \${file%%_*} )" 2>/dev/null || true
            fi
        fi
    done
fi


el_info "Building Initrd's ..."

# INFO NOTE: initrd's always needs to be rebuilt in order to know the new cryptsetup information from where to boot from
# so the correct order is: - configure /etc, - have mounted the disks correctly (like, manually /dev/mapper/sda5_crypt /somwhere, if you mount the partition from thunar it will not work because is mounted from a different disk name (like /dev/mapper/udisks-luks-uuid-something... ), which then the boot will fail (disk not found! of course!) - rebuild initrds to have them, - rebuild grub (update-grub, grub-setup, grub-install) - these steps are verified that works this way from a manual install

if ((is_wheezy)) || [[ -e /usr/sbin/update-initramfs.orig ]] ; then
    mv -f /usr/sbin/update-initramfs.orig /usr/sbin/update-initramfs || true
fi

initrd_status="0"
rm -f /boot/initrd.img* 2>/dev/null || true
LC_ALL="$EL_LC_EN" update-initramfs -k all -d -c 2>&1 | grep -vE '(^mkdir:|mdadm:|Generating /|live-boot)'
initrd_status="\${PIPESTATUS[0]}"

if [[ "\$initrd_status" != 0 ]] || ! ls /boot/initrd.img* 1>/dev/null 2>&1 ; then
    LC_ALL="$EL_LC_EN" update-initramfs -k $( uname -r ) -d -c 2>&1 | grep -vE '(^mkdir:|mdadm:|Generating /|live-boot)'
    initrd_status="\${PIPESTATUS[0]}"

    if [[ "\$initrd_status" != 0 ]] ; then
        LC_ALL="$EL_LC_EN" update-initramfs -k all -d -c 2>&1 | grep -vE '(^mkdir:|mdadm:|Generating /|live-boot)'
        el_error "problem generating initramfs?"
        zenity --error --text="$message_initrd_regeneration_failed"
    fi
fi
rm -f /boot/*old-dkms 2>/dev/null


function_efi_delete_all(){
    # delete dups
    efibootmgr --remove-dups -q

    echo -e "EFI Before:"
    efibootmgr -v

    # list all UUIDs that doesn't exist and delete these entries
    while read -ru 3 line
    do
        [[ -z "\$line" ]] && continue
        for entry in \$( efibootmgr -v | grep "\$line" | awk '{print \$1}' | sed -e 's|\*$||g' | sed -e 's|^Boot||g' | grep "^....$" )
        do
            if [[ -n "\$entry" ]] ; then
                echo "deleting \$entry boot ID:" 1>&2
                efibootmgr -b "\$entry" -B -q
            fi
        done
    done 3<<<"\$( efibootmgr -v | grep -viE "(CD|DVD|Removable|Built-in|Shell|Internal|Network|UiApp|FvVol)" | grep -E "^Boot[[:digit:]]+.*\(.*[[:alnum:]]+-[[:alnum:]]+-[[:alnum:]]+-" | tr ',' '\n' | grep -E "[[:alnum:]]+-[[:alnum:]]+-[[:alnum:]]+-" | sed -e 's|).*\$||g' -e 's|^.*(||g' | sort -u )"

    echo -e "\nEFI After:"
    efibootmgr -v

}

function_efi_delete_unused_entries(){
    # delete dups
    efibootmgr --remove-dups -q

    efibootmgr_list="\$( efibootmgr -v 2>&1 )"
    echo -e "EFI Before:\n\$efibootmgr_list"

    # list all UUIDs that doesn't exist and delete these entries
    while read -ru 3 line
    do
        [[ -z "\$line" ]] && continue
        if ! lsblk -o PARTUUID | grep -qsi "\$line" ; then
            echo "UUID \$line not found, deleting it:\n\$( echo "\$efibootmgr_list" | grep "\$line" )" 1>&2
            for entry in \$( efibootmgr -v | grep "\$line" | awk '{print \$1}' | sed -e 's|\*$||g' | sed -e 's|^Boot||g' | grep "^....$" )
            do
                if [[ -n "\$entry" ]] ; then
                    echo "deleting \$entry boot ID:" 1>&2
                    efibootmgr -b "\$entry" -B -q
                    is_efi_entry_deleted=1
                fi
            done
        fi
    done 3<<<"\$( efibootmgr -v | grep -viE "(CD|DVD|Removable|Built-in|Shell|Internal|Network|UiApp|FvVol)" | grep -E "^Boot[[:digit:]]+.*\(.*[[:alnum:]]+-[[:alnum:]]+-[[:alnum:]]+-" | tr ',' '\n' | grep -E "[[:alnum:]]+-[[:alnum:]]+-[[:alnum:]]+-" | sed -e 's|).*\$||g' -e 's|^.*(||g' | sort -u )"

    echo -e "EFI After:"
    efibootmgr -v

}




# rebuild grub (we changed the resolution), this needs to be done after to have regenerated the initrd's
el_info 'Building Grub configuration'

# check
if ! grep -qs "grub" /root/preseeds.d/* 2>/dev/null ; then
    el_warning "no grub preseeds included:\n\$(ls -la /root/preseeds.d/ ; cat /root/preseeds.d/* )"
fi

if ! grep -qsi "^GRUB_DISABLE_OS_PROBER=.*false" "/etc/default/grub" ; then
    sed -i -e '/^GRUB_DISABLE_OS_PROBER=/d' "/etc/default/grub"
    echo -e "# allow other operating systems appear in your boot menu\nGRUB_DISABLE_OS_PROBER=\"false\"" >> "/etc/default/grub"
fi

if [[ -n "$HD_PART_EFI_dev" ]] ; then
    el_info 'GRUB in EFI mode'

    if ! (($is_efibootmgr_available)) ; then
        rm -rf /boot/efi/EFI/$efi_name 2>/dev/null
    fi

    #grub-install

    # note: reconfigures are needed to use the forced efi conf from preseeds
    # EFI mode:

    # always reconfigure manually if failed previously
    if (($is_efi_64)) ; then
        # enable listing of other os's, update: doesn't seems to work
        echo -e "grub-efi-amd64\tgrub2/enable_os_prober\tboolean\ttrue" | debconf-set-selections
        dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub-efi-amd64 2>/dev/null 1>&2
    fi
    if (($is_efi_32)) ; then
        # enable listing of other os's
        echo -e "grub-efi-ia32\tgrub2/enable_os_prober\tboolean\ttrue" | debconf-set-selections
        dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub-efi-ia32 2>/dev/null 1>&2
    fi
    # enable os-prober
    echo -e "grub-pc\tgrub2/enable_os_prober\tboolean\ttrue" | debconf-set-selections
    dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub-pc 2>/dev/null 1>&2

    # important: to set the folder name for the bootloader-id, the command must be run and after it, the normal grub-install
    # grub-install --bootloader-id=elive

    # fix for sparsefiles in grub
    if (($is_fix_grub_sparsefiles_wanted)) ; then
        sed -i -e 's|^.*GRUB_DEFAULT=.*$|GRUB_DEFAULT=0|g' "/etc/default/grub"
        sed -i -e 's|^.*GRUB_SAVEDEFAULT=.*$|GRUB_SAVEDEFAULT="false"|g' "/etc/default/grub"
        # re-update
        # update-grub
    fi

    # always delete unused efi entries, let's keep it clean and without memory exhausted states
    if (($is_efibootmgr_available)) ; then
        function_efi_delete_unused_entries
    fi

    # cleanup the entire EFI entries to free it from other systems, if user wants to
    if (($is_previous_installation_failed)) && (($is_efibootmgr_available)) ; then
        if zenity --question --text="$message_efi_delete_all" ; then
            function_efi_delete_all
        fi
    fi

    # install grub and check its logs
    logs="\$( update-grub 2>&1 ; grub-install 2>&1 )"


    # fixes
    # do a few loops in order to finish all fixes (one fixed can lead to another found one)
    for i in \$(seq 3)
    do
        # no space left errors:
        if echo "\$logs" | grep -qsi "grub-install: error:.*No space left" ; then

            set -x

            if ((is_efi_entry_deleted)) ; then
                zenity --error --text="$message_efi_disk_full"
                urxvt -g 110x42 -e bash -c "efibootmgr -v ; echo ; echo 'ELIVE HELPER: Delete manually the BootXXXX unused entries like:  efibootmgr -b xxx -B' ; bash -l"
            fi

            if (($is_efibootmgr_available)) ; then
                function_efi_delete_unused_entries
            fi

            # update results
            logs="\$( update-grub 2>&1 )"
        fi

        # found a missing grub-pc-bin dependency? install it
        if echo "\$logs" | grep -qs "grub-install: error:.*modinfo.sh" ; then

            set -x

            buf="\$( echo "\$logs" | grep -oE "grub-install: error: .* doesn't exist" | tr ' ' '\n' | grep "/usr/lib/grub/.*/modinfo.sh" | tail -1 )"
            case "\$buf" in
                *"/i386-pc/"*)        package="grub-pc" ; ;;
                *"/i386-coreboot/"*)  package="grub-coreboot" ; ;;
                *"/i386-efi/"*)       package="grub-efi-ia32" ; ;;
                *"/x86_64-efi/"*)     package="grub-efi-amd64" ; ;;
                *"/x86_64-emu/"*)     package="grub-emu" ; ;;
                *"/i386-ieee1275/"*)  package="grub-ieee1275" ; ;;
                *"/x86_64-xen/"*)     package="grub-xen" ; ;;
                *"/i386-xen"*)       package="grub-xen" ; ;;
                *)                  el_error "unknown grub package for: \$buf" ; ;;
            esac

            echo -e "Installing missing dependency for grub: \$package" 1>&2


            # install missing dependency
            if apt-get install -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confnew" -y \$package ; then
                apt-get -y autoremove
                logs="\$( update-grub 2>&1 )"
            else
                el_error "failed installing missing dependency for grub: \$package\nfrom buf: \$buf\nfrom logs: \$logs"
            fi
        fi

    done

    set +x

    # show final results of grub installation so that the next step can show possible errors
    echo "update-grub logs:" 1>&2
    echo "\$logs" 1>&2

    # install grub
    grub-install

    if ! grep -qsi "debian" /tmp/os-probed && [[ -d "/boot/efi/EFI/debian" ]] && [[ ! -d "/boot/efi/EFI/elive" ]] ; then
        el_warning "EFI was installed as debian but no debian system is installed, there's not an elive one. Why?:\n\$( ls -1 /boot/efi/EFI/ )"
        echo -e "secureboot: '$is_secureboot_enabled', is_efi_32 = '$is_efi_32', is_efi_64 = '$is_efi_64', is_i386 = '$is_i386', is_amd64 = '$is_amd64'"

    else
        echo -e "EFI dirs:\n\$( ls -1 /boot/efi/EFI/ )"
    fi

    # debug
    echo -e "Installed grub packages:\$( dpkg -l | grep -iE "grub" | awk '{print \$2}' | sort -u )"

else

    el_info 'GRUB in MBR/BIOS (legacy) mode'

    # enable listing of other OS's
    echo -e "grub-pc\tgrub2/enable_os_prober\tboolean\ttrue" | debconf-set-selections

    # MBR mode or wheezy:
    dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub2
    dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub-common
    # FIXME: add a report in case that the normal grub install fails (error messages), this one:
    dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub-pc

    # force to install grub if the previous one fails (seems like a check about reiserfs don't wants it)
    el_info 'Installing GRUB in $mbr_device_install '

    # grub-setup only exists in wheezy
    if (($is_wheezy)) ; then
        grub-setup "$mbr_device_install" || grub-setup --skip-fs-probe "$mbr_device_install" || grub-setup --force --skip-fs-probe "$mbr_device_install" || dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub-pc
    fi

    #grub-install --no-floppy --recheck "$mbr_device_install" # dont use recheck, it will update the device.map (and it's supposed that is correct, or it will skip any possible our-modification on it)
    grub-install --no-floppy "$mbr_device_install"
    grub-install "$mbr_device_install"

    # just run again the one that should be correct:
    dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub2
    dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub-common
    dpkg-reconfigure -fnoninteractive -pcritical --no-reload grub-pc

    update-grub
    grub-install "$mbr_device_install"
    sync
fi


# reinstall refind customized for elive
if (($is_apple)) ; then
    /usr/lib/eliveinstaller/refind-install-elive-customization
fi

# cleanup apt
el_info 'APT autoremoving'
if ! (($is_packages_remove_ignored)) ; then
    DEBIAN_FRONTEND=noninteractive  apt-get -f --purge -s autoremove | grep -E "^(Remv|Purg)\s+" | awk '{print \$2}' | sort -u > "/tmp/.packages-to-remove_autoremoved"
    DEBIAN_FRONTEND=noninteractive  apt-get -y autoremove
fi

# generate a list of installed packages
mkdir -p "/etc/elive/system/packages"
COLUMNS=1000 dpkg -l  > "/etc/elive/system/packages/packages-installed-clean"

# generate machine-id for non-systemd machines
if (($is_sysvinit_wanted)) ; then
    rm -f /etc/machine-id /var/lib/dbus/machine-id
    dbus-uuidgen --ensure=/etc/machine-id
    ln -fs /etc/machine-id /var/lib/dbus/machine-id
fi

echo -e "\nDONE"
sleep 1

touch "/tmp/.eliveinstaller-chroot-tasks-done"

EOF
    chmod +x "$TARGET/tmp/temp"

    # second pass run !
    # TODO move this to a "last" script-run action in hooks_end
    if [[ -s "$TARGET/tmp/temp" ]] ; then
        sync
        # {{{
        if ! [[ -n "$HD_PART_EFI_dev" ]] ; then
            if ! [[ "${HD_DEV}" = "$mbr_device_install" ]] ; then
                el_explain 0 "Going to install the bootloader in $mbr_device_install which is not the same as ${HD_DEV}, if you think that this is not correct please report the bug (probably user selected it manually)"
            fi

            if [[ "$HD_DEV" != "/dev/mmcblk"* ]] && [[ "$HD_DEV" != "/dev/nvme"* ]] && [[ "$HD_DEV" != "/dev/md"*p* ]] && echo "$HD_DEV" | grep -q '[0-9]$' ; then
                el_warning "Looks like you are going to install the bootloader in a partition and not in the main disk?"
            fi

        fi

        # }}}

        echo "85" > "$TMP_PROGRESS_f"
        echo -e "# $( eval_gettext "Finishing configuring your system..." )" >> "$TMP_PROGRESS_f"

        # mount with bind first
        EL_DEBUG=0 mounts-manager --chroot-layers-mount "$TARGET" "eliveinstaller" bind 1>/dev/null

        # run all !
        #if { ((is_tmux_unwanted)) || ((conf_is_novice_user)) || ((conf_is_extra_help_wanted)) ; } && ! ((is_tmux_wanted)) ; then
        if ((is_tmux_unwanted)) ; then
            chroot "$TARGET" "/tmp/temp" 2>&1 | tee "${TMP_LOGS_REPORT_f}"
        else
            # test if we can run in a detached mode, so the user can close terminal without problems
            if [[ -x /usr/bin/tmux ]] && [[ -x /usr/bin/tmux-attach-jobs ]] ; then
                sed -i -e 's|^if-shell.*message_welcome.*$|#&|g' "/root/.tmux.conf"
                sed -i -e 's|^if-shell.*message_welcome.*$|#&|g' "/home/$eliveuser/.tmux.conf"
                timeout 10 tmux-attach-jobs eliveinstaller start "touch '${TMP_MAIN_d}/test-tmux-attach.txt'" 1>/dev/null
                sleep 1 ; sync
            fi
            # run it, undetached or direct if we don't have tmux feature
            if [[ -e "${TMP_MAIN_d}/test-tmux-attach.txt" ]] ; then
                rm -f "${TMP_MAIN_d}/test-tmux-attach.txt"

                sed -i -e 's|^if-shell.*message_welcome.*$|#&|g' "/root/.tmux.conf"
                sed -i -e 's|^if-shell.*message_welcome.*$|#&|g' "/home/$eliveuser/.tmux.conf"
                tmux-attach-jobs eliveinstaller rebuilds "chroot '$TARGET' bash -c '/tmp/temp' 2>&1 | tee '${TMP_LOGS_REPORT_f}' " 1>/dev/null
                sleep 2 ; sync

                # visualize tasks if we have it already not opened
                if ! LC_ALL="$EL_LC_EN" ps aux | grep -v grep | grep -qs "rxvt.*tmux attach" ; then
                    urxvt -e bash -c "unset TMUX ; tmux attach -t eliveinstaller" &
                fi
            else
                urxvt -e bash -c "chroot '$TARGET' bash -c '/tmp/temp' 2>&1 | tee '${TMP_LOGS_REPORT_f}' " &
            fi
        fi

        # continue after to have finished the tasks
        while ! test -e "$TARGET/tmp/.eliveinstaller-chroot-tasks-done"
        do
            sleep 2
        done
        rm -f "$TARGET/tmp/.eliveinstaller-chroot-tasks-done"
        rm -f "$TARGET/tmp/temp"

        # something in grub install (uefi probably) switches screen to blanking, so re-enable it here
        screen_dpms_on

        # report debug, and check
        if [[ -s "$TMP_LOGS_REPORT_f" ]] ; then
            cat "${TMP_LOGS_REPORT_f}" 1>&2

            # nvidia drivers failed?
            if cat "${TMP_LOGS_REPORT_f}" | colors-remove | grep -qs "nvidia.*deb" ; then
                if cat "${TMP_LOGS_REPORT_f}" | colors-remove | grep -qsE "(Driver compilation seems to have failed)" ; then
                    guitool warning "$( eval_gettext "The Nvidia driver installation appears to have failed. Refer to our forum for a detailed tutorial. You may need to follow manual steps to set up your system." )"
                fi
            fi

            # grub install failed?
            if ! cat "${TMP_LOGS_REPORT_f}" | colors-remove | grep -qsE "Installation finished. No error reported." ; then
                if cat "${TMP_LOGS_REPORT_f}" | colors-remove | grep -qsE "(grub.*: error:|grub.*: warning:|grub.*command not found|package.*grub.*is not installed)" ; then
                    # get error messages
                    # if [[ -n "$LIVE_MOUNT_MEDIA_DEV_v" ]] ; then
                        # we just need to remove this type of entry from device.map ? (hd1)   /dev/disk/by-id/usb-SanDisk_Cruzer_4C531001431122101253-0:0
                        buf="$( cat "${TMP_LOGS_REPORT_f}" | grep -vE "(grub-probe.*cannot find a GRUB drive.*${LIVE_MOUNT_MEDIA_DEV_v})" | grep -E "(grub.*: error:|grub.*: warning:|grub.*command not found|package.*grub.*is not installed)" | sort -u )"
                    # else
                        # buf="$( cat "${TMP_LOGS_REPORT_f}" | grep -E "(grub.*: error:|grub.*command not found|package.*grub.*is not installed)" | sort -u )"
                        # buf="$( cat "${TMP_LOGS_REPORT_f}" | grep -E "(grub.*: error:|grub.*: warning:|grub.*command not found|package.*grub.*is not installed)" | sort -u )"
                    # fi

                    # fix cryptodisk error
                    if echo "$buf" | grep -qsE "(attempt to install to encrypted disk without cryptodisk enabled|GRUB_ENABLE_CRYPTODISK=y)" ; then
                        # add the required option
                        set -x
                        echo -e "GRUB_ENABLE_CRYPTODISK=y" >> "$TARGET/etc/default/grub"
                        urxvt -e bash -c "chroot '$TARGET' bash -c 'update-grub ; grub-install' 2>&1 | tee '${TMP_LOGS_REPORT_f}' "

                        # el_explain 0 "grub failed due to cryptodisk, re-running with the required option:"
                        el_warning 0 "grub failed due to cryptodisk, re-running with the required option:"
                        cat "${TMP_LOGS_REPORT_f}" 1>&2

                        # update error list
                        buf="$( cat "${TMP_LOGS_REPORT_f}" | grep -vE "(grub-probe.*cannot find a GRUB drive.*${LIVE_MOUNT_MEDIA_DEV_v})" | grep -E "(grub.*: error:|grub.*: warning:|grub.*command not found|package.*grub.*is not installed)" | sort -u )"
                        set +x
                    fi

                    # show grub error messages if needed
                    if [[ -n "$buf" ]] ; then
                        if echo "$buf" | grep -qs "not proceed with blocklists" && ((is_disk_table_gpt)) ; then
                            guitool warning "$( eval_gettext "The GRUB installation encountered issues due to the absence of a BIOS type partition. If you cannot boot your installed system, you must add one and reinstall Elive." )"
                        else
                            el_error "grub failed to install (user informed):\n$buf"
                            if ((is_windows_installed)) ; then
                                guitool warning "$( eval_gettext "The GRUB installation reported problems configuring the boot system. If you are unable to boot Elive, try reinstalling with different options. Remember, having Windows installed on the same computer often causes issues for other operating systems due to its monopolistic practices." )\n\nDetails:\n${buf}"
                            else
                                guitool warning "$( eval_gettext "The GRUB installation reported problems configuring the boot system. If you are unable to boot Elive, try reinstalling with different options." )\n\nDetails:\n${buf}"
                            fi
                        fi
                    fi
                fi
            fi

            # if [[ -d "$TARGET/boot/efi/" ]] ; then
            #     el_warning "temporal verification:\n$( tree $TARGET/boot/efi/ )"
            # fi

            rm -f "${TMP_LOGS_REPORT_f}"
        fi

        # warn about windows problems booting with efi / mbr
        if grep -qs ":Windows" /tmp/os-probed ; then
            guitool warning "$( eval_gettext "Your computer has Windows installed. If your new Elive refuses to boot, it is because that greedy system imposes its own boot sequence to keep its monopoly. You can also try booting from EFI or in 'Legacy' mode. However, since you can also use many Windows programs in Elive, we suggest getting rid of that problematic and insecure system from hell." )"

            # fix: sometimes windows efi forces to boot windows only instead of show the grub menu, tell the user how to fix that
            if ((is_efi_booted)) ; then
                efi_fix_windows_monopol
            fi
        fi
    fi


    # }}}
    # services and configurations {{{

    # acpid should be always enabled
    # update: moved to iso build, it is now always enabled by default
    #if guitool yesnoquestion "$( eval_gettext "Enable ACPI? We assume you should always select this option. However, we want to ask about this feature to ensure that enabling it does not cause any issues for users. If you experience issues with power management when it is enabled, please inform us in our forums." )" ; then
        #chroot "$TARGET" systemctl enable acpid.service
    ##else
        ##chroot "$TARGET" systemctl disable acpid.service
    #fi

    if [[ -n "$HD_DEV" ]] ; then
        if device_is_ssd "$HD_DEV" && ! ((is_sysvinit_wanted)) ; then
            chroot "$TARGET" systemctl enable fstrim.timer
        fi

    else
        el_error "deleteme temporal debug: HD_DEV variable is not set, we should always have it"
    fi

    if grep -qs btrfs "$TARGET/etc/fstab" && [[ -e "$TARGET/var/lib/dpkg/info/btrfsmaintenance.list" ]] && ! ((is_sysvinit_wanted)) ; then
        el_explain 0 "Enabling btrfs maintenance (systemd services) tools"
        chroot "$TARGET" systemctl enable btrfs-balance.timer
        chroot "$TARGET" systemctl enable btrfs-defrag.timer
        chroot "$TARGET" systemctl enable btrfs-scrub.timer
        chroot "$TARGET" systemctl enable btrfs-trim.timer
        # TODO: add into elive-pm checks
    fi


    if ((is_autologin_wanted)) && [[ -n "$autologin_username" ]] ; then

        if grep -qs "autologin-user=" "$TARGET/etc/lightdm/lightdm.conf" ; then
            sed -i -e "s|^\(#\)*autologin-user.*=.*$|autologin-user=$autologin_username|g" "$TARGET/etc/lightdm/lightdm.conf"
        else
            el_warning "unable to set autologin user, we didnt found the specific entry section where to put it"
        fi
        # set our conf too
        sed -i "/autologin-user/d" "$TARGET/etc/elive-version"
        # deprecated:
        #echo -e "autologin-user: $autologin_username" >> "$TARGET/etc/elive-version"
    fi

    # always show users list in login
    if grep -qs "greeter-hide-users=" "$TARGET/etc/lightdm/lightdm.conf" ; then
        sed -i -e "s|^\(#\)*greeter-hide-users.*=.*$|greeter-hide-users=false|g" "$TARGET/etc/lightdm/lightdm.conf"
    else
        el_warning "unable to set lightdm greeter-hide-users to 'false', which needs to be in the specific section"
    fi

    # change the BG to a blurred version of our actual one
    el_explain 1 "Creating background for lightdm..."
    rm -rf "$TARGET/var/tmp/wallpaper"
    mkdir -p "$TARGET/var/tmp/wallpaper"
    chmod 1777 "$TARGET/var/tmp/wallpaper"
    if [[ -x "$( which convert-7 )" ]] ; then
        if ! convert-7 -resize 2000x2000 -limit time 600 "$( find /etc/elive/wallpaper/ -type f | sort | tail -1 )" /tmp/wallpaper.jpg ; then
            cp -f "$( find /etc/elive/wallpaper/ -type f | sort | tail -1 )" /tmp/wallpaper.jpg
        fi
        convert-7 "/tmp/wallpaper.jpg" -blur 0x30 -brightness-contrast -8x2 -limit time 1000 "$TARGET/var/tmp/wallpaper/wallpaper.jpg"
    fi
    rm -f /tmp/wallpaper.jpg
    sed -i -e "s|^\(#\)*background.*=.*$|background=/var/tmp/wallpaper/wallpaper.jpg|g" "$TARGET/etc/lightdm/lightdm-gtk-greeter.conf"

    # set lightdm gtk theme
    if ((is_gtk_theme_default_wanted)) ; then
        sed -i -e "s|^\(#\)*theme-name.*=.*$|theme-name = Adwaita|g" "$TARGET/etc/lightdm/lightdm.conf"
    else
        sed -i -e "s|^\(#\)*theme-name.*=.*$|theme-name = $( grep gtk-theme-name /home/eliveuser/.gtkrc-2.0 | sed -e 's|^.*=||g' -e 's|"||g' | tail -1 )|g" "$TARGET/etc/lightdm/lightdm.conf"
    fi

    # required conf for nvidia-optimus cards:
    if [[ -e "/tmp/.nvidia_optimus_conf_onlynvidia" ]] ; then
        # create the script to enable nvidia by default:
        mkdir -p "$TARGET/etc/lightdm"
        echo -e "#!/bin/sh\nxrandr --setprovideroutputsource modesetting NVIDIA-0\nxrandr --auto" > "$TARGET/etc/lightdm/display_setup.sh"
        chmod +x "$TARGET/etc/lightdm/display_setup.sh"
        # set lightdm to run the script when login
        sed -i -e "s|^\(#\)*display-setup-script=.*$|display-setup-script=/etc/lightdm/display_setup.sh|g" "$TARGET/etc/lightdm/lightdm.conf"
    fi
    # copy the nvidia template in case is needed (after installing nvidia), or delete it in case is not needed
    if [[ -s "/etc/X11/xorg.conf.d/20-nvidia.conf" ]] ; then
        cp -f "/etc/X11/xorg.conf.d/20-nvidia.conf" "$TARGET/etc/X11/xorg.conf.d/20-nvidia.conf"
    else
        rm -f "$TARGET/etc/X11/xorg.conf.d/20-nvidia.conf"
    fi



    # set the computer's dpi:
    dpi="$( el_dpi_get )"
    if [[ "${dpi}" = *"x"* ]] ; then
        dpi_h="${dpi%x*}"

        if [[ "${dpi_h}" -gt "180" ]] ; then
            # update: this one doesn't works
            # sed -i -e "s|^\(#\)*xserver-command=.*$|xserver-command=X -dpi ${dpi_h}|g" "$TARGET/etc/lightdm/lightdm.conf"
            sed -i -e "s|^\(#\)*greeter-wrapper=.*$|greeter-wrapper=/etc/lightdm/pre-session-scaled.sh|g" "$TARGET/etc/lightdm/lightdm.conf"
            echo -e "#!/bin/sh\nexport GDK_SCALE=2\nexec \"\$@\"" > "$TARGET/etc/lightdm/pre-session-scaled.sh"
            chmod +x "$TARGET/etc/lightdm/pre-session-scaled.sh"

        fi
    fi

    # always enable onboard keyboard in case that no keyboard are found:
    #if [[ -e "/tmp/.no-keyboard" ]] ; then
        #sed -i -e "s|^#keyboard=.*$|keyboard=onboard|" "$TARGET/etc/lightdm/lightdm-gtk-greeter.conf"
    #fi
    # update: onboard doesn't take resources and is good to enable it by default
    if [[ -x "$TARGET/usr/bin/onboard" ]] ; then
        sed -i -e "s|^#keyboard=.*$|keyboard=onboard|" "$TARGET/etc/lightdm/lightdm-gtk-greeter.conf"
    fi

    # always enable these services if they are installed
    if ! ((is_sysvinit_wanted)) ; then
        for service in ssh urfkill systemd-journald acer-aspireone-fand smbd nmbd
        do
            chroot "$TARGET" systemctl unmask "$service" 1>/dev/null 2>&1 || true
            chroot "$TARGET" systemctl enable "$service" 1>/dev/null 2>&1 || true
        done
    fi

    # reconfigure ssh keys if any of them is not available (ubuntu 20.04 bug)
    if ((is_sysvinit_wanted)) || chroot "$TARGET" systemctl is-enabled "sshd.service" 1>/dev/null 2>&1 ; then
        #if ! ls $TARGET/etc/ssh/*_key* 1>/dev/null 2>&1 ; then
        ssh-keygen -A
        #fi
    fi

    # allow network to be used for any user by default, this solves a small issue where after install there was no network
    sed -i -e '/^permissions=/d' "$TARGET/etc/NetworkManager/system-connections/"*  2>/dev/null || true
    # allow all interfaces (wifi device names can change)
    sed -i -e '/^interface-name=/d' "$TARGET/etc/NetworkManager/system-connections/"*  2>/dev/null || true

    # - enable / disable systemd services }}}

    # end {{{

    space_left_verify

    # close the gui
    #rm -f "$TMP_PROGRESS_CONFIGURING_f"


    # }}}
}
#########################################################################
hooks_users_normal(){
    # pre {{{
    # Steps of new users and stuff (No Upgrade mode)
    functions_pre
    el_debug "\targs '$*'"
    el_debug "Setting up the root and user on the new system"
    el_check_variables "dhome"

    # - pre }}}

    echo 10 >> "$TMP_PROGRESS_USERS_f"
    # user's configurations {{{
    # save the audio card conf so that will not be asked again

    if [[ -s "/home/$eliveuser/.asoundrc" ]] ; then
        cp -f "/home/$eliveuser/.asoundrc" "$TARGET/etc/skel/"
        cp -f "/home/$eliveuser/.asoundrc" "$TARGET/root/"
    fi


    # - user's configurations }}}
    #ident
    if [[ -s "$TARGET/tmp/temp" ]] ; then
        el_warning "file '$TARGET/tmp/temp' was not empty, but we are going to remove it before to run its previous contents:\n$(cat $TARGET/tmp/temp)"
    fi
    cat > "$TARGET/tmp/temp" << EOF
#!/bin/bash
set -E
export TERM=xterm-256color

# ignore accidentals Ctrl + C
trap '' SIGINT


export LC_ALL="$EL_LC_EN"

export EL_DEBUG="$EL_DEBUG"
[[ "$EL_DEBUG" -gt 3 ]] && set -x || true


# root configurations
EL_DEBUG=0 elive-skel user root

sync


EOF
    chmod +x "$TARGET/tmp/temp"

    # Warning: these logs shows the root password, we don't want that if we are receiving logs

    # just in case it is not mounted (auto management of flags)
    EL_DEBUG=0 mounts-manager --chroot-layers-mount "$TARGET" "eliveinstaller" bind 1>/dev/null

    echo 20 >> "$TMP_PROGRESS_USERS_f"

    # TODO move this to a "last" script-run action in hooks_end
    chroot "$TARGET" bash "/tmp/temp" 1>/dev/null || exit_error
    #zenity --warning --text="Thana: here you must try a script to create the user from users-manager run from the chroot: $TARGET , template already made on: $TARGET/tmp/temp - DO NOT CLOSE this dialog until you did it, which stops the process"

    # debug and report
    if [[ -s "$TMP_LOGS_REPORT_f" ]] ; then
        cat "${TMP_LOGS_REPORT_f}" 1>&2
        rm -f "${TMP_LOGS_REPORT_f}"
    fi

    # change passwords (from here so we don't debug this)
    if [[ -n "$rootpass" ]] ; then
        chroot "$TARGET" bash -c 'printf "%s:%s\\n" "$1" "$2" | chpasswd -m' -- "root" "$rootpass"
    else
        el_error "root pass empty?"
        # reconfigure for now instead of fail:
        password_root_ask
        chroot "$TARGET" bash -c 'printf "%s:%s\\n" "$1" "$2" | chpasswd -m' -- "root" "$rootpass"
    fi

    echo 60 >> "$TMP_PROGRESS_USERS_f"

    if [[ -e "$TARGET/etc/sudoers" ]] ; then
        el_debug "Deleting eliveuser from sudo"
        sed -i 's|^.*eliveuser.*$||g' "$TARGET/etc/sudoers"
        chmod 0440 "$TARGET/etc/sudoers"
    fi

    echo 80 >> "$TMP_PROGRESS_USERS_f"

    # first-user confs
    sed -i '/^first-user:.*$/d' "$TARGET/etc/elive-version"
    echo -e "first-user: $username" >> "$TARGET/etc/elive-version"

    if ((is_autologin_wanted)) ; then
        autologin_username="$username"
    fi

    # recyclate xchat's username
    cp -f /home/$eliveuser/.xchat2/xchat.conf "$TARGET$dhome/$username/.xchat2/xchat.conf" 2>/dev/null || true
    cp -f /home/$eliveuser/.config/hexchat/hexchat.conf "$TARGET$dhome/$username/.config/hexchat/hexchat.conf" 2>/dev/null || true
    # change status to installed:
    sed -i 's|Elive_user_|Elive_inst_|g' "$TARGET$dhome/$username/.xchat2/xchat.conf" 2>/dev/null || true
    sed -i 's|Elive_user_|Elive_inst_|g' "$TARGET$dhome/$username/.config/hexchat/hexchat.conf" 2>/dev/null || true
    sed -i "s|${eliveuser}|${username}|g" "$TARGET$dhome/$username/.config/hexchat/hexchat.conf" 2>/dev/null || true
    sed -i "s|${eliveuser}|${username}|g" "$TARGET$dhome/$username/.config/hexchat/hexchat.conf" 2>/dev/null || true

    uid="$( awk -v user="$username" 'BEGIN{FS=":"}{if ($1 == user) print $3 }' "$TARGET/etc/passwd" )"
    el_check_variables "uid"
    chown "${uid}" "$TARGET$dhome/$username/.xchat/xchat.conf" 2>/dev/null || true
    chown "${uid}" "$TARGET$dhome/$username/.config/hexchat/hexchat.conf" 2>/dev/null || true


    rm -f "$TMP_PROGRESS_USERS_f"
    rm -f "$TARGET/tmp/temp"
    guitool "working_start"
}
#########################################################################
hooks_users_migrate(){
    functions_pre
    el_debug "\targs '$*'"
    el_debug "Adding old users to the new system..."
    local conf count uid user userlist
    el_check_variables "dhome"

    if [[ -s "$TARGET/tmp/temp" ]] ; then
        el_warning "file '$TARGET/tmp/temp' was not empty, but we are going to remove it before to run its previous contents:\n$(cat $TARGET/tmp/temp)"
    fi
    cat > "$TARGET/tmp/temp" << EOF
#!/bin/bash
set -E
# do not fail here:
set +e
export LC_ALL="$EL_LC_EN"
sync

# ignore accidentals Ctrl + C
trap '' SIGINT


export TERM=xterm-256color

export EL_DEBUG="$EL_DEBUG"
[[ "$EL_DEBUG" -gt 3 ]] && set -x || true

# add confs for root
EL_DEBUG=0 elive-skel user root

# migrations for new versions
# adduser hal plugdev

EOF
    chmod +x "$TARGET/tmp/temp"

    echo 20 >> "$TMP_PROGRESS_USERS_f"

    # TODO move this to a "last" script-run action in hooks_end
    chroot "$TARGET" bash "/tmp/temp" 1>/dev/null #2>&1

    # debug and report
    #if [[ "$EL_DEBUG" -gt 2 ]] ; then
        #cat "$TARGET/tmp/temp" 1>&2
    #fi


    echo 30 >> "$TMP_PROGRESS_USERS_f"

    # first migrate root user
    user_old_adapt root

    echo 40 >> "$TMP_PROGRESS_USERS_f"
    #   for i in $(ls $TARGET/home)
    count=0

    el_debug "/home contents:\n$( ls -l "$TARGET$dhome/" )"
    # show if the user dir has common user files to verify when the reimport fails
    el_debug "/home real users:\n$( find "$TARGET$dhome" -maxdepth 1 -type d | grep "/\." | sed -e 's|^.*/\.|.|g' | tail -n 30 | tr '\n' ' ' ; echo )"

    while read -ru 3 i
    do
        [[ "$user_old_adapt" = "lost+found" ]] && continue
        user_old_adapt "${i##*/}"
        count="$(( $count + 1 ))"
    done 3<<< "$( find "${TARGET}${dhome}/" -mindepth 1 -maxdepth 1 -type d )"

    # special user confs
    user_first="$( cat "$TARGET/etc.old-to-${eliveversion}/elive-version" 2>/dev/null | grep "first-user:" | sed -e 's|^.*first-user: ||g' | tail -1 )"

    # if there's only one user (very probably), count it as the default username
    if [[ "${count}" -eq 1 ]] ; then
        username="$( LC_ALL="$EL_LC_EN" find "$TARGET/$dhome/" -mindepth 1 -maxdepth 1 -type d | grep "$dhome" | grep "[[:alpha:]]" | sort -u | tail -1 )"
        username="${username##*/}"

        if [[ -n "$user_first" ]] && ! [[ "$user_first" = "$username" ]] ; then
            el_warning "detected first user from etc '${user_first}' != '${username}' from home unique dir "
        fi
        user_first="$username"

        if ! ((is_autologin_wanted)) ; then
            # configure autologin if we don't have it
            local message_want_autologin
            message_want_autologin="$( printf "$( eval_gettext "Would you like to log in automatically as user '%s'?" )" "$username" )"

            if guitool yesnoquestion "$message_want_autologin" ; then
                is_autologin_wanted=1
                autologin_username="$username"
            fi
        fi

        # deprecated
        #if ((is_autologin_wanted)) ; then
            #sed -i "/autologin-user/d" "$TARGET/etc/elive-version"
            #echo -e "autologin-user: $username" >> "$TARGET/etc/elive-version"
        #else
            #echo -e "#autologin-user: not_wanted_autologin" >> "$TARGET/etc/elive-version"
        #fi
    fi


    # reimport sudo confs for each user
    cp -a "$TARGET/etc.old-to-${eliveversion}/sudoers.d/"*  "$TARGET/etc/sudoers.d/" 2>/dev/null || true
    chmod 0440 "$TARGET/etc/sudoers.d/"* 2>/dev/null || true
    # fix possible broken ones:
    rm -f "$TARGET/etc/sudoers.d/"sudo_nopasswd_generic_ 2>/dev/null || true

    if [[ -n "$user_first" ]] ; then
        # generic sudoers needed
        el_debug "adding first-user '$user_first' to sudo-generic confs"
        chroot "$TARGET" /usr/sbin/user-manager --quiet --user-sudo-generic-add="${user_first}"

        el_debug "adding first-user '$user_first' to sudo-apt confs"
        chroot "$TARGET" /usr/sbin/user-manager --quiet --user-sudo-apt-add="${user_first}"
    fi

    # fix user NAMES from passwd file:
    cat "$TARGET/etc/passwd" | awk -v FS=":" -v OFS=":" '{ if ($5 == "Elive_User" || $5 == "Elive User") $5 = $1 ; print $0}' > "$TARGET/etc/passwd.new"
    mv -f "$TARGET/etc/passwd.new" "$TARGET/etc/passwd"


    # wanted autologin? ask which user to use for that
    #if ((is_autologin_wanted)) && ! [[ -n "$autologin_username" ]] && ! grep -qs "autologin-user: not_wanted_autologin" "$TARGET/etc/elive-version" ; then
    if ((is_autologin_wanted)) && ! [[ -n "$autologin_username" ]] ; then
        while read -ru 3 line
        do
            uid="$( echo "$line" | awk -v FS=":" '{print $3}' )"
            user="$( echo "$line" | awk -v FS=":" '{print $1}' )"

            if [[ -d "$TARGET/$dhome/$user" ]] ; then
                userlist+=("$user")
            fi
        done 3<<< "$( cat "$TARGET/etc/passwd" )"

        if [[ "${#userlist[@]}" = 1 ]] && [[ -d "$TARGET/$dhome/${userlist[*]}" ]] ; then
            username="${userlist[@]}"
        else
            username="$( printf "%s\n" "${userlist[@]}" | tr ' ' '\n' | sort -u | $guitool --title='Elive Installer' --list --column="$( eval_gettext "User" )" --text="$( eval_gettext "Choose the user to auto-login with." )" 2>/dev/null || echo cancel )"
            read -r username <<< "$username"
        fi

        # set autologin
        if [[ -n "$username" ]] ; then
            autologin_username="$username"
            #sed -i "/autologin-user/d" "$TARGET/etc/elive-version"
            #echo -e "autologin-user: $username" >> "$TARGET/etc/elive-version"
        fi
    fi

    # set the autologin
    if ((is_autologin_wanted)) && [[ -n "$autologin_username" ]] ; then
        sed -i -e "s|^\(#\)*autologin-user=.*$|autologin-user=$autologin_username|g" "$TARGET/etc/lightdm/lightdm.conf"
    fi

    echo 70 >> "$TMP_PROGRESS_USERS_f"
    if ! grep -q 'users:x:100:' "$TARGET/etc/group" ; then
        echo 'users:x:100:' >> "$TARGET/etc/group"
    fi

    if [[ -e "$TARGET/etc/sudoers" ]] ; then
        #echo "$username    ALL=(ALL) ALL" >> $TARGET/etc/sudoers
        # important, this is for delete the elive's users password
        sed -i "s|^.*eliveuser.*$||g" "$TARGET/etc/sudoers"
        chmod 0440 "$TARGET/etc/sudoers"
    fi

    echo 80 >> "$TMP_PROGRESS_USERS_f"
    # copy some confs from the old /etc like web services
    # update: this is not reliable and makes things broken, better ot have the old confs in separated dirs so the user can meld them
    # for conf in exim4 fail2ban apache apache2 lighttpd nginx monit php thin iptables.rules cups sane.d
    # do
    #     if [[ -e "$TARGET/etc.old-to-${eliveversion}/$conf" ]] ; then
    #         rm -rf "$TARGET/etc/${conf}" 2>/dev/null || true
    #         cp -a "$TARGET/etc.old-to-${eliveversion}/$conf" "$TARGET/etc/"
    #     fi
    # done

    # backup old possible installer-modules
    if [[ -s "$TARGET/etc.old-to-${eliveversion}/elive/installer-module" ]] ; then
        mkdir -p "$TARGET/etc/elive/installer-modules.bak"
        cp -f "$TARGET/etc.old-to-${eliveversion}/elive/installer-module" "$TARGET/etc/elive/installer-modules.bak/installer-module_${old_eliveversion}.cod" 2>/dev/null || true
    fi

    # set kernel version
    sed -i -e '/^kernel:/d' "$TARGET/etc/elive-version"
    echo -e "kernel: $(uname -r)" >> "$TARGET/etc/elive-version"

    # machine ID
    machine_id="$( cat "$TARGET/etc.old-to-${eliveversion}/elive-version" 2>/dev/null | grep "machine-id:" | sed -e 's|^.*machine-id: ||g' | tail -1 )"
    if [[ -n "$machine_id" ]] ; then
        echo "machine-id: $machine_id" >> "$TARGET/etc/elive-version"
    else
        echo -e "machine-id: $( el_get_machine_id )" >> "$TARGET/etc/elive-version"
    fi

    # make sure that root has a password:
    password_root_ask

    #echo 90 >> "$TMP_PROGRESS_USERS_f"
    rm -f "$TMP_PROGRESS_USERS_f"
    rm -f "$TARGET/tmp/temp"
    guitool "working_start"
}
#########################################################################
user_old_adapt(){
    # this function is run with USER as parameter and sets modifications to the user
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    local user file entry diruser uid member group_entry_line group listgroups numbergroups

    user="$1"

    el_check_variables "dhome"

    cat > "$TARGET/tmp/temp" << EOF
#!/bin/bash
set -E
export LC_ALL="$EL_LC_EN"
export TERM=xterm-256color

# ignore accidentals Ctrl + C
trap '' SIGINT

export EL_DEBUG="$EL_DEBUG"
[[ "$EL_DEBUG" -gt 3 ]] && set -x || true
true

EOF
    chmod +x "$TARGET/tmp/temp"

    # - pre }}}

    # return if user don't exist
    if ! awk -v FS=":" -v user="$user" '{if($1 == user) print}' "$TARGET/etc.old-to-${eliveversion}/passwd" | grep -qs "$user" ; then
        el_debug "user $user doesn't exist in old install, skipping it..."
        return 0
    fi

    echo "# $( eval_gettext "Adapting old user ") $user" >> "$TMP_PROGRESS_USERS_f"

    el_debug "Importing user: $user"

    diruser="$( awk -v user="$user" 'BEGIN{FS=":"}{if ($1 == user) print $6 }' "$TARGET/etc.old-to-${eliveversion}/passwd" )"
    uid="$( awk -v user="$user" 'BEGIN{FS=":"}{if ($1 == user) print $3 }' "$TARGET/etc.old-to-${eliveversion}/passwd" )"
    gid="$( awk -v user="$user" 'BEGIN{FS=":"}{if ($1 == user) print $4 }' "$TARGET/etc.old-to-${eliveversion}/passwd" )"
    if ! [[ -d "${TARGET}${diruser}" ]] ; then
        mkdir -p "${TARGET}${diruser}"
        chown "${uid}:${gid}" "${TARGET}${diruser}"
    fi


    for file in passwd group shadow gshadow
    do
        [[ -s "$TARGET/etc.old-to-${eliveversion}/$file" ]] || continue

        entry="$( awk -v FS=":" -v user="$user" '{if($1 == user) print}' "$TARGET/etc.old-to-${eliveversion}/${file}" | tail -1 )"

        if [[ -n "$entry" ]] ; then
            if grep -qE "^${user}:" "$TARGET/etc/${file}" ; then
                if [[ "$entry" = "$(grep "^${user}:" "$TARGET/etc/$file" )" ]] ; then
                    # is the same, we don't need to change it
                    true
                else
                    if [[ "$EL_DEBUG" -gt 2 ]] ; then
                        el_debug "re-importing line for '${user}' in '$file' from:\n${entry}, to:\n$(grep "^${user}:" "$TARGET/etc/${file}" )"
                    fi
                    sed -i "s|^${user}:.*$|${entry}|g" "$TARGET/etc/${file}"
                fi
            else
                echo "$entry" >> "$TARGET/etc/${file}"
                el_debug "added line for '${user}' in file '/etc/${file}' :  $( grep -E "^${user}" "$TARGET/etc/$file" | tail -1)"
            fi
        else
            # create new entries
            # sometimes is not included
            # case "$file" in
            #     group)
            #         # update: don't do that, gid is correctly set so sometimes there's no group for the user but uses the "users" group instead
            #         echo -e "${user}:x:${uid}:" >> "$TARGET/etc/group"
            #         ;;
            #     gshadow)
            #         echo -e "${user}:!::" >> "$TARGET/etc/gshadow"
            #         ;;
            # esac
            el_warning "entry for $user in $file is not included"
        fi

        sed -i '/^$/d' "$TARGET/etc/$file"
        sed -i 's|:,$|:|g' "$TARGET/etc/$file"
    done

    # clenaups
    # is important to always remove the efreet cache for big changes
    for i in \
        .cache/efreet \
        .cache/fontconfig \
        .cache/qtsharedcache \
        .config/fontconfig \
        .config/systemd \
        .dbus \
        .dmrc \
        .icons \
        .Xauthority \
        $NULL
    do
        if [[ "$user" = "root" ]] ; then
            rm -rf "$TARGET/root/$i" 2>/dev/null || true
        else
            rm -rf "$TARGET$dhome/$user/$i" 2>/dev/null || true
        fi
    done
        # .local/share/keyrings \

    # USER only {{{
    if [[ "$user" != "root" ]] ; then
        # set groups {{{
        # root don't need groups
        unset listgroups
        for group_entry_line in $( cat "$TARGET/etc.old-to-${eliveversion}/group" )
        do
            unset member
            for i in $( echo "$group_entry_line" | sed 's/^.*://g' | tr ',' ' ' )
            do
                # loop each user (i)
                [[ -z "$i" ]] && continue

                if echo "$i" | grep -qs "$user" ; then
                    member=yes
                fi
            done

            if [[ -n $member ]] ; then
                listgroups="$listgroups $( echo "$group_entry_line" | awk -v FS=: '{print $1}')"
            fi
        done

        # special confs for the privileged (first) user
        if grep -qs "first-user: ${user}$" "$TARGET/etc.old-to-${eliveversion}/elive-version" 2>/dev/null ; then
            # add user to sudoers if is the first one
            #listgroups="$listgroups sudo"
            # instead, we should use our own list of groups
            listgroups="$( echo "$listgroups $groups" | tr ' ' '\n' | sort -u | tr '\n' ' ' )"
            echo -e "first-user: $user" >> "$TARGET/etc/elive-version"
        fi

        # migrate sudo users
        cat "$TARGET/etc.old-to-${eliveversion}/sudoers.d/"* "$TARGET/etc.old-to-${eliveversion}/sudoers" 2>/dev/null | grep "^${user}\s*" | sort -u >> "$TARGET/etc/sudoers.d/users-elive-migrated"
        if ! [[ -s "$TARGET/etc/sudoers.d/users-elive-migrated" ]] ; then
            rm -f "$TARGET/etc/sudoers.d/users-elive-migrated"
        fi

        listgroups="$( echo "$listgroups" | sed 's/^\ //' )"
        numbergroups="$( echo "$listgroups" | wc -w )"
        el_debug "old groups to import: $listgroups ($numbergroups)"

        #   user2=${user//\$/\\\$}
        if [[ "$numbergroups" -ge 3 ]] ; then
            for group in $listgroups $groups
            do
                echo -e "adduser '$user' '$group' 2>/dev/null 1>&2 || true" >> "$TARGET/tmp/temp"
            done
        else
            for group in $groups
            do
                echo -e "adduser '$user' '$group' 2>/dev/null 1>&2 || true" >> "$TARGET/tmp/temp"
            done
        fi
        # - set groups }}}
        # incremental migration / migrate upgrades by versions {{{

        while read -ru 3 conf
        do
            case "$conf" in
                .xchat|.gtk-bookmarks|.conkyrc|.elxstrt|.e16|.config|*thanatests*|*electrum*|.local)
                    # confs that we don't want to change
                    continue
                    ;;
                .cache)
                    # remove it entirely, but save a few things
                    # echo "EL_DEBUG=0 rm -rf '${dhome}/$user/${conf}.old' " >> "$TARGET/tmp/temp"
                    # echo "EL_DEBUG=0 mv '${dhome}/$user/${conf}' ${dhome}/$user/${conf}.old 2>/dev/null " >> "$TARGET/tmp/temp"
                    # echo "EL_DEBUG=0 mkdir -p '${dhome}/$user/${conf}' " >> "$TARGET/tmp/temp"
                    # echo "EL_DEBUG=0 mv '${dhome}/$user/${conf}.old/zsh' '${dhome}/$user/${conf}/' 2>/dev/null " >> "$TARGET/tmp/temp"
                    # echo "EL_DEBUG=0 chown -R '${user}' '${dhome}/$user/${conf}' 2>/dev/null " >> "$TARGET/tmp/temp"
                    # el_explain 2 "~/.cache for user '$user' entirely removed"
                    echo "EL_DEBUG=0 rm -rf '${dhome}/$user/${conf}/thumbnails' " >> "$TARGET/tmp/temp"
                    echo "EL_DEBUG=0 rm -rf '${dhome}/$user/${conf}/shotwell' " >> "$TARGET/tmp/temp"
                    echo "EL_DEBUG=0 rm -rf '${dhome}/$user/${conf}/google-chrome' " >> "$TARGET/tmp/temp"
                    echo "EL_DEBUG=0 rm -rf '${dhome}/$user/${conf}/chromium' " >> "$TARGET/tmp/temp"
                    echo "EL_DEBUG=0 rm -rf '${dhome}/$user/${conf}/mozilla' " >> "$TARGET/tmp/temp"
                    echo "EL_DEBUG=0 rm -rf '${dhome}/$user/${conf}/bkp' " >> "$TARGET/tmp/temp"
                    echo "EL_DEBUG=0 rm -rf '${dhome}/$user/${conf}/efreet' " >> "$TARGET/tmp/temp"
                    echo "EL_DEBUG=0 rm -rf '${dhome}/$user/${conf}/mesa_shader_cache' " >> "$TARGET/tmp/temp"
                    el_explain 2 "~/.cache for user '$user' cleaned up"
                    ;;
                .bash*)
                    if ((is_shared_home_partition)) ; then
                        # do not touch bash'es in shared home's, because all the other systems uses bash and they want to still use it
                        continue
                    else
                        if ((is_user_confs_upgrade_wanted)) ; then
                            echo "EL_DEBUG=0 elive-skel upgrade '$conf' '$user'" >> "$TARGET/tmp/temp"
                            el_explain 2 "upgraded user '$user' conf: $conf  (generic upgrade)"
                        fi
                    fi
                    ;;
                *)
                    if ((is_user_confs_upgrade_wanted)) ; then
                        echo "EL_DEBUG=0 elive-skel upgrade '$conf' '$user'" >> "$TARGET/tmp/temp"
                        el_explain 2 "upgraded user '$user' conf: $conf  (generic upgrade)"
                    fi
                    ;;
            esac
        done 3<<< "$( ls -a1 /etc/skel | awk 'NR > 2')"


        # TODO: for user .desktop files (.local/share/applications) created, show a menu with checkboxes asking which ones wants to re-import, see what is going on with the icons: .local/share/icons

        # make sure that we upgrade these ones, even if they don't exist in skel:
        # note: eliveversion != old_eliveversion, make sure to know which one are you searching for
        #if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" "lt" "2.9.23" ; then
        if grep -qs "package_e17conf_installed_version_prev" "${TARGET}${dhome}/${user}/.config/elive/elive-tools/el_config/%%usr%%bin%%e_wrapper_conf.sh" ; then
            # since the old upgrades can have went wrong, better to tell the user about upgrade it
            # update is always good to renew the desktop configuration, so there's nothing wrong by asking that to the user
            #if [[ -n "$old_eliveversion" ]] && LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "lt" "2.9.98" ; then
                rm -f "${TARGET}${dhome}/${user}/.config/elive/elive-tools/el_config/%%usr%%bin%%e_wrapper_conf.sh"
            #fi
        else
            # elive-skel should upgrade that better, but since we don't know which exactly we have, better to change it entirely
            #echo "EL_DEBUG=0 elive-skel upgrade '.e' '$user'" >> "$TARGET/tmp/temp"
            mv -f "$TARGET$dhome/$user/.e" "$TARGET$dhome/$user/.e.old-${old_eliveversion}" 2>/dev/null || true
            echo "chown -R '${user}' \"$dhome/$user/.e.old-${old_eliveversion}\" 2>/dev/null" >> "$TARGET/tmp/temp"
            rm -rf "$TARGET$dhome/$user/.e" 2>/dev/null || true
        fi

        # remove Downloads & other (empty) dirs to be less annoying, they are going to be re-created when running new e17 conf, also remove the migrator mark
        if [[ -n "$E_START" ]] ; then
            echo "rmdir \"$dhome/$user/\"* 2>/dev/null || true" >> "$TARGET/tmp/temp"
            echo "rm \"$dhome/$user/.config/elive/migrator/xdg-default-dirs-language-upgraded.state\" 2>/dev/null || true" >> "$TARGET/tmp/temp"
            el_warning "removed '~/.config/elive/migrator/xdg-default-dirs-language-upgrade.state' file, migrated references will be regenerated, betatest it"
        fi

        # always restart the counter of desired donations:
        echo "rm -f \"$dhome/$user/.cache/elive-donations-support/suggested_last\" 2>/dev/null || true" >> "$TARGET/tmp/temp"

        # remove startup sound so the user can pick it again
        echo "rm -f \"$dhome/$user/.config/elive-tools/el_config/\"*elive-startup-sound* 2>/dev/null || true" >> "$TARGET/tmp/temp"

        # for some reason the .config/elive dir is owned by root, simply fix it in every upgrade:
        echo "chown ${user} \"$dhome/$user/.config/elive\" 2>/dev/null || true" >> "$TARGET/tmp/temp"


        #
        #
        # MIGRATIONS
        #
        #

        # Before 2.2.1  (first instalable versions)
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "2.2.1" ; then
            # remove /mnt from gtk bookmarks
            sed -i '/file:\/\/\/mnt /d' "$TARGET$dhome/$user/.gtk-bookmarks" 2>/dev/null || true
        fi
        # remove deprecated and uncompatible .gtkrc files
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "2.7.1" ; then
            rm -f "$TARGET$dhome/$user/.gtkrc"*  2>/dev/null
        fi
        # needed updates
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "3.7.8" ; then
            echo "EL_DEBUG=0 elive-skel upgrade '.config/cairo-dock' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.config/gtk-3.0' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.gtkrc-2.0' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.local/share/applications/mimeapps.list' '$user'" >> "$TARGET/tmp/temp"
        fi
        # uca.xml is not used from older versions, so we will miss these features otherwise
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "3.7.16" ; then
            echo "EL_DEBUG=0 elive-skel upgrade '.config/Thunar' '$user'" >> "$TARGET/tmp/temp"
        fi
        #conkyrc needs to be updated and e16 too because it got black before 3.8.9
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "3.8.9" ; then
            echo "EL_DEBUG=0 elive-skel upgrade '.e16' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.conkyrc' '$user'" >> "$TARGET/tmp/temp"
        fi
        # migration (crossgrade) needs many new confs:
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "3.8.10" ; then
            echo "EL_DEBUG=0 elive-skel upgrade '.config/audacious' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.config/GIMP' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.config/smplayer' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.config/xfce4' '$user'" >> "$TARGET/tmp/temp"
        fi
        # fix pulseaudio in buster -> newer upgrade
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "3.8.27" ; then
            # switch to hexchat
            echo "EL_DEBUG=0 elive-skel upgrade '.config/hexchat' '$user'" >> "$TARGET/tmp/temp"
            # delete broken files for pulseaudio
            echo "su -c 'rm -rf ~/.config/pulse/* ~/.pulse  2>/dev/null || true' '$user'" >> "$TARGET/tmp/temp"
            # a reconfigure is needed to fix the audio
            echo "su -c 'audio-configurator --quiet --auto --smart --no-messages 2>/dev/null ' '$user'" >> "$TARGET/tmp/temp"
            #echo "su -c 'setvolume defaults' '$user'" >> "$TARGET/tmp/temp"
            #echo "su -c 'timeout 5 aplay /dev/null 2>/dev/null || true' '$user'" >> "$TARGET/tmp/temp"

            # upgrade .conkyrc because of the wrong memory showing up
            echo "EL_DEBUG=0 elive-skel upgrade '.conkyrc' '$user'" >> "$TARGET/tmp/temp"
        fi
        # This version includes many changes, like ulauncher etc
        if LC_ALL=C dpkg --compare-versions "$old_eliveversion" "lt" "3.8.32" ; then
            echo "EL_DEBUG=0 elive-skel upgrade '.config/cairo-dock' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.config/ulauncher' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.local/share/ulauncher' '$user'" >> "$TARGET/tmp/temp"
        fi
        # gnome-keyring needs to be autostarted in order to have wifi correctly asking for the password:
#         # TODO: many upgrades needs to be made, for example apps launcher associations
#         if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "3.8.39" ; then
#             cat >> "$TARGET/tmp/temp" << EOF
#             if ! grep -qs 'gnome-keyring-secrets' "/home/${user}/.e16/startup-applications.list" ; then
#                 echo "/etc/xdg/autostart/gnome-keyring-pkcs11.desktop" >> "/home/${user}/.e16/startup-applications.list"
#                 echo "/etc/xdg/autostart/gnome-keyring-secrets.desktop" >> "/home/${user}/.e16/startup-applications.list"
#                 echo "/etc/xdg/autostart/gnome-keyring-ssh.desktop" >> "/home/${user}/.e16/startup-applications.list"
#             fi
#             set +x
# EOF
#         fi
#         set +x
        #
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "3.8.40" ; then
            echo "EL_DEBUG=0 elive-skel upgrade '.config/smplayer' '$user'" >> "$TARGET/tmp/temp"
            echo "rm -f '$dhome/$user/.screenlayout/'*sh" >> "$TARGET/tmp/temp"
        fi
        if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "3.8.42" ; then
            echo "EL_DEBUG=0 elive-skel upgrade '.tmux.conf' '$user'" >> "$TARGET/tmp/temp"
        fi

        # we need to change the themes and designs between special versions
        if ((is_retrowave)) || ((is_old_special_version)) || ((is_desktop_upgrade_wanted)) ; then
            echo "EL_DEBUG=0 elive-skel upgrade '.config/cairo-dock' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.config/ulauncher' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.local/share/ulauncher' '$user'" >> "$TARGET/tmp/temp"

            echo "rm -f '$dhome/$user/.gtkrc-2.0.mine'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.config/gtk-3.0/settings.ini' '$user'" >> "$TARGET/tmp/temp"
            #echo "EL_DEBUG=0 elive-skel upgrade '.gtkrc-2.0.mine' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.xsettingsd' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.Xdefaults' '$user'" >> "$TARGET/tmp/temp"

            echo "EL_DEBUG=0 elive-skel upgrade '.e16' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.conkyrc' '$user'" >> "$TARGET/tmp/temp"

            echo "EL_DEBUG=0 elive-skel upgrade '.e' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.ecomp' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.emerald' '$user'" >> "$TARGET/tmp/temp"

            echo "EL_DEBUG=0 elive-skel upgrade '.config/audacious/playlists/1000.audpl' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.config/terminology' '$user'" >> "$TARGET/tmp/temp"

            # very needed on each distro upgrade and especially on migration (crossgrade) mode
            echo "EL_DEBUG=0 elive-skel upgrade '.config/mimeapps.list' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.local/share/applications/defaults.list' '$user'" >> "$TARGET/tmp/temp"
            echo "EL_DEBUG=0 elive-skel upgrade '.local/share/applications/mimeapps.list' '$user'" >> "$TARGET/tmp/temp"

            echo "rm -rf '$dhome/$user/.config/elementary'" >> "$TARGET/tmp/temp"
            echo "rm -rf '$dhome/$user/.elementary'" >> "$TARGET/tmp/temp"
            echo "rm -f '$dhome/$user/.config/Thunar/uca.xml'" >> "$TARGET/tmp/temp"
        fi

        echo "rm -f '$dhome/$user/.xsession-errors'" >> "$TARGET/tmp/temp"

        # this forces to reconfigure audio on next desktop start, and also restart the volumes, required in every upgrade for compatibility reasons
        echo "su -c 'touch ~/.asoundrc-reconfigureme' '$user'" >> "$TARGET/tmp/temp"

        # when user wants to keep its desktop settings, update at least it to use the new provided wallpaper by default
        if ! ((is_desktop_upgrade_wanted)) ; then
            echo "su -c 'touch ~/.wallpaper-reconfigureme' '$user'" >> "$TARGET/tmp/temp"
        fi

        # - incremental upgrades by versions }}}
        # generic sudoers needed
        el_debug "adding user '$user' to sudo-generic confs"
        chroot "$TARGET" /usr/sbin/user-manager --quiet --user-sudo-generic-add="${user}"

    fi
    # USER only }}}

    # always add confs that didnt existed before for this user
    el_debug "adding missing confs for user '$user':"
    echo "EL_DEBUG=0 elive-skel update '$user'" >> "$TARGET/tmp/temp"

    # always create new configurations for everything if we are migrating from other system
    if ((is_migration_otheros)) ; then
        echo "EL_DEBUG=0 elive-skel user '$user'" >> "$TARGET/tmp/temp"
    fi
    # }}}

    # change default bash to new zsh
    #if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "le" "2.2.1" ; then
    if ! ((is_shell_in_bash_wanted)) ; then
        chroot "$TARGET" /usr/bin/chsh -s /bin/zsh "$user"
    fi
    #fi

    # run it all!
    if [[ -s "$TARGET/tmp/temp" ]] ; then
        EL_DEBUG=0 mounts-manager --chroot-layers-mount "$TARGET" "eliveinstaller" bind 1>/dev/null
        # if [[ "$EL_DEBUG" -gt 2 ]] ; then
        #     cat "$TARGET/tmp/temp" 1>&2
        # fi
        if ! chroot "$TARGET" bash "/tmp/temp" ; then
            echo -e "\nE: running script, contents are:\n" 1>&2
            cat "$TARGET/tmp/temp" 1>&2
        fi

        rm -f "$TARGET/tmp/temp"
    fi


}
#########################################################################
hooks_end(){
    functions_pre
    el_debug "\targs '$*'"

    _dummy="$( eval_gettext "Error: You are logged in as root. Working in admin mode is dangerous, and your permissions should be more restricted. Please log in as a normal user for a more compatible experience." )"
    cat >> "$TARGET/root/.elxstrt" << EOF
if [[ "$UID" = "0" ]] && { ! cat /proc/cmdline | grep -q nurse ; } ; then
    zenity --error --text="$_dummy" )" &
fi
EOF
    chmod +x "$TARGET/root/.elxstrt"

    # this takes space and we don't need it at all for root
    rm -rf "$TARGET"/root/.gimp*

    # FIXME: re-implement:
    #mv "$TARGET/usr/share/eliveinstaller/postinstall-elive" "$TARGET/etc/init.d/"

    # clean dummy installer modules
    [[ -f "$TARGET/etc/elive/installer-module" ]] && rm -f $TARGET/etc/elive/elive_installer_key_*cod 2>/dev/null || true

    # Ugly ubuntu voohdooh requires ugly hacks
    if ((is_ubuntu_root_pass_empty)) || ((is_shared_home_partition)) ; then
        password_root_ask

        chroot "$TARGET" bash -c 'printf "%s:%s\\n" "$1" "$2" | chpasswd -m' -- "root" "$rootpass"
    fi


    if [[ -d "$TARGET/elive" ]] || ls "$TARGET/"DSS* 1>/dev/null 2>&1 ; then
        el_error "Some old directories found like /elive or /DSS*"
    fi

    # save installation status, so that upgrade mode can know if we really installed correctly
    mkdir -p "$TARGET/elive"
    echo "Status: dumped" > "$TARGET/etc/elive/installation-status"

    # wanted cleanups
    if ((is_cleanup_wanted_manpages)) ; then
        cat > "$TARGET/etc/dpkg/dpkg.cfg.d/02_nodoc_manpages" << EOF
path-exclude /usr/share/man/*
path-exclude /usr/share/groff/*
path-exclude /usr/share/info/*
EOF
    fi
    if ((is_cleanup_wanted_dir_usrsharedoc)) ; then
        rm -rf "$TARGET/usr/share/doc"
        mkdir -p "$TARGET/usr/share/doc"
        mkdir -p "$TARGET/etc/dpkg/dpkg.cfg.d"
        # do not install future packages with that:
        cat > "$TARGET/etc/dpkg/dpkg.cfg.d/01_nodoc" << EOF
path-exclude /usr/share/doc/*
path-exclude /usr/share/man/*
path-exclude /usr/share/groff/*
path-exclude /usr/share/info/*
# lintian stuff is small, but really unnecessary
path-exclude /usr/share/lintian/*
path-exclude /usr/share/linda/*
EOF
        true
    else
        # always delete this file if we want to run them
        rm -f "$TARGET/etc/dpkg/dpkg.cfg.d/01_nodoc"
    fi

    # delete softer warning messages
    rm -f "$TARGET/lib/lsb/init-functions.d/80-left-info-blocks-live"

    # delete old /etc's, only keep the last upgrade
    for dir in "$TARGET/etc.old-to-"*
    do
        [[ -d "$dir" ]] || continue
        if [[ "$dir" != "$TARGET/etc.old-to-${eliveversion}" ]] ; then
            rm -rf "$dir" 2>/dev/null
        fi
    done

    # always remove the old timezones conf because it depends of the correctly-configured other confs (/etc/localtime), so we need to have it set up cleanly again
    rm -rf "$TARGET/etc/elive-tools/geolocation/timezones.conf"

    # bug seems fixed
    #if ((is_buster)) ; then
        #if grep -qs "none.*swap.*sw" "$TARGET/etc/fstab" ; then
            #sed -i -e '/^.*none.*swap.*sw.*$/s|^\(.*\)$|# SWAP usage is EXTREMELY discouraged in these versions, by some uknown reason they are used (filled) when there is no need (lots of free space in ram), in fact we also consider that swap has never been something useful, since the total lagged and blocked resulting OS when is needed to use can lead to a forced shudown, which is worse than the kernel killing the application due to ram usage, in fact the only usefulness of it would be to being able to hibernate the computer, we disable it for now and if you want, enable it yourself but you are warned about a resuling slow and lagged experience, if you know a way to make the swap not being used at all unless the ram is filled (this was not the case in Elive 3.0, which was always empty unless ram is filled), please suggest to the Elive Forums how to make it working this way\n#\1|' "$TARGET/etc/fstab"
        #fi
    #fi

    if [[ "$EL_DEBUG" -gt 2 ]] ; then
        el_debug "Showing resulting fstab:"
        cat "$TARGET/etc/fstab" 1>&2
        if [[ "$( cat "$TARGET/etc/crypttab" | wc -l )" -gt 1 ]] ; then
            el_debug "Showing resulting crypttab:"
            cat "$TARGET/etc/crypttab" 1>&2
        fi
    fi

    if ((is_lightdm_disable_wanted)) ; then
        if ((is_wheezy)) || ((is_sysvinit_wanted)) ; then
            rm -f "$TARGET/etc/rc2.d/"*lightdm*
        else
            chroot "$TARGET" systemctl disable lightdm.service 2>/dev/null 1>&2 || chroot "$TARGET" systemctl disable lightdm 2>/dev/null 1>&2 || true
        fi
    fi

    # copy possibly needed packages
    #cp -n /var/tmp/cached-debs/*deb "$TARGET/var/cache/apt/archives/" 2>/dev/null || true



    # allow the user to hack the inside system if failed
    if ((is_previous_installation_failed)) ; then
        if guitool yesnoquestion "$( eval_gettext "It seems your previous installation attempt failed. If you know how to use the Terminal, you can modify or run things like installing the Grub boot loader or fixing other issues. Do you want to open a Terminal inside the installed system?" )" ; then
            guitool info "$( eval_gettext "Now you can do anything from this terminal and everything will be run inside your installed system. You can do things such as installing any missing packages required for an UEFI GRUB structure or editing the fstab. When you're done, simply exit the terminal." )"
            urxvt -e bash -c "chroot '$TARGET' "
        fi

        guitool info "$( eval_gettext "Note: If you find something that needs improvement in the installer, it would be very helpful for Elive if you comment about it in our forums." )"
    fi

    # betatest / debug:
    #if [[ -n "$HD_PART_EFI_dev" ]] ; then
        #guitool info "TEMPORAL TERMINAL: to configure EFI manually for what could be needed, for example ,do we need grub-install (and without parameters?), do we need to setup any grub thingy at all?"
        #urxvt -e bash -c "chroot '$TARGET' "
    #fi

    # Bootloader sanity checks:
    #if ((is_efi_booted)) ; then
        ## booted in UEFI mode
        ## EFI partition required
        #if [[ -b "${HD_PART_EFI_dev}" ]] ; then
            ## all good
            #true
        #else
            #guitool warning "$( eval_gettext "For your system to be bootable with your setup, an EFI partition is necessary. Not including one could result in an unbootable system." )"
            #el_warning "Sanity booloader check failed: UEFI booted + not EFI partitions used"
        #fi
    #else
        ## booted in Legacy mode
        ## must require BIOS partition
        #if ((is_disk_table_gpt)) ; then
            #if [[ -b "$HD_DEV" ]] ; then
                #if ! gdisk -l "$HD_DEV" | grep -qs "EF02" ; then
                    #guitool info "$( eval_gettext "Your selected setup uses GPT and LEGACY BOOT. It is recommended to have an extra BIOS type partition to act as the MBR for correctly installing the bootloader and booting the computer." )"
                #fi
            #fi
        #fi
        ## must install in MBR
        #if ((is_disk_table_msdos)) ; then
            ## all good
            #true
        #fi
    #fi

    # fix for root confs: since elive-skel always upgrade .tmux.conf we don't want this feature for root, this also breaks elive-upgrader UX:
    # note: updated on the srv building
    sed -i -e 's|^if-shell.*message_welcome.*$|#&|g' "$TARGET/root/.tmux.conf"

    # cleanups
    # needed because some files are left on the home of the user because HOME is set to it when we chroot for user-manager and similar
    rm -rf "$TARGET/home/$eliveuser" 1>/dev/null 2>&1 || true
    rm -rf "$TARGET/root/preseeds"* 1>/dev/null 2>&1 || true

    # checks
    # temporal fix for grub until the bug is found:
    # show the entire grub resulting conf to see that everything is good (like our other options)
    # cat "$TARGET/etc/default/grub" 1>&2

    # always re-fix grub to include other OS's because it seems like this setting is not well set previously
    if ! grep -qsi "^GRUB_DISABLE_OS_PROBER=.*false" "$TARGET/etc/default/grub" ; then
        # el_error "GRUB_DISABLE_OS_PROBER entry not found: $( grep GRUB_DISABLE_OS_PROBER "$TARGET/etc/default/grub" )"
        sed -i -e '/^GRUB_DISABLE_OS_PROBER=/d' "$TARGET/etc/default/grub"
        echo -e "# allow other operating systems appear in your boot menu\nGRUB_DISABLE_OS_PROBER=\"false\"" >> "$TARGET/etc/default/grub"
    fi


    # umount temporal swap if we had one and delete it
    swapfile_installer remove

    echo "90" > "$TMP_PROGRESS_f"
}
#===  FUNCTION  ================================================================
#          NAME:  hooks_end_post_umount
#   DESCRIPTION:  hooks to run after to have umounted target
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
hooks_end_post_umount(){
    # pre {{{
    local part dev point
    el_debug "\targs '$*'"

    # }}}

    # disable time-based fsck checks {{{
    # make bugged bioses to still boot disks (live mode updates date, bios thinks that is in the past, system cannot boot due to fsck report)
    while read -ru 3 line
    do
        [[ -z "$line" ]] && continue
        dev="$( echo "$line" | awk '{print $1}' )"
        point="$( echo "$line" | awk '{print $2}' )"

        if ! [[ -e "$TARGET/$point" ]] ; then
            el_debug "created missing directory mountpoint '$point' for '${dev}'"
            mkdir -p "$TARGET/$point"
        fi

        entry="$( hd_uuid_decode "$dev" )"
        if [[ -b "$entry" ]] ; then
            el_debug "disabling time-based fsck's for $entry"
            tune2fs -C 0 -i 0 "$entry" 1>/dev/null 2>&1 || true
            sync
        fi
    done 3<<< "$( cat "${TMP_MAIN_d}/installed-fstab" | grep -vE "\s*swap\s*" | grep -v "^#" | sort -u )"

    # debug
    if ((is_reiserfs_warn)) ; then
        el_debug "REISERFS dmesg issues?\n$(dmesg | tail -n 100)"
    fi

    # - disable time-based fsck checks }}}
}
#########################################################################
help_extra_show(){
    functions_pre
    el_debug "\targs '$*'"
    if ((ignore_show_helpers)) || [[ -n "$oldelive" ]] ; then
        return 0
    fi

    #echo -e "$HELP_MAIN" | $guitool --title='Elive Installer' --text-info --editable --width=500 --height=470  1>/dev/null
    local message_HELP_MAIN
    message_HELP_MAIN="$( printf "$( eval_gettext "To install Elive you need to have a free partition for it, if you don't know how to repartition your hard disk just select the automated partitioning which also includes the optimal options, but if you want to keep other systems like Windows with your Elive together in the computer you need to resize your Windows partition to make it smaller and then create a new one for Elive, you can do this with the Gparted application, please search a more detailed tutorial on internet about how to do that or the help of a friend. So basically Elive needs a single partition to install the system and your user files and we also suggest to add a SWAP partition to have hibernation features.\n\nEncryption is suggested so it will allow you to have your data protected even if your computer is stolen, nobody could read your files and specially your personal passwords, this will require to insert your password at every boot and is very important to not forget it, otherwise there will be no way to access to your data.\n\nThe password will be used for both your username and a special user called Root which is for admin purposes, also for the encrypted partition if you select it.\n\nThe hostname is a name that you give to your computer and is useful to know which one is your computer while browsing your network, we suggest to give it a funny name.\n\nFinally, if you think that we miss any important information on this guide please suggest it to us an improving of it. We hope you enjoy Elive as much as we did making it!" )" "" )"

    echo -e "$message_HELP_MAIN" | $guitool --title='Elive Installer' --text-info --editable --width=500 --height=470  1>/dev/null 2>&1 || true
}
#########################################################################
#state_development(){
    #functions_pre
    #el_debug "\targs '$*'"

    #if ! ((is_thanatests)) && ((conf_is_extra_help_wanted)) ; then
        ## run an hexchat instance
        #guitool "working_stop"

        #if $guitool --title='Elive Installer' --question --text="$( eval_gettext "Note: This is a beta version. If you find any issues in Elive, we need to know about them so we can address them for you and everyone else. If you don't report these issues, we won't be aware of them. Do you want to open our IRC chat channel now?" )" ; then

            #if el_verify_internet 2>/dev/null && ! LC_ALL="$EL_LC_EN" ps aux | grep -v grep | grep -v $$ | grep hexchat ; then
                #if [[ -n "$E_START" ]] ; then
                    #su -c "enlightenment_remote -exec-cmd \"/usr/bin/hexchat\"" $eliveuser 2>/dev/null
                #else
                    #su -c "/usr/bin/hexchat" $eliveuser 2>/dev/null &
                #fi
            #fi
        #fi
        #guitool "working_start"
    #fi
#}
########################################################################
passwords_verify(){
    # pre {{{
    local a b
    functions_pre
    #el_debug "\targs '$*'" # do not enable this, we don't want to send user's passwords over internet
    a="$1"
    b="$2"

    #el_check_variables "a|b"
    # - pre }}}
    if [[ -z "$a" ]] ; then
        guitool error "$( eval_gettext "Password fields cannot be empty." )"
        return 1
    fi
    if [[ "$a" = "$b" ]] ; then
        is_root_pass_validated=1
        return 0
    else
        guitool error "$( eval_gettext "Passwords do not match. Try again..." )"
        # return false
        return 1
    fi
}
#########################################################################
password_root_ask(){
    functions_pre
    # checks {{{
    # do we have already a root pass set? continue
    if echo "$rootpass" | grep -qs "[[:alnum:]]" && ((is_root_pass_validated)) ; then
        return 0
    fi

    # - checks }}}
    # thanatests {{{
    if ((is_thanatests)) ; then
        rootpass="asdf"
        return 0
    fi
    # }}}

    if ((is_migration)) ; then
        ubuntu_root_pass="$( grep "^root" $TARGET/etc.old-to-${eliveversion}/shadow | awk 'BEGIN{FS=":"}{print $2}' 2>/dev/null)"
        if [[ -z "$ubuntu_root_pass" ]] || [[ "$(echo "$ubuntu_root_pass" | wc -c)" -lt "4" ]] || [[ "$ubuntu_root_pass" = "x" ]] ; then
            is_ubuntu_root_pass_empty=1
            unset rootpass
            el_explain 0 "root password in previous system was empty, ask for a new one"
        fi
    fi

    # migration with correctly password: no more to do
    if ((is_migration)) && ! ((is_ubuntu_root_pass_empty)) ; then
        return 0
    fi

    el_debug "Setting the system's password"


    rootpass="$( $guitool --title='Elive Installer' --entry --hide-text --text="$( eval_gettext "Insert a password for your computer." )" || echo cancel )"

    # empty
    if [[ -z "$rootpass" ]] || [[ "$rootpass" = "cancel" ]] ; then
        guitool error "$( eval_gettext "You need to insert a password to continue. This password of your choice will be used to secure your Elive system." )"
        rootpass="$( $guitool --title='Elive Installer' --entry --hide-text --text="$( eval_gettext "Insert a password for your computer." )" || echo cancel )"
        verify_result "$rootpass"
    fi

    # request valid chars {{{
    if printf "%s" "$rootpass" | grep -qs '\\' ; then
        guitool error "$( eval_gettext "Sorry, you cannot use the slash character, use a different password" )"" (\\)."
        rootpass="$( $guitool --title='Elive Installer' --entry --hide-text --text="$( eval_gettext "Insert a password for your computer." )" )"
    fi
    if printf "%s" "$rootpass" | grep -qs '\\' ; then
        unset rootpass
    fi

    if printf "%s" "$rootpass" | grep -qs -- "-" ; then
        guitool error "$( eval_gettext "Sorry, you cannot use this character, use a different password:" )"" (-)."
        rootpass="$( $guitool --title='Elive Installer' --entry --hide-text --text="$( eval_gettext "Insert a password for your computer." )" )"
    fi
    if printf "%s" "$rootpass" | grep -qs -- "-" ; then
        unset rootpass
    fi


    if printf "%s" "$rootpass" | grep -qs "'" ; then
        guitool error "$( eval_gettext "Passwords cannot contain quotes. Please choose a different password." )"" (')."
        rootpass="$( $guitool --title='Elive Installer' --entry --hide-text --text="$( eval_gettext "Insert a password for your computer." )" )"
    fi
    if printf "%s" "$rootpass" | grep -qs "'" ; then
        unset rootpass
    fi

    if printf "%s" "$rootpass" | grep -qs '"' ; then
        guitool error "$( eval_gettext "Sorry, you cannot use the double-quote character, use a different password" )"" (\")."
        rootpass="$( $guitool --title='Elive Installer' --entry --hide-text --text="$( eval_gettext "Insert a password for your computer." )" )"
    fi
    if printf "%s" "$rootpass" | grep -qs '"' ; then
        unset rootpass
    fi

    # }}}

    # clean the ' char # update: not needed anymore since we don't accept them
    #rootpass="$( printf "%s" "$rootpass" | sed "s|'|'\\\''|g" )"


    LC_ALL=C sleep 0.2
    repeatrootpass="$( $guitool --title='Elive Installer' --entry --hide-text --text="$( eval_gettext "Again, please..." )" )"

    if ! passwords_verify "$rootpass" "$repeatrootpass" ; then
        password_root_ask
    fi
}
#########################################################################
username_user_verify(){
    functions_pre
    el_debug "\targs '$*'"

    if [[ -z "$username" ]] ; then
        guitool warning "$( eval_gettext "No name entered. Enter a user name to continue." )"
        username_user_ask
    fi
    if [[ "$username" = "eliveuser" ]] ; then
        guitool warning "$( eval_gettext "The username 'eliveuser' is not valid for compatibility reasons. Please choose a different username." )"
        username_user_ask
    fi
}
#########################################################################
username_user_ask(){
    functions_pre
    local answer

    if ((is_migration)) || ((is_shared_home_partition)) ; then
        return 0
    fi
    if [[ -n "$username" ]] && ! [[ "$username" = "eliveuser" ]] ; then
        return 0
    fi

    el_debug "Setting up the user"

    # thanatests {{{
    if ((is_thanatests)) ; then
        username="asdf"
        return 0
    fi
    # }}}

    answer="$( $guitool --title='Elive Installer' --entry --text="$( eval_gettext "Enter a short login name for your new user using English characters. (nickname)" )" || echo cancel )"
    if [[ "$answer" = "cancel" ]] || [[ -z "$answer" ]] ; then
        answer="eliveuser"
        local message_auto_username
        message_auto_username="$( printf "$( eval_gettext "Your username will be %s" )" "elive" )"

        guitool info "$message_auto_username"
    else
        verify_result "$answer"
    fi

    username="$( echo "$answer" | sed "s|'|'\\\''|g" )"

    # cleanup to valid username
    username="${username// /_}"
    username="${username//[![:alnum:]]/-}"

    # remove non-ascii characters
    username="$( echo "$username" | perl -p -e 's|[^[:ascii:]]||g' )"
    username="${username##+(-)}"
    username="${username%%+(-)}"

    if [[ -z "$username" ]] ; then
        username="user"
    fi

    if [[ -n "$answer" ]] && [[ "$username" != "$answer" ]] ; then
        answer="$( $guitool --title='Elive Installer' --entry --entry-text="$username" --text="$( eval_gettext "Warning: Invalid user name due to invalid characters, but it has been converted to a working one. Proceed with this or choose another name." )" )"
        verify_result "$answer"
        username="$answer"
        username="${username// /_}"
        username="${username//[![:alnum:]]/-}"
        if [[ -n "$answer" ]] && [[ "$username" != "$answer" ]] ; then
            local message_loginname
            message_loginname="$( printf "$( eval_gettext "Important: Your login name is %s" )" "$username" )"
            guitool info "$message_loginname"
        fi
    fi


    username_user_verify
}
#########################################################################
username_autologin_ask(){
    functions_pre
    el_debug "\targs '$*'"

    if ((is_cryptsetup_configured)) ; then
        is_autologin_wanted=1
        autologin_username="$username"
    fi

    # not already set, so ask
    if ! ((is_autologin_wanted)) ; then
        if guitool yesnoquestion "$( eval_gettext "Would you like to log in automatically without having to enter your password each time?" )" ; then
            is_autologin_wanted=1
            autologin_username="$username"
        fi
    fi

}
#########################################################################
swapfile_add_auto(){
    functions_pre
    el_debug "\targs '$*'"

    # create a swap file to use as swap, TODO see if we can use this for hibernation

    freespaceroot="$( LC_ALL="$EL_LC_EN" df -m | grep ${HD_PART_ROOT_dev} | awk '{print $4}' | tail -1 )"

    if [[ "$freespaceroot" -ge "$(( $RAM_TOTAL_SIZE_mb + 20000 ))" ]] ; then
        swapfilesize="$(( $RAM_TOTAL_SIZE_mb + 1200 ))"
    elif [[ "$freespaceroot" -ge "$(( $RAM_TOTAL_SIZE_mb + 10000 ))" ]] ; then
        swapfilesize="$(( $RAM_TOTAL_SIZE_mb + 600 ))"
    elif [[ "$freespaceroot" -ge "$(( $RAM_TOTAL_SIZE_mb + 3000 ))" ]] ; then
        swapfilesize="$(( $RAM_TOTAL_SIZE_mb + 300 ))"
    else
        return 1
    fi

    if [[ "$swapfilesize" -gt 12000 ]] ; then
        el_explain 0 "swapfile size was '$swapfilesize' MB but it was reduced to a max of 12000 MB"
        swapfilesize="12000"
    else
        echo -e "    swapfile better size is $swapfilesize"
    fi

    # TODO: move it to /boot or /var
    ( echo ; dd if=/dev/zero of=$TARGET/swapfile.swp bs=1M count=$swapfilesize 2>/dev/null ) | ( $guitool --title='Elive Installer' --progress --pulsate --auto-close --auto-kill || echo cancel )
    chmod 0600 "$TARGET/swapfile.swp" 1>/dev/null

    # TODO: implement: https://btrfs.readthedocs.io/en/latest/Swapfile.html
    mkswap "$TARGET/swapfile.swp" 1>/dev/null
    echo -e "/swapfile.swp    swap    swap        defaults        0    0" >> $TARGET/etc/fstab

    excluded_dirs_for_rsync="$excluded_dirs_for_rsync /swapfile.swp"
}
#########################################################################
grub_recover(){ # not cleaned yet
    functions_pre
    el_debug "\targs '$*'"
    unset point
    #   [[ -f /tmp/.macbook-detected ]] && return 0
    el_error "function not cleaned yet: $FUNCNAME"
    #if ! ((is_efi_booted)) ; then

    if guitool question "There's no Grub in the MBR of your harddisk: "$( get_disk_from_partition_complex "$1" ), "if you have installed Microsoft Windows it may have deleted your nice Grub menu. Do you want to (re)install Grub to your MBR? If you don't know what this step is for, then select Cancel." ; then
        point="$(awk -v dev=$1 '{if( $1 !~ /^#/ && $1 == dev) print $2}' /etc/fstab | grep -v "^#" | head -1)"
        # TODO: move to mounts-manager
        timeout 40 mount -o dev $point 2>/dev/null 1>&2
        if ! [[ -d $point/boot ]]
        then
            possible_boot_partition="$(awk '{if( $1 !~ /^#/ && $2 ~ /boot/) print $1}' $point/etc/fstab | grep -v "^#" )"
            [[ -z $possible_boot_partition ]] || timeout 25 mount -o dev $possible_boot_partition $point/boot

            if ! [[ -d $point/boot ]]
            then
                # FIXME: Add the feature for root partitions in XFS
                guitool error "The directory $point/boot doesn't exist"
                exit_error
            fi
        fi
        echo -e "Recovering grub to $( get_disk_from_partition_complex $i ) from $i, system:"
        echo -e "$( cat $point/etc/elive-version )"
        dev="${1}"
        dev="${dev#/dev/mapper/}"
        dev="${dev#/dev/md/}"
        dev="${dev#/dev/}"
        grub-install --no-floppy --recheck --root-directory=$point /dev/$( get_disk_from_partition_complex /dev/$dev ) 1>/dev/null || \
            grub-install --no-floppy --recheck --root-directory=$point/boot /dev/$( get_disk_from_partition_complex /dev/$dev ) 1>/dev/null
        echo grub-install --no-floppy --recheck --root-directory=$point $( get_disk_from_partition_complex $dev )
        guitool info "$( eval_gettext "Grub is re-installed in the MBR. You can now reboot to see if it works." )"
        exit
    fi
}
#########################################################################
host_ask(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    local answer hostsfile hostname_suffix
    if [[ -n "$username" ]] ; then
        hostname_suffix="$username"
    else
        #hostname_suffix="$(dmidecode -s system-manufacturer | awk '{print $1}' )"
        hostname_suffix="spaceship"
    fi
    # thanatests {{{
    if ((is_thanatests))  ; then
        hostname="Thana-hostname-auto"
    fi
    # }}}
    # }}}
    # do we have a name to recycle?
    if [[ -n "$old_hostname" ]] ; then
        hostname="$old_hostname"
        # cleanup hostname
        hostname="${hostname// /-}"
        hostname="${hostname//_/-}"
        hostname="${hostname//[![:alnum:]]/-}"
        # remove non-ascii characters
        hostname="$( echo "$hostname" | perl -p -e 's|[^[:ascii:]]||g' )"
        hostname="${hostname##+(-)}"
        hostname="${hostname%%+(-)}"
    fi

    if [[ -z "$hostname" ]] ; then
        answer="$( $guitool --title='Elive Installer' --entry --entry-text="Elive-${hostname_suffix}" --text="Hostname: $( eval_gettext "Choose a name for this computer using English characters only.\n\nThe hostname is a simple name that you give to this computer to identify it, it is very common to use a funny name for it." )" || echo cancel )"

        verify_result "$answer"
        hostname="$answer"
        read -r hostname <<< "$hostname"
    fi

    # cleanup hostname
    hostname="${hostname// /-}"
    hostname="${hostname//_/-}"
    hostname="${hostname//[![:alnum:]]/-}"
    # remove non-ascii characters
    hostname="$( echo "$hostname" | perl -p -e 's|[^[:ascii:]]||g' )"
    hostname="${hostname##+(-)}"
    hostname="${hostname%%+(-)}"

    if [[ -n "$answer" ]] && [[ "$hostname" != "$answer" ]] ; then
        answer="$( $guitool --title='Elive Installer' --entry --entry-text="$hostname" --text="$( eval_gettext "Warning, the name that you inserted was not valid because it included invalid characters to use as hostnames, I have cleaned it up to this one, you can continue or change it again." )" )"
        verify_result "$answer"
        hostname="$answer"
        hostname="${hostname// /-}"
        hostname="${hostname//_/-}"
        hostname="${hostname//[![:alnum:]]/-}"
        # remove non-ascii characters
        hostname="$( echo "$hostname" | perl -p -e 's|[^[:ascii:]]||g' )"
        hostname="${hostname##+(-)}"
        hostname="${hostname%%+(-)}"
    fi

    # indent
    # note: we copy to .orig because the mounting layers in chroot moved it
    if [[ -f "$TARGET/etc/hosts.orig" ]] ; then
        hostsfile="$TARGET/etc/hosts.orig"
    else
        hostsfile="$TARGET/etc/hosts"
    fi
    cat > "$hostsfile" << EOF
127.0.1.1 $hostname $hostnamefqdn
127.0.0.1 localhost localhost.localdomain
# INFO: 127.0.0.1 is localhost, and 127.0.1.1 is the hostname (machine-name, without domain) and also the FQDN one, some applications like gnome requires this to be mapped when there's no conection to internet

# The following lines are desirable for IPv6 capable hosts
# (added automatically by netbase upgrade)

::1     ip6-localhost ip6-loopback
fe00::0 ip6-localnet
ff00::0 ip6-mcastprefix
ff02::1 ip6-allnodes
ff02::2 ip6-allrouters
ff02::3 ip6-allhosts

EOF

    cp "$hostsfile" "$TARGET/etc/hosts.bak"

    # anything back to import?
    if ((is_migration)) ; then
        if [[ -s "$TARGET/etc.old-to-${eliveversion}/hosts" ]] ; then
            cat  "$TARGET/etc.old-to-${eliveversion}/hosts" | grep -v "^#" | grep -v "^::" | grep -v "^f" | grep -v "^127" | grep "[[:digit:]]" | sed '/^$/d' >> "$hostsfile"
        fi
    fi


    if [[ -f "$TARGET/etc/hostname.orig" ]] ; then
        echo "$hostname" > "$TARGET/etc/hostname.orig"
    else
        echo "$hostname" > "$TARGET/etc/hostname"
    fi

    echo "$hostname" > "$TARGET/etc/mailname"

}
#########################################################################
loopback_add(){ # not cleaned yet
    # TODO: needed?
    functions_pre
    el_debug "\targs '$*'"
    if cat $1 | egrep -q "auto.*lo" && cat $1 | egrep -q "iface lo inet loopback"
    then
        echo "Loopback correct on network interfaces"
        return 0
    fi

    echo -e "auto lo\niface lo inet loopback\n" > $1
}
#########################################################################
nvidia_bug_fix(){ # not cleaned yet
    functions_pre
    el_debug "\targs '$*'"
    # TODO: needed?
    # Fix #421028
    # Actually added to deliver
    if ( test -f $TARGET/usr/lib/xorg/modules/drivers/nvidia_drv.o ) && ( test ! -f $TARGET/usr/lib/xorg/modules/drivers/nvidia_drv.so )
    then
        mkdir -p $TARGET/usr/lib/xorg/modules/drivers
        cp -a /usr/lib/xorg/modules/drivers/nvidia_drv.so $TARGET/usr/lib/xorg/modules/drivers
    fi
}
#########################################################################
hd_errors_check(){
    functions_pre
    #el_debug "\targs '$*'"
    local size HD_DEV_BADBLOCK_SECTORS

    if ! [[ -b "${HD_DEV}" ]] ; then
        return 0
    fi
    if el_flag check badblocks_hd_repaired ; then
        return 0
    fi
    if el_flag check badblocks_notified ; then
        return 0
    fi


    HD_DEV_BADBLOCK_SECTORS="$( LC_ALL="$EL_LC_EN" smartctl -s on -A "${HD_DEV}" | grep -i "Reallocated_Sector_C" | tr ' ' '\n' | grep "[[:digit:]]" | sed -e 's/[^0-9]*//g' | tail -1 )"

    if [[ -n "$HD_DEV_BADBLOCK_SECTORS" ]] && [[ "$HD_DEV_BADBLOCK_SECTORS" -gt 0 ]] ; then

        local message_badblocks_found
        #message_badblocks_found="$( printf "$( eval_gettext "Elive detected that your hard disk contains badblocks, errors of unaccessible inodes, normally you should replace your hard disk with a new one, they becomes unusable and sometimes unrecoverable, sometimes zeroing the entire disk can make it to work, but stills not a reliable option, please don't save your important data on it so you can lose it in the future with big probability. The installer can give you the option to erase it, the process is very slow and is not a guaranteed process." )" "" )"
        message_badblocks_found="$( printf "$( eval_gettext "Elive has detected that your hard disk contains bad blocks. You should be able to continue using your hard disk by performing a new installation where all its files will be renewed. If you encounter problems in the future, such as applications crashing or data loss, you should consider replacing your hard disk. It is strongly advised to always keep a backup of your important files." )" "" )"

        el_warning "badblocks in your HD '$HD_DEV' found ($HD_DEV_BADBLOCK_SECTORS, by smartctl), user is informed"
        guitool warning "$message_badblocks_found"
        el_flag  add  badblocks_notified

        if ((is_virtualized)) ; then
            return 0
        fi

        local message_badblocks_repair
        message_badblocks_repair="$( printf "$( eval_gettext "Would you like to fill your entire hard disk with zeroes? This will help identify all bad sectors but is a risky process as it could render your disk completely unusable if there is extensive damage. We recommend this if the quality of your data is important; otherwise, you can skip this step. This extremely slow process can multiple hours. All data on this hard disk WILL BE ERASED." )" "" )"

        if guitool yesnoquestioninvbutcontinue "$message_badblocks_repair" ; then
            if ! guitool question "$( eval_gettext "You are about to erase your ENTIRE hard disk. Continue?" )" ; then
                exit
            fi

            # Erase the disk
            { echo 10 ; dd if=/dev/zero of=${HD_DEV} bs=4M  2>/dev/null || true ; sync ; echo 100 ; } | $guitool --title='Elive Installer' --progress --text="$( eval_gettext "Overwriting your hard disk with zeroes. Be patient; this may take some good time." )" --pulsate --auto-close
                el_flag  add  badblocks_hd_repaired

            # re-read partition table, remove ghost partitions
            device_check_is_busy "$HD_DEV"

            #echo 20 > "$TMP_PROGRESS_CONFIGURING_f"

            # do a new check of the bad blocks:
            buf="$( LC_ALL="$EL_LC_EN" smartctl -s on -A "${HD_DEV}" | grep -i "Reallocated_Sector_C" | tr ' ' '\n' | grep "[[:digit:]]" | sed -e 's/[^0-9]*//g' | tail -1 )"
            local message_new_badblocks
            message_new_badblocks="$( printf "$( eval_gettext "Previously, you had %s bad blocks found, and now there are %s ones." )" "$HD_DEV_BADBLOCK_SECTORS" "$buf" )"
            guitool info "$message_new_badblocks"

            # report this useful information
            el_error "INFO REPAIR:  HD had badblocks ($HD_DEV_BADBLOCK_SECTORS) and after to ZEROING the disk it has now $buf "


            #if [[ "$HD_DEV_BADBLOCK_SECTORS" -gt 0 ]] ; then
                ## note: seems like smartctl continue showing errors even if we have them removed, so do not check it again
                #guitool info "$( eval_gettext "Hard disk has been erased, but there still seem to be errors. Please reboot your system and try again." )"

            #else
                ## if no smartmontools supported we need to make sure that the number is zero
                #if [[ "$HD_DEV_BADBLOCK_SECTORS" -eq 0 ]] ; then
                    #local message_badblocks_repaired_good
                    #message_badblocks_repaired_good="$( printf "$( eval_gettext "Congratulations! Your hard disk has been repaired! Installer will be run again..." )" "" )"

                    #guitool info "$message_badblocks_repaired_good"
                #fi
            #fi


            # exit in any of the cases, to run again the installer
            ( sync ; sleep 6 ; sudo IGNOREALREADYRUNNING=1 eliveinstaller-wrapper ; sleep 2 ; ) &
            exit

        fi
    fi

}
#########################################################################
media_errors_check(){
    functions_pre
    el_debug "\targs '$*'"
    local file line buf

    # already reported
    if ((is_errors_in_live_reported)) && [[ "$1" != "force" ]] ; then
        return
    fi

    # fetch messages
    buf="$( { LC_ALL=C tail -n 100 /var/log/messages /var/log/dmesg /var/log/syslog /var/log/kern.log 2>/dev/null || true ; dmesg 2>&1 | tail -n 400 ; } | LC_ALL=C grep -vE "(usb.*set config.*error)" | grep "error" || true )"
    if [[ -z "$buf" ]] ; then
        return
    fi

    # demo for debug:
    #buf="
#Feb 23 08:13:30 localhost kernel: [  889.523048] SQUASHFS error: Unable to read page, block 608e5893, size f5b5
#Feb 23 08:13:30 localhost kernel: [  889.524290] SQUASHFS error: squashfs_read_data failed to read block 0x608c0490
#Feb 23 08:13:30 localhost kernel: [  889.524291] SQUASHFS error: Unable to read data cache entry [608c0490]
#Feb 23 08:13:30 localhost kernel: [  889.524293] SQUASHFS error: Unable to read page, block 608c0490, size d261
#Feb 23 08:13:30 localhost kernel: [  889.528884] SQUASHFS error: squashfs_read_data failed to read block 0x60a4f56b
#Feb 23 08:13:30 localhost kernel: [  889.528889] SQUASHFS error: Unable to read fragment cache entry [60a4f56b]
#Feb 23 08:13:30 localhost kernel: [  889.528891] SQUASHFS error: Unable to read page, block 60a4f56b, size 9c3a
#Feb 23 08:13:30 localhost kernel: [  889.529284] SQUASHFS error: Unable to read fragment cache entry [60a4f56b]
#Feb 23 08:13:30 localhost kernel: [  889.529285] SQUASHFS error: Unable to read page, block 60a4f56b, size 9c3a
#[ 352.092275] usb 2-2: device descriptor read/64, error -71

#"

    while read -ru 3 line
    do
        if ((is_errors_in_live_reported)) ; then
            break
        fi

        #if echo "$line" | LC_ALL=C grep -iE "error" | LC_ALL=C grep -iE "(squashfs|sr0|sr1|sr2|sg0|sg1|sg2|dvd|scd)" | LC_ALL=C grep -qsiE "(sector|Unable to read|failed to read)" ; then
        #if echo "$line" | LC_ALL=C  grep -qsiE "(SQUASHFS.*error|I/O error.*dev loop|I/O error.*sector|critical medium error.*sector|Medium Error|usb.*read.*error)" ; then
        if echo "$line" | LC_ALL=C  grep -qsiE "(SQUASHFS.*error|I/O error.*dev loop)" ; then

            if ((is_live_from_usb)) ; then
                # errors from Squashfs (usb)
                el_error "errors in the live media (usb)"
                guitool error "$( eval_gettext "The integrity check for Elive showed errors. Do not install Elive with corrupted data. The issue could be due to an interrupted download, or a damaged DVD or USB drive. Please try again with a new download and use a different storage medium. We recommend using a USB drive instead of a DVD." )"
                is_errors_in_live_reported=1
                break 1
            else
                # errors from Squashfs (dvd)
                el_error "errors in the live media (cdrom)"
                guitool error "$( eval_gettext "The integrity check for Elive showed errors. Do not install Elive with corrupted data. The issue could be due to an interrupted download, or a damaged DVD or USB drive. Please try again with a new download and use a different storage medium. We recommend using a USB drive instead of a DVD." )"
                is_errors_in_live_reported=1
                break 1
            fi

        fi
    done 3<<< "$( echo "$buf" )"

    true
}

#########################################################################
grub_mbr_search(){ # not cleaned yet
    # TODO: re-implement
    functions_pre
    el_debug "\targs '$*'"

    # FIXME: Feature to recover grub not yet implemented
    el_step_requires_fixme "/tmp" "Feature to recover grub not yet implemented"


    if ! cat $( get_disk_from_partition_complex $1 ) 2>/dev/null | head -6 | strings | grep -qUi grub ; then
        echo -e "    Not Grub found in the MBR of $( get_disk_from_partition_complex $1 )"
        el_debug "grub recover in $1"
        grub_recover "$1"
    fi
    ((is_grub_recovery_asked)) && grub_recover "$1"
}
#########################################################################
#calculate_disks_and_sizes(){
    ## FIXME: move to partitions-list tool, this is very not reliable, dont include vda, or raids or lvm or others
    #functions_pre
    #el_debug "\targs '$*'"
    ## Calculate Disks and sizes

    #DISKS_AND_SIZES="$(awk '{
    #if ($4 ~ /^[hs]d.*[^[:digit:]]+$/)
        #{
            #dev=$4
            #size=$3
            #printf("/dev/%s %d\n", dev, int(size) / 1024)
        #}
    #}' /proc/partitions)"

    #echo -e "    Disks and Sizes:\n$DISKS_AND_SIZES" 1>&2
#}
#########################################################################
root_partition_select(){
    # pre {{{
    functions_pre
    el_debug

    local entry part menu answer ignore_this_format m_part

    # - pre }}}
    # list partitions and ask which one to use {{{
    if [[ -n "${HD_PART_ROOT_dev}" ]] ; then
        # we may have already this partition set, from old-system functions for example
        answer="${HD_PART_ROOT_dev}"
    else
        el_explain 2 "Selection of the root partition"

        guitool "working_start"

        unset menu
        while read -ru 3 entry
        do
            part="$( echo "$entry" | awk -v FS="::" '{print $1}' )"
            if ! [[ -b "$part" ]] ; then
                continue
            fi
            # update: nothing wrong by usings disks, and also: if the partition is not formated it doesn't has UUID
            #if ! LC_ALL="$EL_LC_EN" udevadm info --query=all --name="$part" 2>&1 | grep -qs "ID_FS_UUID=" ; then
                #el_debug "ignoring $entry because is a disk"
                #continue
            #fi

            if el_array_member_check "$part" "${HD_PARTITIONS_USED_array[@]}" || el_array_member_check "$part" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; then
                el_debug "$part already in our array of partitions used, skipping it"
                el_debug "partitions used array is: ${HD_PARTITIONS_USED_array[*]}"
            else
                # skip invalid ones:
                # show partitions of at least 2 gb, which will skip /boot or EFI ones making it easier for the end user to pick
                if ! [[ "$( echo "$entry" | awk -v FS="::" '{print $6}' )" -gt 2000000000 ]] ; then
                    el_debug "skipping '$entry' because is too small"
                    continue
                fi
                # Include device name
                menu+=("$( echo "$entry" | awk -v FS="::" '{print $1}' )")

                # get the name of the OS for the menu in priority
                if [[ -s /tmp/os-probed ]] && grep -qs "^$( echo "$entry" | awk -v FS="::" '{print $1}' ):" /tmp/os-probed ; then

                    menu+=("$( awk -v FS=":" -v m_part="$(echo "$entry" | awk -v FS="::" '{print $1}' )"  '{if ($1 == m_part ) print $2}' /tmp/os-probed ) - $( echo "$entry" | awk -v FS="::" '{print "("$3")"}' )")

                else

                    # fix names for the menus:
                    entry="$( echo "$entry" | awk -v FS="::" -v OFS="::" '{if ( $2 == "extended" ) $5 = "(extended)" ; if ( $2 == "disk" ) $5 = "(disk)" ; if ( $3 == "crypto_LUKS" ) $5 = "(crypto_LUKS)"  ; print $0}' )"

                    case "$( echo "$entry" | awk -v FS="::" '{print $5}' )" in
                        "(no name)")
                            menu+=("$( echo "$entry" | awk -v FS="::" '{print "("$3")"}' )")
                            ;;
                        "(swap)")
                            menu+=("$( echo "(Swap partition)" )")
                            ;;
                        "(extended)")
                            menu+=("$( eval_gettext "(Extended partition)" )")
                            ;;
                        "(disk)")
                            menu+=("$( eval_gettext "(Entire disk)" )")
                            ;;
                        "(unformatted)")
                            menu+=("$( eval_gettext "(Unformatted)" )")
                            ;;
                        "(crypto_LUKS)")
                            menu+=("$( eval_gettext "(LUKS Encryption layer)" )")
                            ;;
                        *)
                            menu+=("$( echo "$entry" | awk -v FS="::" '{print $5" - ("$3")"}' )")
                            ;;
                    esac
                fi

                # include size
                size="$( echo "$entry" | awk -v FS="::" '{print $7}' )"
                if [[ -z "$size" ]] || [[ "$size" = "uknown" ]] ; then
                    size="$( eval_gettext "unknown" )"
                fi
                menu+=("$( echo "$size" )")

                # do not use el_ arrays, they uses a different algorithm that doesn't allow duplcates
            fi
        done 3<<< "$( EL_DEBUG=0 partitions-list --show-all --show-raw | grep -viE "::(extended|crypto_LUKS|persistence|udf|iso9660|squashfs|EFI|Elive_)::" || true )"

        # TODO: add here and all the other "partitioners selectors" a "Help" entry where explains the users what they needs to do here (full-dumb mode, we want ppl be able to install elive)

        # menu is generated, show it in the GUI
        #guitool "working_stop"
        answer="$( $guitool --title='Elive Installer' --list --height=400 --width=620 --text="$( eval_gettext "Choose where to install Elive (Root partition)" )" --column="$( eval_gettext "Device" )" --column="$( eval_gettext "Name" )" --column="$( eval_gettext "Size" )" "${menu[@]}" 2>/dev/null || echo cancel )"

        if [[ "$answer" = "cancel" ]] || [[ -z "$answer" ]] ; then
            guitool error "$( eval_gettext "No partition selected for Elive installation. Installation aborted. Restart the installer after repartitioning your hard disk as desired." )"
            guitool info "$( eval_gettext "If you're unsure how to partition your hard disk, it's recommended to use the automated partition mode, which will handle it for you with optimal settings." )"
            exit
        fi

        # check and assign
        verify_result "$answer"

        guitool "working_start"
    fi

    HD_PART_ROOT_dev="$answer"
    el_explain 0 "Selected root partition is $HD_PART_ROOT_dev"
    el_umount "$HD_PART_ROOT_dev" 1>/dev/null

    if echo "$HD_PART_ROOT_dev" | grep -qsE "/dev/sd" ; then
        if ! echo "$HD_PART_ROOT_dev" | grep -qsE "/dev/sd.*[[:digit:]]$" ; then
            if ! guitool question "$( eval_gettext "Your selected root partition appears to be an entire disk rather than a partition. It is recommended to repartition your hard disk to create the necessary partitions first. Would you like to proceed with your selection?" )" ; then
                exit
            fi
        fi
    fi

    # get the name of the device for this partition, no matter which structure this partition has
    HD_DEV="$( get_disk_from_partition_complex "$HD_PART_ROOT_dev" )"


    if [[ -n "$HD_DEV" ]] && [[ -b "$HD_DEV" ]] ; then
        # check
        device_check_is_busy "$HD_DEV"

        # enable smartmon to scan for possible bad blocks on the disk
        if ! smartctl -s on "$HD_DEV" 2>&1 | grep -qsiE "(model|SMART Enabled)" ; then
            echo -e "smartmontools" >> "/tmp/.packages-to-remove.tmp"
            echo -e "hddtemp" >> "/tmp/.packages-to-remove.tmp" || true
        fi

        HD_SIZE_mb="$( hd_device_get_size_mb "$HD_DEV" )"

        # determine if MSDOS or GPT
        if ! ((is_disk_table_msdos)) && ! ((is_disk_table_gpt)) ; then
            header_type="$( LC_ALL="$EL_LC_EN" parted -ms -a optimal -l "$HD_DEV" 2>/dev/null | awk -v HD_DEV="$HD_DEV" -v FS=":" '{if ($1 == HD_DEV) print $6}' )"

            case "$header_type" in
                msdos)
                    is_disk_table_msdos=1
                    el_debug "Disk has table: MSDOS"
                    if [[ "$EL_DEBUG" -gt 2 ]] ; then
                        el_debug "result of fdisk:"
                        fdisk -l 1>&2
                    fi
                    ;;
                gpt)
                    is_disk_table_gpt=1
                    el_debug "Disk has table: GPT"

                    # note: do not run gdisk on a msdos-type disk, it will fail and maybe broke it!
                    if [[ "$EL_DEBUG" -gt 2 ]] ; then
                        el_debug "result of gdisk:"
                        timeout 14 gdisk -l "$HD_DEV" 1>&2
                    fi
                    ;;
                *)
                    if ((is_efi_booted)) ; then
                        is_disk_table_gpt=1
                    else
                        is_disk_table_msdos=1
                    fi
                    el_warning "Unknown partitions table for $HD_DEV ($header_type), not gpt or msdos? asigning autodetection...\ndemo data, column 6:\n$( LC_ALL="$EL_LC_EN" parted -ms -a optimal -l "$HD_DEV" 2>/dev/null )"
                    ;;
            esac
        fi

        if [[ "$EL_DEBUG" -gt 2 ]] ; then
            el_debug "result of parted:"
            LC_ALL="$EL_LC_EN" parted -ms -a optimal -l "$HD_DEV" 2>&1 | grep -vE "(sr0|Error)" 1>&2

            el_debug "result of partitions-list:"
            EL_DEBUG=0 partitions-list --show-all-raw 1>&2 || true

            el_debug "result of: sfdisk -d:"
            LC_ALL="$EL_LC_EN" sfdisk -d "${HD_DEV}" 1>&2 || true
            1>&2



        fi

        # check and repair errors in HD (fast check)
        hd_errors_check
    else
        unset HD_DEV
    fi


    if LC_ALL="$EL_LC_EN" udevadm info --query=all --name="$HD_PART_ROOT_dev" 2>/dev/null | grep -qs "E: ID_USB_DRIVER=usb-storage" && ! ((is_autoupgrade)) ; then
        guitool warning "$( eval_gettext "Warning: The partition you have selected is on a USB thumb drive. Installing an operating system on a USB thumb drive is not recommended due to its slower performance and limited lifespan from write operations. We advise using it only as a live system or with persistence enabled, which can be set up using the tool from the Elive menu. You can proceed with the installation, but please be aware of these risks." )"
    fi

    # Check the key / installer-module here
    #instcod_request

    # - list partitions and ask which one to use }}}
    # checks {{{

    # Check speeds
    verify_speeds "$( get_disk_from_partition_complex "${HD_PART_ROOT_dev}" )"

    # Check if SSD
    if device_is_ssd "$HD_PART_ROOT_dev" && ! ((is_partitioning_auto)) && ! ((ignore_show_suggestions)) && ! ((is_migration)) ; then
        local message_device_is_ssd
        # translators: example data for the %s entries: Ext4, Mib, gparted
        message_device_is_ssd="$( printf "$( eval_gettext "Your hard disk is an SSD. It is recommended to use the %s file system for all your partitions. This will improve the performance of your SSD. Also, make sure that your partitions are correctly aligned as '%s' in the %s application." )" "Ext4" "MiB" "gparted" )"

        guitool info "$message_device_is_ssd"
    fi

    # Check alignment
    partition_alignment_check "$HD_PART_ROOT_dev" report

    if [[ "$HD_PART_ROOT_dev" = "/dev/mapper/"* ]] ; then
        is_fs_mapper=1
    fi

    # Check if there's a system that the user would want to migrate to an Elive system
    if ! ((is_partitioning_auto)) ; then
        old_system_prepare
    fi

    # }}}
    # ask for filesystem {{{

    if ((is_migration)) ; then
        el_check_variables "OLD_HD_PART_ROOT_FS_v" # TODO: put variable checkers everywhere in the installer when they are really needed
        HD_PART_ROOT_FS_v="$OLD_HD_PART_ROOT_FS_v"
    else

        # ask if wants to format
        if ((ignore_format_disks)) ; then
            HD_PART_ROOT_FS_v="$( hdtools_get_fs "$HD_PART_ROOT_dev" )"
            ignore_this_format=1
        else
            local message_erase
            message_erase="$( printf "$( eval_gettext "About to ERASE your partition '%s' for '%s'" )" "${HD_PART_ROOT_dev}" "/ (root)" )"

            # going to format, are you sure?
            if [[ -z "$HD_PART_ROOT_dev" ]] ; then
                if ! guitool question "$message_erase" ; then
                    if guitool question "$( eval_gettext "You chose not to format the partition, but the installer expects it to be formatted. Continue without formatting?" )" ; then
                        ignore_this_format=1
                        HD_PART_ROOT_FS_v="$( hdtools_get_fs "$HD_PART_ROOT_dev" )"
                        el_explain 2 "User wanted to NOT format the partition"
                    fi
                fi
            fi
        fi

        # we may have already it, like from automated partitioning
        if [[ -z "${HD_PART_ROOT_FS_v}" ]] ; then
            answer="$( select_filesystem "${HD_PART_ROOT_dev}" )"
            verify_result "$answer"

            : ${HD_PART_ROOT_FS_v=$answer}
            : ${HD_PART_ROOT_FS_v="ext4"}
        fi
    fi


    # - ask for filesystem }}}
    # format partition {{{
    el_check_variables "HD_PART_ROOT_dev|HD_PART_ROOT_FS_v"

    if ! ((is_autoupgrade)) ; then
        el_umount "$HD_PART_ROOT_dev" 1>/dev/null
    fi


    #if ((is_disk_table_gpt)) ; then
        ## gpt mode: UPDATE this doesn't works: they are set to "ee" and as a single partition
        ## this can be a way to get the ID, but is not very reliable, so just better set it as we wanted it to be
        ##HD_PART_ROOT_dev_id="$( echo -e "p\nq" | LC_ALL="$EL_LC_EN" timeout 10 gdisk "$HD_DEV" 2>/dev/null | tail -n "+10" | grep "^/dev/" | awk -v HD_PART_ROOT_dev="$HD_PART_ROOT_dev" 'if ($1 == HD_PART_ROOT_dev ) {print $5}' | tail -1 )"
        #HD_PART_ROOT_dev_id="$(
        #while read -ru 3 line
        #do
            #if [[ "$line" = "Number"* ]] || ((is_reached)) ; then
                #is_reached=1
            #else
                #continue
            #fi

            #echo "$line" | awk -v id="$( get_partition_number_from_partition "$HD_PART_ROOT_dev" )" 'if ($1 == id) {print $6}'

        #done 3<<< "$( echo -e "p\nq" | LC_ALL="$EL_LC_EN" timeout 10 gdisk "$HD_DEV" 2>/dev/null )" | head -1
        #)"
        #HD_PART_ROOT_dev_id="${HD_PART_ROOT_dev_id:0:2}"
    #else

        #HD_PART_ROOT_dev_id="$( sfdisk --print-id $( get_disk_from_partition_complex ${HD_PART_ROOT_dev} ) $( get_partition_number_from_partition ${HD_PART_ROOT_dev} ) )"
    #fi

    # set it "linux partition" if is already not set

        #&& [[ "$HD_PART_ROOT_dev_id" != 83 ]] ; then
    if ! ((is_partitioning_auto)) ; then

        # we must convert to gpt
        if ((is_disk_table_msdos)) && [[ -b "$HD_DEV" ]] ; then
            if ((is_efi_booted)) ; then
                # warn about the requirement
                guitool warning "$( eval_gettext "You must convert your disk to a GPT structure to work with UEFI. Alternatively, you can configure your BIOS to boot in Legacy mode." )"

                # convert to gpt
                if guitool question "$( eval_gettext "Would you like to attempt converting your hard disk table from an MSDOS partition table to the GPT structure? This operation can be risky for your data, so if you have any important data, you must back it up before proceeding. CONTINUE AT YOUR OWN RISK." )" ; then

                    # convert
                    # note: this call will only show errors in standard-output
                    # first: get status, if gpt is corrupted create a new one:
                    result="$( echo -e "q" | LC_ALL=C gdisk "$HD_DEV" 2>&1 1>/dev/null | colors-remove )"
                    if echo "$result" | grep -qsi "invalid main GPT header" ; then
                        # this commands selects the first option (use mbr), and write
                        el_explain 0 "converting disk table MSDOS -> GPT using forced mode (was corrupted, so recreating)"
                        result="$( echo -e "1\nw\nY\nq" | LC_ALL=C gdisk "$HD_DEV" 2>&1 | colors-remove )"
                        unset result
                    else
                        el_explain 0 "converting disk table MSDOS -> GPT using normal"
                        result="$( echo -e "w\nY\nq" | LC_ALL=C gdisk "$HD_DEV" 2>&1 | colors-remove )"
                    fi
                    if echo "$result" | grep -qsi "partition table overlaps" ; then
                        guitool error "$( eval_gettext "Conversion failed due to overlapping partitions. This issue can be resolved by shrinking the problematic partition's end by approximately 8 MB and then retrying the installer. This can be done using Gparted. We strongly recommend backing up your important data before proceeding. Aborting installation..." ):\n\n$( echo "$result" | tail -n 10 )"
                        exit
                    fi
                    result="$( echo "$result" | grep -vE "^(GPT fdisk \(gdisk|Partition table scan:|  MBR:|  BSD:|  APM:|  GPT:|\*\*\*\*\*|Found invalid|in memory. |typing 'q'|to GPT|Command \(|Final checks|PARTITIONS\!\!|Do you want|The operation has)" | sed -e '/^$/d' )"

                    if [[ -n "$result" ]] ; then
                        el_warning "disk table conversion (msdos -> gpt) problem?:\n$result"
                    fi

                    el_explain 0 "conversion done, reboot needed now (user told)..."
                    guitool warning "$( eval_gettext "IMPORTANT: Conversion complete. You must reboot the computer to recognize the new disk structure, then run the installer again." )"
                    touch /tmp/reboot-needed
                    exit

                    partprobe "$HD_DEV" || true
                    is_disk_table_gpt=1
                    unset is_disk_table_msdos
                else

                    # user don't want to convert
                    if ! guitool question "$( eval_gettext "Continue with the installation anyway?" )" ; then
                        exit
                    fi
                fi
            #else
                ## info about improvement
                # update: not needed
                #guitool info "$( eval_gettext "Your hard disk was originally formated with the old MSDOS table, we recommend it to convert to the new GPT structure if you can boot your computer in UEFI mode." )"
            fi
        fi

        # UPDATE: we should not need this code:
        #if echo "$HD_PART_ROOT_dev" | grep -qsE "((xvd|sd|hd|vd)[a-z]|nvme[[:digit:]]+n[[:digit:]]+|mmcblk[[:digit:]]+)" ; then

            #if ((is_disk_table_msdos)) ; then
                ##el_warning "$HD_PART_ROOT_dev is not set as a linux-partition-type (83), It should set it? details: device: $HD_PART_ROOT_dev fs: $HD_PART_ROOT_FS_v , $(file -s "$HD_PART_ROOT_dev" )"
                #sfdisk --force -c "$( get_disk_from_partition_complex ${HD_PART_ROOT_dev} )" "$( get_partition_number_from_partition ${HD_PART_ROOT_dev} )" 83
            #else
                ##parted -ms -a optimal "$( get_disk_from_partition_complex ${HD_PART_ROOT_dev} )" mkfs "$( get_partition_number_from_partition ${HD_PART_ROOT_dev} )" reiserfs
                ##parted -ms -a optimal "$( get_disk_from_partition_complex ${HD_PART_ROOT_dev} )" mkfs "$( get_partition_number_from_partition ${HD_PART_ROOT_dev} )" reiserfs
                ## change the type of partition
                ## TODO: add 1>/dev/null to silent it (debug purposes)
                #el_debug "Going to send the next gdisk commands for write root partition as a linux FS: \nt\n$( get_partition_number_from_partition ${HD_PART_ROOT_dev} )\n8300\nw\nY\nq"
                #if ! echo -e "t\n$( get_partition_number_from_partition ${HD_PART_ROOT_dev} )\n8300\nw\nY\nq" | LC_ALL="$EL_LC_EN" timeout 20 gdisk "$( get_disk_from_partition_complex ${HD_PART_ROOT_dev} )" ; then
                    ## use "2" before to change the id, so it fails it can be because it asks which mode (msdos/gpt/blank-gpt) to use, and we want to select "2" in this case
                    #el_debug "Going to send the next gdisk commands for write root partition as a linux FS: \n2\nt\n$( get_partition_number_from_partition ${HD_PART_ROOT_dev} )\n8300\nw\nY\nq"
                    #echo -e "2\nt\n$( get_partition_number_from_partition ${HD_PART_ROOT_dev} )\n8300\nw\nY\nq" | LC_ALL="$EL_LC_EN" timeout 20 gdisk "$( get_disk_from_partition_complex ${HD_PART_ROOT_dev} )"
                #fi
            #fi
        #fi



        device_check_is_busy "$HD_PART_ROOT_dev"

        if ((ignore_this_format)) || ((ignore_format_disks)) ; then
            HD_PART_ROOT_FS_v="$( hdtools_get_fs "$HD_PART_ROOT_dev" )"
        fi

        case "${HD_PART_ROOT_FS_v}" in
            reiserfs)
                if ! ((is_migration)) && ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    partition_format "${HD_PART_ROOT_dev}" "${HD_PART_ROOT_FS_v}" "Elive_OS"
                fi
                HD_PART_ROOT_FS_OPTS_v="defaults,commit=300,relatime,notail"
                is_reiserfs_used=1
                ;;
            xfs)
                if ! ((is_migration)) && ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    partition_format "${HD_PART_ROOT_dev}" "${HD_PART_ROOT_FS_v}" "Elive_OS"
                fi
                #needs_boot_partition=1
                ;;
            ext3)
                if ! ((is_migration)) && ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    if device_is_ssd "$HD_PART_ROOT_dev" ; then
                        HD_PART_ROOT_FS_v="ext4"
                    fi
                    partition_format "${HD_PART_ROOT_dev}" "${HD_PART_ROOT_FS_v}" "Elive_OS"
                    HD_PART_ROOT_FS_OPTS_v="defaults,commit=300,relatime"
                fi
                ;;
            ext4)
                if ! ((is_migration)) && ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    partition_format "${HD_PART_ROOT_dev}" "${HD_PART_ROOT_FS_v}" "Elive_OS"

                    HD_PART_ROOT_FS_OPTS_v="defaults,commit=300,relatime"
                fi
                ;;
            ext2)
                if ! ((is_migration)) && ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    partition_format "${HD_PART_ROOT_dev}" "${HD_PART_ROOT_FS_v}" "Elive_OS"
                fi
                ;;
            jfs)
                if ! ((is_migration)) && ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    partition_format "${HD_PART_ROOT_dev}" "${HD_PART_ROOT_FS_v}" "Elive_OS"
                fi
                ;;
            reiser4)
                if ! ((is_migration)) && ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    partition_format "${HD_PART_ROOT_dev}" "${HD_PART_ROOT_FS_v}" "Elive_OS"
                fi
                if device_is_ssd "$HD_PART_ROOT_dev" ; then
                    HD_PART_ROOT_FS_OPTS_v="defaults,relatime,txmod=wa"
                else
                    HD_PART_ROOT_FS_OPTS_v="defaults,relatime"
                fi
                needs_boot_partition=1
                el_array_member_add "$HD_PART_ROOT_FS_v" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")
                ;;
            reiser5)
                el_error "reiser5 needs to implement"
                exit_error
                ;;
            btrfs)
                if ! ((is_migration)) && ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    partition_format "${HD_PART_ROOT_dev}" "${HD_PART_ROOT_FS_v}" "Elive_OS"
                fi

                if ((is_fs_compression_wanted)) ; then
                    HD_PART_ROOT_FS_OPTS_v="compress=${compressor_algorithm},commit=300,lazytime"
                else
                    HD_PART_ROOT_FS_OPTS_v="commit=300,lazytime"
                fi
                # there's an error on the boot in grub that is unable to write on btrfs partitions, so we need to create a separate /boot partition
                needs_boot_partition=1
                el_array_member_add "$HD_PART_ROOT_FS_v" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")
                ;;
            *)
                local message_wrong_fs

                if ((is_migration)) ; then
                    message_wrong_fs="$( printf "$( eval_gettext "The filesystem '%s' for the root partition is not detected correctly. Please report this issue to Elive so it can be resolved." )" "$HD_PART_ROOT_FS_v" )"

                    guitool error "$message_wrong_fs"
                else
                    message_wrong_fs="$( printf "$( eval_gettext "Invalid file system '%s'. Please restart the installer." )" "$HD_PART_ROOT_FS_v" )"

                    guitool error "$message_wrong_fs"
                fi

                el_error "FS for root partition not detected correctly: root fs: '$HD_PART_ROOT_FS_v' , root part '$HD_PART_ROOT_dev', migration mode name '$systemname'"

                exit_error

                ;;
        esac

    fi


    if [[ -z "$HD_PART_ROOT_FS_OPTS_v" ]] ; then
        HD_PART_ROOT_FS_OPTS_v="defaults"
    fi

    unset ignore_this_format

    # add partition in the ignore list
    el_array_member_add "$HD_PART_ROOT_dev" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")

    # - format partition }}}
    # encryption vars  {{{
    partition_crypttab_add_if_needed "$HD_PART_ROOT_dev"

    # }}}
    # fix first-partition shift for grub2 {{{
    if [[ -n "$HD_DEV" ]] && [[ -b "${HD_DEV}1" ]] && ((is_disk_table_msdos)) && ! ((is_automated_partitioned)) ; then

        # our first partition is in the first cylinder, we will be unable to install grub
        if LC_ALL="$EL_LC_EN" fdisk -l "$HD_DEV" 2>/dev/null | grep "^/dev/" | tr '*' ' ' | awk '{print $2}' | grep -qsE "^1$" ; then

            # and first partition is swap:
            if EL_DEBUG=0 partitions-list --show-only="${HD_DEV}1" | grep -qs ":swap::swap:" && echo "$HD_DEV" | grep -qsE "((xvd|sd|hd|vd)[a-z]|nvme[[:digit:]]+n[[:digit:]]+|mmcblk[[:digit:]]+|md[[:digit:]]+)" ; then

                swapoff "${HD_DEV}1" 2>/dev/null || true

                if LC_ALL=C swapon --noheadings --raw | grep -qs "^${HD_DEV}1" ; then
                    el_error "unable to mount first swap partition: ${HD_DEV}1"
                    guitool error "$( eval_gettext "Your swap partition cannot be unmounted. Please unmount it manually, then create the needed partition at the beginning of your disk as the first partition; it should be 8MB in size. Otherwise, GRUB cannot be installed in your MBR, and your system won't be able to boot. You can also let the installer automatically repartition your hard disk." )"
                    exit
                fi

                swap_uuid="$( blkid -s UUID "${HD_DEV}1" | sed -e 's|^.*UUID="||g' -e 's|".*$||g' )"
                swap_size_end="$( LC_ALL="$EL_LC_EN" parted "${HD_DEV}1" print | grep linux-swap | awk '{print $3}' | grep -E "[[:digit:]]*[[:alpha:]]" | tail -1 )"

                # re-create swap partition with a shifted space
                if [[ -n "$swap_size_end" ]] && [[ -n "$swap_uuid" ]] ; then
                    parted -ms -a optimal "$HD_DEV" rm 1
                    parted -ms -a optimal "$HD_DEV" mkpart primary linux-swap 8MB $swap_size_end
                    sync ; sleep 1
                    mkswap -U $swap_uuid "${HD_DEV}1" 1>/dev/null
                    el_explain 0 "Swap partition in ${HD_DEV}1 has been shifted with 8 preceding Mb, needed for install the new grub later"
                    is_upgrade_shifted_first_part=1
                fi
            else
                el_warning "required to delete and create again the first partition, $FUNCNAME"
                guitool error "$( eval_gettext "Delete the first partition and create it again, selecting the option to align to MiB. Otherwise, your system will be unable to boot. This message appears because your first partition starts on the first cylinder. The new Grub2 requires some free space before the beginning of the first partition for correct installation. Without this, your system cannot boot." )"
                if ! guitool question "$( eval_gettext "Would you like to proceed with the installation now?" )" ; then
                    exit
                fi
            fi
        fi
    fi


    # - fix first-partition shift }}}
}
#########################################################################
swap_select(){
    # pre {{{
    # TODO: can we hibernate with a file instead of a swap partition?
    functions_pre
    local entry part menu answer is_swap_entries_enough disk buf

    # - pre }}}
    # auto add existing swap partitions {{{
    guitool "working_start"

    # add them to our used array in order to exclude the results
    while read -ru 3 entry
    do
        if ! [[ -b "$entry" ]] ; then
            continue
        fi

        # enough? we don't want to add swaps from other systems unless we don't have any one
        #if ((is_swap_entries_enough)) ; then
            #el_warning "swap_select: we had already enough swap partitions added, we should stop here? TODO break here"
            ##break
        #fi

        # add it if encrypted but comes from the same encrypted layer
        if check_part_is_from_encrypted "$entry" ; then

            if [[ -n "$( get_encrypted_partition_from_partition "$entry" )" ]] && [[ "$( get_encrypted_partition_from_partition "$entry" )" = "$( get_encrypted_partition_from_partition "$HD_PART_ROOT_dev" )" ]] ; then

                el_debug "adding swap partition '$entry' because comes from the same encrypted layer (disk is $HD_DEV)"

                # always reformat swap partitions because sometimes they don't work:
                uuid="$( hd_uuid_encode "$entry" )"
                [[ -z "$uuid" ]] && uuid="$( uuidgen )"
                swapoff "$entry" 2>/dev/null || true
                mkswap -q -f -U "$uuid" "$entry"  2>/dev/null || true
                sync ; sleep 1

                echo -e "$( hd_uuid_encode_fstab "${entry}" )  none        swap        sw        0    0" >> "$TMP_FSTAB_EXTRA_f"
                el_array_member_add "$entry" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")

                SWAPDEVICES="$entry $SWAPDEVICES"
                read -r SWAPDEVICES <<< "$SWAPDEVICES"

                is_swap_entries_enough=1
            else
                el_debug "ignoring Swap partition '$entry' because it comes from an encrypted layer that is not ours"
            fi

        else
            # comes from the same disk, so we should not want to have more partitions
            if [[ -b "$HD_DEV" ]] && [[ -b "$entry" ]] && echo "$entry" | grep -qs "$HD_DEV" ; then
                # add normal swap detected partitions
                el_debug "adding swap partition '$entry' (automatically detected, from a non-encrypted layer)"

                # always reformat swap partitions because sometimes they don't work:
                uuid="$( hd_uuid_encode "$entry" )"
                [[ -z "$uuid" ]] && uuid="$( uuidgen )"
                swapoff "$entry" 2>/dev/null || true
                el_debug "reformating swap partition '$entry' with uuid '$uuid'"
                mkswap -q -f -U "$uuid" "$entry"  2>/dev/null || true
                sync ; sleep 1

                echo -e "$( hd_uuid_encode_fstab "${entry}" )  none        swap        sw        0    0" >> "$TMP_FSTAB_EXTRA_f"
                el_array_member_add "$entry" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")

                SWAPDEVICES="$entry $SWAPDEVICES"
                read -r SWAPDEVICES <<< "$SWAPDEVICES"

                el_debug "swap_select: we have added the swap partition from the same disk where Elive is installed, mark that we don't want more: $entry | $HD_DEV"
                is_swap_entries_enough=1
            fi
        fi

    done 3<<< "$( EL_DEBUG=0 partitions-list --show-all | awk -v FS="::" '{if ( $3 == "swap" ) print $1}' || true )"

    # upgrade mode doesnt require to select swap interactively, automated mode: no swap possible, too small hd
    if ((is_autoupgrade)) || [[ "$part_size_swap" = "no_swap" ]] || ((is_migration)) ; then
        return 0
    fi


    # - auto add existing swap partitions }}}
    # request swap partition if we dont have {{{
    # only show dialog if we have not found existing swap partitions, so it is suggested to have one
    if [[ -n "$SWAPDEVICES" ]] ; then
        true
        #if ! ((ignore_show_suggestions)) && ((conf_configure_options)) && ! ((is_migration)) && ! ((is_automated_partitioned)) ; then
            #local message_swap_auto_added
            #message_swap_auto_added="$( printf "$( eval_gettext "Swap partitions automatically included to the install: %s" )" "$SWAPDEVICES" )"
            #guitool info "$message_swap_auto_added" || true &
        #fi
    else
        # suggest how much swap and why {{{

        if [[ "$RAM_TOTAL_SIZE_mb" -lt 1200 ]] ; then
            local message_swap_small_ram
            message_swap_small_ram="$( printf "$( eval_gettext "You have less than 2 GB of RAM. it is recommended to create a SWAP partition, also necessary for full RAM or hibernation. Suggested size is slightly larger than RAM. Select automated partitioning for optimal results if unsure." )" )"
            guitool info "$message_swap_small_ram"
        else
            # FIXME: this is an info message, which leads to another question... we want just one, "you dont have a swap, add one? (list to add)", thats all!
            local message_swap_needed_for
            message_swap_needed_for="$( printf "$( eval_gettext "There is no swap partition on your hard disk. Create one if you want to use hibernation. We suggest using the installer's automated partitioning for optimal results." )" )"

            guitool info "$message_swap_needed_for"
        fi


        # - suggest how much swap and why }}}
        # list partitions and ask which one to use {{{
        unset menu
        guitool "working_start"
        while read -ru 3 entry
        do
            part="$( echo "$entry" | awk -v FS="::" '{print $1}' )"
            if ! [[ -b "$part" ]] ; then
                el_debug "Skipping $part for swap possibilities"
                continue
            fi
            # update: nothing wrong by usings disks, and also: if the partition is not formated it doesnt has UUID
            #if ! LC_ALL="$EL_LC_EN" udevadm info --query=all --name="$part" 2>&1 | grep -qs "ID_FS_UUID=" ; then
                #el_debug "ignoring $entry because is a disk"
                #continue
            #fi

            if ! el_array_member_check "$part" "${HD_PARTITIONS_USED_array[@]}" && ! el_array_member_check "$part" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; then
                # create an array for use in the menu, with one per line
                menu+=("$( echo "$entry" | awk -v FS="::" '{print $1}' )")

                # get the name of the OS for the menu in priority
                if [[ -s /tmp/os-probed ]] && grep -qs "^$( echo "$entry" | awk -v FS="::" '{print $1}' ):" /tmp/os-probed ; then

                    menu+=("$( awk -v FS=":" -v m_part="$(echo "$entry" | awk -v FS="::" '{print $1}' )"  '{if ($1 == m_part ) print $2}' /tmp/os-probed ) - $( echo "$entry" | awk -v FS="::" '{print "("$3")"}' )")

                else

                    # fix names for the menus:
                    entry="$( echo "$entry" | awk -v FS="::" -v OFS="::" '{if ( $2 == "extended" ) $5 = "(extended)" ; if ( $2 == "disk" ) $5 = "(disk)" ; if ( $3 == "crypto_LUKS" ) $5 = "(crypto_LUKS)"  ; print $0}' )"

                    case "$( echo "$entry" | awk -v FS="::" '{print $5}' )" in
                        "(no name)")
                            menu+=("$( echo "$entry" | awk -v FS="::" '{print "("$3")"}' )")
                            ;;
                        "(swap)")
                            menu+=("$( echo "(Swap partition)" )")
                            ;;
                        "(extended)")
                            menu+=("$( eval_gettext "(Extended partition)" )")
                            ;;
                        "(disk)")
                            menu+=("$( eval_gettext "(Entire disk)" )")
                            ;;
                        "(unformatted)")
                            menu+=("$( eval_gettext "(Unformatted)" )")
                            ;;
                        "(crypto_LUKS)")
                            menu+=("$( eval_gettext "(LUKS Encryption layer)" )")
                            ;;
                        *)
                            menu+=("$( echo "$entry" | awk -v FS="::" '{print $5" - ("$3")"}' )")
                            ;;
                    esac
                fi

                # include size
                size="$( echo "$entry" | awk -v FS="::" '{print $7}' )"
                if [[ -z "$size" ]] || [[ "$size" = "uknown" ]] ; then
                    size="$( eval_gettext "unknown" )"
                fi
                menu+=("$( echo "$size" )")

                # do not use el_ arrays, they uses a different algorithm that doesnt allow duplcates
            fi
        done 3<<< "$( EL_DEBUG=0 partitions-list --show-all | grep -viE -- "::(extended|crypto_LUKS|persistence|udf|iso9660|squashfs|EFI|Elive_)::" || true )"

        if [[ -z "${menu[*]}" ]] ; then
            el_debug "no partitions found that can be used for swap: $(EL_DEBUG=0 partitions-list --show-all --show-raw || true )"
            return 0
        fi

        # menu is generated, show it in the GUI
        #guitool "working_stop"
        answer="$( $guitool --title='Elive Installer' --list --height=300 --width=500 --text="$( eval_gettext "Choose a SWAP partition\n\nor press Cancel, if you do not want a SWAP partition." )" --column="$( eval_gettext "Device" )" --column="$( eval_gettext "Name" )" --column="$( eval_gettext "Size" )" "${menu[@]}" 2>/dev/null || echo cancel )"

        # check
        if ! [[ -b "$answer" ]] ; then
            if [[ "${answer}" != "cancel" ]] && [[ -n "${answer}" ]] ; then
                el_error "selection '$answer' is not valid for a swap partition"
            fi
            return 0
        fi

        guitool "working_start"

        # - list partitions and ask which one to use }}}
        # confirm and create {{{

        local message_erase
        message_erase="$( printf "$( eval_gettext "About to ERASE your partition '%s' for '%s'" )" "${answer}" "Swap" )"

        if guitool question "$message_erase" ; then
            disk="$( get_disk_from_partition_complex "$answer" )"

            if [[ -b "$disk" ]] && [[ -b "$answer" ]] ; then
                el_explain 0 "Selected swap partition is '$answer', disk is '$disk' "
                el_umount "$answer" 1>/dev/null

                # pre-format first
                mkswap -f "$answer" 1>/dev/null 2>&1 || true

                header_type="$( LC_ALL="$EL_LC_EN" parted -ms -a optimal -l "$disk" 2>/dev/null | awk -v disk="$disk" -v FS=":" '{if ($1 == disk) print $6}' )"

                buf="$( get_partition_number_from_partition "${answer}" )"
                if [[ -b "$buf" ]] ; then
                    case "$header_type" in
                        gpt)
                            if gdisk -l "$disk" | grep -E "(GPT: present)" ; then
                                el_debug "Setting the type of partition 82 (swap) for '$answer' (GPT mode)"
                                echo -e "t\n${buf}\n8200\nw\nY\nq" | LC_ALL="$EL_LC_EN" timeout 40 gdisk "$disk"
                            else
                                el_error "GPT not present in '$disk' ?\n$( gdisk -l "$disk" 2>&1 )"
                            fi
                            ;;
                        msdos)
                            el_debug "Setting the type of partition 82 (swap) for '$answer' (MBR/msdos mode)"
                            sfdisk --force -c "$disk ${buf}" 82  2>/dev/null || true
                            ;;
                        *)
                            el_error "unknown header type: $header_type\nparted -ms -a optimal -l '$disk' 2>&1"
                            ;;
                    esac
                fi

                # TODO: move to partition_format function
                if ! mkswap -f "$answer" 1>/dev/null 2>&1 ; then
                    sync ; sleep 1
                    el_warning "unable to make swap partition on '$answer':"
                    mkswap -f "$answer" || true
                    return
                fi
                # re-read partition table
                device_check_is_busy "$disk"

                # Check alignment
                partition_alignment_check "$answer" report


                SWAPDEVICES="$answer"
                el_debug "SWAPDEVICES = $SWAPDEVICES"
                echo -e "$( hd_uuid_encode_fstab "${answer}" )  none        swap        sw    0    0" >> "$TMP_FSTAB_EXTRA_f"
                # TODO: move to partition_format function (and remove this next line too)
                el_array_member_add "$answer" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")
            fi
        fi

        # - confirm and create }}}
    fi
    # - request swap partition if we don't have }}}
}
#########################################################################

partition_select_other_ask(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    local answer result


    # - pre }}}
    # auto partitioned, needed {{{
    if ((is_partitioning_auto)) ; then

        # FIXME: all the partitions should be mounted in pre_copying instead
        if [[ -n "$HD_PART_HOME_dev" ]] ; then
            # we only have a possible home partition when auto parted
            mkdir -p "$TARGET/home"
            el_umount "$HD_PART_HOME_dev" 1>/dev/null

            # FIXME: all the partitions should be mounted in pre_copying instead
            mount -o dev "${HD_PART_HOME_dev}" "$TARGET/home"
        fi

        # return because we are auto-partitioned
        return 0
    fi
    # - auto partitioned }}}
    # we don't want to ask for partitions if we are migration or autoupgrade {{{
    if ((is_migration)) || ((is_autoupgrade))  ; then
        # FIXME: so what happens with migrations and autoupgrades? we dont include mounted dependencies ? this sounds like a little rewrite for 4.0
        # FIXME: is /boot included too? we will see...
        return 0
    fi

    el_debug "Selection of other partitions ?"
    # }}}
    # ask about if add extra partitions, mount them, and loop

    trap_signal_error_debug "disable"

    # see if we need a /boot for this partition structure
    if { echo "$HD_PART_ROOT_dev" | grep -qs "/dev/mapper/" && ! ((is_migration)) && ! [[ -b "$HD_PART_BOOT_dev" ]] ; } || { ((needs_boot_partition)) && { ! [[ -b "$HD_PART_BOOT_dev" ]] && ! [[ -b "$OLD_HD_PART_BOOT_dev" ]] ; } ; } ; then

        $yad --question --image=gtk-dialog-question --text="$( eval_gettext "Your selected root partition requires an additional /boot partition. Do you want to add it now?" )" --button="$( eval_gettext "Yes" ):100" --button="$( eval_gettext "No" ):101" --button="$( eval_gettext "Cancel installation" ):102"
        result=$?
        case "$result" in
            100)
                needs_boot_partition=1
                partitions_select_other_list_ask "/boot"
                ;;
            101)  true ; ;;
            102)
                guitool warning "$( eval_gettext "You exited the installer." )"
                exit
                ;; # this exits the installer
        esac
    else

        local message_extra_partition_ask
        if [[ -n "$HD_PART_HOME_dev" ]] ; then
            message_extra_partition_ask="$( printf "$( eval_gettext "Would you like to add another partition?\n\nThis is optional and not required for a functional Elive installation." )" "" )"
        else
            message_extra_partition_ask="$( printf "$( eval_gettext "Would you like to add another partition, such as for /home?\n\nThis is optional, and Elive functions well without it." )" "" )"
        fi


        $yad --question --image=gtk-dialog-question --text="$message_extra_partition_ask" --button="$( eval_gettext "Yes" ):100" --button="$( eval_gettext "No" ):101" --button="$( eval_gettext "Cancel installation" ):102"
        result=$?
        case "$result" in
            100)
                partitions_select_other_list_ask
                ;;
            101)  true ; ;;
            102)  exit ; ;; # this exits the installer
        esac
    fi

    trap_signal_error_debug "enable"
}
#########################################################################
partitions_select_other_list_ask(){
    # ask details and mount into target
    functions_pre
    el_debug "\targs '$*'"
    local answer answer_given menu entry part partition_extra_dev partition_extra_dir partition_extra_fs ignore_this_format _point
    answer_given="$1"

    # Select mountpoint {{{
    if [[ -n "$answer_given" ]] ; then
        # if we pass directly the name of the mountpoint that we want/need to add...
        answer="$answer_given"

    else
        if [[ -n "$HD_PART_HOME_dev" ]] || [[ -n "$OLD_HD_PART_HOME_dev" ]] ; then

            el_debug "/home partition found"
            answer="$( $guitool --title='Elive Installer' --width=280 --height=340  --list --text="$( eval_gettext "Choose an additional mount point" )" --column="$( eval_gettext "Type" )" "/boot" "/var" "/usr" "/usr/local" "/opt" "/srv" "EFI" "BIOS" 2>/dev/null || echo cancel )"
        else

            el_debug "/home partition NOT found"
            answer="$( $guitool --title='Elive Installer' --width=280 --height=360 --list --text="$( eval_gettext "Choose an additional mount point" )" --column="$( eval_gettext "Type" )" "/home" "/boot" "/var" "/usr" "/usr/local" "/opt" "/srv" "EFI" "BIOS" 2>/dev/null || echo cancel )"
        fi
    fi

    guitool "working_start"

    [[ "$answer" = "cancel" ]] && return 0
    partition_extra_dir="$answer"
    el_debug "selected '$partition_extra_dir' extra mountpoint"

    # - Select mountpoint }}}
    # tell the user that can abort in any moment {{{
    if ! ((is_user_notified_can_abort_repartition)) ; then
        guitool info "$( eval_gettext "Help tip: If you wish to create a new partition on your hard disk, you must abort the installation on the next screen and start again. The initial steps of the installer will provide the chance to use the partitioner at the appropriate time to modify your disks." )"
        is_user_notified_can_abort_repartition=1
    fi
    # - tell the user that can abort in any moment }}}
    # list partitions and ask which one to use {{{
    local message_partition_extra_dir
    message_partition_extra_dir="$( printf "$( eval_gettext "Choose a partition to use for %s" )" "${partition_extra_dir}" )"


    unset menu
    while read -ru 3 entry
    do
        part="$( echo "$entry" | awk -v FS="::" '{print $1}' )"
        if ! [[ -b "$part" ]] ; then
            el_debug "skipping $part as a partition candidate"
            continue
        fi

        if el_array_member_check "$part" "${HD_PARTITIONS_USED_array[@]}" ||  el_array_member_check "$part" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; then
            el_debug "$part already in our array of partitions used, skipping it"
            el_debug "partitions used array is: ${HD_PARTITIONS_USED_array[*]}"
        else
            # Include device name
            menu+=("$( echo "$entry" | awk -v FS="::" '{print $1}' )")

            # get the name of the OS for the menu in priority
            if [[ -s /tmp/os-probed ]] && grep -qs "^$( echo "$entry" | awk -v FS="::" '{print $1}' ):" /tmp/os-probed ; then

                menu+=("$( awk -v FS=":" -v m_part="$(echo "$entry" | awk -v FS="::" '{print $1}' )"  '{if ($1 == m_part ) print $2}' /tmp/os-probed ) - $( echo "$entry" | awk -v FS="::" '{print "("$3")"}' )")

            else

                # fix names for the menus:
                entry="$( echo "$entry" | awk -v FS="::" -v OFS="::" '{if ( $2 == "extended" ) $5 = "(extended)" ; if ( $2 == "disk" ) $5 = "(disk)" ; if ( $3 == "crypto_LUKS" ) $5 = "(crypto_LUKS)"  ; print $0}' )"

                case "$( echo "$entry" | awk -v FS="::" '{print $5}' )" in
                    "(no name)")
                        menu+=("$( echo "$entry" | awk -v FS="::" '{print "("$3")"}' )")
                        ;;
                    "(swap)")
                        menu+=("$( echo "(Swap partition)" )")
                        ;;
                    "(extended)")
                        menu+=("$( eval_gettext "(Extended partition)" )")
                        ;;
                    "(disk)")
                        menu+=("$( eval_gettext "(Entire disk)" )")
                        ;;
                    "(unformatted)")
                        menu+=("$( eval_gettext "(Unformatted)" )")
                        ;;
                    "(crypto_LUKS)")
                        menu+=("$( eval_gettext "(LUKS Encryption layer)" )")
                        ;;
                    *)
                        menu+=("$( echo "$entry" | awk -v FS="::" '{print $5" - ("$3")"}' )")
                        ;;
                esac
            fi

            # Include size
            size="$( echo "$entry" | awk -v FS="::" '{print $7}' )"
            if [[ -z "$size" ]] || [[ "$size" = "uknown" ]] ; then
                size="$( eval_gettext "unknown" )"
            fi
            menu+=("$( echo "$size" )")

            # do not use el_ arrays, they uses a different algorithm that doesnt allow duplcates
        fi
    done 3<<< "$( EL_DEBUG=0 partitions-list --show-all | awk -v FS="::" '{if ($3 != "swap") print $0}' | grep -viE -- "::(extended|crypto_LUKS|persistence|udf|iso9660|squashfs|EFI|Elive_)::"  || true )"

    menu+=("(ABORT)")
    menu+=("exit")
    menu+=("$( eval_gettext "Abort the installation" )")

    # menu is generated, show it in the GUI
    #if [[ "$EL_DEBUG" -gt 2 ]] ; then
        #printf 'menu is:\n%s\n' "${menu[@]}" 1>&2
    #fi
    #guitool "working_stop"
    answer="$( $guitool --title='Elive Installer' --list --height=340 --width=600 --cancel-label="$( eval_gettext "Ignore partition and continue" )" --ok-label="$( eval_gettext "Select" )" --text="$message_partition_extra_dir" --column="$( eval_gettext "Device" )" --column="$( eval_gettext "Name" )" --column="$( eval_gettext "Size" )" "${menu[@]}" 2>/dev/null || echo cancel )"

    # check {{{
    [[ "$answer" = *"ABORT"* ]] && exit
    [[ "$answer" = "cancel" ]] && return 0

    if ! [[ -b "$answer" ]] ; then
        if [[ -n "$answer" ]] ; then
            el_error "You selected $answer which is not a block device"
        fi

        return 0
    fi

    verify_result "$answer"
    el_explain 0 "Selected extra partition is $answer"

    guitool "working_start"
    # }}}

    partition_extra_dev="$answer"

    if [[ "$partition_extra_dev" = "/dev/mapper/"* ]] ; then
        is_fs_mapper=1
    fi

    # Check alignment
    partition_alignment_check "$answer" report


    # - list partitions and ask which one to use }}}

    # specific info cases {{{
    if [[ ${partition_extra_dir} = "/var" ]] ; then
        HD_PART_VAR_dev="${partition_extra_dev}"
    fi
    # already existing data in EFI partition? do not format it
    if [[ ${partition_extra_dir} = "EFI" ]] ; then
        if [[ "$( hdtools_get_fs "$partition_extra_dev" )" = "vfat" ]] ; then
            EL_DEBUG=0 mounts-manager --mount "$partition_extra_dev" 1>/dev/null
            _point="$( EL_DEBUG=0 mounts-manager --get-mountpoint "$partition_extra_dev" )"
            if [[ -d "$_point/efi" ]] || [[ -d "$_point/EFI" ]] ; then
                ignore_this_format=1
            fi
            el_umount "$partition_extra_dev" 1>/dev/null
        fi
    fi

    # - specific info }}}
    # format and mount {{{
    el_umount "$partition_extra_dev" 1>/dev/null

    local message_erase
    message_erase="$( printf "$( eval_gettext "About to ERASE your partition '%s' for '%s'" )" "${partition_extra_dev}" "$partition_extra_dir" )"

    # going to format, are you sure?
    if ! ((ignore_format_disks)) && ! ((ignore_this_format)) ; then

        if guitool question "$message_erase" ; then
            true
        else
            #el_warning "User not wanted to format extra partition $partition_extra_dev as $partition_extra_dir ; is this expected? mount problems? deleteme warning"

            # if /home then warn about problems
            if [[ "$partition_extra_dir" = "/home" ]] ; then

                if guitool yesnoquestion "$( eval_gettext "If you are using this user partition with another OS, Elive can import existing users into your new Elive system. You need to provide the name of the root partition of that OS so it can fetch and recreate the users' settings in a compatible way. Do you want to proceed? If not, your home partition will be added, and you will need to manually add users to the system with their correct UIDs and other data." )" ; then

                    # ask which root partition is referenced by the other system {{{

                    guitool info "$( eval_gettext "Important: Using this user partition with another operating system is not recommended due to differing settings and application versions. This can lead to incompatibility issues, causing malfunctions or inability to launch certain applications. Understanding this, you can identify the cause if such issues arise." )"
                    # TODO: to make the process simpler, we could eventually scan the found partitions in order to see if we have the needed files and in such case, if only one, just use it, otherwise, only list the found matches

                    local message_partition_extra_shared_root
                    message_partition_extra_shared_root="$( printf "$( eval_gettext "Choose the root partition of your other system" )" "" )"

                    # select partition {{{
                    unset menu
                    while read -ru 3 entry
                    do
                        part="$( echo "$entry" | awk -v FS="::" '{print $1}' )"
                        if ! [[ -b "$part" ]] ; then
                            el_debug "skipping $part as a partition candidate"
                            continue
                        fi

                        if ! el_array_member_check "$part" "${HD_PARTITIONS_USED_array[@]}" && ! el_array_member_check "$part" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; then
                            # create an array for use in the menu, with one per line
                            menu+=("$( echo "$entry" | awk -v FS="::" '{print $1}' )")

                            # get the name of the OS for the menu in priority
                            if [[ -s /tmp/os-probed ]] && grep -qs "^$( echo "$entry" | awk -v FS="::" '{print $1}' ):" /tmp/os-probed ; then

                                menu+=("$( awk -v FS=":" -v m_part="$(echo "$entry" | awk -v FS="::" '{print $1}' )"  '{if ($1 == m_part ) print $2}' /tmp/os-probed ) - $( echo "$entry" | awk -v FS="::" '{print "("$3")"}' )")

                            else

                                # fix names for the menus:
                                entry="$( echo "$entry" | awk -v FS="::" -v OFS="::" '{if ( $2 == "extended" ) $5 = "(extended)" ; if ( $2 == "disk" ) $5 = "(disk)" ; if ( $3 == "crypto_LUKS" ) $5 = "(crypto_LUKS)"  ; print $0}' )"

                                case "$( echo "$entry" | awk -v FS="::" '{print $5}' )" in
                                    "(no name)")
                                        menu+=("$( echo "$entry" | awk -v FS="::" '{print "("$3")"}' )")
                                        ;;
                                    "(swap)")
                                        menu+=("$( echo "(Swap partition)" )")
                                        ;;
                                    "(extended)")
                                        menu+=("$( eval_gettext "(Extended partition)" )")
                                        ;;
                                    "(disk)")
                                        menu+=("$( eval_gettext "(Entire disk)" )")
                                        ;;
                                    "(unformatted)")
                                        menu+=("$( eval_gettext "(Unformatted)" )")
                                        ;;
                                    "(crypto_LUKS)")
                                        menu+=("$( eval_gettext "(LUKS Encryption layer)" )")
                                        ;;
                                    *)
                                        menu+=("$( echo "$entry" | awk -v FS="::" '{print $5" - ("$3")"}' )")
                                        ;;
                                esac
                            fi

                            # Include size
                            size="$( echo "$entry" | awk -v FS="::" '{print $7}' )"
                            if [[ -z "$size" ]] || [[ "$size" = "uknown" ]] ; then
                                size="$( eval_gettext "unknown" )"
                            fi
                            menu+=("$( echo "$size" )")

                            # do not use el_ arrays, they uses a different algorithm that doesnt allow duplcates
                        fi
                    done 3<<< "$( EL_DEBUG=0 partitions-list --show-all --show-raw | grep -viE -- "::(extended|crypto_LUKS|persistence|udf|iso9660|squashfs|iso9660|EFI|Elive_|swap)::" || true )"

                    # menu is generated, show it in the GUI
                    #guitool "working_stop"
                    answer="$( $guitool --title='Elive Installer' --list --height=340 --width=600 --text="$message_partition_extra_shared_root" --column="$( eval_gettext "Device" )" --column="$( eval_gettext "Name" )" --column="$( eval_gettext "Size" )" "${menu[@]}" 2>/dev/null || echo cancel )"

                    # }}}
                    # check {{{
                    [[ "$answer" = "cancel" ]] && return 0

                    if ! [[ -b "$answer" ]] ; then
                        if [[ -n "$answer" ]] ; then
                            el_error "You selected $answer which is not a block device"
                        fi

                        return 0
                    fi

                    verify_result "$answer"

                    guitool "working_start"
                    # }}}

                    if [[ -b "$answer" ]] ; then
                        partition_shared_home_referenced_root_partition="$answer"
                        el_debug "selected root partition for the shared home system (where to import users) is $partition_shared_home_referenced_root_partition"
                        ignore_this_format=1
                        excluded_dirs_for_rsync="$excluded_dirs_for_rsync /home"
                        is_shared_home_partition=1
                        partition_extra_fs="$( hdtools_get_fs "$partition_extra_dev" )"
                        # set a very old version of elive so we can run all the migration triggers
                        old_eliveversion="1.0.0"

                        el_explain 0 "Selected __shared home__ mode: partition is '${partition_extra_dev}' with fs '${partition_extra_fs}'"

                    else
                        # user didnt give the relative root partition (maybe not exists?) but he wanted to share a home partition, since we dont know what to do we exit
                        if guitool question "$( eval_gettext "To import existing users into your new Elive system, the installer requires the root partition of the other operating system to access user configurations. Canceling will exit the installer, allowing you to restart with different options. Continuing will add your user partition, but you'll need to manually re-add users with correct UID, GID, and other data for proper functionality." )" ; then
                            # not format /home, and this /home is not shared with other systems
                            ignore_this_format=1
                            excluded_dirs_for_rsync="$excluded_dirs_for_rsync /home"
                            is_shared_home_partition=1
                            partition_extra_fs="$( hdtools_get_fs "$partition_extra_dev" )"
                            # set a very old version of elive so we can run all the migration triggers
                            old_eliveversion="1.0.0"

                            el_explain 0 "Selected __import existing home__ mode: partition is '${partition_extra_dev}' with fs '${partition_extra_fs}'"
                        else
                            exit
                        fi
                    fi

                    # - list partitions and ask which one to use }}}
                else
                    # not format /home, and this /home is not shared with other systems
                    ignore_this_format=1
                    excluded_dirs_for_rsync="$excluded_dirs_for_rsync /home"
                    is_shared_home_partition=1
                    partition_extra_fs="$( hdtools_get_fs "$partition_extra_dev" )"
                    # set a very old version of elive so we can run all the migration triggers
                    old_eliveversion="1.0.0"

                    el_explain 0 "Selected __import existing home__ mode: partition is '${partition_extra_dev}' with fs '${partition_extra_fs}'"
                fi
            else

                # if is not a /home mountpoint
                if guitool yesnoquestion "$( eval_gettext "You chose not to format the partition. This should only be done if you know what you are doing; the installer expects it to be formatted for everything to work correctly. Continue without formatting?" )" ; then
                    ignore_this_format=1

                    if [[ "$partition_extra_dir" != "EFI" ]] ; then
                        el_warning "User confirmed that doesn't wants to format the extra partition $partition_extra_dev as $partition_extra_dir ; is this expected? mount problems? deleteme warning"
                    fi
                else
                    exit
                fi
            fi
        fi
    fi


    # ask which filesystem to use {{{
    if [[ -n "$partition_extra_fs" ]] ; then
        el_explain 0 "preselected FS: $partition_extra_fs"
    else

        if [[ ${partition_extra_dir} = "/boot" ]] ; then
            # best option, no need to ask
            partition_extra_fs="ext4"
            HD_PART_BOOT_dev="${partition_extra_dev}"
        fi

        if [[ ${partition_extra_dir} = "EFI" ]] || [[ ${partition_extra_dir} = "BIOS" ]] ; then
            # best option, no need to ask
            partition_extra_fs="$partition_extra_dir"
        fi

        # select if we don't have any assigned
        if [[ -z "$partition_extra_fs" ]] ; then
            partition_extra_fs="$( select_filesystem "${partition_extra_dev}" )"
            verify_result "$partition_extra_fs"
        fi

    fi


    # - ask for filesystem to use }}}

    guitool "working_start"

    # just set a predefined default
    partition_extra_fs_opts_v="defaults"

    if ((ignore_this_format)) || ((ignore_format_disks)) ; then
        partition_extra_fs="$( hdtools_get_fs "$partition_extra_dev" )"
    fi

    case "${partition_extra_fs}" in
        reiserfs)
            if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"
            fi
            partition_extra_fs_opts_v="defaults,commit=300,relatime,notail"
            is_reiserfs_used=1
            ;;
        ext4)
            if device_is_ssd "$partition_extra_dev" ; then
                if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"
                fi

                partition_extra_fs_opts_v="defaults,commit=300,relatime"
            else
                if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                    partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"
                fi

                partition_extra_fs_opts_v="defaults,commit=300,relatime"
            fi
            ;;
        ext3)
            if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                if device_is_ssd "$partition_extra_dev" ; then
                    el_warning "Your partition is an SSD, using ext4 instead because includes special features for it"
                    partition_extra_fs="ext4"

                    partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"
                    partition_extra_fs_opts_v="defaults,commit=300,relatime"
                else
                    partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"

                    partition_extra_fs_opts_v="defaults,commit=300,relatime"
                fi
            else
                # no format
                partition_extra_fs_opts_v="defaults,commit=300,relatime"
            fi
            ;;
        ext2)
            if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"
            fi
            ;;
        xfs)
            if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"
            fi
            ;;
        jfs)
            if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"
            fi
            ;;
        reiser4)
            if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"
            fi

            if device_is_ssd "$partition_extra_dev" ; then
                partition_extra_fs_opts_v="defaults,relatime,txmod=wa"
            else
                partition_extra_fs_opts_v="defaults,relatime"
            fi
            el_array_member_add "$partition_extra_fs" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")
            ;;
        reiser5)
            el_error "reiser5 needs to implement"
            exit_error
            ;;
        btrfs)
            if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                partition_format "${partition_extra_dev}" "${partition_extra_fs}" "Elive_${partition_extra_dir##*/}"

            fi
            if ((is_fs_compression_wanted)) ; then
                partition_extra_fs_opts_v="compress=${compressor_algorithm},commit=300,lazytime"
            else
                partition_extra_fs_opts_v="commit=300,lazytime"
            fi
            el_array_member_add "$partition_extra_fs" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")
            ;;
        EFI|BIOS)
            if ! ((ignore_this_format)) && ! ((ignore_format_disks)) ; then
                partition_format "${partition_extra_dev}" "${partition_extra_fs}" "${partition_extra_dir}"
            fi
            # assign dev:
            if [[ "$partition_extra_fs" = "EFI" ]] ; then
                HD_PART_EFI_dev="$partition_extra_dev"
                HD_PART_EFI_FS_v="vfat"
            fi

            ;;
        *)
            local message_wrong_fs
            message_wrong_fs="$( printf "$( eval_gettext "Invalid file system %s. If this is an upgrade of your existing Elive system (not a clean install), this indicates a bug and should be reported. For a clean installation, restart the Elive installer and select the appropriate file system." )" "$partition_extra_fs" )"

            el_error "wrong FS for '$partition_extra_dev':  $partition_extra_fs"
            guitool error "$message_wrong_fs"
            exit_error
            ;;
    esac


    # set defaults
    if [[ "$partition_extra_fs" != "BIOS" ]] && [[ "$partition_extra_fs" != "EFI" ]] ; then
        if [[ -z "$partition_extra_fs_opts_v" ]] ; then
            partition_extra_fs_opts_v="defaults"
        fi

        el_debug "disabling time-based fsck's for $partition_extra_dev"


        #before is launched "pre_copying" for mount the root dir, and then, mount this extra partitions on it
        mkdir -p "$TARGET/usr/local" "$TARGET/boot" "$TARGET/home" "$TARGET/var" "$TARGET/usr" "$TARGET/opt" "$TARGET/srv"
        # needs to create, if not, not mount


        # FIXME: all the partitions should be mounted in pre_copying instead
        # try to mount it with the GIVEN options
        if ! mount -o dev -t "${partition_extra_fs}" -o "${partition_extra_fs_opts_v}" "${partition_extra_dev}" "$TARGET${partition_extra_dir}" ; then
            el_error "Unable to mount extra partition with the given options, please report it: "
            echo mount -o dev -t "${partition_extra_fs}" -o "${partition_extra_fs_opts_v}" "${partition_extra_dev}" "$TARGET${partition_extra_dir}" 1>&2

            el_warning "Unable to mount '$partition_extra_dev' partition with fs: '$partition_extra_fs' and options (-o $partition_extra_fs_opts_v ), device is: $(file -s "$partition_extra_dev" )"


            el_warning "fallbacking to normal mounting and removing default options"
            partition_extra_fs_opts_v="defaults"

            if ! mount -o dev -t "${partition_extra_fs}" "${partition_extra_dev}" "$TARGET${partition_extra_dir}" ; then
                el_error "Still unable to mount extra partition"
                guitool error "Unable to mount '$partition_extra_dev' partition with fs: '$partition_extra_fs' and options (-o $partition_extra_fs_opts_v ), device is: $(file -s "$partition_extra_dev" )"
                exit_error
            fi
        fi

        # do not mount /boot partitions, we will do that later
        if [[ "$partition_extra_dir" = "/boot" ]] ; then
            el_umount "$partition_extra_dev" 1>/dev/null 2>&1 || true
            HD_PART_BOOT_fs="$partition_extra_fs"
        fi

        if [[ ${partition_extra_dir} = "/home" ]] ; then
            HD_PART_HOME_FS_v="$partition_extra_fs"
        fi



        echo -e "$( hd_uuid_encode_fstab "${partition_extra_dev}" )  ${partition_extra_dir}     ${partition_extra_fs}    ${partition_extra_fs_opts_v}    0    2" >> "$TMP_FSTAB_EXTRA_f"
        el_debug "$( hd_uuid_encode_fstab "${partition_extra_dev}" )  ${partition_extra_dir}     ${partition_extra_fs}    ${partition_extra_fs_opts_v}    0    2"

        el_array_member_add "$partition_extra_dev" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")


        partition_crypttab_add_if_needed "$partition_extra_dev"

    fi

    # - format }}}
    partition_select_other_ask
}

#===  FUNCTION  ================================================================
#          NAME:  partition_required_ask
#   DESCRIPTION:  check for required partitions remaining for our install
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
partition_required_ask(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"

    # }}}

    # boot partition required and not added
    if ((needs_boot_partition)) && ! [[ -b "$HD_PART_BOOT_dev" ]] ; then
        guitool info "$( eval_gettext "Your selected setup requires an extra BOOT partition. Please add one; it should not be inside an encrypted layer, the recommended size for it is 300MB in size." )"
        partitions_select_other_list_ask "/boot"
    fi

    if ((is_efi_booted)) ; then
        efi_select

        if ! [[ -b "$HD_PART_EFI_dev" ]] ; then
            #if ((is_migration)) ; then
            guitool info "$( eval_gettext "Your selected setup requires an EFI partition. Please add one; it should not be inside an encrypted layer, the recommended size for it is 100MB." )"
            partitions_select_other_list_ask "EFI"
        fi
        # extra check
        #if ((is_disk_table_msdos)) ; then
            #el_warning "detected MSDOS disk table but booted in UEFI mode, this is not expected and should not happen"
        #fi
    else
        # booted in Legacy mode
        if ((is_disk_table_gpt)) ; then
            if [[ -b "$HD_DEV" ]] ; then
                if gdisk -l "$HD_DEV" | grep -qs "EF02" ; then
                    # all good, we have it
                    true
                else
                    if guitool yesnoquestion "$( eval_gettext "Your selected setup uses GPT and LEGACY BOOT. It is strongly recommended to add an extra partition to act as the BIOS MBR; it should not be inside an encrypted layer and should be 8MB in size. Do you want to add it now?" )" ; then
                        partitions_select_other_list_ask "BIOS"
                    fi
                fi
            else
                el_warning "No HD_DEV variable set in $FUNCNAME"
            fi
        fi
        # must install in MBR
        if ((is_disk_table_msdos)) ; then
            # all good
            # note: in msdos table shows as EF00:
            # note: do not run gdisk in a msdos table ever!
            true
        fi
    fi
}
#########################################################################
detect_vista_recovery(){ # not cleaned yet
    functions_pre
    # FIXME: wrong code
    cat /etc/fstab | while read line
    do
        unset dev mount
        dev="$( echo "$line" | awk '{if ($3 == "ntfs-3g") print $1}' )"
        mount="$( echo "$line" | awk '{if ($3 == "ntfs-3g") print $2}' )"
        if [[ -n "$mount" && -n "$dev" ]] ; then
            if ! mount | grep -q "^$dev" ; then
                # TODO: move to mounts-manager
                timeout 30 mount -t ntfs-3g "$dev" "$mount"
            fi

            if ls -1 "$mount" | grep -iq -- "WinREPartition.ini" ; then
                echo "$dev $mount"
            fi

        fi
    done
    unset dev mount

    # FIXME: hacer un detector (si no se ha detectado) que consista en mirar el tamaño de la particion de windows si es menor que X
    for label in $( find /dev/disk/by-label -type l )
    do
        if echo "$label" | grep -iqE "(recover|WinRE)" ; then
            label_dev="$( readlink -f "$label" )"
            label_dev="/dev/${label_dev##*/}"
            label_mountpoint="$( cat /etc/fstab | awk -v label_dev="$label_dev" '{if ($1 == label_dev) print $2}' | tail -1 )"
            if [[ -n "$label_dev" && -n "$label_mountpoint" ]] ; then
                echo "$label_dev" "$label_mountpoint"
            fi
        fi
    done
    unset label label_dev label_mountpoint
}

########################################################################
space_left_full(){
    functions_pre
    el_debug "\targs '$*'"
    local spaceleft

    if ((is_installation_canceled)) ; then
        return 0
    fi

    spaceleft="$( LC_ALL="$EL_LC_EN" df | awk '{ if($5 == "100%" || $5 == "99%" || $5 == "98%" ) {print $1 " " $6} ; } ' | grep "$TARGET" | awk '{print $1}' || true )"
    el_debug "space left is $spaceleft"

    if [[ -n "$spaceleft" ]] ; then
        local message_nospaceleft
        message_nospaceleft="$( printf "$( eval_gettext "No space left on %s. Use a larger partition or resize this one. The installation cannot continue and has been aborted." )" "$spaceleft" )"
        el_error "no space left (reported)"
        guitool error "$message_nospaceleft"

        # close stayd dialogs
        killall zenity 1>/dev/null 2>&1 || killall -9 zenity 1>/dev/null 2>&1
        killall yad 1>/dev/null 2>&1 || killall -9 yad 1>/dev/null 2>&1

        # free some space, or we may were not able to umount it
        rm -rf "$TARGET/usr/share/doc" "$TARGET/usr/share/man"

        sync

        exit_error
    else
        #el_error "error found in $FUNCNAME : $(df -m )"
        if ((is_reiserfs_used)) ; then
            guitool error "$( eval_gettext "An error occurred during the installation process. The likely cause is the use of ReiserFS, which is no longer maintained. Please use a different filesystem. Alternatively, your download might have been corrupted; try downloading Elive again and recording it to a different storage medium (USB or DVD)." )"
        else
            guitool error "$( eval_gettext "An error occurred during the installation process. The likely cause is a broken download or a faulty storage medium (USB or DVD) used for recording Elive. Please download Elive again and record it to a different medium. We also recommend trying the automated partitioning mode for optimal disk usage." )"
        fi

        el_error "error installing system, rsync failed, REPORTED to the user"

        df -m | grep -vE "^(rootfs|udev|tmpfs|/dev/loop|aufs)" | $guitool --title='Elive Installer' --text-info --width=500 --height=420 2>/dev/null
        exit_error
    fi

    copy_canceled_stopping
}
#########################################################################
space_left_verify(){
    functions_pre
    el_debug "\targs '$*'"
    local answer spaceleft

    # warng if low, exit if not enough {{{
    spaceleft="$( LC_ALL="$EL_LC_EN" df -m  2>/dev/null | awk -v TARGET=$TARGET '$6 == TARGET {print $4}' )"
    el_debug "Free Space left is: $spaceleft"

    if [[ "$spaceleft" -lt "200" ]]; then
        el_error "no space left on partition $HD_PART_ROOT_dev"
        guitool error "$( eval_gettext "No space left on partition %s." "${HD_PART_ROOT_dev}" )"
        if guitool yesnoquestion "$( eval_gettext "You have run out of space on your hard disk. Stop the installation now?" )" ; then
            exit 187
        fi
    fi

    if [[ "$spaceleft" -lt "3000" ]] && [[ -z "$MACHINE_VIRTUAL_TYPE" ]] ; then
        guitool warning "$( eval_gettext "You now have less than 3 GB of free space on your root partition. You can still use Elive, but risk running out of disk space soon, making your system unusable. Next time you install Elive, ensure you have at least 20 GB of free space available." )" &
    fi
    # - warng if low, exit if not enough }}}
}
########################################################################
swaptab_add(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    local toadd dev uuid

    # - pre }}}
    # This function seems to be needed for a correct hibernation restoring and update, was implemented in topaz, i will leave it available for compatibility for now, check hte old hibnerate package too for it
    for toadd in "$@"
    do
        dev="$( hd_uuid_decode "$toadd" )"
        uuid="$( LC_ALL="$EL_LC_EN" blkid "$toadd" | grep "$toadd" | tr ' ' '\n' | grep "UUID=" | head -1 )"

        if ! grep -h "$dev" "$TARGET/etc/swaptab" 2>/dev/null 1>&2 | grep -q "$uuid" ; then
            echo -e "    Adding $dev with $uuid to swap-tab"

            if LC_ALL="$EL_LC_EN" mount | grep -qs "$TARGET" ; then
                echo "$dev $uuid" >> "$TARGET/etc/swaptab"
            else
                echo -e "WARNING: target not mounted while trying to add a swaptab entry"
            fi
        fi
    done
}
########################################################################
copying_main_system_pre(){
    # pre {{{
    local url

    #}}}
    # play a game during install {{{

    if ((is_virtualbox)) && [[ -x "$( which chromium || true )" ]] ; then
        browser_preference="google-chrome chrome chromium"
    fi

    # TODO: add a flag that, if the user selects to donate, put a pause in the installation process until the user has finished the donation process, and a timetout of max 10 minutes for that
    # TODO: add a more convincing way to donate, maybe with a revslider run in python-html or similar when there's at least X amount of ram (try semi-transparent bg? )

    # run this in BG
    # if cost-free version or stable (cost-free):
    #if ((is_stable)) && [[ "$DISPLAY" != "localhost:10"* ]] ; then
    if [[ "$DISPLAY" != "localhost:10"* ]] && ! ((is_thanatests)) && ! ((is_premium_user)) ; then
        (
        url="https://www.elivecd.org/donate/?id=installer-donation"

        sleep 10
        #guitool info "$( eval_gettext "The installation will take a while to complete. Unfortunately, we don't have a good way to fund this project and the sum of donations is currently very low to sustain this project in the long term. If you can contribute any amount to this project, it would be magnificent. Thanks to people like you, Elive has been able to exist since 2005 without any restrictions, limitations, or annoying ads from hell. Elive offers an amazing system to people all over the world and completely free of charge." )"
        guitool info "$( eval_gettext "The installation will take some time to complete. If you appreciate this project, any contribution would be wonderful. Thanks to supporters like you, Elive has been available since 2005 without restrictions, limitations, or annoying ads. And it will always remain an amazing free system for everyone worldwide." )"

        if guitool question "$( eval_gettext "This project is sustained solely by contributions from our users. Help us to make it grow even more!" ) 🙏" ; then

            #su -c "web-launcher --incognito --temp-profile \"$url\" ${browser_preference}" "$eliveuser" 1>/dev/null 2>&1  &
            su -c "web-launcher \"$url\"" "$eliveuser" 1>/dev/null 2>&1  &
        else

            ## play super mario war
            if ! ((is_thanatests)) && [[ "$RAM_TOTAL_SIZE_mb" -gt "1200" ]] ; then
                if ! LC_ALL="$EL_LC_EN" ps aux | grep -v grep 2>/dev/null | grep -v $$ 2>/dev/null | grep -qs "smw" 2>/dev/null && [[ -x "/usr/games/smw" ]] && ! ((is_virtualized)) && [[ -n "$E_START" ]]  ; then
                    if ! ((is_virtualized)) ; then
                        el_notify normal znes "$( eval_gettext "Maybe the next time" )" "$( eval_gettext "Enjoy it" )"
                        su -c "sleep 4 ; enlightenment_remote -exec-cmd \"/usr/games/smw\"" "$eliveuser" 2>/dev/null
                    fi
                fi
            fi
        fi
        ) &
    fi


    # }}}
    # mounts: shared /home needs to get /etc confs for the migrate step {{{
    rm -rf "$TARGET/etc.old-to-"*  2>/dev/null

    if ((is_shared_home_partition)) ; then
        local _point

        EL_DEBUG=0 mounts-manager --mount "$partition_shared_home_referenced_root_partition" 1>/dev/null
        _point="$( EL_DEBUG=0 mounts-manager --get-mountpoint "$partition_shared_home_referenced_root_partition" )"
        cp -a "$_point/etc" "$TARGET/etc.old-to-${eliveversion}"
        el_umount "$_point"

        # we need also to add this mark for special configurations, like to not create the .asoundrc file
        echo "do not delete this file, created by elive, to know that this home is shared between more than one OS" > "$TARGET/$dhome/.shared-home"

        # audio configuration should use alsa.conf, remove all .asoundrc too:
        # update: we use .asoundrc
        #rm -f "$TARGET/etc/skel/.asoundrc" 2>/dev/null
    else
        cp -a "$TARGET/etc" "$TARGET/etc.old-to-${eliveversion}" 2>/dev/null || true
    fi


    # - shared /home needs to get /etc confs for the migrate step }}}
    # add a temporal swap file in low resources {{{
    if [[ "$RAM_TOTAL_SIZE_mb" -lt "1200" ]]; then
        guitool "working_start"

        swap_partitions_all_do "activate"

        # if the system has very low ram, also set a default empty wallpaper to save some memory
        if [[ "$RAM_TOTAL_SIZE_mb" -lt 800 ]] && [[ -n "$E_START" ]] ; then
            #guitool info "$( eval_gettext "Since you have very little RAM: Using a white wallpaper will use less resources." )"
            # this sets a white/blank wallpaper, best memory saving option:
            su -c "enlightenment_remote -desktop-bg-set wkljlwejkodwijmwoiejowei 2>/dev/null || true" $eliveuser 2>/dev/null || true
        fi

    fi

    # - add a temporal swap file in low resources }}}
    # fixes {{{
    if [[ -d "$TARGET/snap" ]] ; then
        maintain_installed_packages_extra="snapd $maintain_installed_packages_extra"
        is_snaps_installed=1
        _app_snaps="TRUE"
    fi
    # - fixes }}}
}
########################################################################
copying_main_system_post(){
    # always delete this hack for the live mode
    rm -f "$TARGET/etc/dpkg/dpkg.cfg.d/01_nodoc" # FIXME seems like not needed anymore: its not meant to be copied from the exclude list, needs verifying

    # fix for plymouth that was modified for hte live mode:
    sed -i -e '/^Wants=.*drivers-install/d' -e '/^After=.*drivers-install/d' "$TARGET/usr/lib/systemd/system/plymouth-quit.service" 1>/dev/null 2>&1 || true
}
#########################################################################
copy_canceled_stopping(){
    functions_pre
    el_debug "\targs '$*'"

    is_installation_canceled=1

    # dont use function here, since we need to get its pid
    $guitool --title='Elive Installer' --error --text="$( eval_gettext "Installation cancelled." )" || true &
    pid_tmp_gui="$!"

    # kill rsync and say that is ok {{{
    for i in $( seq 3 )
    do
        #killall -9 cp 2>/dev/null || true
        killall -9 rsync 2>/dev/null || true
        sync
        LC_ALL="$EL_LC_EN" sleep 0.4
    done
    rm -f "$TMP_PROGRESS_WORKING_COPYING_f"

    kill "$pid_tmp_gui" 2>/dev/null || kill -9 "$pid_tmp_gui" 2>/dev/null || true
    $guitool --title='Elive Installer' --info --text="$( eval_gettext "All processes have been closed correctly. You can restart the installation now..." )" 2>/dev/null || true
    # - kill rsync and say that is ok }}}
    exit
}
#########################################################################
partition_crypttab_add_if_needed(){
    local partition p_num hd_part_luks_name partition_number_prefix partition_disk_from
    el_debug "\targs '$*'"
    partition="$1"

    if ! [[ -b "$partition" ]] ; then
        el_error "partition is not a block device: $partition"
        return 0
    fi

    # check
    partition_disk_from="$( get_disk_from_partition_complex "$partition" )"
    #if [[ "$HD_DEV" != "$partition_disk_from" ]] ; then
        #el_warning "HD_DEV is not the same as partition_disk_from: $HD_DEV | $partition_disk_from"
    #fi

    if [[ "$partition" = "/dev/mapper/"* ]] ; then
        if [[ "$HD_DEV" = /dev/nvme* ]] || [[ "$HD_DEV" = /dev/mmcblk* ]] || { [[ "$HD_DEV" = /dev/md* ]] && ls /dev/md*p* 1>/dev/null 2>&1 ; } ; then
            el_debug "Device $HD_DEV is SSD over PCIe (nvme) or eMMC, or md"
            partition_number_prefix="p"
        else
            partition_number_prefix=""
        fi

        if check_part_is_from_encrypted "$partition" ; then

            p_num="$( get_partition_number_from_partition "$partition" )"
            hd_part_luks_name="${HD_DEV##*/}${partition_number_prefix}${p_num}_crypt"

            # already added in default crypttab so we don't need this entry
            p_enc_uuid="$( LC_ALL="$EL_LC_EN" cryptsetup luksDump "$partition" 2>/dev/null | grep "^UUID:" | head -1 | awk '{print $2}' )"
            if ((is_migration)) || ((is_autoupgrade)) ; then
                if [[ -s "$TARGET/etc.old-to-${eliveversion}/crypttab" ]] && grep -qs "UUID=$p_enc_uuid" "$TARGET/etc.old-to-${eliveversion}/crypttab" ; then
                    return 0
                fi
            else
                if [[ -s "$TARGET/etc/crypttab" ]] && grep -qs "UUID=$p_enc_uuid" "$TARGET/etc/crypttab" ; then
                    return 0
                fi
            fi


            if device_is_ssd "$HD_DEV" ; then
                el_array_member_add "${hd_part_luks_name}    UUID=$( LC_ALL="$EL_LC_EN" cryptsetup luksDump "${HD_DEV}${partition_number_prefix}${p_num}" | grep "^UUID:" | head -1 | awk '{print $2}' )    none    luks,discard" "${hd_dev_to_add_in_crypttab[@]}" ; hd_dev_to_add_in_crypttab=("${_out[@]}")
            else
                el_array_member_add "${hd_part_luks_name}    UUID=$( LC_ALL="$EL_LC_EN" cryptsetup luksDump "${HD_DEV}${partition_number_prefix}${p_num}" | grep "^UUID:" | head -1 | awk '{print $2}' )    none    luks" "${hd_dev_to_add_in_crypttab[@]}" ; hd_dev_to_add_in_crypttab=("${_out[@]}")
            fi
            is_encryption_wanted=1

            el_debug "Encrypted partition added to crypttab list: '${HD_DEV}${partition_number_prefix}${p_num}' "
        fi
    fi

}
#########################################################################
old_system_prepare(){
    # This function prepares a possible old system for our new install (upgrade mode)
    # We come from root_partition_select, with the var HD_PART_ROOT_dev already set
    # pre {{{
    functions_pre

    # we already scanned to search old elive versions and we asked to the user if wants to upgrade it
    #if ((is_old_elive_upgrade_not_wanted)) ; then
        #return 0
    #fi
    # return if we already said that we don't want to migrate THIS partition
    if el_array_member_check "$HD_PART_ROOT_dev" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; then
        return 0
    fi

    unset OLD_HD_PART_HOME_dev OLD_HD_PART_BOOT_dev old_hostname
    # FIXME: rename all "systemname" to a better one, it is used globally now
    local part mountedin p_fs p_mountpoint p_dev uuid

    el_debug "\targs '$*'"

    # }}}

    # check for another (non-elive) system {{{
    # NOTE: actually os-probe doesnt show Elive systems, this is what we want here, but in any case we added a checker
    if ! ((is_autoupgrade)) && ! ((is_old_elive_upgrade_request_accepted)) ; then
        # get the name of which system we found on the selected-to-use root partition
        systemname="$( cat /tmp/os-probed | awk -v FS=":" -v HD_PART_ROOT_dev="${HD_PART_ROOT_dev}" '{if($1 == HD_PART_ROOT_dev) print $2 }' )"
        if [[ -n "$systemname" ]] && [[ "$systemname" != "Elive"* ]] ; then
            el_explain 2 "The selected Root Partition has already another System called: $systemname"
        else
            el_debug "nothing found (known by os-prober) to migrate"
            unset systemname old_hostname is_migration is_migration_wanted
            return 0
        fi
    fi


    # - check for another (non-elive) system }}}
    # prepare and fixes {{{
    el_umount "${HD_PART_ROOT_dev}" 2>/dev/null 1>&2 || true
    EL_DEBUG=0 mounts-manager --mount "$HD_PART_ROOT_dev" 1>/dev/null
    mountedin="$( EL_DEBUG=0 mounts-manager --get-mountpoint "$HD_PART_ROOT_dev" )"

    #if [[ -z "$mountedin" ]] ; then
        #el_error "no mountpoint get?\nls /media:\n$( ls -1 /media/ )\nmounts:\n$( mount | tail -5 )"
        #el_warning "unable to mount? some debug:\n$( EL_DEBUG=3 mounts-manager --mount "$HD_PART_ROOT_dev" 2>&1 )"
    #fi

    # }}}
    # checkers {{{
    # nothing found to migrate
    if [[ -z "$mountedin" ]] || [[ ! -e "$mountedin/etc/fstab" ]] ; then
        el_debug "nothing found to migrate, mountedin = '$mountedin' & fstab =\n$( cat "$mountedin/etc/fstab" 2>/dev/null || true )\nls / =\n$(ls "$mountedin" )"
        unset systemname old_hostname is_migration is_migration_wanted
        return 0
    fi

    # copy back a last copy of etc if we come from a previous unfinished upgrade mode
    if [[ ! -d "${mountedin}/etc" ]] ; then
        if ls -1 "${mountedin}/" | sort -g | grep -qs "/etc" ; then
            cp -a "${mountedin}/$( ls -1 "$mountedin" | sort -g | grep "/etc" | tail -1 )" "${mountedin}/etc"
        fi
    fi

    el_explain 2 "root partition mounted in ${mountedin}"
    el_explain 2 "FSTAB of old system is:\n${el_c_n}$( cat "${mountedin}/etc/fstab" )"

    # - checkers }}}

    # encryption vars  {{{
    partition_crypttab_add_if_needed "$HD_PART_ROOT_dev"

    # }}}

    # get variables {{{
    #if grep -qs "special version" "$mountedin/etc/elive-version" ; then
    if ! [[ -e "$mountedin/var/lib/dpkg/info/elive-skel-default-all.list" ]] ; then
        is_old_special_version=1
    fi

    old_hostname="$( grep -v "^#" "$mountedin/etc/hostname" 2>/dev/null | tail -1 )"
    if [[ -z "$old_hostname" ]] ; then
        old_hostname="${systemname}-to-Elive"
    fi

    if [[ -z "$old_hostname" ]] ; then
        el_explain 2 "hostname empty, ignoring migration...."
        guitool error "$( eval_gettext "The hostname configuration for your old system cannot be found. We will proceed with a fresh installation instead of migration mode." )"
        unset systemname old_hostname is_migration is_migration_wanted
        return 0
    else
        hostname="$old_hostname"
        el_explain 2 "old hostname is $old_hostname"
    fi

    if [[ "$old_hostname" = "localhost"*"localdomain" ]] ; then
        el_explain 2 "previous system was not installed correctly (old_hostname is '$old_hostname'), ignoring migration..."
        unset systemname old_hostname is_migration is_migration_wanted
        guitool warning "$( eval_gettext "Your selected system to migrate doesn't appear to have a correct previously installed status. We will do a NEW normal, clean installation method instead." )"
        return 0
    fi
    # - get hostname }}}

    # ask if want to migrate or clean {{{
    if [[ -n "$systemname" ]] && ! ((is_old_elive_upgrade_request_accepted)) ; then
        systemname="$( cat /tmp/os-probed | grep "$systemname" | awk -v FS=":" '{print $3" ("$2")" }' | tail -1 )"
        el_explain 2 "found another OS which we can migrate named '$systemname'"

        local translated_message
        translated_message="$( printf "$( eval_gettext "There is a system '%s' installed on '%s' named %s.\n\nDo you want to migrate it to an Elive system while keeping the existing users and their files?" )" "$systemname" "${HD_PART_ROOT_dev}" "${old_hostname}" )"

        if ((is_migration_not_wanted)) ; then
            el_explain 2 "migration not wanted, ignoring migration...."
            unset systemname old_hostname is_migration is_migration_wanted
            el_array_member_add "$HD_PART_ROOT_dev" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; HD_PARTITIONS_USED_NOMIGRATE_array=("${_out[@]}")
            return 0
        fi

        if ((is_migration)) || ((is_migration_otheros)) || ((is_migration_wanted)) ; then
            # just continue with migration
            is_migration=1
            is_migration_otheros=1
            echo -e "\n\nMODE: MIGRATION ($systemname)\n\n" 1>&2
        else
            if guitool yesnoquestion "$translated_message" ; then
                is_migration=1
                is_migration_otheros=1
                echo -e "\n\nMODE: MIGRATION ($systemname)\n\n" 1>&2
            else

                el_explain 2 "migration not wanted, ignoring migration...."
                unset systemname old_hostname is_migration is_migration_wanted
                el_array_member_add "$HD_PART_ROOT_dev" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; HD_PARTITIONS_USED_NOMIGRATE_array=("${_out[@]}")
                is_migration_not_wanted=1
                return 0
            fi
        fi

        # tell the user about possible fail
        if ! ((is_thanatests)) ; then
            guitool info "$( eval_gettext "Important: This is an experimental feature and may not function uniformly across all systems. If installation fails, do a normal install instead" )"
        fi
    fi

    # - ask if upgrade or clean }}}
    # check if root partition is same  {{{
    # needed to get the FS:
    part="$( awk '{if ( $1 !~ /^#/ &&  $2 == "/" ) print $1}' "$mountedin/etc/fstab" | grep -v "^#" | head -1 )"
    if [[ -z "$part" ]] ; then
        part="$( grep "search.*set=root.*\-.*\-.*\-" "$mountedin/boot/grub/grub.cfg" 2>/dev/null | sed -e 's|^.*hint-||g' | grep "\-.*\-.*\-" | head -1   | sed -Ee 's|\s|\n|g' | grep "\-.*\-.*\-" | head -1 )"
        [[ -n "$part" ]] && part="$( hd_uuid_decode "${part}" )"

        if [[ ! -b "$part" ]] ; then
            el_debug "we cannot guess the root partition from the grub.cfg file, trying to use the first partition found"
            part="$HD_PART_ROOT_dev"
        fi
    fi

    if [[ -n "$part" ]] ; then
        part="$( hd_uuid_decode "${part}" )"
    else
        guitool error "$( eval_gettext "The root partition from your old system cannot be found. The installation will proceed as a new system, not a migration." )"
        unset systemname old_hostname is_migration is_migration_wanted
        return 0
    fi

    # Note: arch linux doesn't has a fstab at all, only the root partition uuid on the grub.cfg file and the /home i don't know (systemd?)
    # if [[ -e "$mountedin/etc/arch-release" ]] && [[ -z "$part" ]] ; then
    #     guitool warning "$( eval_gettext "Your Arch system doesn't has an fstab file so we cannot migrate it into an Elive one. This is normally caused when gefstab has not been run after the installation." )"
    # fi

    if [[ "$( readlink -f "${HD_PART_ROOT_dev}" )" != "$( readlink -f "${part}" )" ]] ; then
        if [[ -b "$HD_PART_ROOT_dev" ]] ; then
            part="$( readlink -f "$HD_PART_ROOT_dev" )"
        fi
    fi

    if [[ "$( readlink -f "${HD_PART_ROOT_dev}" )" != "$( readlink -f "${part}" )" ]] ; then
       if [[ -n "$part" ]] ; then
           local translated_message
           translated_message="$( printf "$( eval_gettext "The ROOT partition on your old system does not match the one in the fstab file. Have you changed the hard disk cables or reformatted the partition?\n\nOld root partition: %s\nCurrent root partition: %s\n\nContinuing will result in a NEW FRESH INSTALLATION, not an upgrade." )" "$part" "${HD_PART_ROOT_dev}" )"

           if guitool question "$translated_message" ; then
               el_explain 2 "user wanted to do a clean installation instead because the ROOT partition didn't match the fstab one:"
               readlink -f "${HD_PART_ROOT_dev}"
               readlink -f "${part}"
               unset systemname old_hostname is_migration is_migration_wanted
               return 0
           else
               guitool info "$( eval_gettext "If the root partition cannot be found, you need to manually edit the FSTAB entries to point the installer to the correct devices. You can do this from within the Live-CD mode, then rerun the installer." )"
               exit
           fi
       fi
   fi

    # - check if root partition is same  }}}

    # get other partitions references {{{
    OLD_HD_PART_ROOT_FS_v="$( awk '{if ( $1 !~ /^#/ && $2 == "/" )print $3}' "$mountedin/etc/fstab" | grep -v "^#" | head -1 )"
    if [[ -z "$OLD_HD_PART_ROOT_FS_v" ]] ; then
        OLD_HD_PART_ROOT_FS_v="$( hdtools_get_fs "$part" )"
    fi
    if [[ -z "$OLD_HD_PART_ROOT_FS_v" ]] ; then
        el_error "unable to detect the root partition FS: $( cat "$mountedin/etc/fstab" )"

        guitool warning "$( eval_gettext "Unable to determine the root file system of your old system. The installation will be continued as a new installation instead." )"
        echo -e "\n\nMODE: NEW CLEAN INSTALLATION\n\n" 1>&2
        unset systemname old_hostname is_migration is_migration_wanted
        return 0
    fi

    OLD_HD_PART_ROOT_FS_OPTS_v="$( awk '{if ( $1 !~ /^#/ && $2 == "/" )print $4}' "$mountedin/etc/fstab" | grep -v "^#" | head -1 )"
    if [[ "$OLD_HD_PART_ROOT_FS_OPTS_v" = *"compress="* ]] ; then
        is_fs_compression_used=1
    fi

    el_explain 2 "root partition for old system is $part with FS $OLD_HD_PART_ROOT_FS_v"


    havehome="$( awk '{if ( $1 !~ /^#/ && $2 == "/home") print $1}' "$mountedin/etc/fstab" | grep -v "^#" | head -1 )"
    if [[ -n "$havehome" ]] ; then
        OLD_HD_PART_HOME_dev="$( awk '{if ( $1 !~ /^#/ &&  $2 == "/home" )print $1}' "$mountedin/etc/fstab" | grep -v "^#" | head -1 )"
        OLD_HD_PART_HOME_dev="$( hd_uuid_decode "$OLD_HD_PART_HOME_dev" )"
        OLD_HD_PART_HOME_fs="$( awk '{if ( $1 !~ /^#/ &&  $2 == "/home" )print $3}' "$mountedin/etc/fstab" | grep -v "^#" | head -1 )"

        el_explain 2 "old home partition for old system is $OLD_HD_PART_HOME_dev with FS $OLD_HD_PART_HOME_fs"
    fi
    unset havehome

    has_old_partition_boot="$( awk '{if ( $1 !~ /^#/ && $2 == "/boot") print $1}' "$mountedin/etc/fstab" | grep -v "^#" | head -1 )"

    if [[ -n "$has_old_partition_boot" ]] ; then

        OLD_HD_PART_BOOT_dev="$( awk '{if ( $1 !~ /^#/ &&  $2 == "/boot" )print $1}' "$mountedin/etc/fstab" | grep -v "^#" | head -1 )"
        OLD_HD_PART_BOOT_dev="$( hd_uuid_decode "$OLD_HD_PART_BOOT_dev" )"
        OLD_HD_PART_BOOT_fs="$( awk '{if ( $1 !~ /^#/ &&  $2 == "/boot" )print $3}' "$mountedin/etc/fstab" | grep -v "^#" | head -1 )"
        if [[ "$OLD_HD_PART_BOOT_fs" = "auto" ]] || [[ "$OLD_HD_PART_BOOT_fs" = "none" ]] ; then
            OLD_HD_PART_BOOT_fs="$( hdtools_get_fs "${OLD_HD_PART_BOOT_dev}" )"
        fi

        el_explain 2 "old boot partition for old system is $OLD_HD_PART_BOOT_dev with FS $OLD_HD_PART_BOOT_fs"
    fi
    unset has_old_partition_boot

    # more partitions to add, include all of them
    while read -ru 3 entry
    do
        if [[ -z "$entry" ]] ; then
            continue
        fi

        # convert /dev to UUID or other
        if ! echo "$entry" | grep -qsE "^/dev/(fd0|fd1|sr0|sr1|sr2|sg0|sg1|sg2|dvd|scd)" ; then
            if echo "$entry" | grep -qsE "^/dev/" ; then
                p_dev="$( echo "$entry" | awk '{print $1}' )"
                p_dev="$( hd_uuid_encode_fstab "$p_dev" )"
                entry="$( echo "$entry" | awk -v first="$p_dev" '{ $1 = first ; print $0 }' )"
                p_mountpoint="$( echo "$entry" | awk -v first="$p_dev" '{ $1 = first ; print $2 }' )"
                p_fs="$( echo "$entry" | awk -v first="$p_dev" '{ $1 = first ; print $3 }' )"

            else
                if echo "$entry" | grep -qsE "^/" ; then
                    p_dev="$( echo "$entry" | awk '{print $1}' )"
                    entry="$( echo "$entry" | awk -v first="$p_dev" '{ $1 = first ; print $0 }' )"
                    p_mountpoint="$( echo "$entry" | awk -v first="$p_dev" '{ $1 = first ; print $2 }' )"
                fi
            fi

            if [[ -n "$entry" ]] ; then
                el_debug "FSTAB entry to migrate ($p_mountpoint):  $entry"

                # add to the array
                el_array_member_add "$entry" "${FSTAB_MIGRATE_ADD_array[@]}" ; FSTAB_MIGRATE_ADD_array=("${_out[@]}")
                #excluded_dirs_for_rsync="$excluded_dirs_for_rsync $p_mountpoint"
            fi
        fi

    done 3<<< "$( awk '{if ( $1 !~ /^#/ && $2 != "/" && $2 !~ /tmpfs|\/boot|\/boot\/efi|\/boot\/EFI|\/home|none|\/media\// && $3 !~ /proc|sysfs|swap|\/tmp/ && $1 ~ /^\/dev|^UUID|^LABEL|^\// ) print $0}' "$mountedin/etc/fstab" )"
    #done 3<<< "$( awk '{if ( $1 !~ /^#/ ) print $0}' "$mountedin/etc/fstab" )"


    # checks
    if [[ "$EL_DEBUG" -gt 2 ]] ; then
        el_debug "Old fstab contents:"
        cat "$mountedin/etc/fstab" | grep -v "^#"  1>&2
    fi



    # - get other partitions references }}}

    # backup original etc {{{
    etc_backuped="${mountedin}/etc.old-to-${eliveversion}"

    if [[ ! -d "${mountedin}/etc" ]] ; then
        el_warning "Where is etc in the old system?"
    fi

    if [[ -d "${mountedin}/etc" ]] && [[ ! -d "$etc_backuped" ]] ; then
        rm -rf "${mountedin}/"etc.old-to-*
        cp -a "${mountedin}/etc" "$etc_backuped"
    fi

    # - backup original etc }}}
    # hooks and fixes pre {{{
    # remove conflicting things like rpm databases
    rm -rf "$mountedin/root/.rpmdb" "$mountedin/.rpmdb" "$mountedin/home/"*"/.rpmdb"

    # - hooks and fixes pre }}}
    # checks and reports {{{
    # checks of compatibility
    for i in $OLD_HD_PART_ROOT_FS_v $OLD_HD_PART_HOME_fs
    do
        case "$i" in
            reiserfs)
                if ! ((is_reiserfs_available)) ; then
                    guitool error "$( eval_gettext "ReiserFS appears to be poorly supported on this kernel version as it's being phased out. This may lead to instability, system crashes, or data loss. Backup your data and switch to a different file system. Stability might improve with a newer kernel or different Elive version." )"

                    if guitool yesnoquestion "$( eval_gettext "Upgrading your previous Elive is not possible with this kernel. Exit now to backup your important files and perform a new installation on another filesystem. Do you wish to proceed with a NEW and CLEAN install from scratch? This will result in the loss of all your users, files, and configurations." )" ; then
                        unset systemname old_hostname is_migration is_migration_otheros
                        echo -e "\n\nMODE: NEW CLEAN INSTALLATION\n\n" 1>&2
                        return 0
                    else
                        # this will trigger the message about a user canel exit
                        exit
                    fi
                fi
                ;;
            ext4)
                true
                #if ((is_btrfs_available)) && ! ((is_btrfs_available_notified)) ; then
                    #guitool info "$( eval_gettext "This new version of Elive supports experimental filesystems like BTRFS or REISER4 that can make your system faster and featuring transparent compression, this also means that all your data will also take less space. We will continue using your previous filesystem which is also more safe, but the next time you perform a new installation you could be interested to try these options." )"
                    #is_btrfs_available_notified=1
                #fi
                ;;
        esac
    done

    # - checks and reports }}}
    # end and cleanups {{{
    # umount it when we have finished
    el_umount "$mountedin" 1>/dev/null 2>/dev/null

    # make sure to have an old version of elive in order to run specific upgrade hooks
    if [[ -z "$old_eliveversion" ]] ; then
        old_eliveversion="0.0.1"
    fi

    is_migration=1
    # }}}
}
#########################################################################
old_elive_scan_disks(){
    # pre {{{
    functions_pre
    if ((is_autoupgrade)) || ((ignore_disks_scan_for_elives)) ; then
        return 0
    fi
    el_debug "\targs '$*'"


    local point
    guitool "working_start"

    # - pre }}}
    # try to mount, scan for other elives and make a list {{{

    while read -ru 3 dev
    do
        unset point old_hostname old_eliveversion old_elivecodename
        if [[ -z "$dev" ]] ; then
            continue
        fi
        if ! [[ -b "$dev" ]] ; then
            el_debug "skipping $dev for scanning"
            continue
        fi

        if el_array_member_check "$dev" "${HD_PARTITIONS_USED_array[@]}" || el_array_member_check "$dev" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; then
            el_debug "ignoring '$dev' because is a partition set to ignore"
            continue
        fi

        if EL_DEBUG=0 mounts-manager --mount "$dev" 1>/dev/null ; then

            point="$( EL_DEBUG=0 mounts-manager --get-mountpoint "$dev" )"

            if [[ -z "$point" ]] ; then
                el_debug "unable to get mountpoint for $dev, it is mounted? skipping..."
                continue
            fi

            el_debug "Scanning $dev for old elive"
            if [[ -f "$point/etc/elive-version" ]] && [[ -n "$point" ]] && [[ -d "$point" ]] && [[ "$point" != "/" ]] ; then

                old_hostname="$( grep -v "^#" "$point/etc/hostname" 2>/dev/null | tail -1 )"
                old_eliveversion="$(cat "$point/etc/elive-version" | grep elive-version | sed -e 's|^.*version: ||g' -e 's|+$||g' )"
                old_elivecodename="$(cat "$point/etc/elive-version" | grep elive-codename | sed 's|^.*codename: ||g')"
                #old_autologin_user="$(cat "$point/etc/elive-version" | grep "^autologin-user:" | sed 's|^.*autologin-user: ||g' | grep "[[:alpha:]]*" | tail -1 )"
                old_autologin_user="$( grep "^autologin-user=" "$point/etc/lightdm/lightdm.conf" 2>/dev/null | awk -v FS="=" '{print $2}' | tail -1 )"
                oldelivesize="$( LC_ALL="$EL_LC_EN" df -h | awk -v TARGET=$point '$6 == TARGET {print $2}' | tail -1 )"

                # only really-installed systems are considered to be upgradeable
                if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" "ge" "2.9.9" ; then
                    # should be "installed" (a first correct boot):
                    if [[ "$( grep "Status: " "$point/etc/elive/installation-status" 2>/dev/null | sed -e 's|^.*Status: ||g' | tail -1 )" = "installed" ]] ; then
                        is_was_installed_correctly=1
                    else
                        if ((is_thanatests)) || ((is_was_installed_correctly)) ; then
                            is_was_installed_correctly=1
                        else
                            local message_body
                            message_body="$( eval_gettext "The state of your ancient installed Elive is not marked as correctly installed. Continuing with this upgrade mode may lead to bugs. Do you want to proceed anyway?" )"
                            local message_no
                            message_no="$( printf "$( eval_gettext "No, do a new clean installation instead." )" "" )"
                            local message_yes
                            message_yes="$( printf "$( eval_gettext "Yes, upgrade my old system." )" "" )"
                            if $guitool --title='Elive Installer' --question --cancel-label="$message_no" --ok-label="$message_yes" --text="$message_body" 2>/dev/null ; then
                                is_was_installed_correctly=1
                                el_explain 2 "User selected to continue with upgrade-mode even if the system is not marked as correctly installed"
                            fi
                        fi
                    fi

                    if ! ((is_was_installed_correctly)) ; then
                        #guitool warning "$( eval_
                        #gettext "An Elive system has been detected in your hard disk but it looks like it was not installed correctly, so we will ignore it option as an upgrade mode, if you think that there's something wrong that we can fix please report the issue to Elive so we can fix it." )"
                        el_array_member_add "$dev" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; HD_PARTITIONS_USED_NOMIGRATE_array=("${_out[@]}")
                        unset \
                            OLD_HD_PART_BOOT_dev \
                            OLD_HD_PART_BOOT_fs \
                            OLD_HD_PART_HOME_dev \
                            OLD_HD_PART_HOME_fs \
                            OLD_HD_PART_ROOT_FS_OPTS_v \
                            OLD_HD_PART_ROOT_FS_v \
                            old_autologin_user \
                            old_eliveversion \
                            old_hostname \
                            old_elivecodename

                        # only add this mark if was an elive system
                        if [[ -s "$point/etc/elive-version" ]] || [[ -d "$point/etc/elive" ]] ; then
                            is_previous_installation_failed=1
                        fi

                        # always umount, do not leave partitions mounted
                        el_umount "$point" 2>/dev/null || true

                        continue
                    fi
                fi
                el_debug "Old elive system found in $dev"

                # copy possible installer-modules
                #instcod_files_scan_copy

                if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$old_eliveversion" "ge" "1.9.51" && [[ "$old_hostname" != "localhost"*"localdomain" ]] ; then
                    local message_old_elive_list
                    message_old_elive_list="$( printf "$( eval_gettext "Name: '%s', Version: '%s %s', Size: '%s'" )" "${old_hostname}" "${old_elivecodename}" "${old_eliveversion}" "${oldelivesize}" )"

                    OLD_ELIVE_LIST="${OLD_ELIVE_LIST[@]}\n${dev}\n${message_old_elive_list}"
                    OLD_ELIVE_LIST="$( echo -e "${OLD_ELIVE_LIST[@]}" | sed -e 's/^\ //g' -e '/^$/d' )"
                else
                    unset \
                        OLD_HD_PART_BOOT_dev \
                        OLD_HD_PART_BOOT_fs \
                        OLD_HD_PART_HOME_dev \
                        OLD_HD_PART_HOME_fs \
                        OLD_HD_PART_ROOT_FS_OPTS_v \
                        OLD_HD_PART_ROOT_FS_v \
                        old_autologin_user \
                        old_eliveversion \
                        old_hostname \
                        old_elivecodename

                    el_array_member_add "$dev" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; HD_PARTITIONS_USED_NOMIGRATE_array=("${_out[@]}")
                fi
            fi

            el_umount "$point" 2>/dev/null || true
        else
            el_debug "Unable to mount $dev for scan to search old elive's, details of partition: $(EL_DEBUG=0 partitions-list --show-only="$dev" || true )"
        fi
    done 3<<< "$( EL_DEBUG=0 partitions-list --show-all | awk -v FS="::" '{ if ($3 ~ /^(ext2|ext3|ext4|reiserfs|btrfs|reiser4|reiser5|jfs|xfs)$/  ) print $1}' || true )"

    # - try to mount, scan for other elives and make a list }}}
    # suggest migration if other non-elive systems were found {{{
    if [[ -n "$OLD_ELIVE_LIST" ]] ; then
        el_explain 2 "List of other Elive systems found: ${OLD_ELIVE_LIST}"

        old_elive_upgrademode_ask

    else
        # check if we have other systems installed useful and suggest the user to try the migration mode
        os_probed="$( cat /tmp/os-probed 2>/dev/null | grep -v "Elive" | grep ":linux$" || true )"
        if [[ -n "$os_probed" ]] ; then
            local message_migrate_suggest m_part m_name message_migrate_wanted

            m_part="$( cat /tmp/os-probed | grep ":linux" | psort -- -p "UnknownLSB" | awk -v FS=":" '{print $1}' | tail -1 )"
            m_name="$( cat /tmp/os-probed | grep ":linux" | psort -- -p "UnknownLSB" | awk -v FS=":" '{print $3" ("$2")" }' | tail -1 )"
            m_name="${m_name/UnknownLSB /}"

            if [[ "$( echo "$os_probed" | wc -l )" -eq 1 ]] ; then
                el_explain 2 "Another OS found in '$m_part' named '$m_name'"

                message_migrate_wanted="$( printf "$( eval_gettext "Would you like to migrate your previous system to an Elive system while preserving the same partitioning structure and existing user files?" )" "" )"

                if ((is_migration_wanted)) ; then
                    HD_PART_ROOT_dev="$( echo "$m_part" )"
                    is_migration=1
                    is_migration_otheros=1
                else
                    if guitool yesnoquestion "${message_migrate_wanted}\n\nSystem '$m_name' on '$m_part'" ; then
                        # tell the user about possible fail
                        HD_PART_ROOT_dev="$( echo "$m_part" )"
                        is_migration=1
                        is_migration_otheros=1
                    else
                        is_migration_not_wanted=1
                    fi
                fi
            else
                el_explain 2 "Other operating systems found:\n$( cat /tmp/os-probed )"
                message_migrate_suggest="$( printf "$( eval_gettext "Other operating systems have been found on your computer. If you don't repartition your hard disk and select a partition with an existing system, you can convert it to an Elive system by migrating its users and files.\n\nFor your '%s', you need to select '%s', for example." )" "$m_name" "$m_part" )"
                guitool info "$message_migrate_suggest"
            fi
        else
            el_explain 2 "Not ancient Elive systems installed where found automatically, continuing..."
        fi
    fi

    # try to help the user in case his previous installation failed
    if ((is_previous_installation_failed)) ; then
        # if we have a windows system in EFI mode and we are installing again because it failed, is probably because it didn't booted
        list_others_systems
        if grep -qs ":Windows" /tmp/os-probed ; then
            if ((is_efi_booted)) ; then
                local message_none
                message_none="$( printf "$( eval_gettext "None of the above listed" )" "" )"
                local message_win_alone
                message_win_alone="$( printf "$( eval_gettext "Your computer booted to Windows (only) without showing the menu to select between Elive or Windows" )" "" )"

                unset result
                while [[ -z "$result" ]] ; do
                    result="$( echo -e "1\n${message_win_alone}\n2\n${message_none}" | zenity --list --width=400 --height=200 --column="" --column="Case" --hide-column=1 || echo cancel )"
                done

                case "$result" in
                    1)
                        guitool info "$( eval_gettext "We can resolve the boot menu issue. Take a photo of the next message with your phone, then reboot your computer to return to Windows and fix the issue as explained:" )"
                        efi_fix_windows_monopol
                        reboot
                        ;;
                esac

            fi
        fi
        if ((is_efi_booted)) ; then
            efi_fix_debian_dir
            efi_fix_sparsefiles
        fi
    fi


    # - suggest migration if other non-elive systems were found }}}
}
#########################################################################
old_elive_upgrademode_ask(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    el_check_variables "OLD_ELIVE_LIST"
    local oldelivelistnum answer

    # The user select if want to upgrade mode or not, we set a few variables and then we continue with the normal process of the installation for now
    oldelivelistnum="$(echo -e "${OLD_ELIVE_LIST[@]}" | wc -l)"
    el_debug "oldelivelistnum = $oldelivelistnum"

    # }}}
    # request to use if want to upgrade {{{
    local message_upgrade_or_not
    # FIXME: do not show this message (or more exactly calling this function) if the we dont detect that the installation was finished correctly (post-install script stuff), because the user could want to try to install it multiple (failed) times
    message_upgrade_or_not="$( printf "$( eval_gettext "Your system can be upgraded instead of being reinstalled. This will save all your settings and keep all users and passwords.\n\nDo you want to continue in upgrade mode? (recommended)" )" )"


    if [[ "$oldelivelistnum" -ge 4 ]] ; then
        # More than one system found

        if ! guitool yesnoquestion "$( eval_gettext "Multiple versions of Elive have been found on your system." )""\n\n${message_upgrade_or_not}" 1>/dev/null 2>&1 ; then
            unset \
                OLD_HD_PART_BOOT_dev \
                OLD_HD_PART_BOOT_fs \
                OLD_HD_PART_HOME_dev \
                OLD_HD_PART_HOME_fs \
                OLD_HD_PART_ROOT_FS_OPTS_v \
                OLD_HD_PART_ROOT_FS_v \
                old_autologin_user \
                old_eliveversion \
                old_hostname \
                old_elivecodename \
                point
            #is_old_elive_upgrade_not_wanted=1
            while read -ru 3 line
            do
                if [[ "$line" = "/dev/"* ]] ; then
                    el_explain 0 "unwanted upgrade for: '${line}' "
                    el_array_member_add "$line" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; HD_PARTITIONS_USED_NOMIGRATE_array=("${_out[@]}")
                fi
            done 3<<< "$( printf "%s\n" "${OLD_ELIVE_LIST[@]}" )"
            return 0
        fi

        answer="$( printf "%s\n" "${OLD_ELIVE_LIST[@]}" | $guitool --title='Elive Installer' --list --height=220 --width=580 --text="$( eval_gettext "Select the Elive system you want to upgrade." )" --column="$( eval_gettext "Partition" )" --column="$( eval_gettext "System" )" 2>/dev/null || echo cancel )"

        if [[ "$answer" = "cancel" ]] || [[ -z "$answer" ]] ; then
            #is_old_elive_upgrade_not_wanted=1
            while read -ru 3 line
            do
                if [[ "$line" = "/dev/"* ]] ; then
                    el_explain 0 "unwanted upgrade for: '${line}' "
                    el_array_member_add "$line" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; HD_PARTITIONS_USED_NOMIGRATE_array=("${_out[@]}")
                fi
            done 3<<< "$( printf "%s\n" "${OLD_ELIVE_LIST[@]}" )"
            return 0
        fi
        verify_result "$answer"
        dev="$answer"

        EL_DEBUG=0 mounts-manager --mount "$dev" 1>/dev/null
        point="$( EL_DEBUG=0 mounts-manager --get-mountpoint "$dev" )"

        # get some data
        if [[ -s "$point/etc/elive-version" ]] && [[ -n "$point" ]] && [[ -d "$point" ]] ; then
            old_hostname="$(cat "$point/etc/hostname" 2>/dev/null | tail -1 )"
            old_eliveversion="$(cat "$point/etc/elive-version" | grep elive-version | sed -e 's|^.*version: ||g' -e 's|+$||g' )"
            old_elivecodename="$(cat "$point/etc/elive-version" | grep elive-codename | sed 's|^.*codename: ||g' )"
            #old_autologin_user="$(cat "$point/etc/elive-version" | grep "^autologin-user:" | sed 's|^.*autologin-user: ||g' | grep "[[:alpha:]]*" | tail -1 )"
            old_autologin_user="$( grep "^autologin-user=" "$point/etc/lightdm/lightdm.conf" 2>/dev/null | awk -v FS="=" '{print $2}' | tail -1 )"
        fi
    else

        # get some data
        dev="$( printf "%s\n" "${OLD_ELIVE_LIST[@]}" | grep "/dev/" | head -1)"
        EL_DEBUG=0 mounts-manager --mount "$dev" 1>/dev/null
        point="$( EL_DEBUG=0 mounts-manager --get-mountpoint "$dev" )"
        el_explain 2 "Old elive found mounted in $point"

        if [[ -s "$point/etc/elive-version" ]] && [[ -n "$point" ]] && [[ -d "$point" ]] ; then
            old_hostname="$(cat "$point/etc/hostname" 2>/dev/null | tail -1 )"
            old_eliveversion="$(cat "$point/etc/elive-version" | grep elive-version | sed -e 's|^.*version: ||g' -e 's|+$||g' )"
            old_elivecodename="$(cat "$point/etc/elive-version" | grep elive-codename | sed 's|^.*codename: ||g')"
            #old_autologin_user="$(cat "$point/etc/elive-version" | grep "^autologin-user:" | sed 's|^.*autologin-user: ||g' | grep "[[:alpha:]]*" | tail -1 )"
            old_autologin_user="$( grep "^autologin-user=" "$point/etc/lightdm/lightdm.conf" 2>/dev/null | awk -v FS="=" '{print $2}' | tail -1 )"
        else
            el_array_member_add "$dev" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; HD_PARTITIONS_USED_NOMIGRATE_array=("${_out[@]}")
            el_error "Failed to mount: $OLD_ELIVE_LIST ($dev in $point)"
        fi

        # not valid elive system found, skip it
        if [[ "$old_hostname" = "localhost"*"localdomain" ]] ; then
            unset old_hostname hostname

            if ! ((is_thanatests)) && ! ((is_was_installed_correctly)) ; then
                unset \
                    OLD_HD_PART_BOOT_dev \
                    OLD_HD_PART_BOOT_fs \
                    OLD_HD_PART_HOME_dev \
                    OLD_HD_PART_HOME_fs \
                    OLD_HD_PART_ROOT_FS_OPTS_v \
                    OLD_HD_PART_ROOT_FS_v \
                    old_autologin_user \
                    old_eliveversion \
                    old_hostname \
                    old_elivecodename \
                    point

                is_previous_installation_failed=1
                el_array_member_add "$dev" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; HD_PARTITIONS_USED_NOMIGRATE_array=("${_out[@]}")
                #is_old_elive_upgrade_not_wanted=1
                return 0
            fi
        fi

        # Only one system found
        local message_got_old_elive
        message_got_old_elive="$( printf "$( eval_gettext "An older version of Elive is on your hard disk. Version '%s - %s' named '%s'." )" "$( echo "$old_elivecodename" | awk '{print $1}' )" "$old_eliveversion" "$old_hostname" )"

        if ! ((is_old_elive_upgrade_request_accepted)) ; then
            if ! guitool yesnoquestion "${message_got_old_elive}\n\n${message_upgrade_or_not}"  ; then
                unset \
                    OLD_HD_PART_BOOT_dev \
                    OLD_HD_PART_BOOT_fs \
                    OLD_HD_PART_HOME_dev \
                    OLD_HD_PART_HOME_fs \
                    OLD_HD_PART_ROOT_FS_OPTS_v \
                    OLD_HD_PART_ROOT_FS_v \
                    old_autologin_user \
                    old_eliveversion \
                    old_hostname \
                    old_elivecodename \
                    point
                el_array_member_add "$dev" "${HD_PARTITIONS_USED_NOMIGRATE_array[@]}" ; HD_PARTITIONS_USED_NOMIGRATE_array=("${_out[@]}")
                #is_old_elive_upgrade_not_wanted=1
                return 0
            fi
        fi
    fi

    # we want to continue with the migration / upgrade mode

    # set real home directory
    if [[ -s "$point/etc/adduser.conf" ]] ; then
        dhome="$( grep "^DHOME" "$point/etc/adduser.conf" | sed -e 's|^.*=||g' )"
    fi
    if [[ -z "$dhome" ]] ; then
        dhome="/home"
    fi


    el_umount "$dev" 1>/dev/null


    # - request to use if want to upgrade }}}
    # set upgrade mode vars {{{

    # if we have not returned yet is that we want to upgrade our elive
    if ! echo "$dev" | grep -qsE "^/dev" ; then
        # FIXME deleteme
        el_warning "dev is $dev, that's correct?"
    fi
    HD_PART_ROOT_dev="$( echo "$dev" | awk '{print $1}' | head -1 )"

    # TODO
    #grub_mbr_search $dev

    el_array_member_add "$HD_PART_ROOT_dev" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")
    el_explain 2 "Selected partition '$HD_PART_ROOT_dev' for use as upgrade-mode of old Elive system"

    echo -e "\n\nMODE: UPGRADE (elive)  from ${old_eliveversion}\n\n" 1>&2

    hostname="$old_hostname"

    if [[ -n "$old_autologin_user" ]] ; then
        is_autologin_wanted=1
        autologin_username="$old_autologin_user"
    fi

    is_migration=1
    is_old_elive_upgrade_request_accepted=1
    unset is_old_elive_upgrade_not_wanted

    # - set upgrade mode vars }}}
}
#############################################################


#===  FUNCTION  ================================================================
#          NAME:  functions_pre
#   DESCRIPTION:  hooks for every function (previously)
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
functions_pre(){
    # pre {{{
    local arg
    #el_debug "\targs '$*'"

    arg="$1"
    #el_check_variables "arg"

    # }}}
    # check pwd {{{
    # always go to /tmp
    if [[ "$(pwd)" != /tmp ]] ; then
        if ((is_thanatests)) ; then
            el_warning "PWD is not /tmp: $(pwd)"
        fi

        # set it back, not needed if this code is stable
        cd /tmp
    fi

    # - check pwd }}}
    # something makes the screen blanking, so force a screen on everytime
    screen_dpms_on
}

#===  FUNCTION  ================================================================
#          NAME:  el_step_requires_fixme
#   DESCRIPTION:  temporal "step" to put in experimental/development code, only
#    PARAMETERS:  $1 = dir to login (optional)
#		  $2 = message (optional)
#===============================================================================
el_step_requires_fixme(){
    # pre {{{
    local dir prev_dir message

    dir="$1"
    shift
    message="$1"
    shift

    prev_dir="$(pwd)"

    : ${dir:="$prev_dir"}

    #el_set_display_variables

    el_check_dirs "${dir}"

    # }}}
    # only for my own tests mode
    if ! ((is_thanatests)) ; then
        return
    fi

    # info
    echo -e "We have set a breakpoint here, report to Elive this message if needed:"
    echo -e "${FUNCNAME} for $0"
    echo -e "Function: ${FUNCNAME[1]}  |  Line: $LINENO"
    if [[ -n "$message" ]] ; then
        el_explain 0 "message: XX${message}XX"
        echo -e ""
    fi

    # message
    guitool info "FIXME: el_step_requires_fixme: $message"

    # open interactive
    cd "$dir"
    #urxvt
    /etc/alternatives/x-terminal-emulator

    cd "${prev_dir}"
}

#===  FUNCTION  ================================================================
#          NAME:  target_umount
#   DESCRIPTION:  Umounts target directory entirely
#    PARAMETERS:  $1 = target
#       RETURNS:  -
#===============================================================================
target_umount(){
    # pre {{{
    local TARGET mounted
    el_debug "\targs '$*'"

    TARGET="$1"
    el_check_variables "TARGET"

    sync
    # }}}

    EL_DEBUG=0 mounts-manager --chroot-layers-umount "$TARGET" "eliveinstaller" bind || true

    # umount any possible efi partition
    umount "$TARGET/boot/efi" 2>/dev/null 1>&2 || true
    sync


    # umount all sub-mountpoints
    # they are sorted by lengh to correctly umount them in the correct order
    while read -ru 3 entry
    do
        if [[ "$entry" = "$TARGET" ]] || [[ -z "$entry" ]] ; then
            # do not umount target itself until the end
            continue
        fi

        el_debug "umount \"$entry\""
        umount "$entry" 2>/dev/null 1>&2 || true
        sync

    done 3<<< "$( { LC_ALL="$EL_LC_EN" mount | sed -e 's|^.* on | |g' -e 's| type .*$||g' ; } | grep -E "($TARGET|/media/.*Elive|/mnt/.*Elive)" | sort -u | sort | sort -nr )"
    sync


    if ((is_autoupgrade)) ; then
        # set it back to ro
        mount -o remount -o ro "$TARGET"

    else
        el_umount "$TARGET" 2>/dev/null 1>&2 || true

        if LC_ALL="$EL_LC_EN" mount | grep -qs "$TARGET" ; then
            for mounted in $(mount | awk -v target="$TARGET" '{if ($3 ~ target && $3 != target) print $3}')
            do
                el_debug "Umounting $mounted"
                umount "$mounted" 2>/dev/null 1>&2 || true
                sync
            done

            el_umount "$TARGET" 2>/dev/null 1>&2 || true

            if LC_ALL="$EL_LC_EN" mount | grep -qs -- "$TARGET" ; then
                # btrfs needs to finish flushing things
                sync ; sleep 2
                el_umount "$TARGET" 2>/dev/null 1>&2 || true
                LC_ALL="$EL_LC_EN" mount | grep -- "$TARGET" 1>&2

                if LC_ALL="$EL_LC_EN" mount | grep -qs -- "$TARGET" ; then
                    # if we run dpkg-reconfigure with the gnome interface, we probably have dev/pts blocked:
                    umount -l "$TARGET/dev/pts" || true
                    umount -l "$TARGET/dev" || true
                    umount -l "$TARGET" || true

                    if LC_ALL="$EL_LC_EN" mount | grep -qs -- "$TARGET" ; then
                        el_error "It stills mounted target?"
                        LC_ALL="$EL_LC_EN" mount | grep -- "$TARGET"
                        el_processes_using_mountpoint "$TARGET" 1>&2
                    fi
                fi
            fi
        fi
    fi

    if LC_ALL=C swapon --noheadings --raw | grep -qs "${TARGET}" ; then
        el_warning "we are using a swap inside TARGET:\n$( swapon --noheadings --raw )"
    fi
}

error_signal_debug(){
    trapped="$?"

    if ! ((is_run_again_installer_wanted)) ; then
        if [[ -n "$trapped" ]] && ! ((ignore_error_signal_debug)) ; then
            el_warning "trapped signal: '${trapped}' from '${FUNCNAME[1]}' in lines '${BASH_LINENO[*]}'"
            is_trapped_signal=1
        fi
    fi
}

#===  FUNCTION  ================================================================
#          NAME:  exit_error
#   DESCRIPTION:  exit program with an error and cleanups
#    PARAMETERS:  ####$1 = message
#       RETURNS:  -
#===============================================================================
exit_error(){
    # pre {{{
    trapped="$?"
    if [[ -n "$trapped" ]] ; then
        el_error "trapped signal: '${trapped}' from '${FUNCNAME[1]}' in lines '${BASH_LINENO[*]}'"
    fi

    # remove the trap ^C so we can use it as normal
    trap - INT

    functions_pre
    el_debug "\targs '$*'"

    guitool "working_stop"
    # }}}
    rm -f "$TMP_PROGRESS_USERS_f" "$TMP_PROGRESS_f" "$TMP_PROGRESS_CONFIGURING_f" || true

    # cleanups
    if ((is_migration)) ; then
        # copy back /etc if the migration failed
        if EL_DEBUG=0 mounts-manager --check-is-mounted "$TARGET" 1>/dev/null 2>&1 ; then
            if [[ -d "$TARGET/etc" ]] && [[ -d "$TARGET/etc.old-to-${eliveversion}" ]] ; then
                if [[ "$( du -s "$TARGET/etc" | awk '{print $1}' )" -lt "$( du -s "$TARGET/etc.old-to-${eliveversion}" | awk '{print $1}' )" ]] ; then
                    el_explain 0 "copying back '$TARGET/etc.old-to-${eliveversion}' to '$TARGET/etc'"
                    rm -rf "$TARGET/etc"
                    cp -fa "$TARGET/etc.old-to-${eliveversion}" "$TARGET/etc" 2>/dev/null || true
                fi
            fi
        fi
    fi

    # close dialogs
    killall zenity 1>/dev/null 2>&1 || killall -9 zenity 1>/dev/null 2>&1
    killall yad 1>/dev/null 2>&1 || killall -9 yad 1>/dev/null 2>&1

    # debug info {{{
    if [[ $EL_DEBUG -ge 3 ]] ; then
        el_explain 2 "__Debug Info that can be useful__"

        # disk structures:
        el_explain 0 "Fdisk info:"
        if ((is_disk_table_gpt)) && [[ -b "$HD_DEV" ]] ; then
            timeout 10 gdisk -l "$HD_DEV"
        else
            fdisk -l
        fi
        echo ""
        el_debug "Partitions is now: \n$( EL_DEBUG=0 partitions-list --show-all --show-raw --show-disks || true )"

        # dump variables
        #el_explain 2 "Status of variables:"
        ##LC_ALL="$EL_LC_EN" set | grep -v grep | grep -E "^(EL_|ELIVE|INST|LIST_|inst_|is_|has_|is_mode_|_array)"
        #LC_ALL="$EL_LC_EN" declare -p | grep -v grep | grep -E "(elive|elinst|alignment|resolution|rsync|HARDWARE_ID|HD_|is_|want_|has_|have|needs_|speed_disk|username|freespace|spaceleft|autopart|luks|ignore_|LIVE_MOUNT|mbr_|number_of|^old|OLD_|RAM_|SWAP)" | grep -vE "(rootpass|password|key_|BASH_SOURCE|FUNCNAME|DISK_SECTORS|guitool|LOCKFILE|message_|TMP_|OLDPWD|SUDO|TEXTDOMAIN|XAUTHORITY|conf_send_debug_reports)" | sort -u
        #echo ""

        # show flags
        el_explain 2 "Flags status when exited:"
        tree -a "${TMP_FLAGS_d}" 2>/dev/null || true
        echo ""

        el_explain 2 "Done debugging info"
    fi

    # show version of elive installer
    apt-cache policy "$elinst_package_name" | tail -n +1 | head -n +3 || true

    # }}}
    # show message {{{

    if el_check_version_development_is_days_recent 20 && ! el_flag check exit_error_reported && [[ "$EL_DEBUG" -lt 3 ]] ; then
        $guitool --title='Elive Installer' --warning --text="$( eval_gettext "Installation cancelled. If you believe there's an issue with the Elive installer, please report it to https://bugs.elivecd.org with all necessary details to reproduce the problem. Issues remain unknown unless reported, and unfortunately, today's fewer people are reporting bugs. Your contribution is for the benefit of all users." )" 2>/dev/null || true
        $guitool --title='Elive Installer' --info --text="$( eval_gettext "You may need to reboot your computer if the installation fails. This can occur due to kernel issues detecting the partition structure." )" 2>/dev/null || true
        el_flag add exit_error_reported
    fi

    # }}}
    media_errors_check 1>/dev/null 2>&1

    if ! ((is_thanatests)) ; then
        $guitool --title='Elive Installer' --error --text="$( eval_gettext "An error occurred during installation. Run the installer again and include your email in the reports so we can receive the error messages to fix them in the next update. Otherwise, use our forums for manual assistance. The installation has been aborted." )" 2>/dev/null || true
    fi

    if [[ -x "$( which hexchat || true )" ]] && ! ((is_thanatests)) && ! el_flag check chat_online_suggested  ; then
        if el_verify_internet 2>/dev/null && ! LC_ALL="$EL_LC_EN" ps aux | grep -v grep | grep -v $$ | grep hexchat ; then
            if guitool question "$( eval_gettext "Would you like to open the online chat for better assistance?" )" ; then
                if [[ -n "$E_START" ]] ; then
                    su -c "enlightenment_remote -exec-cmd \"/usr/bin/hexchat\"" $eliveuser 2>/dev/null
                else
                    su -c "/usr/bin/hexchat" $eliveuser 2>/dev/null &
                fi
                sleep 2
            fi
            el_flag add  chat_online_suggested
        fi
    fi

        #guitool info "$( eval_gettext "You exited or canceled the installation." )"
        #guitool "working_stop"
        #el_info "user canceled installation"

    rm -f "$LOCKFILE"

    el_notify normal logo-elive "Installation Aborted" "$( eval_gettext "The Elive installer is complete. You can run it again or reboot your computer." )" 2>/dev/null

    # mark a failed state
    touch /tmp/.eliveinstaller-failed

    exit "$trapped"
}

#===  FUNCTION  ================================================================
#          NAME:  exit_installer
#   DESCRIPTION:  finish the full work of the installer and run the cleanups
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
exit_installer(){
    # pre {{{
    trapped="$?"
    local var mounted
    functions_pre
    el_debug "\targs '$*'"

    # if we are not in live (like checking installer module), do not proceed with everything else
    if ! grep -q "boot=live" /proc/cmdline ; then
        exit 0
    fi

    guitool "working_stop"
    # }}}
    # old code {{{
    #[[ ! -d "${LIVE_MOUNT_MAIN_d}"  ]] && exit

    target_umount "$TARGET" || true

    rm -f "$TMP_FILE_1_f" "$TMP_FILE_2_f"

    # deleteme? what this does? i dont think that we need this at all actually
    while read -ru 3 line
    do
        [[ -z "$line" ]] && continue
        pid="$(echo "$line" | awk '{print $2}' )"
        el_debug "killing "$pid" -> "$(ps aux | grep "$pid" )""
        kill "$pid" 2>/dev/null || kill -9 "$pid" 2>/dev/null || true
    done 3<<< "$( LC_ALL="$EL_LC_EN" ps aux | grep -v grep | grep "tail .*/var/log" || true )"

    #rmdir "$TARGET" 2>/dev/null
    #true
    #exit $1
    # - old code }}}
    # checks {{{
    media_errors_check force 1>/dev/null 2>&1

    HD_DEV_BADBLOCK_SECTORS="$( LC_ALL="$EL_LC_EN" smartctl -s on -A "${HD_DEV}" | grep -iE "Reallocated_Sector_C" | tr ' ' '\n' | grep "[[:digit:]]" | sed -e 's/[^0-9]*//g' | tail -1 )"
    if [[ "$HD_DEV_BADBLOCK_SECTORS" -gt 0 ]] ; then
        #el_warning "errors in HD by smartctl: HD_DEV_BADBLOCK_SECTORS is $HD_DEV_BADBLOCK_SECTORS (FIXME/improve)"
        $guitool --title='Elive Installer' --warning --text="$( eval_gettext "Your hard disk seems to have bad blocks, making it unreliable. You should not save important data to this disk because you risk losing it. Use the 'Disks Tool' to examine your hard disk for bad blocks." )" 2>/dev/null || true
    fi

    # don't tell the user anything which is not needed since we have the reports sent
    #if ((is_trapped_signal)) ; then
        #guitool warning "$( eval_gettext "The installer detected a fail signal during the process, this should not happen or maybe it has only been a minor check test failed. In any case if you see any problem in your installed system just reinstall the system with a clean instalation with a better repartitioning, or if is possible, simply use the automated partitioning mode." )"
    #fi
    # - checks }}}
    # pending tasks {{{
    rm -f "$TMP_PROGRESS_USERS_f" "$TMP_PROGRESS_f" "$TMP_PROGRESS_CONFIGURING_f" || true

    # close integrity data checkers
    killall -9 gpg    1>/dev/null 2>&1 || true
    # download of connection_average_speed
    killall -9 wget   1>/dev/null 2>&1 || true
    # help dialogs
    # killall -9 zenity 1>/dev/null 2>&1 || true

    sync
    for i in 1 2 3 4
    do
        if jobs | grep -qs "[[:alpha:]]" ; then
            # echo -e "waiting jobs: $(jobs | grep "[[:alpha:]]" )"
            sleep 1
        else
            break
        fi
    done
    # if [[ "$EL_DEBUG" -gt 2 ]] && jobs | grep -qs "[[:alpha:]]" ; then
    #     el_explain 0 "\nWaiting some processes to finish:"
    #     jobs
    # fi

    wait
    sync

    # remove dm devices
    dmsetup remove_all || true

    # - pending tasks }}}
    # re-enable things like focus in E17, annoying for install {{{
    if [[ -n "$E_START" ]] ; then
        su -c "enlightenment_remote -focus-policy-set SLOPPY 2>/dev/null || true" $eliveuser
        su -c "enlightenment_remote -focus-policy-set NEWALWAYS 2>/dev/null || true" $eliveuser
        su -c "xhost - 2>/dev/null 2>/dev/null || true" $eliveuser || true
    fi
    if [[ -n "$EROOT" ]] ; then
        su -c "eesh focus mode sloppy 1>/dev/null || true" $eliveuser 2>/dev/null || true
        su -c "eesh set focus.all_new_windows_get_focus 1  1>/dev/null || true" $eliveuser 2>/dev/null || true
    fi

    # re-enable cairo-dock, pretty needed and required to run the installer again
    if [[ -n "$EROOT" ]] ; then
        if ! pidof -q cairo-dock ; then
            ( cairo-dock  2>/dev/null 1>&2 ) &
        fi
    fi
    # - }}}
    # show time passed {{{
    seconds_stop="$(date +%s)"
    seconds_total="$( echo -e "( $seconds_stop - $seconds_start )" | LC_ALL="$EL_LC_EN" bc -l | sed -e 's|\..*$||g' )"
    echo -e "  ${el_c_m}I: ${el_c_g}Elapsed: ${el_c_n}                                 ${el_c_m}Total${el_c_n}  $(show_time_elapsed $seconds_total )"
    #echo -e "  ${el_c_m}I: ${el_c_g}Elapsed: ${el_c_n}                                 ${el_c_m}Total${el_c_n}"

    # always remove the lock
    rm -f "$LOCKFILE"

    # }}}
    # end {{{
    if ((is_run_again_installer_wanted)) ; then
        unset is_run_again_installer_wanted
        rm -f "$LOCKFILE"

        #( sync ; sleep 4 ; sudo eliveinstaller ; sleep 2 ; ) &
        ( sync ; sleep 4 ; sudo IGNOREALREADYRUNNING=1 eliveinstaller-wrapper ; sleep 2 ; ) &

        exit 69
    fi

    guitool "working_stop"

    # close remaining gui'$ (like the one asking for donations that was run in BG)
    # for i in $(seq 5)
    # do
    #     killall zenity 1>/dev/null 2>&1 || true
    #     LC_ALL=C sleep 0.5
    # done

    # }}}
}

#===  FUNCTION  ================================================================
#          NAME:  show_time_elapsed
#   DESCRIPTION:  show the time elapsed in special format
#    PARAMETERS:  time in seconds, $2 = plain|normal
#===============================================================================
show_time_elapsed(){
    # vars {{{
    local seconds hours minutes
    seconds="$1"
    hours="$((seconds / 3600))"
    seconds="$((seconds % 3600))"
    minutes="$((seconds / 60))"
    seconds="$((seconds % 60))"


    # }}}
    # better readability {{{
    #if [[ "$( echo "$seconds" | wc -m )" -lt 3 ]] ; then
    #   seconds="0$seconds"
    #fi


    # }}}
    # show, in plain or colored {{{
    if [[ "$2" = "plain" ]] ; then
        if [[ "$hours" -gt 0 ]] ; then
            echo -n "${hours}:${minutes}:${seconds}"
        else
            if [[ "$minutes" -gt 0 ]] ; then
                echo -n "${minutes} m.  ${seconds} s."
            else
                echo -n "${seconds} s."
            fi
        fi
    else
        if [[ "$hours" -gt 0 ]] ; then
            echo -n "${C_Y}${hours}${C_M}:${C_Y}${minutes}${C_M}:${C_Y}${seconds}${C_N}"
        else
            if [[ "$minutes" -gt 0 ]] ; then
                echo -n "${C_Y}${minutes}${C_N}m. ${C_Y}${seconds}${C_N}s."
            else
                echo -n "${C_Y}${seconds} ${C_G}seconds${C_N}"
            fi
        fi
    fi


    # }}}
}

#===  FUNCTION  ================================================================
#          NAME:  hd_uuid_decode
#   DESCRIPTION:  get the dev name of a given uuid
#    PARAMETERS:  $1 = uuid
#       RETURNS:  device address
#       example:  HD_PART_ROOT_dev=$(hd_uuid_decode ${HD_PART_ROOT_dev})
#===============================================================================
hd_uuid_decode(){
    functions_pre
    local uuid_partition buf

    el_debug "\targs '$*'"

    uuid_name="$1"
    uuid_name="${uuid_name//\"}"

    if echo "$uuid_name" | grep -qs "^/" \
        && ! echo "$uuid_name" | grep -qs "^/dev" ; then
        el_debug "UUID '$uuid_name' is an address but not a block device, fallbacking it"
        echo "$uuid_name"
        return 0
    fi


    if echo "$uuid_name" | grep -qsiE "(UUID=|/by-uuid/)" ; then
        buf="$uuid_name"
        buf="${buf#/dev/disk/by-uuid/}"
        buf="${buf#UUID=}"
        if [[ -b "/dev/disk/by-uuid/$buf" ]] ; then
            uuid_name="$buf"
        fi
    fi
    if echo "$uuid_name" | grep -qsiE "(LABEL=|/by-label/)" ; then
        buf="$uuid_name"
        buf="${buf#/dev/disk/by-label/}"
        buf="${buf#LABEL=}"
        if [[ -b "/dev/disk/by-label/$buf" ]] ; then
            uuid_name="/dev/disk/by-label/$buf"
            uuid_name="$(readlink -f "$uuid_name" 2>/dev/null )"
        fi
    fi

    if [[ -z "$1" ]] || [[ "$1" = "#"* ]] ; then
        el_error "argument passed to $FUNCNAME is not valid: '$1'"
        return 0
    fi

    if echo "$uuid_name" | grep -qs "^/dev" ; then
        el_debug "UUID '$uuid_name' is already a device, no need to decode"
        echo "$uuid_name"
        return 0
    fi

    uuid_name="$(readlink -f "/dev/disk/by-uuid/$uuid_name" 2>/dev/null || true )"

    if [[ -n "$uuid_name" ]] ; then
        #echo "/dev/${uuid_name##*/}"
        echo "$uuid_name"
    else
        el_warning "$FUNCNAME: '$uuid_name' for '$uuid_name' invalid | restart udev..."
        if [[ "$2" = "secondtime" || "$uuid_name" = "secondtime" ]] ; then
            guitool error "$( eval_gettext "An error occurred with the installer. Reboot your computer and try again. If the problem persists, try using different options or the automated mode." )"
            exit_error
        fi
        service udev restart 1>&2
        # just in case
        sync ; sleep 1

        $FUNCNAME "$uuid_name" "secondtime"
    fi
}
#===  FUNCTION  ================================================================
#          NAME:  hd_uuid_encode_fstab
#   DESCRIPTION:  gives the needed entry recognized as device for a fstab first entry
#    PARAMETERS:  $1 = partition
#       RETURNS:  valid partition address for fstab
#===============================================================================
hd_uuid_encode_fstab(){
    # pre {{{
    local part uuid orig

    el_debug "\targs '$*'"

    part="${1}"
    orig="${1}"

    # fallback
    if ! echo "$part" | grep -qsE "^(/dev/|UUID=|LABEL=)" ; then
        echo "$orig"
        return 0
    fi

    part="${part#/dev/disk/by-uuid/}"
    part="${part#UUID=}"
    part="$( echo "$part" | sed -e 's|"||g' )"
    el_check_variables "part"

    # }}}
    # this is the prefered option, so just return it
    if [[ "${part}" = "/dev/mapper/"* ]] ; then
        echo "$part"
        return 0
    fi

    # get uuid
    uuid="$( hd_uuid_encode "$part" )"

    if [[ -n "$uuid" ]] ; then
        # return uuid valid entry if we have
        echo "UUID=$uuid"
    else
        # otherwise just return the normal partition entry
        if ! [[ -b "$part" ]] ; then
            el_warning "$FUNCNAME is called with argument $part which is not a valid partition (or block device)"
        fi
        echo "$orig"
    fi
}
#===  FUNCTION  ================================================================
#          NAME:  hd_uuid_encode
#   DESCRIPTION:  get the UUID of a device
#    PARAMETERS:  $1 = device
#       RETURNS:  uuid clean value (without UUID=)
#===============================================================================
hd_uuid_encode(){
    # pre {{{
    local dev uuid part_type errors_result orig

    el_debug "\targs '$*'"

    dev="${1}"
    orig="${1}"

    # fallback
    if ! echo "$dev" | grep -qsE "^(/dev/|UUID=|LABEL=)" ; then
        echo "$orig"
        return 0
    fi

    #dev="${dev#/dev/disk/by-uuid/}"
    dev="$( echo "$dev" | sed -e 's|"||g' )"
    if [[ "$dev" = UUID=* ]] ; then
        dev="${dev#UUID=}"
        if [[ -b "/dev/disk/by-uuid/$dev" ]] ; then
            dev="$( readlink -f "/dev/disk/by-uuid/$dev" )"
        fi
    fi
    if [[ "$dev" = LABEL=* ]] ; then
        dev="${dev#LABEL=}"
        if [[ -b "/dev/disk/by-label/$dev" ]] ; then
            dev="$( readlink -f "/dev/disk/by-label/$dev" )"
        fi
    fi
    el_check_variables "dev"

    if [[ "${dev}" != "/dev/"* ]] && [[ -b "/dev/$dev" ]] ; then
        dev="/dev/$dev"
    fi

    # }}}
    # first check if value is already an uuid
    if [[ -n "$dev" ]] && [[ -b "/dev/disk/by-uuid/$dev" ]] ; then
        uuid="$dev"
        echo "$uuid"
        return 0
    fi

    if ! [[ -b "$dev" ]] ; then
        el_error "$FUNCNAME: $dev is not a block device"
        return 0
    fi

    #uuid="$( LC_ALL="$EL_LC_EN" blkid "$dev" | tr ' ' '\n' | grep "^UUID=" | sed -e 's|^UUID="||g' -e 's|"$||g' )"
    uuid="$( LC_ALL="$EL_LC_EN" lsblk -rino PATH,TYPE,UUID | awk -v item="${dev}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' )"

    # trigger /dev for new changes
    if [[ ! -b "/dev/disk/by-uuid/$uuid" ]] && [[ ! -L "/dev/disk/by-uuid/$uuid" ]] ; then
        device_check_is_busy "$dev"

        uuid="$( LC_ALL="$EL_LC_EN" lsblk -rino NAME,TYPE,UUID | awk -v item="${dev##*/}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' | sort -ug | tail -1 )"

        if [[ ! -b "/dev/disk/by-uuid/$uuid" ]] && [[ ! -L "/dev/disk/by-uuid/$uuid" ]] ; then
            sleep 1
            uuid="$( LC_ALL="$EL_LC_EN" lsblk -d -rino UUID "$dev" | head -1 )"
        fi
    fi

    if [[ ! -b "/dev/disk/by-uuid/$uuid" ]] && [[ ! -L "/dev/disk/by-uuid/$uuid" ]] ; then
        device_check_is_busy "${dev}"

        uuid="$( LC_ALL="$EL_LC_EN" udevadm info --query=all --name="$dev" | grep "ID_FS_UUID=" | sed -e 's|^.*UUID=||g' | grep "[[:alnum:]]" | sed -e '/^$/d' | tail -1 )"

        if [[ -n "$uuid" ]] ; then
            if [[ ! -b "/dev/disk/by-uuid/$uuid" ]] && [[ ! -L "/dev/disk/by-uuid/$uuid" ]] ; then
                if el_check_version_development_is_days_recent 20 && ! el_flag check exit_error_reported ; then
                    el_error "Unable to convert device '$dev' to UUID. Extra info: $(blkid "$dev")"
                fi

                el_warning "$FUNCNAME: uuid is $uuid but it doesn't seems to point to a valid device in /dev/disk/by-uuid ? lists is:"
                file /dev/disk/by-uuid/* 1>&2

                unset uuid
            fi
        fi
    fi

    # maybe a file? (note the -b part which checks for /dev )
    if file "$dev" | grep -qs " UUID=" ; then
        uuid="$( file "$dev" | sed -e 's|^.* UUID=||g' | awk '{print $1}' )"
        # check for valid block device
        if [[ ! -b "/dev/disk/by-uuid/$uuid" ]] && [[ ! -L "/dev/disk/by-uuid/$uuid" ]] ; then
            unset uuid
        fi
    fi

    # maybe a swap?
    if swaplabel "$dev" 2>&1 | grep -qs "^UUID:" ; then
        uuid="$( swaplabel "$dev" 2>&1 | awk '{print $2}' )"
        # check for valid block device
        if [[ ! -b "/dev/disk/by-uuid/$uuid" ]] && [[ ! -L "/dev/disk/by-uuid/$uuid" ]] ; then
            unset uuid
        fi
    fi

    # get it from partitions-list:
    uuid="$( EL_DEBUG=0 partitions-list --show-only="$dev" | awk -v FS="::" '{print $8}' | tail -1 || true )"
    if [[ ! -b "/dev/disk/by-uuid/$uuid" ]] && [[ ! -L "/dev/disk/by-uuid/$uuid" ]] ; then
        if [[ -n "$uuid" ]] ; then
            el_warning "we got uuid '${uuid}' from partitions-list but aparently doesn't exist yet in /dev, lets use it anyways for now: \n$(file /dev/disk/by-uuid/* )"
        fi
    fi

    # partition has no UUID, we must create a new one for it, otherwise becomes unusable
    if [[ -b "$dev" ]] && [[ -z "$uuid" ]] && [[ "$2" != "recursive" ]] ; then
        tune2fs "$dev" -U random 2>/dev/null 1>&2 || true
        sync
        # get the new uuid value
        uuid="$( hd_uuid_encode "$dev" recursive )"
    fi

    # if is a wrong swap partition, reformat it
    if [[ -z "$uuid" ]] ; then
        part_type="$( EL_DEBUG=0 partitions-list --show-only="$dev" | awk -v FS="::" '{print $2}' | tail -1 || true )"

        if [[ "$part_type" = "swap" ]] ; then
            local message_swap_wrong
            message_swap_wrong="$( printf "$( eval_gettext "Your SWAP partition '%s' appears to have issues. Do you want to reformat it?" )" "$dev" )"

            if guitool yesnoquestion "$message_swap_wrong" ; then
                swapoff "$dev" 2>/dev/null || true

                errors_result="$( mkswap -c -f "$dev"  2>/dev/null )"
                if [[ -n "$errors_result" ]] ; then
                    el_warning "errors formating swap?\n$errors_result"
                    # TODO: tell the user about these errors
                fi

                uuid="$( swaplabel "$dev" 2>&1 | awk '{print $2}' )"
                if [[ ! -b "/dev/disk/by-uuid/$uuid" ]] && [[ ! -L "/dev/disk/by-uuid/$uuid" ]] ; then
                    unset uuid
                fi
            fi
        fi
    fi

    if [[ -n "$uuid" ]] ; then
        el_debug "UUID for $dev is $uuid"
        echo "${uuid#UUID=}"
        return 0
    else
        el_error "unable to get UUID for $dev"
        return 1
    fi
}

#===  FUNCTION  ================================================================
#          NAME:  device_is_ssd
#   DESCRIPTION:  check if a given device or partition is SSD type
#    PARAMETERS:  $1 = device or partition
#       RETURNS:  true if is SSD, false if is mechanical
#===============================================================================
device_is_ssd(){
    # pre {{{
    local device alignment_type
    #el_debug "\targs '$*'"

    device="$( get_disk_from_partition_complex "$1" )"

    if ! el_check_variables "device" ; then
        el_error "$FUNCNAME called without device argument"
        return 1
    fi

    # maybe is a symbolic link, we need to correct devname
    if [[ -L "$device" ]] ; then
        device="$( readlink -f "$device" )"
    fi

    if ! [[ -b "$device" ]] ; then
        el_error "$device is not a block device?"
        return 1
    fi

    # new PCIe SSDs devices
    if [[ "$device" = /dev/nvme* ]] || [[ "$device" = /dev/mmcblk* ]] || { [[ "$device" = /dev/md* ]] && ls /dev/md*p* 1>/dev/null 2>&1 ; } ; then
        el_debug "Device $device is SSD over PCIe (nvme) or eMMC, or md"
        return 0
    fi




    # }}}

    # lsblk also supports /dev/dm-*
    alignment_type="$( LC_ALL="$EL_LC_EN" lsblk -rino PATH,ROTA | awk -v item="${dev}" '{ if ($1 == item) print $2}' )"


    if [[ -z "$alignment_type" ]] ; then
        alignment_type="$( cat /sys/block/$(basename ${device})/queue/rotational 2>/dev/null )"
    fi

    if [[ -z "$alignment_type" ]] ; then
        el_error "$device is not a block device with rotational info?"
        return 1
    fi


    case $alignment_type in
        0)
            el_debug "Device $device is SSD"
            return 0
            ;;
        1)
            el_debug "Device $device is a Mechanical disk"
            return 1
            ;;
        *)
            el_warning "Device $device unable to detect if is a SSD or mechanical, from /sys/block/$(basename $device )/queue/rotational "
            return 1
            ;;
    esac
}
#===  FUNCTION  ================================================================
#          NAME:  partition_alignment_check
#   DESCRIPTION:  check if the alignment of the given partition is correct
#    PARAMETERS:  $1 = partition , $2 = report (optional)
#       RETURNS:  true if is, false if is not, it also shows a message optionally
#===============================================================================
partition_alignment_check(){
    # pre {{{
    local partition device partition_number is_reportable
    el_debug "\targs '$*'"

    partition="$1"
    device="$( get_disk_from_partition_complex "$partition" )"
    partition_number="$( get_partition_number_from_partition "$partition" )"

    if [[ "$2" = "report" ]] ; then
        is_reportable=1
    fi

    if ! el_check_variables "partition" ; then
        return 1
    fi

    # maybe is a symbolic link, we need to correct devname
    if [[ -L "$partition" ]] ; then
        partition="$( readlink -f "$partition" )"
    fi

    if ! hd_device_is_blockdevice "$partition" ; then
        el_error "$partition is not a block device"
        return 1
    fi

    # }}}

    # check if is aligned
    if [[ "$( LC_ALL="$EL_LC_EN" blockdev --getalignoff "$partition" )" != 0 ]] ; then
        el_debug "Detected wrong align by the tool 'blockdev --getalignoff PART' (should return zero)"
        is_bad_aligned=1
    fi

    if [[ -n "$partition_number" ]] ; then
        if LC_ALL="$EL_LC_EN" parted "$device" align-check optimal "$partition_number" | grep -qs "not aligned" ; then
            el_debug "Detected wrong align by the tool 'parted DEV align-check optimal PART_NUM'"
            is_bad_aligned=1
        fi
    fi

    if ((is_bad_aligned)) ; then
        if ((is_reportable)) && ! ((ignore_show_suggestions)) ; then
            # seems like this is always the best option to use, even for mechanical disks:
            alignment_type="Mib"

            local message_alignment_suggest
            message_alignment_suggest="$( printf "$( eval_gettext "The selected partition is not optimally aligned, which may result in poor performance.\n\nYou can proceed with the installation, and you likely won't notice any issues. However, it's advised to repartition your hard disk using the %s application and recreate your partitions with the alignment option set to '%s'.\n\nAlternatively, use Elive's automated partitioner, which makes the best assumptions for you." )" "gparted" "$alignment_type" )"

            # report to the user if wanted
            if ! ((conf_is_novice_user)) ; then
                if ((is_migration)) ; then
                    guitool info "$message_alignment_suggest"
                else
                    guitool warning "$message_alignment_suggest"
                fi
            fi
        fi

        # not aligned correctly
        # update: we never used an "if" for this fucntions, this is unneded:
        #return 1
        return 0
    else
        # correctly aligned
        el_explain 2 "$partition is correctly aligned"

        return 0
    fi
}

#===  FUNCTION  ================================================================
#          NAME:  partitioner_failed
#   DESCRIPTION:  reports to the user that the partitioner has failed
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
partitioner_failed(){
    # pre {{{
    el_debug "\targs '$*'"

    # }}}
    el_error "partitioner_failed:\n$( cat "$TMP_LOGS_ERR_f" )\n$(dmesg | grep -iE "(error|${HD_DEV})" | tail -n 10 )"
    local message_error
    message_error="$( printf "$( eval_gettext "The partitioning process failed, possibly due to a small or error-prone hard disk, or because you need to reboot to update the kernel with partitioning changes. It is recommended to have a hard disk larger than 20 GB for automated partitioning. Alternatively, you can partition it manually." )" "" )"

    guitool error "${message_error}\n\nError Message:\n$( cat "$TMP_LOGS_ERR_f" )"

    rm -f "$TMP_LOGS_ERR_f"
    exit_error
}

#===  FUNCTION  ================================================================
#          NAME:  partition_format
#   DESCRIPTION:  formats a partition
#    PARAMETERS:  $1 = dev, $2 = fs, $3 name, $@ opts
#       RETURNS:  -
#===============================================================================
partition_format(){
    # pre {{{
    local dev fs name is_ssd result mkfs_opts
    el_debug "\targs '$*'"
    el_security_function_loop || return 0

    dev="$1"
    shift
    fs="$1"
    shift
    name="$1"
    shift

    if ! el_check_variables "dev|fs|name" ; then
        el_error "not enough parameters to format partition given"
        exit_error
    fi

    guitool working_start

    sync

    if ! hd_device_is_blockdevice "$dev" ; then
        sync ; sleep 1
        hd_refresh "$dev"
        sync ; sleep 1
        if ! hd_device_is_blockdevice "$dev" ; then
            el_error "dev '$dev' doesn't looks like to be a block device"
            guitool error "$( eval_gettext "One of the disks we wanted to format doesn't appears to be a valid device." )"
            exit_error
        fi
    fi

    # }}}

    # check
    device_check_is_busy "$dev"

    if device_is_ssd "$dev" ; then
        is_ssd=1
    fi
    # select optimal compression based in cpu power
    case "$(uname -m)" in
        # zstd is the best option, do not change since we also use the compression level in mount options
        i*86) compressor_algorithm="zstd"   ; ;;
        *64)  compressor_algorithm="zstd"  ; ;;
        *)    compressor_algorithm="zstd"  ; ;;
    esac

    # clean a bit the partition headers first
    dd if=/dev/zero of="$dev" bs=4096 count=2500  2>/dev/null || true

    # format specific filesystems
    case "$fs" in
        reiserfs|reiser3)
            if [[ -s "$TMP_BADBLOCKS_f" ]] ; then
                mkfs.reiserfs -q -l "$name" -B "$TMP_BADBLOCKS_f" -b 4096 "$@" "$dev" 1>/dev/null 2>"$TMP_LOGS_ERR_f" || error_found
            else
                mkfs.reiserfs -q -l "$name" "$@" "$dev" 1>/dev/null 2>"$TMP_LOGS_ERR_f" || error_found
            fi
            is_reiserfs_used=1
            ;;
        reiser4|Reiser4*)
            # note: also supports -B for badblocks skipping
            if ((is_ssd)) ; then
                if ((is_fs_compression_wanted)) ; then
                    # compression options: gzip1, lzo1, zstd
                    # NEW 4.19 options:
                    # -d for SSD, -o node=node41 for checksums support, -o create=reg40 for disable compression, -o create=ccreg40,compress=zstd1 for compression,
                    #mkfs_opts="$mkfs_opts --discard -o node=node41,create=ccreg40,compress=${compressor_algorithm}1"
                    mkfs_opts="$mkfs_opts -o node=node41,create=ccreg40,compress=${compressor_algorithm}1"
                else
                    #mkfs_opts="$mkfs_opts --discard -o node=node41,create=reg40"
                    mkfs_opts="$mkfs_opts -o node=node41,create=reg40"
                fi
            else
                if ((is_fs_compression_wanted)) ; then
                    mkfs_opts="$mkfs_opts -o node=node41,create=ccreg40,compress=${compressor_algorithm}1"
                else
                    mkfs_opts="$mkfs_opts -o node=node41,create=reg40"
                fi
            fi

            mkfs.reiser4 -f -y ${mkfs_opts} -L "$name" "$@" "$dev" 1>/dev/null 2>"$TMP_LOGS_ERR_f" || error_found

            el_array_member_add "$fs" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")
            ;;
        reiser5)
            el_error "reiser5 to implement"
            exit_error
            ;;
        ext2)
            mkfs.ext2 -q -L "$name" "$@" "$dev" 1>/dev/null 2>"$TMP_LOGS_ERR_f" || error_found
            ;;
        ext3)
            #if ((is_ssd)) ; then
                #mkfs_opts="$mkfs_opts -E discard"
            #fi

            mkfs.ext3 -q -L "$name" ${mkfs_opts} "$@" "$dev" 1>/dev/null 2>"$TMP_LOGS_ERR_f" || error_found

            el_debug "changing the max percent of FS reserved to 3 instead of 5"
            tune2fs -m 3 "$dev" 1>/dev/null 2>&1 || true
            ;;
        ext4)
            #if ((is_ssd)) ; then
                #mkfs_opts="$mkfs_opts -E discard"
            #fi

            mkfs.ext4 -q -L "$name" ${mkfs_opts} "$@" "$dev" 1>/dev/null 2>"$TMP_LOGS_ERR_f" || error_found

            el_debug "changing the max percent of FS reserved to 4 instead of 5"
            tune2fs -m 4 "$dev" 1>/dev/null 2>&1 || true

            ;;
        xfs)
            mkfs.xfs -f -q -L "$name" "$@" "$dev" 1>/dev/null 2>"$TMP_LOGS_ERR_f" || error_found
            ;;
        jfs)
            mkfs.jfs -q -L "$name" "$@" "$dev" 1>/dev/null 2>"$TMP_LOGS_ERR_f" || error_found
            ;;
        btrfs)
            # note: do NOT enable discard since is not recommended: https://wiki.debian.org/Btrfs
            # note: supports compression, optimal is zstd, specific directories can be set for compression too!
            # note: recommended options for SD cards or slow SSD's:  /dev/sdaX / btrfs x-systemd.device-timeout=0,noatime,compress=lzo,commit=0,ssd_spread,autodefrag 0 0
            # note: autodefrag is not recommended
            # TODO: check the debian wiki link for get a stats to show possible FS errors for elive-health
            # NOTE: compression happens on mount options, not format
            mkfs.btrfs -f -q -L "$name" "$@" "$dev" 1>/dev/null 2>"$TMP_LOGS_ERR_f" || error_found
            el_array_member_add "$fs" "${modules_extra_initramfs[@]}" ; modules_extra_initramfs=("${_out[@]}")
            ;;
        EFI)
            # format in vfat
            result="$( mkfs.vfat -F 32 "$dev" 2>&1 )"
            if echo "$result" | grep -qsi "Not enough clusters" ; then
                result="$( mkfs.vfat -s2 -F 32 "$dev" 2>&1 )"
                if echo "$result" | grep -qsi "Not enough clusters" ; then
                    result="$( mkfs.vfat -s1 -F 32 "$dev" 2>&1 )"
                    if echo "$result" | grep -qsi "Not enough clusters" ; then
                        el_warning "unable to format correctly the vfat fs for EFI partition"
                        is_failed=1
                    fi
                fi
            fi
            # TODO: avoid formating (prev step) if not wanted
            #el_debug "parted -ms -a optimal \"$( get_disk_from_partition "$dev" )\" name \"$( get_partition_number_from_partition "$dev" )\" \"EFI System Partition\""
            # note: the name needs to be scaped on this way in order to work (buggy parted command?)
            parted -ms -a optimal "$( get_disk_from_partition "$dev" )" name "$( get_partition_number_from_partition "$dev" )" \"EFI\ System\ Partition\"

            #el_debug "parted -ms -a optimal \"$( get_disk_from_partition "$dev" )\" set \"$( get_partition_number_from_partition "$dev" )\" boot on"
            #parted -ms -a optimal "$( get_disk_from_partition "$dev" )" set "$( get_partition_number_from_partition "$dev" )" boot on

            #el_debug "parted -ms -a optimal \"$( get_disk_from_partition "$dev" )\" set \"$( get_partition_number_from_partition "$dev" )\" esp on"
            parted -ms -a optimal "$( get_disk_from_partition "$dev" )" set "$( get_partition_number_from_partition "$dev" )" esp on
            ;;
        BIOS)
            # erase contents for safer results
            if [[ -b "$dev" ]] ; then
                dd if=/dev/zero of=$dev bs=4M 1>/dev/null 2>&1 || true
            fi
            #(parted)name 1 grub
            parted -ms -a optimal "$( get_disk_from_partition "$dev" )" name "$( get_partition_number_from_partition "$dev" )" grub
            #(parted)set 1 boot on
            # note: just in case, but the boot flag will be overwritten:
            #parted -ms -a optimal "$( get_disk_from_partition "$dev" )" set "$( get_partition_number_from_partition "$dev" )" boot on
            #(parted)set 1 bios_grub on
            parted -ms -a optimal "$( get_disk_from_partition "$dev" )" set "$( get_partition_number_from_partition "$dev" )" bios_grub on
            ;;
        swap|SWAP)
            swapoff "$dev" 2>/dev/null || true
            if ! mkswap -f "$dev"  2>/dev/null 1>&2 ; then
                sync ; sleep 1
                if ! mkswap -f "$dev" ; then
                    is_failed=1
                fi
            fi
            ;;
        *)
            el_warning "unknown filesystem selected $fs"
            ;;
    esac

    if ! ((is_failed)) ; then
        el_explain 2 "Formatted '$dev' in '$fs' named '$name' "

        # always mark the partitions formated as used
        el_array_member_add "$dev" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")

        sync ; sleep 2
        hd_refresh "$dev"
    fi
}

#===  FUNCTION  ================================================================
#          NAME:  get_encrypted_partition_from_partition
#   DESCRIPTION:  returns the encrypted partition "layer" from partition, example: /dev/mapper/elive-vg_root will be /dev/sda
#    PARAMETERS:  $1 = part
#       RETURNS:  true if is from an encrypted layer, false if not
#===============================================================================
get_encrypted_partition_from_partition(){
    # pre {{{
    local part encrypted_layer
    el_debug "\targs '$*'"
    el_security_function_loop || return 0

    part="$1"

    el_check_variables "part"
    if ! [[ -b "$part" ]] ; then
        el_error "$FUNCTION called without a block paramter: $part - ${FUNCNAME[1]}"
        return 1
    fi

    # }}}
    # cryptsetup directly partition
    if LC_ALL="$EL_LC_EN" cryptsetup status "$part" 2>/dev/null | grep -qs "type:.*LUKS" ; then
        el_debug "Partition '$part' is directly an encrypted layer"
        echo "$part"
        return 0
    fi

    # lvs layer
    # know the VG group name for this partition
    vg_name="$( LC_ALL="$EL_LC_EN" LVM_SUPRESS_FD_WARNINGS=1 lvs --noheadings --separator : -o vg_name "$part" 2>/dev/null )"
    if [[ -n "$vg_name" ]] ; then
        while read -ru 3 line
        do
            # get a PV assigned disk for this name
            if [[ "$line" = "PV Name"* ]] ; then
                pv_name="$( echo "$line" | sed -e 's|^.*/dev/|/dev/|g' )"
            fi

            # if the previosly PV assigned name matches to our wanted VG group (partition), finish here
            if [[ "$line" = "VG Name"*${vg_name} ]] ; then
                break
            fi

        done 3<<< "$( LC_ALL="$EL_LC_EN" pvdisplay -m 2>/dev/null )"

        # so if we have got it, check if comes from an encrypted layer
        if [[ -n "$pv_name" ]] ; then
            if LC_ALL="$EL_LC_EN" cryptsetup status "$pv_name" 2>/dev/null | grep -qs "type:.*LUKS" ; then
                encrypted_layer="$( LC_ALL="$EL_LC_EN" cryptsetup status "$pv_name" | grep "device:" | sed -e 's|^.*device:||g' | tail -1 )"
                # remove extra leading blank chars
                read -r encrypted_layer <<< "$encrypted_layer"

                el_debug "Partition '$part' comes from an encrypted layer over: $encrypted_layer"
                echo "$encrypted_layer"
                return 0
            fi
        fi
    fi

    # default mode, not encrypted
    el_debug "Partition $part doesn't looks like to be assigned to an encrypted layer"

}
#===  FUNCTION  ================================================================
#          NAME:  check_part_is_from_encrypted
#   DESCRIPTION:  checks if a partition comes from an encrypted layer
#    PARAMETERS:  $1 = part
#       RETURNS:  true if is from an encrypted layer, false if not
#===============================================================================
check_part_is_from_encrypted(){
    # pre {{{
    local part
    el_debug "\targs '$*'"
    el_security_function_loop || return 0

    part="$1"

    el_check_variables "part"
    if ! hd_device_is_blockdevice "$part" ; then
        el_error "$FUNCTION called without a block paramter: $part - ${FUNCNAME[1]}"
        return 1
    fi

    # }}}
    # cryptsetup directly partition
    if LC_ALL="$EL_LC_EN" cryptsetup status "$part" 2>/dev/null | grep -qs "type:.*LUKS" ; then
        el_debug "Partition '$part' is directly an encrypted layer"
        return 0
    fi

    # lvs layer
    # know the VG group name for this partition
    vg_name="$( LC_ALL="$EL_LC_EN" LVM_SUPRESS_FD_WARNINGS=1 lvs --noheadings --separator : -o vg_name "$part" 2>/dev/null )"
    if [[ -n "$vg_name" ]] ; then
        while read -ru 3 line
        do
            # get a PV assigned disk for this name
            if [[ "$line" = "PV Name"* ]] ; then
                pv_name="$( echo "$line" | sed -e 's|^.*/dev/|/dev/|g' )"
            fi

            # if the previosly PV assigned name matches to our wanted VG group (partition), finish here
            if [[ "$line" = "VG Name"*${vg_name} ]] ; then
                break
            fi

        done 3<<< "$( LC_ALL="$EL_LC_EN" pvdisplay -m 2>/dev/null )"

        # so if we have got it, check if comes from an encrypted layer
        if [[ -n "$pv_name" ]] ; then
            if LC_ALL="$EL_LC_EN" cryptsetup status "$pv_name" 2>/dev/null | grep -qs "type:.*LUKS" ; then
                el_debug "Partition '$part' comes from an encrypted layer over: $( LC_ALL="$EL_LC_EN" cryptsetup status "$pv_name" | grep "device:" | sed -e 's|^.*device:||g' -e 's| ||g' -e 's| ||g' -e 's| ||g' | tail -1 )"
                return 0
            fi
        fi
    fi

    # default mode, not encrypted
    el_debug "Partition $part doesn't looks like to be assigned to an encrypted layer"
    return 1

}
#===  FUNCTION  ================================================================
#          NAME:  efi_select
#   DESCRIPTION:  select the efi partition which needs to be installed elive
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
efi_select(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"
    local answer efi_partition

    # booted not in UEFI mode
    if ! ((is_efi_booted)) ; then
        return 0
    fi

    # }}}
    # for betatests:
    #el_array_member_add "/dev/sda6" "${efi_partitions_maybe[@]}" ; efi_partitions_maybe=("${_out[@]}")
    #el_array_member_add "/dev/sda1" "${efi_partitions[@]}" ; efi_partitions=("${_out[@]}")
    #el_array_member_add "/dev/sda2" "${efi_partitions[@]}" ; efi_partitions=("${_out[@]}")

    if [[ -b "$HD_PART_EFI_dev" ]] ; then
        answer="$HD_PART_EFI_dev"
    else
        if [[ -b "$HD_DEV" ]] ; then
            # get list of available efi partitions
            efi_detect_partition "$HD_DEV"

            # if we have only one partition in all the system:
            if [[ -z "${efi_partitions[*]}" ]] && [[ -n "${efi_partitions_maybe[*]}" ]] && [[ "${#efi_partitions_maybe[@]}" -eq 1 ]] ; then
                el_array_member_add "${efi_partitions_maybe[*]}" "${efi_partitions[@]}" ; efi_partitions=("${_out[@]}")
            fi

            # only one efi partition found?
            if [[ "${#efi_partitions[@]}" -eq 1 ]] ; then
                efi_partition="${efi_partitions[*]}"

                # check that we don't have other possible partitions with efi
                if [[ "${#efi_partitions_maybe[@]}" = 1 ]] && el_array_member_check "${efi_partition}" "${efi_partitions_maybe[@]}" ; then
                    # same? unlist it
                    unset efi_partitions_maybe
                fi

                if [[ "${#efi_partitions_maybe[@]}" -ge 1 ]] ; then
                    el_explain 0 "we have only one efi partition '${efi_partition}' in our disk '${HD_DEV}', but we have other efi partitions too '${efi_partitions_maybe[*]}' "
                    unset efi_partition
                else
                    el_explain 0 "only one partition for EFI found, using it:  '${efi_partition}' "
                fi
            fi

            # select which partition where to install EFI
            # TODO: improve with a better "description" structure (generate a menu), showing which OSes are shown inside
            if [[ -n "$efi_partition" ]] ; then
                answer="$efi_partition"
            else
                if ! [[ -n "${efi_partitions[*]}" ]] && ! [[ -n "${efi_partitions_maybe[*]}" ]] ; then
                    # no partitions found, let's wait the user adds it before to asking for it
                    return 0
                else
                    unset menu
                    while read -ru 3 entry
                    do
                        unset size name
                        if el_array_member_check "$entry" "${HD_PARTITIONS_USED_array[@]}" || el_array_member_check "$part" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; then
                            el_debug "${entry} already in our array of partitions used, skipping it"
                            continue
                        fi

                        size="$( LC_ALL="$EL_LC_EN" lsblk -rino PATH,TYPE,SIZE | awk -v item="${entry}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' )"
                        [[ -z "$size" ]] && size="(unknown)"

                        name="$( udevadm info --query=property --name="$entry" | grep "ID_VENDOR_ID=" | sed -e 's|^.*VENDOR_ID=||g' | tail -1 )"
                        [[ -z "$name" ]] && name="$( udevadm info --query=property --name="$entry" | grep "ID_MODEL=" | sed -e 's|^.*ID_MODEL=||g' | tail -1 )"
                        [[ -z "$name" ]] && name="$( LC_ALL="$EL_LC_EN" lsblk -rino PATH,TYPE,LABEL | awk -v item="${entry}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' )"
                        [[ -z "$name" ]] && name="partition"

                        menu+=("$( echo "$entry" )")
                        [[ -n "$name" ]] && menu+=("$( echo "$name" )") || menu+=("$( echo "-" )")
                        [[ -n "$size" ]] && menu+=("$( echo "$size" )") || menu+=("$( echo "-" )")
                    done 3<<< "$( echo "${efi_partitions[@]}" "${efi_partitions_maybe[@]}" | tr ' ' '\n' | sort -u )"

                    answer="$( $guitool --title='Elive Installer' --list --height=300 --width=400 --text="$( eval_gettext "Choose the EFI partition from which you boot your systems." )" --column="$( eval_gettext "Partition" )" --column="$( eval_gettext "Name" )" --column="$( eval_gettext "Size" )"  "${menu[@]}"  2>/dev/null || echo cancel )"
                fi
            fi
        else
            el_error "we dont have HD_DEV set? '${HD_DEV}' "
        fi
    fi

    if [[ "$answer" = "cancel" ]] ; then
        el_warning "user selected to NOT use an efi partition from the given list"
        guitool warning "$( eval_gettext "You booted the computer in UEFI mode, but no EFI partitions were found in your system. You need to create one to boot your operating systems." )"
    fi

    if [[ -n "$answer" ]] && [[ -b "$answer" ]] ; then
        HD_PART_EFI_dev="$answer"
        HD_PART_EFI_FS_v="vfat" # always should be fat32
        el_explain 0 "partition where to install EFI is '${HD_PART_EFI_dev}' "
        el_array_member_add "$HD_PART_EFI_dev" "${HD_PARTITIONS_USED_array[@]}" ; HD_PARTITIONS_USED_array=("${_out[@]}")

        # always force this conf or it will fail
        mkdir -p "$TARGET/root/preseeds.d"
        if [[ -n "$HD_PART_EFI_dev" ]] ; then
            if ((is_efi_64)) ; then
                rm -f "$TARGET/root/preseeds.d/grub-efi-amd64"
                echo -e "grub-efi-amd64\tgrub2/force_efi_extra_removable\tboolean\ttrue" >> "$TARGET/root/preseeds.d/grub-efi-amd64"
            fi
            if ((is_efi_32)) ; then
                rm -f "$TARGET/root/preseeds.d/grub-efi-ia32"
                echo -e "grub-efi-ia32\tgrub2/force_efi_extra_removable\tboolean\ttrue" >> "$TARGET/root/preseeds.d/grub-efi-ia32"
            fi
        else
            echo -e "grub-pc\tgrub2/force_efi_extra_removable\tboolean\ttrue" >> "$TARGET/root/preseeds.d/grub-pc"
        fi
    else
        guitool warning "$( eval_gettext "You booted the computer in UEFI mode, but no EFI partitions were found in your system. You need to create one to boot your operating systems." )"
        el_warning "No EFI partitions found, and computer booted in UEFI mode:\n$( EL_DEBUG=0 partitions-list --show-all | psort -- -p "vfat" || true )"
    fi
}




# TOOLS extracted from grub
hdtools_available_ids(){
    local id path

    [ -d /dev/disk/by-id ] || return 0
    cached_available_ids="$(
    for path in /dev/disk/by-id/*; do
        [ -e "$path" ] || continue
        printf '%s %s\n' "$path" "$(readlink -f "$path")"
    done | sort -k2 -s -u | cut -d' ' -f1
    )"
    echo "$cached_available_ids"
}

# Returns non-zero and no output if no mapping can be found.
hdtools_device_to_id(){
    local id

    for id in $(hdtools_available_ids); do
        [[ -z "$id" ]] && continue
        if [ "$(readlink -f "$id")" = "$(readlink -f "$1")" ]; then
            echo "$id"
            el_explain 0 "ID for disk $1 is $id"
            return 0
        fi
    done

    # Fall back to the plain device name if theres no by-id link for it.
    if [ -e "$1" ]; then
        echo "$1"
        if ! ((is_virtualized)) ; then
            el_warning "Unable to get ID disk for ${1} in ${FUNCNAME}, fallback to plain device"
        fi
        return 0
    fi

    el_error "No device ('${1}') passed to get ID from disk in $FUNCNAME"
}

hdtools_devices_to_ids(){
    local device id ids
    ids=
    for device; do
        id="$( hdtools_device_to_id "$device" )"
        [[ -z "$id" ]] && continue
        if [ "$id" ]; then
            ids="${ids:+$ids, }$id"
        fi
    done

    if [[ -n "$ids" ]] ; then
        echo "$ids"
    else
        el_error "No ID values get in $FUNCNAME : $ids"
    fi
}

hdtools_all_disks(){
    local id
    for id in $(hdtools_available_ids); do
        [[ -z "$id" ]] && continue
        case $id in
            *-part*)
                true
                ;;
            *)
                echo "$id"
                ;;
        esac
    done
}

hdtools_get_fs(){
    local dev filesystem ret
    dev="$1"

    if [[ -z "$dev" ]] || [[ "$dev" = "#"* ]] ; then
        el_error "argument passed to $FUNCNAME is not valid: '$dev'"
        return 0
    fi

    # use lsblk
    filesystem="$( LC_ALL="$EL_LC_EN" lsblk -rino PATH,TYPE,FSTYPE | awk -v item="${dev}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' | sort -ug | tail -1 )"

    if [[ -n "$filesystem" ]] ; then
        el_debug "detected filesystem for $dev with lsblk: $filesystem"
        echo "$filesystem"
        return 0
    fi

    # use udevadm
    filesystem="$( LC_ALL="$EL_LC_EN" udevadm info --query=all --name="$dev" | grep -F "ID_FS_TYPE=" | sed -e 's|^.*TYPE=||g' | grep "[[:alnum:]]" | sed -e '/^$/d' | tail -1 )"
    if [[ -z "$filesystem" ]] ; then
        filesystem="$( LC_ALL="$EL_LC_EN" udevadm info --query=all --name="$dev" | grep -F "ID_FS_TYPE=" | sed -e 's|^.*TYPE=||g' | grep "[[:alnum:]]" | sed -e '/^$/d' | tail -1 )"
    fi

    if [[ -n "$filesystem" ]] ; then
        el_debug "detected filesystem for $dev with udevadm: $filesystem"
        echo "$filesystem"
        return 0
    fi

    # use file
    ret="$(file -Ls "$dev" )"
    case "$ret" in
        *HFS*|*hfs*)
            echo "hfsplus"
            ;;
        *"x86 boot sector"*"0xee"*|*"DOS/MBR boot sector"*"0xee"*)
            # this is raw / entire disks
            echo "disk"
            ;;
        *extended\ partition\ table*|*"ID=0x5"*)
            # Note: I have a result like this which looks to be wrong, it is not an extended partition, but in any case is not an usable one:  /dev/sdb: sticky x86 boot sector; partition 1: ID=0xee, starthead 0, startsector 1, 1953525167 sectors, extended partition table (last)\011, code offset 0x0
            # Note: ID=0x5 are represented for extended partitions over a GPT structure, but also are considered /dev/sda as disks, in any case since they are "layers" but not partitions is better to not have them showing up
            echo "extended"
            ;;
        *ext3*)
            echo "ext3"
            ;;
        *ext2*)
            echo "ext2"
            ;;
        *"ReiserFS V3"*)
            echo "reiserfs"
            ;;
        *XFS*)
            echo "xfs"
            ;;
        *ntfs*|*NTFS*|ntfs-3g)
            echo "ntfs-3g"
            ;;
        *vfat*|*FAT*)
            echo "vfat"
            ;;
        *swap*)
            echo "swap"
            ;;
        *btrfs*|*BTRFS*)
            echo "btrfs"
            ;;
        *)
            if [[ -x "$(which fdisk || true )" ]] ; then
                if LC_ALL=C fdisk -l "${dev%%[0-9]*}" 2>/dev/null | grep -qs "^${dev}.*BIOS boot" ; then
                    echo "BIOS"
                else
                    if LC_ALL=C fdisk -l "${dev%%[0-9]*}" 2>/dev/null | grep -qsE "^${dev}\s+.*Extended$" ; then
                        echo "extended"
                    else
                        el_debug "unknown FS for $dev"
                    fi
                    #echo "auto"
                fi
            fi
            ;;
    esac

    if [[ -n "$filesystem" ]] ; then
        el_debug "detected filesystem for $dev with file -s: $filesystem"
        echo "$filesystem"
        return 0
    fi


    # use sfidks
    if [[ "$dev" = /dev/nvme* ]] ; then
        ret="$(sfdisk ${sfdisk_print_id} $( echo "$dev" | sed 's/n[0-9]/ &/') 2>/dev/null )"
        if [[ -z "$ret" ]] ; then
            ret="$(sfdisk ${sfdisk_print_id} $( echo "$dev" | sed -e 's/n[0-9]/ &/' -e 's| p| |g' ) 2>/dev/null )"
        fi
    else
        if [[ "$dev" = /dev/mmcblk* ]] ; then
            ret="$(sfdisk ${sfdisk_print_id} $( echo "$dev" | sed 's/p[0-9]/ &/') 2>/dev/null )"
            if [[ -z "$ret" ]] ; then
                ret="$(sfdisk ${sfdisk_print_id} $( echo "$dev" | sed -e 's/p[0-9]/ &/' -e 's| p| |g' ) 2>/dev/null )"
            fi
        else
            if [[ "$dev" = /dev/md* ]] ; then
                ret="$(sfdisk ${sfdisk_print_id} $( echo "$dev" | sed 's/p[0-9]/ &/') 2>/dev/null )"
                if [[ -z "$ret" ]] ; then
                    ret="$(sfdisk ${sfdisk_print_id} $( echo "$dev" | sed -e 's/p[0-9]/ &/' -e 's| p| |g' ) 2>/dev/null )"
                fi
            else
                ret="$(sfdisk ${sfdisk_print_id} $( echo "$dev" | sed 's/[0-9]/ &/') 2>/dev/null )"
            fi
        fi
    fi

    # force mode
    if [[ -z "$ret" ]] ; then
        if [[ "$dev" = /dev/nvme* ]] ; then
            ret="$(sfdisk --force ${sfdisk_print_id} $( echo "$dev" | sed 's/n[0-9]/ &/') 2>/dev/null )"
            if [[ -z "$ret" ]] ; then
                ret="$(sfdisk --force ${sfdisk_print_id} $( echo "$dev" | sed -e 's/n[0-9]/ &/' -e 's| p| |g' ) 2>/dev/null )"
            fi
        else
            if [[ "$dev" = /dev/mmcblk* ]] ; then
                ret="$(sfdisk --force ${sfdisk_print_id} $( echo "$dev" | sed 's/p[0-9]/ &/') 2>/dev/null )"
                if [[ -z "$ret" ]] ; then
                    ret="$(sfdisk --force ${sfdisk_print_id} $( echo "$dev" | sed -e 's/p[0-9]/ &/' -e 's| p| |g' ) 2>/dev/null )"
                fi
            else
                if [[ "$dev" = /dev/md* ]] ; then
                    ret="$(sfdisk ${sfdisk_print_id} $( echo "$dev" | sed 's/p[0-9]/ &/') 2>/dev/null )"
                    if [[ -z "$ret" ]] ; then
                        ret="$(sfdisk ${sfdisk_print_id} $( echo "$dev" | sed -e 's/p[0-9]/ &/' -e 's| p| |g' ) 2>/dev/null )"
                    fi
                else
                    ret="$(sfdisk --force ${sfdisk_print_id} $( echo "$dev" | sed 's/[0-9]/ &/') 2>/dev/null )"
                fi
            fi
        fi
    fi

    case "$ret" in
        1|4|6|b|c|e|14|16|1b|1c|1e|ef)
            filesytem="vfat"
            ;;
        7|17|86|87)
            filesytem="ntfs-3g"
            ;;
        #5|f|85|ee|ef|fd)
            #filesytem="blacklisted"
            #;;
        #ef)
            #filesytem="efi"
            #;;
        af)
            filesytem="hfsplus"
            ;;
        #*)
            #filesytem="auto"
            #;;
    esac

    if [[ -z "$filesytem" ]] ; then
        filesystem="auto"
    fi

    echo "$filesystem"
    return
}


#===  FUNCTION  ================================================================
#          NAME:  device_check_is_busy
#   DESCRIPTION:  check if the device is busy and suggest to reboot in case of
#    PARAMETERS:  $1 = HD dev
#       RETURNS:  0 | 1
#===============================================================================
device_check_is_busy(){
    # pre {{{
    local disk message_busy part part_swap mountpoint
    #el_debug "\targs '$*'"

    # note: disk is not a disk but can be a partition too
    disk="$1"
    sync

    el_check_variables "disk"

    if ! hd_device_is_blockdevice "$disk" ; then
        el_error "$FUNCNAME called with a non-block-device parameter?: $disk"
        return 0
    fi

    # }}}

    # whitelist skipping {{{

    # ignore USB disk, it will always show as busy and we dont want to use it
    if [[ "$LIVE_MOUNT_MEDIA_DEV_v" = "$disk" ]] ; then
        el_debug "skipping check of device busy because disk is the USB live itself"
        return 0
    fi

    # soft skip for disks that uses swapfiles
    for i in "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}"
    do
        if [[ "$disk" = "$( get_disk_from_partition_complex "$i" )" ]] ; then
            el_debug "skipping check of device busy because disk is used by a temporal swapfile"
            return 0
        fi
    done

    # ignore the busy state if we -want- to install on a LVM2 / cryptsetup system, so, we used /dev/mapper to select something:
    if [[ "${disk}" = "/dev/mapper/"* ]] || \
        ((is_encryption_wanted)) || \
        ((is_fs_mapper)) || \
        [[ "${HD_PART_ROOT_dev}" = "/dev/mapper/"* ]] || \
        [[ "${HD_PART_BOOT_dev}" = "/dev/mapper/"* ]] || \
        [[ "${HD_PART_HOME_dev}" = "/dev/mapper/"* ]] || \
        [[ "${HD_PART_VAR_dev}" = "/dev/mapper/"* ]] ; then

        el_debug "skipping check of device busy because the disk is wanted to be used for /dev/mapper/something"
        return 0
    fi

    if LC_ALL="$EL_LC_EN" mount | grep -qs -- "${disk}.*on ${TARGET}" ; then
        if [[ "${EL_DEBUG}" -ge "3" ]] ; then
            el_warning "Installer tried to check if device is busy after to have mounted it, which is normal, so skipping the check because everything is as expected to be. Funcname is ${FUNCNAME[1]} <- ${FUNCNAME[2]} <- ${FUNCNAME[3]}"
        fi

        # report that is OK, because is normal that is busy (duh!)
        return 0
    fi


    # }}}

    # ignore when we are using one of their swap partitions {{{
    if [[ "$( ls -1 "/dev/mapper/" | grep -vE "^(control|ventoy)$" | wc -l )" -eq 0 ]] ; then
        # if is disk:
        if [[ "$disk" = "$( get_disk_from_partition_complex "$disk" )" ]] ; then
            if swap_partitions_all_do "check" ; then

                for part_swap in $( swapon --noheadings --raw | grep "^/dev/" | awk '{print $1}' | tr '\n' ' ' )
                do
                    if [[ -b "$part_swap" ]] ; then
                        if [[ "$( get_disk_from_partition_complex "$part_swap" )" = "$( get_disk_from_partition_complex "$disk" )" ]] ; then
                            el_debug "skipping check of device busy because we are using a swap partition from the same disk"
                            return 0
                        fi
                    fi
                done
            fi
        fi
    fi

    # }}}

    if LC_ALL="$EL_LC_EN" blockdev --rereadpt "$disk" 2>&1 | grep -qs "BLKRRPART: Device or resource busy" ; then
        partprobe "$disk" || true
        sync ; sleep 1

        # if the partition is busy because we are using it on encrypted layer
        if LC_ALL="$EL_LC_EN" blockdev --rereadpt "$disk" 2>&1 | grep -qs "BLKRRPART: Device or resource busy" ; then
            #if partitions-list --show-raw --show-only="$disk" | grep -qs "::crypto_LUKS::" ; then
            if check_part_is_from_encrypted "$disk" ; then
                el_warning "TEMPORAL: dmsetup remove $disk:"
                dmsetup remove "$disk" 2>/dev/null || true
            fi
        fi


        # if is mounted then we are having a terminal opened on it or something similiar, just kill it
        if LC_ALL="$EL_LC_EN" blockdev --rereadpt "$disk" 2>&1 | grep -qs "BLKRRPART: Device or resource busy" ; then
            if LC_ALL="$EL_LC_EN" mount | grep -qs "^$disk" ; then
                el_debug "going to kill possible processes to free disk '$disk': $(mount | grep "^$disk" )"

                for part in $( LC_ALL="$EL_LC_EN" mount | grep -- "^${disk}" | sed -e 's| on .*$||g' )
                do
                    if [[ -b "$part" ]] ; then
                        #EL_DEBUG=0 mounts-manager --get-mountpoint "$part"
                        # umount first any ntfs one
                        if [[ "$( EL_DEBUG=0 partitions-list --show-only="$part" | awk -v FS='::' '{ if ($3 == "ntfs") print $1}' )" = "ntfs" ]] ; then
                            el_umount "$part"
                        fi
                        # kill the processes that are using the directory (a thunar or terminal opened on it?)
                        mountpoint="$( EL_DEBUG=0 mounts-manager --get-mountpoint "$part" )"
                        if [[ -d "$mountpoint" ]] ; then
                            el_processes_using_mountpoint "$( EL_DEBUG=0 mounts-manager --get-mountpoint "$part" )" kill || true
                        fi
                    fi
                done

                sync
                hd_refresh "$disk"
                sync ; sleep 2
            fi
        fi


        if ! ((is_fs_mapper)) || ! ((is_encryption_wanted)) ; then
            if LC_ALL="$EL_LC_EN" blockdev --rereadpt "$disk" 2>&1 | grep -qs "BLK.*PART: Device or resource busy" ; then
                # disable all dmssetup mapped devices and try again
                # WAIT!!! I understand now: the device looks busy (/dev/sda) because dmsetup (lvm2 or luks for example) uses the device for their sub-systems, so its NORMAL and correct that is busy! so if you enable next line your lvm devices would dissappear. In short: removing the devices is -not- the way to go and this is correct... so the correct is to not show the message at all :P, or ask the user if wants to remove them in order to use them (because he may want to install on disk itself and not lvm stuff
                # note: this also can happen you have it MOUNTED, so on such case simply UMOUNT it to remove the error (we should not check if mounted)

                # disable all /dev/mapper/ (only if user dont want to use it, so ask)
                if [[ "$( ls -1 "/dev/mapper/" | grep -vE "^(control|ventoy)$" | wc -l )" -gt 0 ]] ; then
                    #el_warning "We should not have the disk busy's, is there anything we missed to umount?:\n$(ls -1 /dev/mapper/)"

                    if guitool yesnoquestion "$( eval_gettext "Your Hard Disk '$disk' seems to be busy due to a LVM2 or LUKS structure, do you want to deactivate it? Suggested if you have not configured by yourself this disk structures. It will also avoid the installer to possibly fail." )" ; then

                        dmsetup remove_all || true
                        sync ; sleep 1
                        if [[ "$( ls -1 /dev/mapper/ | grep -vE "(control|ventoy)" | wc -l )" -gt 0 ]] ; then
                            el_warning "We should not have the disk busy's, is there anything we missed to umount? - AFTER 'dmsetup remove_all':\n$(ls -1 /dev/mapper/)"
                        fi
                    fi
                fi
            fi
        fi


        # Ignore special cases, at this point:
        #  if we are installing into an extra partition of ventoy:
        if [[ -e "/dev/mapper/ventoy" ]] ; then
            return 0
        fi


        if LC_ALL="$EL_LC_EN" blockdev --rereadpt "$disk" 2>&1 | grep -qs "BLK.*PART: Device or resource busy" ; then
            # still fails? show some debug info:
            hd_refresh "$disk"
            el_warning "Device looks busy:  blockdev --rereadpt '$disk'"

            # more debug info
            if [[ "$EL_DEBUG" -gt 2 ]] ; then
                echo -e "ls /dev/mapper:"
                ls -1 /dev/mapper/ 2>&1
                echo

                echo -e "swaps:"
                swapon -s | grep  "^/"
                echo

                echo -e "mounts:"
                mount | grep "$part"
                mount | tail -n 4
                echo

                EL_DEBUG=0 partitions-list --show-all --show-raw --show-disks || true
                mount | grep "$disk"
                mountpoint="$( EL_DEBUG=0 mounts-manager --get-mountpoint "$part" )"
                if [[ -d "$mountpoint" ]] ; then
                    el_processes_using_mountpoint "$( EL_DEBUG=0 mounts-manager --get-mountpoint "$part" )"
                fi
            fi

            el_explain 3 "Note: this can be a __false-positive__ if the device is really used for what we need to, like lvm2 or luks"

            message_busy="$( printf "$( eval_gettext "The partitions on your hard disk '%s' have been changed, but the kernel was not updated accordingly. Please reboot the computer for the kernel to read the new structure. This situation should never occur. We suggest reporting this issue to Elive to be fixed. Exit now?" )" "$disk" )"
            el_error "parent function to $FUNCNAME: ${FUNCNAME[1]} ${FUNCNAME[2]}"

            # show the message in terminal
            blockdev --rereadpt "$disk"

            if guitool yesnoquestion "$message_busy" ; then
                exit 0
            fi
        fi
    fi

    return 0
}

#########################################################################
bad_integrity_signature(){
    # pre {{{
    functions_pre
    el_debug "\targs '$*'"

    killall $(basename ${guitool%% *}) 2>/dev/null || true
    sleep 2
    # - pre }}}
    el_error "func: bad_integrity_signature"
    guitool error "$( eval_gettext "Your Elive system appears to be damaged or has been altered. Please try again with a new download and write it to a different medium (USB or DVD). Aborting the installation..." )"
    exit_error
}

#########################################################################

#===  FUNCTION  ================================================================
#          NAME:  installer_requires_newer_kernel_version_than
#   DESCRIPTION:  Warns to the user that this version of hte installer needs a newer version of the kernel
#    PARAMETERS:  version of kernel (uname -r), $2 = exit|continue|silent
#===============================================================================
installer_requires_newer_kernel_version_than(){
    # pre {{{
    local version mode
    #el_debug "\targs '$*'"

    version="$1"
    mode="$2"
    el_check_variables "version"

    # }}}
    if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$(uname -r)" gt "$version" ; then
        return 0
    else
        local message_newer_version_needed
        message_newer_version_needed="$( printf "$( eval_gettext "You are attempting to use a feature in the installer that requires a newer kernel version. If installation fails, reboot and select the highest available kernel version. If none are suitable, download a newer version of Elive." )" )"

        if [[ "${mode}" != "silent" ]] ; then
            guitool warning "$message_newer_version_needed"
        fi

        # exit if needed
        case "$mode" in
            exit)
                exit
                ;;
        esac

        # result is wrong, so return it
        return 1
    fi
}
#===  FUNCTION  ================================================================
#          NAME:  installer_requires_newer_elive_version_than
#   DESCRIPTION:  Warns to the user that this version of hte installer needs a newer version of elive
#    PARAMETERS:  version of elive (like 2.5.6), $2 = exit|continue|silent
#       RETURNS:  -
#===============================================================================
installer_requires_newer_elive_version_than(){
    # pre {{{
    local version
    el_debug "\targs '$*'"

    version="$1"
    mode="$2"
    el_check_variables "version"

    # }}}
    if LC_ALL="$EL_LC_EN" dpkg --compare-versions "$eliveversion" lt "$version" ; then
        local message_newer_version_needed
        message_newer_version_needed="$( printf "$( eval_gettext "You are attempting to use a feature in the installer that requires a newer version of Elive. Installation may fail; if so, download a newer version of Elive and try again." )" )"

        if [[ "${mode}" != "silent" ]] ; then
            guitool warning "$message_newer_version_needed"
        fi

        # exit if needed
        case "$mode" in
            exit)
                exit
                ;;
        esac
    fi
}

#===  FUNCTION  ================================================================
#          NAME:  connection_average_speed_get
#   DESCRIPTION:  get the average speed of our connection to internet, saves result to file
#       RETURNS:  nothing, its saved to a file (so that you can run it in bg)
#===============================================================================
connection_average_speed_get(){
    # pre {{{
    el_debug "\targs '$*'"
    local result

    if ! el_verify_internet 2>/dev/null ; then
        return 0
    fi

    # already running, skip
    if el_flag check  connection_average_speed ; then
        return 0
    fi

    el_flag add  connection_average_speed


    # }}}

    # this would download a dummy file of 10 MB
    # note: the time needed for download 10 MB (11 aparently) at 50 kb/s is 3 min 45 seconds
    result="$( LC_ALL="$EL_LC_EN" wget --timeout=600 -O /dev/null https://github.com/emodyz/SpeedTest/raw/master/10MB.bin  2>&1 | grep "saved" | sed -e 's|^.* (||g' -e 's|).*$||g' )"

    if [[ -n "$result" ]] ; then
        # remove extra leading blank chars
        read -r result <<< "$result"

        case "$result" in
            *GB/s)
                result="$( echo "${result%% *} * 1024 * 1024" | LC_ALL="$EL_LC_EN" bc -l | sed -e 's|\..*$||g' ) KB/s"
                ;;
            *MB/s)
                result="$( echo "${result%% *} * 1024" | LC_ALL="$EL_LC_EN" bc -l | sed -e 's|\..*$||g' ) KB/s"
                ;;
            *KB/s)
                result="$( echo "${result%% *} + 0 " | LC_ALL="$EL_LC_EN" bc -l | sed -e 's|\..*$||g' ) KB/s"
                ;;
            *)
                # this should be bytes :/
                result="$( echo "${result%% *} / 1024" | LC_ALL="$EL_LC_EN" bc -l | sed -e 's|\..*$||g' ) KB/s"
                ;;
        esac

        echo "$result" > "/tmp/.connection-speed-average.txt"
    else
        el_warning "Unable to download the dummy file to know connection's speed, ignoring..."
    fi

    el_flag remove  connection_average_speed



}


filesystems_help_show(){
    if ((ignore_show_helpers)) ; then
        return 0
    fi
    if ! ((conf_is_extra_help_wanted)) ; then
        return 0
    fi

    local message_HELP_FS
    message_HELP_FS="$( printf "$( eval_gettext "ReiserFS (3) is the preferred filesystem for old computers and mechanical hard disks because it is very light and fast, and it is also very resilient against forced reboots, keeping your data safe if your computer fails.\n\nReiser4 is a new and experimental filesystem with features like real-time compression, which is good for computers with very small hard disks.\n\nEXT4 has powerful features for SSD non-mechanical hard disks and is the standard and suggested option. EXT versions 3 and 2 are too old to be considered good options.\n\nJFS and XFS are other filesystems, but it is not suggested to use them.\n\nRemember that the only way to ensure data safety is to make backups, and there's always a possibility that your hardware breaks or gets stolen." )" "" )"



    #echo -e "$HELP_FS" | $guitool --title='Elive Installer' --text-info --editable --width=500 --height=470 1>/dev/null &
    echo -e "$message_HELP_FS" | $guitool --title='Elive Installer' --text-info 1>/dev/null 2>&1 &
    help_fs_pid=$!
    sync ; sleep 1
}

filesystems_help_unshow(){
    kill "$help_fs_pid" 2>/dev/null 1>/dev/null || true
}

nvme_check_percent(){
    local percentage_used

    if [[ -b "$1" ]] && [[ "$1" = /dev/nvme* ]] ; then
        percentage_used="$( nvme smart-log "${1}" -H | grep "^percentage_used" | tr ' ' '\n' | grep -E "^[[:digit:]]+%" | tail -1 | sed -e 's|%$||g' )"
        if [[ "${percentage_used}" -ge 90 ]] ; then
            guitool warning "$( eval_gettext "Your NVMe device is nearing the end of its life and must be replaced. Write operations will slow down over time, and eventually, your NVMe device will become read-only. Your current usage percentage is:" ) ${percentage_used}"
        fi
    fi
}

# detect the amount of hard disks we have on this computer
get_disks_number(){
    HD_DISKS_NUMBER_v="$( LC_ALL="$EL_LC_EN" lsblk -d -rino TYPE | grep -vwE "(loop|rom)" | grep -w "disk" | wc -l )"

    if ((is_live_from_usb)) && [[ "$HD_DISKS_NUMBER_v" -gt 1 ]] ; then

        # this feature is only interesting if the usb is not detected as SDA
        if ((is_live_from_usb_detected_as_sda)) ; then
            el_explain 0 "W: booted usb is detected as sda, so not ignoring one entry for ask about mbr"

        else
            # remove one entry
            el_explain 0 "Number of disks was $HD_DISKS_NUMBER_v, but one of them is our usb bootable stick, removing one entry..."
            HD_DISKS_NUMBER_v="$(( $HD_DISKS_NUMBER_v - 1 ))"
        fi
    fi
}


get_partition_number_from_partition(){
    local num partition vg_name pv_name device line
    partition="$1"
    # just for fallback
    device="$partition"

    #el_debug "\targs '$*'"

    sync ; LC_ALL=C sleep 0.2

    # redirect complex info
    if [[ "$partition" = /dev/mapper/* ]] ; then
        # lvs / cryptsetup
        if echo "$partition" | grep -qs "/dev/mapper/" ; then
            vg_name="$( LC_ALL="$EL_LC_EN" LVM_SUPRESS_FD_WARNINGS=1 lvs --noheadings --separator : -o vg_name "$partition" 2>/dev/null || true )"
            if [[ -n "$vg_name" ]] ; then
                while read -ru 3 line
                do
                    if [[ "$line" = "PV Name"* ]] ; then
                        pv_name="$( echo "$line" | sed -e 's|^.*/dev/|/dev/|g' )"
                    fi

                    if [[ "$line" = "VG Name"*${vg_name} ]] ; then
                        device="$pv_name"
                        el_debug "partition '$partition' is a LVM2 type: '$device' "
                        break
                    fi

                done 3<<< "$( LC_ALL="$EL_LC_EN" pvdisplay -m 2>/dev/null )"
            fi

            # decode cryptsetup info
            if [[ -n "$device" ]] ; then
                # it may be an encrypted one in any case, so lets translate it
                if LC_ALL="$EL_LC_EN" cryptsetup status "$device" 2>/dev/null | grep -qs "type:.*LUKS" ; then
                    device="$( LC_ALL="$EL_LC_EN" cryptsetup status "$device" | grep "device:" | sed -e 's|^.*device:||g' )"
                    read -r device <<< "$device"
                    el_debug "partition '$partition' is a cryptsetup type: '$device' "
                fi
            fi

            # set the new partition
            if [[ -n "$device" ]] && [[ "$partition" != "$device" ]] ; then
                partition="$device"
                el_debug "changed partition '$partition' to '$device' "
            fi
        fi
    fi

    if [[ "$partition" = /dev/nvme*p* ]] || [[ "$partition" = /dev/mmcblk*p* ]] || [[ "$partition" = /dev/md*p* ]] ; then
        num="$partition"
        num="${num##*n[0-9]*p}"
        num="${num##*[0-9]*p}"
        el_debug "partition number for '$partition' is '$num' (nvme, mmcblk, or md126p3-like)"
    else
        num="${partition//[^0-9]}"
        el_debug "partition number for '$partition' is '$num' (normal type)"
    fi

    if [[ -n "$num" ]] ; then
        echo "$num"
    else
        echo "0"
        el_error "Unable to get partition number from partition: $partition"
    fi
}


#===  FUNCTION  ================================================================
#          NAME:  hd_device_get_size_mb
#   DESCRIPTION:  gets the MB size of an HD or partition
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
hd_device_get_size_mb(){
    # pre {{{
    local dev size
    el_debug
    el_security_function_loop || return 0

    dev="$1"
    el_check_variables "dev"

    sync ; LC_ALL=C sleep 0.2
    # }}}
    # get size of disk
    if hd_device_is_blockdevice "$dev" ; then
        size="$( LC_ALL="$EL_LC_EN" lsblk -b -rino PATH,TYPE,SIZE | awk -v item="${dev}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' )"

        if echo "$size" | grep -qsE "[[:digit:]]*" ; then
            size="$( echo "$size / 1024 / 1024" | LC_ALL="$EL_LC_EN" bc -l | sed -e 's|\..*$||g' )"
        fi

        # get size of disk
        if ! echo "$size" | grep -qs "[[:digit:]]" ; then
            size="$( awk -v DEV="${HD_DEV##*/}" '{
            if ($4 == DEV)
                {
                    size=$3
                    printf("%d\n", int(size) / 1024)
                }
            }' /proc/partitions | head -1 )"
        fi
    else
        el_debug "$dev is not a block device? $( file "$dev" )"
    fi

    if [[ -n "$size" ]] && echo "$size" | grep -qs "[[:digit:]]" ; then
        echo "$size"
    else
        el_warning "unable to get size for disk $dev \n$( fdisk -l | grep "$dev" )\n$( partitions-list --show-all --show-disks --show-raw || true )\n$( lsblk -rino PATH,TYPE,FSTYPE,SIZE | grep "$dev" )"
    fi
}

#===  FUNCTION  ================================================================
#          NAME:  hd_device_is_blockdevice
#   DESCRIPTION:  check if the provided HD or partition is a block device
#    PARAMETERS:  $1 = part | dev
#       RETURNS:  true if is, false if is not
#===============================================================================
hd_device_is_blockdevice(){
    # pre {{{
    local dev count

    dev="$1"

    # }}}
    # empty value given, fail
    if ! [[ -n "$dev" ]] ; then
        el_error "dev '$dev' is empty, from ${FUNCNAME[1]}"
        return 1
    fi

    for count in $(seq 12)
    do
        if [[ -b "$dev" ]] ; then
            return 0
        else
            sync ; sleep 1
            partprobe "$dev" 2>/dev/null || true
            partprobe "$HD_DEV" 2>/dev/null || true
        fi
    done

    # unable to get it correctly, so we failed
    el_error "dev '$dev' is not a block device:\n$(ls -la ${dev} )\n$(file ${dev})\n$(file -s ${dev})"
    return 1
}

hd_refresh(){
    sync
    partprobe "$1" 2>/dev/null || true
    blockdev --flushbufs --rereadpt "$1" 2>/dev/null || true
}

warn_experimental_fs(){
    local fs
    fs="$1"
    shift

    if ! ((is_user_warned_fs_experimental)) ; then

        case "$fs" in
            btrfs)
                guitool warning "$( eval_gettext "The selected filesystem is complex. Always backup your important data, as it's the only way to ensure the safety of your data." )"
                ;;
            reiser4|reiser5)
                guitool warning "$( eval_gettext "The selected file system is newer, and we highly recommend it. However, if you store important data, we advise regular backups or selecting a commonly used filesystem like EXT4." )"
                ;;
        esac

        is_user_warned_fs_experimental=1
    fi
}

#warn_windows(){
    #if grep -qs ":Windows" /tmp/os-probed ; then
        #guitool warning "$( eval_gettext "Your computer has a Windows system installed. It is well known that it gives an infinite amount of problems, like modifying the clock time or forcing the computer to only boot windows to keep its monopol. Remember that you can run Windows applications inside Elive, so you can verify yourself if you can get rid of this unneeded system which is full of virus and insecurities. Our personal suggestion is to remove it and use the entire disk to install Elive." )"
    #fi
#}

reboot_needed_check(){
    # just a check that forces the user to reboot before to continue (if we need to)
    if [[ -f "/tmp/reboot-needed" ]] ; then
        guitool error "$( eval_gettext "Your computer must be rebooted before we can continue, otherwise, galactic particles could be affected." )"
        exit
    fi
}

# wrapper for a faster umounting
el_umount(){
    if ((is_wheezy)) ; then
        EL_DEBUG=0 mounts-manager --umount "$@" 1>/dev/null
    else
        if ! umount "$@" 1>/dev/null 2>&1 ; then
            sync
            if ! umount "$@" 1>/dev/null 2>&1 ; then
                EL_DEBUG=0 mounts-manager --umount "$@" 1>/dev/null
            fi
        fi
    fi
}


# small function to debug calls that gives returning results
trap_signal_error_debug(){
    case "$1" in
        enable)
            trap "error_signal_debug" ERR
            ;;
        disable)
            trap - ERR
            ;;
    esac
}

swapfile_temporal_live(){
    local swapfile part
    swapfile="$( grep "swapfile-tmp-deleteme" /etc/fstab | awk '{print $1}' | tail -1 )"
    # no entry? ignore
    if [[ -z "$swapfile" ]] ; then
        return
    fi

    case "$1" in
        "activate")
            if [[ -s "$swapfile" ]] && ! swapon -s | grep -qs "$swapfile" ; then
                #swapon "$swapfile" 1>/dev/null 2>&1 || true
                swapon "$swapfile" 1>/dev/null
            fi

            ;;
        "deactivate")
            if [[ -s "$swapfile" ]] && swapon -s | grep -qs "$swapfile" ; then
                swapoff "$swapfile" 1>/dev/null
            fi

            ;;
        "remove")
            if [[ -s "$swapfile" ]] && swapon -s | grep -qs "$swapfile" ; then
                swapoff "$swapfile" 1>/dev/null
                rm -f "$swapfile"
            fi
            part="$( grep "swapfile-tmp-deleteme" /etc/fstab | sed -e 's|^.*# from partition: ||g' | awk '{print $1}' | tail -1 )"
            if [[ -n "$part" ]] ; then
                el_umount "$part" 1>/dev/null || true
                el_array_member_unset "$part" "${HD_PARTITIONS_IGNORE_UMOUNT_array[@]}" ; HD_PARTITIONS_IGNORE_UMOUNT_array=("${_out[@]}")
            else
                el_error "we don't know the partition for the 'swapfile-tmp-deleteme' temporal swap, which should have been generated on drivers-install deliver, we should instead add it from the installer code which also knows it"
            fi

            ;;
    esac
}

swapfile_installer(){
    case "$1" in
        "activate")
            if [[ -s "$TARGET/swapfile-installer.swp" ]] ; then
                swapon "$TARGET/swapfile-installer.swp" 1>/dev/null
            fi

            ;;
        "deactivate")
            if [[ -s "$TARGET/swapfile-installer.swp" ]] ; then
                swapoff "$TARGET/swapfile-installer.swp" 1>/dev/null
            fi

            ;;
        "remove")
            if [[ -s "$TARGET/swapfile-installer.swp" ]] ; then
                swapoff "$TARGET/swapfile-installer.swp" 1>/dev/null || true
                sync
                rm -f "$TARGET/swapfile-installer.swp" || true
            fi

            ;;
    esac
}


############################## GET Disk
hardware_id_get_disk(){
    # pre {{{
    local disk
    disk="$( get_disk_from_partition_complex "$1" )"

    # obtiene el numero de serie del disco duro y lo pone en HARDWARE_ID_DISK_v
    if [[ -z "$disk" ]] ; then
        HARDWARE_ID_DISK_v="disk2008"
        return 0
    fi

    # - pre }}}

    # hdparm
    HARDWARE_ID_DISK_v="$( LC_ALL="$EL_LC_EN" /sbin/hdparm -I "$disk" 2>/dev/null | grep "Serial Number:" 2>/dev/null | sed 's|^.*Serial.*Number:||g' 2>/dev/null | head -1 2>/dev/null | awk '{print $1}' 2>/dev/null )"

    # sdparm
    if [[ -z $HARDWARE_ID_DISK_v ]] ; then
        HARDWARE_ID_DISK_v="$( LC_ALL="$EL_LC_EN" /usr/bin/sdparm -i "$disk" 2>/dev/null | grep "vendor specific:" 2>/dev/null | sed 's|^.*vendor.*specific:||g' 2>/dev/null | head -1 2>/dev/null | awk '{print $1}' 2>/dev/null )"
    fi

    # use awk
    # update: lshal is deprecated
    # if [[ -z $HARDWARE_ID_DISK_v ]] ; then
    #     # FIXME: move to an external function ?
    #     HARDWARE_ID_DISK_v="$( LC_ALL="$EL_LC_EN" lshal 2>/dev/null | awk '
    #          BEGIN {
    #             skip    = 1
    #             inBlock = 0
    #             FS = " = "
    #          }
    #
    #          function die(msg) {
    #             printf "ERROR on line %d of input: %s\n", NR, msg >> "/dev/stderr"
    #             exit 1
    #          }
    #
    #          function trim(str) {
    #             sub(/^ +/, "", str)
    #             sub(/ +$/, "", str)
    #             return str
    #          }
    #
    #          function startBlock() {
    #             split("", PROPERTIES) # delete all values in the PROPERTIES array
    #             split("", PROPORDER)  # clear the ordered properties list
    #             PROPNUM = 0           # reset the order counter
    #             split("", PROPTYPES)  # clear the type annotations as well
    #             parseBlockLine()
    #          }
    #
    #          function parseBlockLine(   key, val, type) {
    #             key = trim($1)
    #             match($2, /\([^\)]+\)$/) # find parenthesized strings that are last on the line
    #          if (RLENGTH > 0) {
    #             type = substr($2, RSTART+1, RLENGTH-2) # offsets to remove parentheses
    #             val  = trim(substr($2, 1, (length($2) - RLENGTH))) # remove type info
    #          } else {
    #             type = "udi"
    #             val  = trim($2)
    #          }
    #          #printf "!%s! ---- !%s! ---- !%s!\n", key, val, type
    #          PROPERTIES[key] = PROPORDER[PROPNUM++] = val
    #          PROPTYPES[key]  = PROPTYPES[PROPNUM]   = type
    #       }
    #
    #       function endBlock(){
    #          #  if ("block.storage_device" in PROPERTIES && \
    #                #      "storage.model"        in PROPERTIES && \
    #             #      length(PROPERTIES["storage.model"]) > 2) {
    #             #     printf "device:\t%s\nmodel:\t%s\n",
    #             #       PROPERTIES["block.storage_device"],
    #             #       PROPERTIES["storage.model"]
    #
    #             if ("block.device"         in PROPERTIES && \
    #                   "storage.serial"       in PROPERTIES && \
    #                   length(PROPERTIES["storage.serial"]) > 2) {
    #                printf "%s\t%s\n",
    #                PROPERTIES["block.device"],
    #                PROPERTIES["storage.serial"]
    #
    #                #                      printf "%s\n", PROPERTIES["storage.serial"]
    #
    #             }
    #             return
    #          }
    #
    #          ##########################################
    #          # parser code below
    #          ##########################################
    #
    #          $0 ~ /^(-+$|Dumped)/ {skip = ! skip; next}
    #          skip {next}
    #          # skip the header and the footer
    #
    #          /^udi/ && (! inBlock) {
    #             inBlock = 1
    #             startBlock()
    #             next
    #          }
    #          /^  / && inBlock {
    #             parseBlockLine()
    #             next
    #          }
    #          /^$/ && inBlock {
    #             inBlock = 0
    #             endBlock()
    #             next
    #          }
    #
    #          (/^udi/ && inBlock) || (/^  / && (! inBlock)) {
    #             die("Block info line encountered out of sequence, parsing error!")
    #          }
    #      ' 2>/dev/null | tr "'" " " 2>/dev/null | awk -v MYDEV="$1" '{if ($1 == MYDEV) print $2 }' 2>/dev/null )"
    #
    #      HARDWARE_ID_DISK_v="${HARDWARE_ID_DISK_v#*-_}"
    #  fi

     if [[ -z $HARDWARE_ID_DISK_v ]] ; then
         HARDWARE_ID_DISK_v="disk2008"
     fi
}

#===  FUNCTION  ================================================================
#          NAME:  mbr_select
#   DESCRIPTION:  Select where to install the boot loader
#    PARAMETERS:  -
#       RETURNS:  -
#===============================================================================
mbr_select(){
    # pre {{{
    local menu entry disk size answer
    el_debug "\targs '$*'"

    if [[ -n "$HD_PART_EFI_dev" ]] ; then
        return 0
    fi


    # }}}

    # check health of nvme
    nvme_check_percent "$HD_DEV"

    # set default if migration mode
    if ((is_migration)) || ((is_autoupgrade))  ; then
        el_debug "MBR to install grub assigned to: $HD_DEV"
        # will use later the auto mode to HD_DEV
        HD_DISKS_NUMBER_v=1
    fi

    if [[ "$HD_DISKS_NUMBER_v" -eq 1 ]] && [[ -n "$HD_DEV" ]] ; then
        # use the only disk available for install grub
        mbr_device_install="$HD_DEV"
        mbr_device_id_install="$( hdtools_device_to_id "$HD_DEV" )"
        el_explain 0 "Number of disks is ${HD_DISKS_NUMBER_v}, HD where to install grub is $mbr_device_install and with ID $mbr_device_id_install"

    else
        # we have more than one disk ?
        unset menu
        while read -ru 3 entry
        do
            unset disk size name
            [[ "$entry" = "/dev/loop"* ]] && continue
            if ! [[ -b "$entry" ]] ; then
                el_debug "skipping $entry because is not a block device"
                continue
            fi

            # skip usb device
            if ((is_live_from_usb)) && [[ "$entry" = "$LIVE_MOUNT_MEDIA_DEV_v" ]] ; then
                el_explain 0 "$entry is our USB bootable stick, we dont want to install grub here"
                continue
            fi

            size="$( LC_ALL="$EL_LC_EN" lsblk -rino PATH,TYPE,SIZE | awk -v item="${entry}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' )"
            if [[ "$size" = "0B" ]] ; then
                el_explain 0 "$entry has 0B size, ignoring it"
                continue
            fi
            [[ -z "$size" ]] && size="(unknown)"

            name="$( udevadm info --query=property --name="$entry" | grep "ID_VENDOR_ID=" | sed -e 's|^.*VENDOR_ID=||g' | tail -1 )"
            [[ -z "$name" ]] && name="$( udevadm info --query=property --name="$entry" | grep "ID_MODEL=" | sed -e 's|^.*ID_MODEL=||g' | tail -1 )"
            [[ -z "$name" ]] && name="$( LC_ALL="$EL_LC_EN" lsblk -rino PATH,TYPE,LABEL | awk -v item="${entry}" '{ if (($1 == item) && ($2 ~ /(part|crypt|lvm|disk|md|raid)/ ) ) print $3}' )"
            [[ -z "$name" ]] && name="disk"

            menu+=("$( echo "$entry" )")
            menu+=("$( echo "$name" )")
            menu+=("$( echo "$size" )")

        done 3<<< "$( { LC_ALL="$EL_LC_EN" lsblk -d -rino PATH ; ls -1 /dev/[shv]d[a-z] /dev/xvd[a-z] /dev/nvme[0-9]*n[0-9] /dev/md[0-9][0-9][0-9] /dev/mmcblk[0-9] /dev/mmcblk[0-9][0-9]   2>/dev/null || true ; } | sort -u )"


        # if we got only one result for the menu is because we have only one device found, dont ask and just use it
        if [[ "${#menu[@]}" -eq 3 ]] ; then
            answer="$( echo "${menu[0]}" | grep "/dev/" | head -1 | awk '{print $1}' )"
            el_explain 0 "MBR: only one disk found, asigned to __${answer}__"
            # check
            if [[ "${HD_DEV}" != "$answer" ]] ; then
                el_warning "$HD_DEV which is the device for your installed system, doesn't match with the device assigned automatically to install the MBR/grub in, we assume that this is correct on this way"
            fi
        else
            # if Elive was installed on /dev/sda, we should not ask for where to install grub, because sda is the first disk where the computer is going to boot and where we installed Elive
            if [[ "${#menu[@]}" -gt 3 ]] && [[ "$HD_DEV" = "/dev/sda" ]] ; then
                el_explain 0 "MBR: selected __sda__ disk because is the first detected by the bios and the selected one to install Elive, so it should be this one"
                answer="/dev/sda"
            fi
        fi



        if ! [[ -b "$answer" ]] ; then
            # if we have those special types of disks, let the user to really-configure it from grub debconf
            #if LC_ALL="$EL_LC_EN" lsblk -o TYPE | grep -qsE "^(lvm|md|raid)" && ((conf_configure_options)) ; then
            if LC_ALL="$EL_LC_EN" lsblk -rino TYPE | grep -qsE "^(lvm|md|raid)" ; then

                if ! ((is_migration)) ; then
                    # do not ask, we always want to select this

                    # if our Elive is installed on the first disk, thats what we just want, so set it
                    if echo "$HD_DEV" | grep -qsE "/dev/(sda|hda)" ; then
                        el_explain 0 "MBR: selected __sda__ disk (in raid/lvm/md mode) because is the first detected by the bios and the one installed for Elive, so it should be this one"
                        answer="$HD_DEV"
                    else
                        el_explain 0 "MBR: letting the user select where to install grub"
                        is_mbr_grub_elive_mode_wanted=1
                    fi
                fi

            else
                el_explain 0 "MBR: letting the user select where to install grub"
                is_mbr_grub_elive_mode_wanted=1
            fi
        fi

        # ask it
        if ((is_mbr_grub_elive_mode_wanted)) ; then
            answer="$( $guitool --title='Elive Installer' --list --height=300 --width=500 --text="$( eval_gettext "Choose the hard disk where GRUB will be installed to boot your Elive system.\n\nSelect the first one if you're unsure.\nAlternatively, press Cancel to use the Debconf configurator from Debian (advanced mode)." )" --column="$( eval_gettext "Device" )" --column="$( eval_gettext "Name" )" --column="$( eval_gettext "Size" )" "${menu[@]}" 2>/dev/null || echo cancel )"
            el_explain 0 "MBR: user selected '$answer' "
        fi


        # convert id
        if [[ -b "$answer" ]] ; then
            mbr_device_install="$answer"
            mbr_device_id_install="$( hdtools_device_to_id "$answer" )"
            el_explain 0 "MBR: HD where to install grub is $mbr_device_install and with ID $mbr_device_id_install"
        else
            el_warning "user didn't selected a correct option when asking for where to install MBR (it should be asked later on debconf)"
        fi

    fi

    # warn if we could not install grub because we have not enough room before the first partition
    # FIXME: move it, do a shift instead like the migration part code (move to a function then)
    # FIXME: also do a chck when selecting root partition, suggest repartition if we dont have swap at first
    if ((is_disk_table_gpt)) ; then
        if LC_ALL="$EL_LC_EN" gdisk -l "$mbr_device_install" 2>/dev/null | grep -E "^\s+1\s+[[:digit:]]*\s+[[:digit:]]*" | tr '*' ' ' | awk '{print $2}' | grep -qsE "^1$" ; then
            guitool warning "$( eval_gettext "Your first partition starts at the beginning of the disk without initial free space. The GRUB bootloader needs a small empty space for the correct installation. If your system cannot boot, you need to repartition your disk and leave a small space of 8 MB before the first partition. These 8 MB will also give you better performance by having your partitions aligned." )"
            el_warning "First partition doesn't starts with a shift to have space to install grub (GPT structure), we should improve the installer to resize the first partition Thanatermesis"
        fi
    else
        if LC_ALL="$EL_LC_EN" fdisk -l "$mbr_device_install" 2>/dev/null | grep "^/dev/" | tr '*' ' ' | awk '{print $2}' | grep -qsE "^1$" ; then
            guitool warning "$( eval_gettext "Your first partition starts at the beginning of the disk without initial free space. The GRUB bootloader needs a small empty space for the correct installation. If your system cannot boot, you need to repartition your disk and leave a small space of 8 MB before the first partition. These 8 MB will also give you better performance by having your partitions aligned." )"
            el_warning "First partition doesn't starts with a shift to have space to install grub (MBR/msdos structure), we should improve the installer to resize the first partition Thanatermesis"
        fi
    fi


    # set where to install grub
    if [[ -n "$mbr_device_id_install" ]] && [[ -b "$mbr_device_id_install" ]] ; then
        # this will reconfigure grub2, so we dont want that on this step, so just add it as a normal preseed:
        #mkdir -p "/tmp/preseeds.d"
        #echo -e "grub-pc\tgrub-pc/install_devices\tmultiselect\t${mbr_device_id_install}" > "/tmp/preseeds.d/grub-pc.cfg"

        # this preseed is appended to TARGET, which already exists (function called from hooks_post_install after preseeds copy)
        mkdir -p "$TARGET/root/preseeds.d"
        sed -i "s|^.*grub-pc.*grub-pc/install_devices.*$||g" "$TARGET/root/preseeds.cfg" 2>/dev/null || true
        sed -i "s|^.*grub-pc.*grub-pc/install_devices.*$||g" "$TARGET/root/preseeds.d/grub-pc" 2>/dev/null || true
        echo -e "grub-pc\tgrub-pc/install_devices\tmultiselect\t${mbr_device_id_install}" >> "$TARGET/root/preseeds.d/grub-pc"
    fi

    # set boot flag
    # not compatible with gpt mode
    if ((is_disk_table_msdos)) ; then
        if [[ -n "$HD_PART_BOOT_dev" ]] ; then
            parted -ms -a optimal "$( get_disk_from_partition_complex "$HD_PART_BOOT_dev" )" set "$( get_partition_number_from_partition "$HD_PART_BOOT_dev" )" boot on
        else
            # if we have a valid partition number (/dev/mapper/foo is not allowed to set a boot flag), set the boot flag to the root partition
            if echo "$HD_PART_ROOT_dev" | grep -qsE "((xvd|sd|hd|vd)[a-z]|nvme[[:digit:]]+n[[:digit:]]+|mmcblk[[:digit:]]+|md[[:digit:]]+p[[:digit:]]+)" ; then
                parted -ms -a optimal "$( get_disk_from_partition_complex "$HD_PART_ROOT_dev" )" set "$( get_partition_number_from_partition "$HD_PART_ROOT_dev" )" boot on
            fi
        fi
    fi

    # check
    if ! grep -qs "grub-pc/install_devices" "$TARGET/root/preseeds.d/grub-pc" ; then
        el_error "debconf preseeds: grub-pc install_devices is not set"
    fi
}



#
# TODO: reimplement:
#

# from hooks_end (steps_4)

    # TODO: re-implement the hibernation:
    #fbvalue="$fbvalue resume=swap:${SWAP_DEVICE_FIRST}"
        # seems like we need this in /etc/default/grub:
        #GRUB_CMDLINE_LINUX="resume=/dev/sdXY"


    # TODO: nurse
    # hooks for nurse-repair mode things
    #sed -i -e '/^kernel.*nurse.*noentrance/s|splash=silent||g' -e '/^kernel.*nurse.*noentrance/s|  | |g' -e '/^title Elive Nurse/s| Nurse Mode | Reparation |g' -e '/^title Elive Nurse/s| Nurse | Reparation |g'   $TARGET/boot/grub/menu.lst


    # windows recovery:
    # TODO: sometimes it is not found, so, when are found 2 equal "windows boot" (ordenador fran) on the same disk, remove the entry from the SMALLER disk
    #detect_vista_recovery | while read -r line
    #do
        #vista_dev="$( echo "$line" | awk '{print $1}' )"
        #vista_mnt="$( echo "$line" | awk '{print $2}' )"
        #if [[ -n "$vista_dev" && -n "$vista_mnt" ]] ; then
            #rmdir $TARGET/$vista_mnt
            #sed -i "\|${vista_dev}.*${vista_mnt}|d" $TARGET/etc/fstab

            ## change grub menu.lst word to: Recovery
            #cat $TARGET/boot/grub/menu.lst | while read -r line
            #do
                #if test -n "$line" && echo "$line" | awk vista_dev="$vista_dev" '{if ($3 == vista_dev) print $0}' | grep -q "# on $vista_dev" ; then
                    #found=1
                #fi
                #if ((found)) ; then
                    #if echo "$line" | grep -q "^title.*Windows" ; then
                        #echo -e "title       Windows Recovery"
                        #unset found
                    #else
                        #echo "$line"
                    #fi
                #else
                    #echo "$line"
                #fi
            #done > $TARGET/boot/grub/menu.lst2
            #mv $TARGET/boot/grub/menu.lst2 $TARGET/boot/grub/menu.lst
        #fi
    #done


    # FIXME: re-implement:
    # FIXME: when deliver wants to install nvidia, make sure that we have a swap, installing those packages should take a good amount of memory, if we have not enough ram, ask if we want to instlal it now or after installation
    # move to hooks_post_install maybe ?
    #if [[ -e /tmp/.nvidia-71 ]] ; then
        #echo -e "dpkg -i /elive/modules/nvidia/71/* 2>/dev/null 1>&2 " >> "$TARGET/tmp/temp"
        #echo -e "echo nvidia-glx hold | dpkg --set-selections" >> "$TARGET/tmp/temp"
        #nvidia_bug_fix
    #fi

    #if [[ -e /tmp/.nvidia-96 ]] ; then
        #echo -e "dpkg -i /elive/modules/nvidia/96/* 2>/dev/null 1>&2 " >> "$TARGET/tmp/temp"
        #echo -e "echo nvidia-glx hold | dpkg --set-selections" >> "$TARGET/tmp/temp"
        #nvidia_bug_fix
    #fi

    #if [[ -e /tmp/.nvidia-173 ]] ; then
        #echo -e "dpkg -i /elive/modules/nvidia/173/* 2>/dev/null 1>&2 " >> "$TARGET/tmp/temp"
        #echo -e "echo nvidia-glx hold | dpkg --set-selections" >> "$TARGET/tmp/temp"
        #nvidia_bug_fix
    #fi

    #if [[ -e /tmp/.nvidia-new ]] ; then
        #echo -e "dpkg -i /elive/modules/nvidia/new/* 2>/dev/null 1>&2 " >> "$TARGET/tmp/temp"
        #echo -e "echo nvidia-glx hold | dpkg --set-selections" >> "$TARGET/tmp/temp"
        #nvidia_bug_fix
    #fi

    #if [[ -e /tmp/.ati-installed ]] ; then
        #echo -e "dpkg -i /elive/modules/ati/* " >> "$TARGET/tmp/temp"
        #echo -e "echo fglrx-driver hold | dpkg --set-selections" >> "$TARGET/tmp/temp"
        ##      echo -e "echo fglrx-kernel-$(uname -r) hold | dpkg --set-selections" >> "$TARGET/tmp/temp" # No hay que hacerle hold, o a la que actualice el kernel no actualizará este driver (solo al fglrx-driver)
    #fi

    # TODO: put somewhere?
    #sed -i "s/Comment\".*$/Comment\",\"(Elive ${elivecodename%% *})\"/g" "$TARGET/etc/iceweasel/profile/prefs.js"












# vim: set foldmethod=marker : #
